/*
 SRSTFormat.java

 Created on Aug 15, 2012 by Richard Johansson (richard.johansson@gu.se).

 $Log$

 */
package se.lth.cs.nlp.depsrl.format;

import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.depgraph.*;

import java.util.*;
import java.io.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class SRSTFormat {

	private static final String EMPTY = "_";

    private static final int FORM = 1;
    private static final int LEMMA = 2;
    private static final int POS = 3;
    private static final int POSFINE = 4;
    private static final int FEATS = 5;
	private static final int HEAD = 6;
	private static final int DEPREL = 7;

    private static final int SEC = 8;

	private static final int PRED = 9;
	private static final int ARG_START = 10;

	public static Pair<DepGraph, List<PAStructure>> 
	readNextGraph(BufferedReader br) throws IOException {

		String line = br.readLine();
		if(line == null)
			return null;
		line = line.trim();

		ArrayList<String[]> lines = new ArrayList();        

		while(true) {
			lines.add(line.split("\t"));
			line = br.readLine();
			if(line == null)
				break;
			line = line.trim();
			if(line.equals(""))
				break;
		}

		int len = lines.size();
		DepGraph dg = new DepGraph(len);

		int width = lines.get(0).length;
		for(int i = 1; i < len; i++)
		    if(lines.get(i).length != width) {
			for(int j = 0; j < len; j++) {
			    if(i == j)
				System.out.print("-->");
			    System.out.println(Arrays.asList(lines.get(j)));
			}
			throw new RuntimeException("Illegal format!");
		    }
		
		dg.nodes[0] = new DepNode();
		for(int i = 0; i < len; i++) {
			String[] ts = lines.get(i);
			dg.nodes[i+1] = new DepNode(i+1, ts[FORM], ts[POS], ts[LEMMA]);
		}

		for(int i = 0; i < len; i++) {
			String[] ts = lines.get(i);
			int hpos = Integer.parseInt(ts[HEAD]);
			dg.nodes[hpos].addChild(dg.nodes[i+1], ts[DEPREL]);
		}

		for(int i = 0; i < len; i++) {
			String[] ts = lines.get(i);
			if(!ts[SEC].equals(EMPTY)) {
			    for(String ps: ts[SEC].split(",")) {
				int ix = ps.indexOf('/');
				int hpos = Integer.parseInt(ps.substring(0, ix));
				dg.nodes[hpos].addChild(dg.nodes[i+1], ps.substring(ix+1));
			    }
			}
		}

		ArrayList<PAStructure> pas = new ArrayList();

		if(width > PRED) {
		    for(int i = 0; i < len; i++) {
			String[] ts = lines.get(i);
			if(!ts[PRED].equals(EMPTY)) {
			    PAStructure pa = new PAStructure();
			    pa.pred = dg.nodes[i + 1];
			    pa.lemma = ts[PRED];
			    pas.add(pa);
			}
		    }

		    if(lines.get(0).length != ARG_START + pas.size())
			throw new RuntimeException("Illegal number of columns!");
		    
		    for(int i = 0; i < len; i++) {
			String[] ts = lines.get(i);
			for(int j = ARG_START; j < ts.length; j++) {
			    if(!ts[j].equals(EMPTY)) {
				PAStructure pa = pas.get(j - ARG_START);
				pa.args.add(dg.nodes[i + 1]);
				pa.argLabels.add(ts[j]);
			    }
			}
		    }
		}
		
		return new Pair(dg, pas);
	}
	
    public static void printForWW(DepGraph dg, List<PAStructure> pas, PrintWriter pw) {
	StringBuilder[] lines = new StringBuilder[dg.nodes.length-1];

	HashMap<Integer, String>[] seclinks = new HashMap[dg.nodes.length];
	String[] plabels = new String[dg.nodes.length];

	for(int i = 1; i < dg.nodes.length; i++) {
	    DepNode n = dg.nodes[i];
	    lines[i-1] = new StringBuilder();
	    lines[i-1].append(i);
	    lines[i-1].append("\t");
	    lines[i-1].append(dg.nodes[i].word);
	    lines[i-1].append("\t");
	    lines[i-1].append(dg.nodes[i].lemma);
	    lines[i-1].append("\t");
	    lines[i-1].append(dg.nodes[i].pos);
	    lines[i-1].append("\t");
	    lines[i-1].append(dg.nodes[i].pos);
	    lines[i-1].append("\t");
	    lines[i-1].append(dg.nodes[i].word);
	    lines[i-1].append("\t");
	    lines[i-1].append(dg.nodes[i].lemma);
	    lines[i-1].append("\t");
	    lines[i-1].append(dg.nodes[i].pos);
	    lines[i-1].append("\t");
	    lines[i-1].append(dg.nodes[i].parents[0].position);
	    lines[i-1].append("\t");
	    lines[i-1].append(dg.nodes[i].relations[0]);
	    lines[i-1].append("\t");

	    if(n.parents.length > 1) {
		for(int j = 1; j < n.parents.length; j++) {
		    DepNode p = n.parents[j];
		    if(seclinks[p.position] == null)
			seclinks[p.position] = new HashMap();
		    seclinks[p.position].put(i, n.relations[j]);
		    plabels[p.position] = "X";
		}
	    }
	}

	for(PAStructure pa: pas) {
	    plabels[pa.pred.position] = pa.lemma;
	    for(int i = 0; i < pa.args.size(); i++) {
		DepNode a = pa.args.get(i);
		String l = pa.argLabels.get(i);
		if(seclinks[pa.pred.position] == null)
		    seclinks[pa.pred.position] = new HashMap();
		String prevl = seclinks[pa.pred.position].get(a.position);
		if(prevl == null)
		    seclinks[pa.pred.position].put(a.position, l);
		else
		    seclinks[pa.pred.position].put(a.position, prevl+"/"+l);
	    }
	}

	for(int i = 1; i < dg.nodes.length; i++) {
	    if(plabels[i] == null)
		lines[i-1].append(EMPTY);
	    else
		lines[i-1].append(plabels[i]);
	}

	for(int i = 1; i < dg.nodes.length; i++) {
	    if(plabels[i] == null)
		continue;
	    HashMap<Integer, String> m = seclinks[i];
	    if(m == null)
		for(int j = 1; j < dg.nodes.length; j++) 
		    lines[j-1].append("\t" + EMPTY);
	    else
		for(int j = 1; j < dg.nodes.length; j++) {
		    lines[j-1].append("\t");
		    String l = m.get(j);
		    if(l == null)
			lines[j-1].append(EMPTY);
		    else
			lines[j-1].append(l);
		}
	}

	for(int i = 0; i < lines.length; i++)
	    pw.println(lines[i]);
	pw.println();
	//System.exit(0);
    }


    public static void main(String[] argv) {
	    try {
		BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
		PrintWriter pw = new PrintWriter(new OutputStreamWriter(System.out));
		Pair<DepGraph, List<PAStructure>> p = readNextGraph(br);
		while(p != null) {
		    printForWW(p.left, p.right, pw);
		    p = readNextGraph(br);
		}

		pw.close();

	    } catch(Exception e) {
		e.printStackTrace();
	    }
	}
    
}
