/*
 TestNonprojectivity.java
  
 Created on May 6, 2008 by Richard Johansson (richard@cs.lth.se).

 $Log: TestNonprojectivity.java,v $
 Revision 1.1  2009/01/16 14:50:55  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 11:50:36  richard
 Added.

   
 */
package se.lth.cs.nlp.depsrl.format;

import java.util.*;
import java.io.*;

import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.depgraph.*;


/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class TestNonprojectivity {

    private static void findNonprojectiveEdges(DepGraph dg,
            DepNode n,
            HashSet<DepNode> result) {
        if(n.parents.length > 0) {            
            DepNode p = n.parents[0];
            int di = p.position > n.position? 1: -1;
            for(int i = n.position + di; i != p.position; i += di) {
                DepNode inter = dg.nodes[i];
                if(!p.isSameOrAncestorOf(inter)) {
                    //System.out.println("n = " + n);
                    //System.out.println("p = " + p);
                    //System.out.println("inter = " + inter);
                    result.add(n);
                    break;
                }
            }
        }
        for(DepNode c: n.children)
            findNonprojectiveEdges(dg, c, result);
    }
    
    private static HashMap<String, int[]> resultMap = new HashMap();

    private static final String ALL = "All";
    private static final String UNKNOWN = "Unknown";
    private static final String WHMOV = "WH-movement";
    private static final String COORD = "Coordination";
    private static final String PUNC = "Punctuation";
    private static final String PRN = "Parenthetical";
    private static final String NP = "Noun phrase";
    
    private static void initResultMap() {
        resultMap.put(ALL, new int[4]);
        resultMap.put(UNKNOWN, new int[4]);
        resultMap.put(WHMOV, new int[4]);
        resultMap.put(COORD, new int[4]);
        resultMap.put(PUNC, new int[4]);
        resultMap.put(PRN, new int[4]);
        resultMap.put(NP, new int[4]);
    }

    private static void printResultMap() {
        for(Object k: new TreeSet(resultMap.keySet())) {
            int[] r = resultMap.get(k);
            double pr = (double) r[0] / r[2];
            if(Double.isNaN(pr))
                pr = 0;
            double re = (double) r[1] / r[3];
            if(Double.isNaN(re))
                re = 0;
            double f1 = (double) 2*pr*re / (pr + re);
            if(Double.isNaN(f1))
                f1 = 0;
            System.out.println(k + ":");
            System.out.println("p = " + r[0] + "/" + r[2] + " = " + pr);
            System.out.println("r = " + r[1] + "/" + r[3] + " = " + re);
            System.out.println("f1 = " + f1);
        }
    }
    
    private static void check(DepGraph dgGold, HashSet<DepNode> gold, 
            DepGraph dgSys, HashSet<DepNode> sys, boolean labeled) {

        for(DepNode ns: sys) {
            String type2 = getNPType(dgSys, ns);
            resultMap.get(type2)[2]++;
            resultMap.get(ALL)[2]++;
            DepNode ng = dgGold.nodes[ns.position];
            if(ng.parents[0].position == ns.parents[0].position
               && (!labeled || ng.relations[0].equals(ns.relations[0]))) {
                resultMap.get(type2)[0]++;
                resultMap.get(ALL)[0]++;
            }
        }
        
        for(DepNode ng: gold) {
            String type1 = getNPType(dgGold, ng);
            resultMap.get(type1)[3]++;
            resultMap.get(ALL)[3]++;
            DepNode ns = dgSys.nodes[ng.position];
            if(ng.parents[0].position == ns.parents[0].position
               && (!labeled || ng.relations[0].equals(ns.relations[0]))) {
                resultMap.get(type1)[1]++;
                resultMap.get(ALL)[1]++;
            }
        }
        
    }

    private static String getNPType(DepGraph dg, DepNode n) {
        DepNode p = n.parents[0];

        if(n.pos.startsWith("W"))
            return WHMOV;

        DepNode lc = n.getLeftmostDescendent();
        if(lc != null && lc.pos.startsWith("W"))
            return WHMOV;
        DepNode rc = n.getRightmostDescendent();
        if(rc != null && rc.pos.startsWith("W"))
            return WHMOV;

        if(n.pos.equals("IN") && n.word.equals("that")
           && n.relations[0].matches("SBJ|OBJ|TMP|PMOD") && p.position > n.position)
            return WHMOV;
        
        if(n.relations[0].equals("P"))
            return PUNC;

        int di = p.position > n.position? 1: -1;
        boolean sawRoot = false;
        boolean sawPunc1 = false, sawPunc2 = false;
        for(int i = n.position + di; i != p.position; i += di) {
            //System.out.println("i = " + i);
            //System.out.println("Before:");
            //System.out.println("sawRoot = " + sawRoot);
            //System.out.println("sawPunc1 = " + sawPunc1);
            //System.out.println("sawPunc2 = " + sawPunc2);
            DepNode inter = dg.nodes[i];
            //System.out.println("inter = " + inter);
            //System.out.println("inter.parent = " + inter.parents[0]);
            if(inter.relations[0].equals("P"))
                if(!sawRoot)
                    sawPunc1 = true;
                else
                    sawPunc2 = true;
            if(inter.parents[0].position == 0) {
                if(!sawPunc1)
                    break;
                sawRoot = true;
            }
            //System.out.println("After:");
            //System.out.println("sawRoot = " + sawRoot);
            //System.out.println("sawPunc1 = " + sawPunc1);
            //System.out.println("sawPunc2 = " + sawPunc2);
        }
        if(sawRoot && sawPunc2)
            return PRN;
        
        if(n.parents[0].pos.startsWith("N")) {
            //System.out.println("n = " + n);
            //System.out.println(dg);
            //System.out.println();
            return NP;
        }
        
        if(n.relations[0].equals("CONJ"))
            return COORD;       

        return UNKNOWN;
    }
    
    private static Scanner goldInputScanner;
    private static BufferedReader goldInputReader;
    private static Scanner sysInputScanner;
    private static BufferedReader sysInputReader;
    
    private static void initInputFiles(String goldFile, String sysFile) throws Exception {
        BufferedReader gbr = new BufferedReader(new FileReader(goldFile));
        String line = gbr.readLine();
        String[] tokens = line.split("\t");
        if(tokens.length == 8 
           || (tokens.length == 10 && tokens[9].equals("_")))
            goldInputReader = new BufferedReader(new FileReader(goldFile));
        else 
            goldInputScanner = new Scanner(new File(goldFile));
        gbr = null;
        
        BufferedReader sbr = new BufferedReader(new FileReader(sysFile));
        line = sbr.readLine();
        tokens = line.split("\t");
        if(tokens.length == 8 
           || (tokens.length == 10 && tokens[9].equals("_")))
            sysInputReader = new BufferedReader(new FileReader(sysFile));
        else 
            sysInputScanner = new Scanner(new File(sysFile));
        sbr = null;
    }

    private static DepGraph nextGraph(Scanner sc, BufferedReader br) {
        if(sc == null)
            return CoNLLFormat.readNextGraph(br);
        else {
            Triple<DepGraph, DepGraph, List<PAStructure>> p 
                = CoNLL2008Format.readNextGraph(sc);
            if(p == null)
                return null;
            else
                return p.first;
        }
    }
    
    public static void main(String[] argv) {

        String goldFile = argv[0];
        String sysFile = argv[1];

        boolean labeled = false;
        if(argv.length >= 3)
            if(argv[2].equals("labeled"))
                labeled = true;
        
        try {
            initInputFiles(goldFile, sysFile);
            int nNonProjLinks = 0, nLinks = 0;
            int nNonProjSents = 0, nSents = 0;
            initResultMap();
            
            while(true) {
                DepGraph dg1 = nextGraph(goldInputScanner, goldInputReader);
                DepGraph dg2 = nextGraph(sysInputScanner, sysInputReader);

                if(dg2 == null)
                    break;
                
                if(dg1.nodes.length != dg2.nodes.length)
                    throw new RuntimeException("Illegal format");
                
                HashSet<DepNode> result1 = new HashSet();
                findNonprojectiveEdges(dg1, dg1.nodes[0], result1);
                
                HashSet<DepNode> result2 = new HashSet();
                findNonprojectiveEdges(dg2, dg2.nodes[0], result2);

                if(result1.isEmpty()) {
                    if(!dg1.isProjective()) {
                        System.out.println(dg1);
                        throw new RuntimeException("problem 1");
                    }
                } else {
                    if(dg1.isProjective())
                        throw new RuntimeException("problem 2");                    
                }
                
                if(result2.isEmpty()) {
                    if(!dg2.isProjective())
                        throw new RuntimeException("problem 3");
                } else {
                    if(dg2.isProjective())
                        throw new RuntimeException("problem 4");                    
                }

		/*
                if(!result1.isEmpty()) {
                    for(DepNode n: result1) {
                        String type1 = getNPType(dg1, n);
                        if(type1.equals(UNKNOWN)) {
                            System.out.println("result1 = " + result1);
                            System.out.println("result2 = " + result2);

                            DepNode ns = dg2.nodes[n.position];
                            if(ns.position == n.position && ns.parents[0].position == n.parents[0].position)
                                System.out.println("Correct");
                            else
                                System.out.println("Incorrect: ");
                            System.out.println("n = " + n);
                            System.out.println("n.parent = " + n.parents[0]);
                            System.out.println("ns.parent = " + ns.parents[0]);
                            System.out.println(dg1);                            
                            System.out.println(dg2);                            
                        }
                    }
                }*/
                    
                nNonProjLinks += result1.size();
                nLinks += dg1.nodes.length - 1;

                nSents += 1;
                nNonProjSents += result1.isEmpty()? 0: 1;
                
                check(dg1, result1, dg2, result2, labeled);
            }
            
            /*
            double pr = (double) ncorrect / nguess;
            double re = (double) ncorrect / ngold;
            double f1 = 2*pr*re / (pr + re);

            System.out.println("pr = " + ncorrect + "/" + nguess + " = " + pr);
            System.out.println("re = " + ncorrect + "/" + ngold + " = " + re);
            System.out.println("f1 = " + f1);
            */

            double pnplink = (double) nNonProjLinks / nLinks;
            double pnpsen = (double) nNonProjSents / nSents;

            System.out.println("Links: " + nNonProjLinks + "/" + nLinks + " = " + pnplink);
            System.out.println("Sentences: " + nNonProjSents + "/" + nSents + " = " + pnpsen);

            System.out.println();
            printResultMap();
            
        } catch(Exception e) {
            e.printStackTrace();
        }
        

        
    }
    
}
