package mpqa_structlearn.io;

import java.util.*;

import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.depgraph.*;

public class EPEPaths {

    static class Edge {
        DepNode from, to;
        String label;
    }
    
    public static ListMap<DepNode, Edge> makeBidirGraph(DepGraph dg) {
        ListMap<DepNode, Edge> out = new ListMap();
        for(DepNode n: dg.nodes) {
            if(n.parents == null)
                continue;
            for(int i = 0; i < n.parents.length; i++) {
                Edge e = new Edge();
                e.from = (DepNode) n.parents[i];
                e.to = (DepNode) n;
                e.label = n.relations[i];
                out.put(e.from, e);
                out.put(e.to, e);
            }
        }
        return out;
    }
    
    private static DepGraph prepared;
    private static ListMap<DepNode, Edge> preparedEdges;
    private static HashMap<Pair<DepNode, DepNode>, DepPath[]> cached;
    
    public static void prepare(DepGraph dg) {
        if(dg == prepared)
            return;
        prepared = dg;
        preparedEdges = makeBidirGraph(dg);
        cached = new HashMap();
    }
    
    public static DepPath[] findPaths(DepNode from, DepNode to, int k) {
        return findPaths(from, to, k, Double.POSITIVE_INFINITY);
    }
    
    public static DepPath[] findPaths(DepNode from, DepNode to, int k, double upperBound) {
        if(EPEReader.TREE_ONLY)
            return treePath(from, to, upperBound);        
        if(preparedEdges == null)
            throw new IllegalStateException("need to prepare edges");
        try {
            return findPaths(preparedEdges, from, to, k, upperBound);
        } catch(Exception e) {
            e.printStackTrace();
            System.out.println(prepared);
            System.out.println(from);
            System.out.println(to);
            System.exit(1);
            return null;
        }
    }
    
    public static DepPath[] findPaths(ListMap<DepNode, Edge> graph, DepNode from, DepNode to, int k, double upperBound) {        
        Pair<DepNode, DepNode> cacheP = new Pair(from, to);
        DepPath[] outC = cached.get(cacheP);
        if(outC != null)
            return outC;
        
        HashMap<DepNode, DepPath[]> results = new HashMap();

        PriorityQueue<DepPath> q = new PriorityQueue();
        q.add(new DepPath(from));
        
        DepPath[] endList = new DepPath[k];
        results.put(to, endList);
        
        while(endList[k-1] == null && !q.isEmpty()) {
            DepPath p = q.poll();
            
            if(p.cost > upperBound)
                continue;
            
            DepPath[] done = results.get(p.n);
            if(done == null)
                done = new DepPath[k];
            
            int position = 0;
            while(position < k) {
                if(done[position] == null)
                    break;
                position++;
            }

            if(position < k) {
                done[position] = p;
                p.advance(q, graph);
            }            
        }
        
        if(endList[0] == null) {
            DepPath p = new DepPath(from);
            endList[0] = new DepPath(p, upperBound, "<NOT_FOUND>", to);
        }
        
        cached.put(cacheP, endList);
        
        return endList;
        /*
        
        for(int i = 0; i < N; i++) {
            System.out.println(endList[i]);
        }

        */
        
    }
    
    public static DepPath[] treePath(DepNode from, DepNode to, double upperBound) {
        HashMap<DepNode, DepPath> seen = new HashMap();
        
        DepPath p = new DepPath(from, false);
        seen.put(p.n, p);
        
        while(p.n.position > 0 && p.cost <= upperBound) { // what if tree isn't connected?
            p = new DepPath(p, 1, p.n.relations[0], p.n.parents[0]);
            p.isParent = true;
            seen.put(p.n, p);
        }

        DepPath p2 = new DepPath(to, false);
        while(!seen.containsKey(p2.n)) {
            p2 = new DepPath(p2, 1, p2.n.relations[0], p2.n.parents[0]);
            if(p2.cost > upperBound || p2.n.position == 0)
                break;
        }

        DepPath a = seen.get(p2.n);
        if(a == null) {
            DepPath dummy = new DepPath(from, false);
            dummy = new DepPath(dummy, upperBound, "<NOT_FOUND>", to);
            return new DepPath[] { dummy };
        }
        
        while(p2.prev != null) {
            a = new DepPath(a, 1, p2.prevRel, p2.prev.n);
            p2 = p2.prev;
        }
        
        //return new Pair(seen.get(p2.n), p2);
        return new DepPath[] { a };
    }
    
}
