package mpqa_structlearn.io;

import java.io.*;
import java.util.*;

import mpqa_structlearn.io.EPEPaths.Edge;
import net.sf.json.*;
import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.depgraph.*;

import se.lth.cs.nlp.depsrl.format.CoNLL2008Format;
import se.lth.cs.nlp.depsrl.format.PAStructure;

public class EPEReader {

    private static final boolean FIX_STANFORD = false;
    private static final boolean REMOVE_PB = false;
    public static boolean TREE_ONLY = false;
    private static final boolean CORRUPT = false;
    private static final double CORRUPT_LABEL_PROB = 1.0;
    private static final double CORRUPT_EDGE_PROB = 0;
    private static final String DEFAULT_ROOT_LABEL = "<ROOT>";
    
    private static final String NOT_FOUND = "<NOT_FOUND>";
    
    public static DepGraph readDepGraph(BufferedReader br, EPEFeatureSpec featureSpec) throws IOException {
        JSONObject json = readNextJSON(br);
        if(json == null)
            return null;

        //if(true)
        //    return null;
        
        JSONArray nodeArray = json.getJSONArray("nodes");

        if(FIX_STANFORD)
            nodeArray = fixStanfordTokens(nodeArray);

        HashMap<Object, EPEDepNode> m = new HashMap();

        ArrayList<EPEDepNode> nodes = new ArrayList();
        for(int i = 0; i < nodeArray.size(); i++) {
            EPEDepNode n = makeDepNode(nodeArray.getJSONObject(i), m, featureSpec);
            if(n != null) {
                nodes.add(n);
                n.position = nodes.size();
            }
        }

        DepGraph dg = new DepGraph(nodes.size());
        EPEDepNode root = new EPEDepNode();
        dg.nodes[0] = root;
        root.start = -1;
        root.end = -1;
        for(EPEDepNode n: nodes)
            dg.nodes[n.position] = n;
        
        // 
        ListMap<EPEDepNode, Pair<EPEDepNode, String>> edges = new ListMap();
        
        for(int i = 0; i < nodeArray.size(); i++)
            link(nodeArray.getJSONObject(i), root, m, edges);

        removeDuplicateEdges(edges);
        
        if(FIX_STANFORD && TREE_ONLY)
            edges = spanningTree(edges, dg.nodes.length-1);
        
        for(int i = 0; i < dg.nodes.length; i++) {
            EPEDepNode n = (EPEDepNode) dg.nodes[i];
            //n.position = i;            
            List<Pair<EPEDepNode, String>> es_n = edges.get(n);
            if(es_n == null)
                continue;
            n.parents = new DepNode[es_n.size()];
            n.relations = new String[es_n.size()];
            for(int j = 0; j < es_n.size(); j++) {
                Pair<EPEDepNode, String> p = es_n.get(j);
                n.parents[j] = p.left;
                n.relations[j] = p.right;
            }
        }

        for(int i = 1; i < dg.nodes.length; i++) {
            DepNode n = dg.nodes[i];
            if(n.parents.length == 0) {
                n.parents = new DepNode[] { root };
                n.relations = new String[] { DEFAULT_ROOT_LABEL };
            }
        }
        
        dg.collectChildren();

        if(TREE_ONLY) {
            for(DepNode n: dg.nodes)
                if(n.position == 0 && n.parents.length > 0)
                    throw new RuntimeException("root has parent");
                else if(n.position > 0 && n.parents.length != 1)
                    throw new RuntimeException("node does not have single parent");
            HashSet<DepNode> seen = new HashSet();
            dg.nodes[0].collectSubtree(seen);
            if(seen.size() != dg.nodes.length)
                throw new RuntimeException("tree not connected");
        }
        
        //System.out.println(dg);
        
        //System.exit(0);
        
        return dg;
    }

    private static void removeDuplicateEdges(
            ListMap<EPEDepNode, Pair<EPEDepNode, String>> edges) {
        for(EPEDepNode c: edges.keySet()) {
            HashSet seen = new HashSet();
            for(Iterator<Pair<EPEDepNode, String>> it = edges.get(c).iterator(); it.hasNext(); ) {
                Pair<EPEDepNode, String> e = it.next();
                if(seen.contains(e))
                    it.remove();
                else
                    seen.add(e);
            }
        }
    }

    private static ListMap<EPEDepNode, Pair<EPEDepNode, String>> spanningTree(ListMap<EPEDepNode, Pair<EPEDepNode, String>> edges,
            int N) {

        /*for(EPEDepNode c: edges.keySet()) {
            System.out.println(c);
            for(Pair<EPEDepNode, String> e: edges.get(c)) {
                System.out.println(" -> " + e);
            }
        }
        System.exit(0);*/
        
        ArrayList<Integer>[] heads = new ArrayList[N];
        ArrayList<Double>[] scores = new ArrayList[N];
        for(int i = 0; i < N; i++) {
            heads[i] = new ArrayList();
            scores[i] = new ArrayList();
        }
            
        for(EPEDepNode c: edges.keySet()) {
            HashSet<EPEDepNode> seen = new HashSet();
            for(Pair<EPEDepNode, String> e: edges.get(c)) {
                EPEDepNode h = e.left;
                if(seen.contains(h))
                    continue;
                seen.add(e.left);
                heads[c.position-1].add(h.position);
                scores[c.position-1].add(-1.0 * Math.abs(c.position - h.position));
            }
        }
        int[][] heads2 = new int[N][];
        double[][] scores2 = new double[N][];
        for(int i = 0; i < N; i++) {
            heads2[i] = CollectionUtils.toIntArray(heads[i]);
            scores2[i] = CollectionUtils.toDoubleArray(scores[i]);
        }

        List<Pair<Double, int[]>> max = ChuLiuEdmonds.findNMaxSums(scores2, heads2, 1);
        int[] bestHeads = max.get(0).right;
        
        ListMap<EPEDepNode, Pair<EPEDepNode, String>> out = new ListMap();
        for(EPEDepNode c: edges.keySet()) {
            int selectedHead = bestHeads[c.position-1];
            for(Pair<EPEDepNode, String> e: edges.get(c)) {
                if(e.left.position != selectedHead)
                    continue;
                out.put(c, e);
                break;
            }
        }
        
        return out;
    }
    
    private static JSONArray fixStanfordTokens(JSONArray arr) {
        ArrayList<Integer> toRemove = new ArrayList();
        for(int i = 0; i < arr.size(); i++) {
            if(toRemove.contains(i))
                continue;
            JSONObject t = arr.getJSONObject(i);
            int id1 = t.getInt("id");
            
            for(int j = i + 1; j < arr.size(); j++) {
                JSONObject t2 = arr.getJSONObject(j);
                int id2 = t2.getInt("id");
                if(id1 == id2) {
                    mergeJSON(t, t2);
                    toRemove.add(j);
                }
            }
        }
        
        JSONArray out = new JSONArray();
        for(int i = 0; i < arr.size(); i++)
            if(!toRemove.contains(i))
                out.add(arr.get(i));
        return out;
    }

    private static void mergeJSON(JSONObject t1, JSONObject t2) {
        if(!t2.has("edges"))
            return;
        JSONArray e1 = t1.getJSONArray("edges");
        JSONArray e2 = t2.getJSONArray("edges");
        e1.addAll(e2);
    }
    
    private static Random rand = new Random(0);
    
    private static void link(JSONObject json, EPEDepNode root,
            HashMap<Object, EPEDepNode> m, 
            ListMap<EPEDepNode, Pair<EPEDepNode, String>> edges) {

        //if(json.toString().contains("ne_"))
        //    System.out.println(json);
        
        EPEDepNode n = m.get(json.get("id"));
        
        if(n == null)
            return;
        
        if(json.has("top"))
            edges.put(n, new Pair(root, "ROOT"));

        if(json.has("edges")) {
            JSONArray edgeArr = json.getJSONArray("edges");
            for(int i = 0; i < edgeArr.size(); i++) {
                JSONObject e = edgeArr.getJSONObject(i);
                String label = e.getString("label");                
                
                if(REMOVE_PB && label.matches("A[0-9]|AM-(.*?)|C-(.*?)|R-(.*?)|SU"))
                    continue;

                if(label.equals("edge_list_sum"))
                    continue;
                
                EPEDepNode t = m.get(e.get("target"));
                
                if(t == n || t == null) {
                    continue;
                }
                
                if(CORRUPT) { 
                    if(rand.nextDouble() < CORRUPT_LABEL_PROB)
                        label = "CORRUPTED";
                    if(rand.nextDouble() < CORRUPT_EDGE_PROB)
                        edges.put(t, new Pair(root, label));
                    else
                        edges.put(t, new Pair(n, label));
                } else
                    edges.put(t, new Pair(n, label));
            }
        }
    }

    private static EPEDepNode makeDepNode(JSONObject json, HashMap<Object, EPEDepNode> m, EPEFeatureSpec featureSpec) {        
        if(!json.has("properties"))
            return null;
        
        EPEDepNode n = new EPEDepNode();
        
        if(!json.has("start")) {
            n.start = 0;
            n.end = 0;
        } else {
            n.start = json.getInt("start");
            n.end = json.getInt("end");
        }
        if(json.has("form"))
            n.word = json.getString("form");
        else
            n.word = NOT_FOUND;
        
        Object id = json.get("id");
        m.put(id, n);
        
        JSONObject props = json.getJSONObject("properties");
        
        //System.out.println("json = " + json);
        //System.out.println("props = " + props + " / " + props.getClass().getName());
        
        if(featureSpec.cposLabel != null) {
            if(!props.has(featureSpec.cposLabel)) {
                //throw new RuntimeException("coarse pos id not found: [" + featureSpec.cposLabel + "] : " + props);
                System.err.println("coarse pos id not found: [" + featureSpec.cposLabel + "] : " + props);
                n.pos = NOT_FOUND;
            }
            else
                n.pos = props.getString(featureSpec.cposLabel);
        }
        if(featureSpec.fposLabel != null && props.has(featureSpec.fposLabel))
            n.posFine = props.getString(featureSpec.fposLabel);
        else
            n.posFine = NOT_FOUND;
        if(featureSpec.lemmaLabel.equals("form"))
            n.lemma = n.word.toLowerCase();
        else if(featureSpec.lemmaLabel != null && props.has(featureSpec.lemmaLabel)) {
            n.lemma = props.getString(featureSpec.lemmaLabel);
        } else
            n.lemma = NOT_FOUND;

        n.features = makeFeatureString(props, featureSpec);
        
        //System.out.println("end");

        /*
        System.out.println(featureSpec);
        System.out.println(json);
        System.out.println(n);
        System.exit(0);
        */
        
        return n;
    }
    
    private static String makeFeatureString(JSONObject props, EPEFeatureSpec s) {
        ArrayList<String> fs = new ArrayList();
        for(Object e: props.entrySet()) {
            Map.Entry<String, String> ee = (Map.Entry) e;
            String k = ee.getKey();
            if(!k.equals(s.cposLabel) && !k.equals(s.fposLabel) && !k.equals(s.lemmaLabel))
                fs.add(ee.getKey() + "=" + ee.getValue());
        }
        if(fs.isEmpty())
            return null;
        Collections.sort(fs);
        return Strings.join(fs, "|");
    }
    
    private static JSONObject readNextJSON(BufferedReader br) throws IOException {
        String l = br.readLine();
        if(l == null)
            return null;
        StringBuilder sb = new StringBuilder(l);        
        sb.append('\n');
        int opened = Strings.count(l, '{'), closed = Strings.count(l, '}');        
        while(opened == 0 || opened > closed) {
            l = br.readLine();
            if(l == null)
                break;
            sb.append(l);
            sb.append('\n');
            opened += Strings.count(l, '{');
            closed += Strings.count(l, '}');
        }
        if(opened == 0)
            return null;
        try {
            return JSONObject.fromObject(sb.toString());            
        } catch(Exception e) {
            System.err.println(sb);
            throw new RuntimeException(e);
        }
    }

    public static DepGraph read2008(BufferedReader brEPE, BufferedReader brCoNLL) throws IOException {

        Triple<DepGraph, DepGraph, List<PAStructure>> g = CoNLL2008Format.readNextGraph(brCoNLL);

        if(g == null)
            return null;
        
        DepGraph dg = g.first;
        
        DepGraph out = new DepGraph(dg.nodes.length-1);
        out.nodes[0] = new EPEDepNode();
        
        for(int i = 1; i < dg.nodes.length; i++) {
            String l = brEPE.readLine();
            String[] ss = l.split("\t");
            EPEDepNode n = new EPEDepNode();
            n.start = Integer.parseInt(ss[0]);
            n.end = Integer.parseInt(ss[1]);
            if(!dg.nodes[i].word.equals(ss[2]))
                throw new RuntimeException("!!");
            n.word = dg.nodes[i].word;
            n.pos = dg.nodes[i].pos;
            if(n.pos.length() > 2)
                n.pos = n.pos.substring(0, 2);
            n.posFine = dg.nodes[i].pos;
            n.lemma = dg.nodes[i].lemma;
            n.position = i;
            out.nodes[i] = n;
        }
        if(!brEPE.readLine().isEmpty())
            throw new RuntimeException("!");

        // dep -> (head, label)
        ListMap<EPEDepNode, Pair<EPEDepNode, String>> edges = new ListMap();

        if(true)
        for(int i = 1; i < dg.nodes.length; i++) {
            DepNode n = dg.nodes[i];
            edges.put((EPEDepNode) out.nodes[i], 
                      new Pair(out.nodes[n.parents[0].position], n.relations[0]));
        }
        
        if(true)
            for(PAStructure pa: g.third)
                for(int i = 0; i < pa.args.size(); i++) {
                    DepNode a = pa.args.get(i);
                    String l = pa.argLabels.get(i);
                    if(a != pa.pred)
                        edges.put((EPEDepNode) out.nodes[a.position], 
                                new Pair(out.nodes[pa.pred.position], pa.argLabels.get(i)));
                }
        
        for(int i = 0; i < out.nodes.length; i++) {
            EPEDepNode n = (EPEDepNode) out.nodes[i];
            List<Pair<EPEDepNode, String>> es_n = edges.get(n);
            if(es_n == null)
                continue;
            n.parents = new DepNode[es_n.size()];
            n.relations = new String[es_n.size()];
            for(int j = 0; j < es_n.size(); j++) {
                Pair<EPEDepNode, String> p = es_n.get(j);
                n.parents[j] = p.left;
                n.relations[j] = p.right;
            }
        }
        out.collectChildren(); // does this include others beside the first one? yes
        
        //System.out.println(out);

      //for(int i = 1; i < out.nodes.length; i++)
        //    System.out.println(out.nodes[i] + " -> " + edges.get((EPEDepNode) out.nodes[i]));

        //for(int i = 1; i < out.nodes.length; i++)
        // System.out.println(i + " -> " + Arrays.toString(out.nodes[i].children));
        
        //System.exit(0);

        return out;
    }
    
    public static String toJSON(DepGraph dg) {
        ArrayList nodes = new ArrayList();
        for(int i = 1; i < dg.nodes.length; i++) {
            EPEDepNode n = (EPEDepNode) dg.nodes[i];
            nodes.add(makeNodeJSON(n));
        }
        
        //System.out.println("*****");
        
        for(int i = 1; i < dg.nodes.length; i++) {
            EPEDepNode d = (EPEDepNode) dg.nodes[i];

            //System.out.println("d = " + d);
            
            for(int j = 0; j < d.parents.length; j++) {
                EPEDepNode h = (EPEDepNode) d.parents[j];
                if(h.position == 0)
                    continue;
                
                //System.out.println(" h = " + h);
                
                HashMap d_h = (HashMap) nodes.get(h.position-1);
                ArrayList edges = (ArrayList) d_h.get("edges");
                HashMap edge = new HashMap();                
                edge.put("label", d.relations[j]);
                edge.put("target", d.position);
                edges.add(edge);
            }
        }
        
        HashMap top = new HashMap();
        top.put("id", 0);
        top.put("nodes", nodes);
        JSONObject json = JSONObject.fromObject(top);        
        return json.toString();
    }
    
    public static HashMap makeNodeJSON(EPEDepNode n) {
        HashMap out = new HashMap();
        HashMap props = new HashMap();
        out.put("id", n.position);
        out.put("start", n.start);
        out.put("end", n.end);
        out.put("form", n.word);
        for(DepNode p: n.parents)
            if(p.position == 0)
                out.put("top", true);
        out.put("properties", props);
        props.put("upos", n.pos);
        props.put("xpos", n.posFine);
        props.put("lemma", n.lemma);
        if(n.children.length > 0) {
            ArrayList edges = new ArrayList();

            // seems we can't do this here...
            /*for(DepNode dep: n.children) {
                HashMap edge = new HashMap();
                edge.put("target", dep.position);
                int ix = CollectionUtils.indexOf(dep.parents, n);
                if(ix == -1)
                    throw new RuntimeException("?");
                edge.put("label", dep.relations[ix]);
                edges.add(edge);
            }*/
            
            out.put("edges", edges);
        }        
        return out;
    }
    
    public static void main0(String[] argv) {
        try {
            BufferedReader br = Util.openFileReader(argv[1], "UTF-8");

            EPEFeatureSpec fs = new EPEFeatureSpec("TODO");
            
            DepGraph dg = readDepGraph(br, fs);
            while(dg != null) {
                
                System.out.println(dg);

                // check single-parent property
                /*for(DepNode n: dg.nodes)
                    if(n.position > 0 && n.parents.length != 1) {
                        throw new RuntimeException("!");
                    }
                    */
                
                dg = readDepGraph(br, fs);
            }
            
        } catch(Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }
    
    public static void main1(String[] argv) {
        try {
            BufferedReader brEPE = Util.openFileReader(argv[1], "UTF-8");
            BufferedReader brCoNLL = Util.openFileReader(argv[2], "UTF-8");

            DepGraph dg = read2008(brEPE, brCoNLL);
            while(dg != null) {
                System.out.println(toJSON(dg));
                //System.exit(0);
                dg = read2008(brEPE, brCoNLL);
            }
            
        } catch(Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }
    
    public static void dump(String[] argv) {
        try {
            String txt = Util.readFile(argv[1]);
            int len = txt.length();
            for(int i = 0; i < len; i++) {
                char c = txt.charAt(i);
                if(Character.isWhitespace(c))
                    System.out.println(i + ": [" + (int) c + "]");
                else
                    System.out.println(i + ": " + c);
            }
            System.exit(1);
            
        } catch(Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }
    
    public static void testPath(String[] argv) {
        try {
            BufferedReader br = Util.openFileReader(argv[1], "UTF-8");

            EPEFeatureSpec fs = new EPEFeatureSpec("TODO");

            final int N = 50;
            final int K = 8;
            
            DepGraph dg = readDepGraph(br, fs);

            int ntotal = 0;
            long ttotal = 0;
            
            int n = 0;

            Random rand = new Random(0);

            double totalCost = 0;
            
            while(dg != null) {
                //System.out.println(dg);
                n += 1;
                //System.out.println(n);
                
                /*for(DepNode nn: dg.nodes) {
                    if(nn.parents.length > 1)
                        throw new RuntimeException("multiple: " + Arrays.toString(nn.relations));
                }*/
                
                long l0 = System.currentTimeMillis();

                ListMap<DepNode, Edge> graph = EPEPaths.makeBidirGraph(dg);
                
                for(int i = 0; i < N; i++) {
                    
                    //DepNode n1 = dg.nodes[rand.nextInt(dg.nodes.length)];
                    //DepNode n2 = dg.nodes[rand.nextInt(dg.nodes.length)];

                    DepNode n1 = dg.nodes[50];
                    DepNode n2 = dg.nodes[50];
                    
                    DepPath[] p = EPEPaths.findPaths(graph, n1, n2, K, Double.POSITIVE_INFINITY);

                    if(p[0] == null)
                        throw new RuntimeException();
                    
                    totalCost += p[0].cost;
                    
                    System.out.println(dg);
                    System.out.println(n1);
                    System.out.println(n2);                    

                    for(DepPath pp: p)
                        if(pp == null)
                            break;
                        else
                            System.out.println(pp + " / " + pp.cost + " / " + pp.n);
                    
                    //System.out.println(p[0].cost);*/
                    
                    //Pair<Path, Path> pp = EPEPaths.treePath(n1, n2);
                    //System.out.println(pp.left.cost + ", " + pp.right.cost);
                    //totalCost += pp.left.cost + pp.right.cost;
                    
                    DepPath[] p2 = EPEPaths.treePath(n1, n2, 256);

                    System.out.println(p2[0] + " / " + p2[0].cost + " / " + p2[0].n);
                    
                    System.exit(0);
                    
                }
                ttotal += System.currentTimeMillis() - l0;
                ntotal += N;
                dg = readDepGraph(br, fs);
            }
            
            System.out.println(ntotal);
            System.out.println(1000 * ((double)ttotal)/ntotal);
            System.out.println(totalCost);
            
        } catch(Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }
    
    public static void printJSON(String[] argv) {
        try {
            BufferedReader br = Util.openFileReader(argv[1], "UTF-8");
            JSONObject json = readNextJSON(br);
            while(json != null) {
                System.out.println(json.toString(4));
                json = readNextJSON(br);
            }
        } catch(Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }
    
    public static void main(String[] argv) {
        if(argv[0].equals("-merge"))
            main1(argv);
        else if(argv[0].equals("-test"))
            main0(argv);
        else if(argv[0].equals("-dump"))
            dump(argv);
        else if(argv[0].equals("-printJSON"))
            printJSON(argv);
        else if(argv[0].equals("-testPath"))
            testPath(argv);
        else
            throw new RuntimeException("?");
    }
    
}
