package mpqa_structlearn.io;

import lkformat.reader.*;
import mpqareader.SubjectivityLexicon;

import java.io.*;
import java.util.*;

import se.lth.cs.nlp.depsrl.format.*;
import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.depgraph.*;

public class LKReader implements Iterator<MPQAText> {

	private SubjectivityLexicon subjLex;
	private LKCollectionReader lkreader;
	
	public LKReader(String dir, String subjLexFile) throws IOException {
		subjLex = new SubjectivityLexicon(subjLexFile);
		lkreader = new LKCollectionReader(dir);
	}
	
	public LKReader(String dir, String subjLexFile, 
			List<String> selectedFiles) throws IOException {
		subjLex = new SubjectivityLexicon(subjLexFile);
		lkreader = new LKCollectionReader(dir, selectedFiles);
	}

	private void createPAs(LKAnnotationLayer predLayer,
			LKAnnotationLayer argLayer,
			HashMap<DepNode, MPQASentence> nodeToSen,
			HashMap<LKAnnotationEntity, DepNode> tokenToNode) {
		
		HashMap<LKAnnotationEntity, PAStructure> tokenToPA = new HashMap();

		for(LKAnnotationEntity pe: predLayer.entityList) {
			if(pe.referred == null || pe.referred.size() != 1)
				throw new RuntimeException("!!!");
			LKAnnotationEntity token = pe.referred.get(0);
			PAStructure pa = new PAStructure();
			pa.pred = tokenToNode.get(token);
			pa.lemma = pe.data.getText();
			pa.predPos = pa.pred.pos; // for simplicity
			MPQASentence sen = nodeToSen.get(pa.pred);
			sen.pas.add(pa);
			tokenToPA.put(token, pa);
		}

		for(LKAnnotationEntity ae: argLayer.entityList) {
			PAStructure pa = tokenToPA.get(ae.from);
			pa.args.add(tokenToNode.get(ae.to));
			pa.argLabels.add(ae.data.getText());
		}
	}
	
	// ([MPQAsentence], { LK-token-entity -> depnode })
	// this map is used because other annotations are defined on top of tokens?
	private static Pair<ArrayList<MPQASentence>, 
	                      HashMap<LKAnnotationEntity, DepNode>> createSentences(LKAnnotationLayer senLayer, 
			LKAnnotationLayer dgLayer,
			HashMap<LKAnnotationEntity, LKAnnotationEntity> posMap,
			HashMap<LKAnnotationEntity, LKAnnotationEntity> lemmaMap) {

		ArrayList<MPQASentence> out = new ArrayList();
		
		HashMap<LKAnnotationEntity, LKAnnotationEntity> tokenToLink = new HashMap();
		
		if(dgLayer != null) {
			for(LKAnnotationEntity le: dgLayer.entityList)
				if(le.to != null)
					tokenToLink.put(le.to, le);
				else if(le.referred.size() == 1)
					tokenToLink.put(le.referred.get(0), le);
				else
					throw new RuntimeException("No link!");
		}

		HashMap<LKAnnotationEntity, MPQASentence> lsToMs = new HashMap();
		HashMap<LKAnnotationEntity, DepNode> tokenToNode = new HashMap();
		
		for(LKAnnotationEntity se: senLayer.entityList) {

			ArrayList<LKAnnotationEntity> tokens = new ArrayList();
			ArrayList<LKAnnotationEntity> poss = new ArrayList();
			ArrayList<LKAnnotationEntity> lemmas = new ArrayList();
			
			for(LKAnnotationEntity te: se.referred) {
				LKAnnotationEntity pose = posMap.get(te);
				if(pose == null)
					continue;
				LKAnnotationEntity leme = lemmaMap.get(te);
				tokens.add(te);
				poss.add(pose);
				lemmas.add(leme);
			}			
			DepGraph dg = new DepGraph(tokens.size());
			dg.nodes[0] = new DepNode();
			for(int i = 0; i < tokens.size(); i++) {
				dg.nodes[i+1] = new DepNode();
				dg.nodes[i+1].position = i+1;
				tokenToNode.put(tokens.get(i), dg.nodes[i+1]);
			}			
			for(int i = 0; i < tokens.size(); i++) {
				LKAnnotationEntity te = tokens.get(i);
				DepNode n = tokenToNode.get(te);
				n.word = te.data.getText();
				n.pos = poss.get(i).data.getText();
								
				LKAnnotationEntity leme = lemmas.get(i);
				if(leme != null)
					n.lemma = leme.data.getText();

				LKAnnotationEntity le = null;
				if(dgLayer != null)
					le = tokenToLink.get(te);
				
				if(dgLayer == null) {
					dg.nodes[0].addChild(n, "ROOT");
				} else if(le.to != null) {
					DepNode p = tokenToNode.get(le.from);
					if(p == null)
						throw new RuntimeException("parent not found!");
					p.addChild(n, le.data.getText());
				} else if(le.referred.size() == 1)
					dg.nodes[0].addChild(n, le.data.getText());
				else
					throw new RuntimeException("No link!");
			}
			dg.collectChildren();
			MPQASentence mse = new MPQASentence();
			mse.dg = dg;
			out.add(mse);
			lsToMs.put(se, mse);
		}
		
		return new Pair(out, tokenToNode);
	}

	private void createExpressions(LKAnnotationLayer exprLayer,
			HashMap<DepNode, MPQASentence> nodeToSen,
			HashMap<String, MPQAAgent> agentMap,
			HashMap<LKAnnotationEntity, DepNode> tokenToNode) {

		for(LKAnnotationEntity ee: exprLayer.entityList) {
			MPQAExpression ex = new MPQAExpression();
			
			DataElementNode n1 = (DataElementNode) ee.data;
			DataNode n2 = n1.children.get(0);
			DataElementNode dn2 = (DataElementNode) n2;			
			ex.type = dn2.name;
			
			String ns = ee.data.getTopAttribute("ns");
			if(ns != null)
				ns = ns.replaceAll("\\s", "");

			ex.polarity = ee.data.getTopAttribute("pol");
			ex.intensity = ee.data.getTopAttribute("int");
			
			if(ns != null && !ns.isEmpty()) {
				ex.nestedSource = new ArrayList();
				for(String src: ns.split(","))
					ex.nestedSource.add(agentMap.get(src));	
			}
			if(ee.referred != null && !ee.referred.isEmpty()) {
				ex.span = new ArrayList();
				MPQASentence sen = null;
				for(LKAnnotationEntity t: ee.referred) {
					DepNode n = tokenToNode.get(t);
					if(n == null)
						throw new RuntimeException("node not found");
					if(sen == null)
						sen = nodeToSen.get(n);
					ex.span.add(n);
				}
				ex.containingSentence = sen;
				sen.exprs.add(ex);
				shrinkToSentence(ex.span, sen); // because some bugs...
			}
		}				
	}
	
	private void shrinkToSentence(ArrayList<DepNode> nodes, MPQASentence sen) {
		nodes.retainAll(Arrays.asList(sen.dg.nodes));
	}
	
	private Pair<ArrayList<MPQAAgent>, HashMap<String, MPQAAgent>> createAgents(LKAnnotationLayer agentLayer,
			HashMap<DepNode, MPQASentence> nodeToSen,
			HashMap<LKAnnotationEntity, DepNode> tokenToNode) {
		ArrayList<MPQAAgent> out = new ArrayList();
		HashMap<String, MPQAAgent> agentMap = new HashMap();
		
		for(LKAnnotationEntity ae: agentLayer.entityList) {
			String mid = ae.data.getTopAttribute("mid");
			if(mid == null)
				continue;
			mid = mid.trim();
			if(mid.isEmpty())
				continue;
			
			String id = ae.localURI;
			if(id == null)
				throw new RuntimeException("id == null");

			MPQAAgent ag = new MPQAAgent();
			ag.mpqaID = mid;
			
			if(ae.data.getTopAttribute("imp") != null && ae.data.getTopAttribute("imp").equals("true"))
				ag.isImplicit = true;

			if(ae.data.getTopAttribute("w") != null && ae.data.getTopAttribute("w").equals("true"))
				ag.isWriter = true;

			agentMap.put("#" + id, ag);
			out.add(ag);
		}
		
		for(LKAnnotationEntity ae: agentLayer.entityList) {
			MPQAAgentInstance ai = new MPQAAgentInstance();
			String ns = ae.data.getTopAttribute("ns");
			if(ns != null)
				ns = ns.replaceAll("\\s", "");
			MPQAAgent ag;
			if(ns == null || ns.isEmpty()) {
				ag = new MPQAAgent();
				out.add(ag);
			} else {
				ai.nestedSource = new ArrayList();
				for(String src: ns.split(",")) {
					MPQAAgent srcAgent = agentMap.get(src);
					if(srcAgent == null) {
						srcAgent = new MPQAAgent();
						out.add(srcAgent);
						agentMap.put(src, srcAgent);
					}
					ai.nestedSource.add(srcAgent);
				}
				ag = ai.nestedSource.get(ai.nestedSource.size() - 1);
			}
			ai.mpqaNbr = ae.data.getTopAttribute("mnbr");
			/*if(ai.mpqaNbr == null)
				throw new RuntimeException("!!!");*/
			ai.agent = ag;
			ag.instances.add(ai);
			if(ae.referred != null && !ae.referred.isEmpty()) {
				ai.span = new ArrayList();
				MPQASentence sen = null;
				for(LKAnnotationEntity t: ae.referred) {
					DepNode n = tokenToNode.get(t);
					if(n == null)
						throw new RuntimeException("node not found");
					if(sen == null)
						sen = nodeToSen.get(n);
					ai.span.add(n);			
				}
				ai.containingSentence = sen;
				shrinkToSentence(ai.span, sen); // because some bugs...
			}
		}				
		return new Pair(out, agentMap);
	}
	
	public MPQAText next() {
		if(!hasNext())
			return null;
		
		LKAnnotatedText text = lkreader.next();
		
		LKAnnotationLayer senLayer = text.getLayer("SENTENCES");
		LKAnnotationLayer posLayer = text.getLayer("POS");
		LKAnnotationLayer lemmaLayer = text.getLayer("LEMMA");

		LKAnnotationLayer dgLayer = text.getLayer("DEP-SYNTAX");
		
		HashMap<LKAnnotationEntity, LKAnnotationEntity> posMap = createTokenMap(posLayer);
		HashMap<LKAnnotationEntity, LKAnnotationEntity> lemmaMap = createTokenMap(lemmaLayer);
		
		Pair<ArrayList<MPQASentence>, 
        HashMap<LKAnnotationEntity, DepNode>> p1 = createSentences(senLayer,
        		dgLayer, posMap, lemmaMap);

		HashMap<DepNode, MPQASentence> nodeToSen = new HashMap();
		for(MPQASentence sen: p1.left)
			for(DepNode n: sen.dg.nodes)
				nodeToSen.put(n, sen);
		
		LKAnnotationLayer agentLayer = text.getLayer("MPQA-agents");		
		Pair<ArrayList<MPQAAgent>, HashMap<String, MPQAAgent>> p = createAgents(agentLayer, nodeToSen, p1.right);

		LKAnnotationLayer predLayer = text.getLayer("SRL-PB-preds");
		LKAnnotationLayer argLayer = text.getLayer("SRL-PB-args");
		if(predLayer == null)
			if(argLayer != null)
				throw new RuntimeException("found arglayer but not predlayer");

		if(predLayer != null)
			createPAs(predLayer, argLayer, nodeToSen, p1.right);
		
		LKAnnotationLayer exprSubjLayer = text.getLayer("MPQA-expressive-subjectivity");
		LKAnnotationLayer objSpeechLayer = text.getLayer("MPQA-objective-speech-event");
		LKAnnotationLayer dirSubjLayer = text.getLayer("MPQA-direct-subjective");

		createExpressions(exprSubjLayer, nodeToSen, p.right, p1.right);
		createExpressions(dirSubjLayer, nodeToSen, p.right, p1.right);
		createExpressions(objSpeechLayer, nodeToSen, p.right, p1.right);
		
		for(MPQASentence sen: p1.left)
			createSubjClues(sen);
		
		MPQAText out = new MPQAText();
		out.source = text.metaInfo.get("source");
		out.agents = p.left;
		out.sentences = p1.left;
		return out;
	}
	
	private HashMap<LKAnnotationEntity, LKAnnotationEntity> createTokenMap(LKAnnotationLayer otherLayer) {
		HashMap<LKAnnotationEntity, LKAnnotationEntity> out = new HashMap();
		for(LKAnnotationEntity e: otherLayer.entityList) {
			if(e.referred == null || e.referred.size() != 1)
				throw new RuntimeException("!!!");
			out.put(e.referred.get(0), e);
		}
		return out;
	}	
	
	private void createSubjClues(MPQASentence sen) {
		sen.slClues = new ArrayList();
		sen.slClues.add(null);
		for(int i = 1; i < sen.dg.nodes.length; i++) {
			DepNode n = sen.dg.nodes[i];
			sen.slClues.add(subjLex.lookup(n.word, n.pos, n.lemma));
		}
	}

	public boolean hasNext() {
		return lkreader.hasNext();
	}

	public static void main(String[] argv) {
		
	}
	
}
