package mpqa_structlearn.io;

import java.io.*;
import java.util.*;

import mpqareader.MPQAReader;
import mpqareader.SubjectivityLexicon;
import se.lth.cs.nlp.nlputils.annotations.*;
import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.depgraph.*;

public class MPQAReaderEPE implements Iterator<MPQAText> {

    private SubjectivityLexicon subjLex;
    private Iterator<AnnotatedText> iter;
    //private Collection<String> selectedFiles;
    private String epeDir;
    private EPEFeatureSpec featureSpec;
    
    public MPQAReaderEPE(String dir, String subjLexFile, 
            Collection<String> selectedFiles,
            String epeDir,
            EPEFeatureSpec featureSpec) throws IOException {
        //this.selectedFiles = selectedFiles;
        this.epeDir = epeDir;
        subjLex = new SubjectivityLexicon(subjLexFile, false);
        iter = MPQAReader.processDirectory(dir, selectedFiles);
        this.featureSpec = featureSpec;
    }
    
    private MPQAText next;
    
    public boolean hasNext() {
        if(next != null)
            return true;
        while(iter.hasNext()) {
            next = getNext();
            if(!next.isCorrupt)
                return true;
        }
        return false;
    }
    
    public MPQAText next() {
        if(!hasNext())
            throw new NoSuchElementException("no more elements");
        MPQAText out = next;
        next = null;
        return out;
    }
    
    private ArrayList<DepGraph> readEPE(String epeFileName) {
        try {
            if(!new File(epeFileName).exists())
                return null;
            ArrayList<DepGraph> out = new ArrayList();
            BufferedReader br = Util.openFileReader(epeFileName, "UTF-8");
            DepGraph dg = EPEReader.readDepGraph(br, featureSpec);
            while(dg != null) {
                out.add(dg);
                dg = EPEReader.readDepGraph(br, featureSpec);
            }            
            return out;
        } catch(Exception e) {
            throw new RuntimeException(e);
        }
    }
    
    private Pair<ArrayList<MPQASentence>, HashMap<DepNode, MPQASentence>> createSentences(ArrayList<DepGraph> dgs) {
        HashMap<DepNode, MPQASentence> nodeToSen = new HashMap();
        ArrayList<MPQASentence> out = new ArrayList();
        for(DepGraph dg: dgs) {
            MPQASentence sen = new MPQASentence();
            sen.dg = dg;
            createSubjClues(sen);
            out.add(sen);
            for(DepNode n: dg.nodes)
                nodeToSen.put(n, sen);
        }
        return new Pair(out, nodeToSen);
    }
    
    private void createSubjClues(MPQASentence sen) {
        sen.slClues = new ArrayList();
        sen.slClues.add(null);
        for(int i = 1; i < sen.dg.nodes.length; i++) {
            DepNode n = sen.dg.nodes[i];
            sen.slClues.add(subjLex.lookup(n.word, n.pos, n.lemma));
        }
    }

    private ArrayList<DepNode> createSpan(Span expr, EPEDepNode[] starts, EPEDepNode[] ends,
            HashMap<DepNode, MPQASentence> nodeToSen, String text) {

        int startIndex = expr.start;
        if(starts[startIndex] == null)
            if(Character.isWhitespace(text.charAt(startIndex)))
                while(starts[startIndex] == null && startIndex < expr.end)
                    startIndex++;
            else
                while(starts[startIndex] == null && startIndex > 0)
                    startIndex--;

        int endIndex = expr.end;
        
        if(endIndex >= ends.length) {
            System.err.println("*** WARNING: endIndex >= ends.length");
            endIndex = ends.length-1;
        }
        
        if(ends[endIndex] == null)
            if(Character.isWhitespace(text.charAt(endIndex-1)))
                while(ends[endIndex] == null && endIndex > startIndex)
                    endIndex--;
            else
                while(ends[endIndex] == null && endIndex < ends.length - 1)
                    endIndex++;
        
        if(startIndex == endIndex)
            throw new RuntimeException("?");
        if(starts[startIndex] == null) {
            //throw new RuntimeException("??: start[startIndex] = null; " + expr + " startIndex = " + startIndex);
            System.err.println("*** WARNING: start[startIndex] = null; " + expr + " startIndex = " + startIndex);
            while(starts[startIndex] == null)
                startIndex++;
        }
        if(ends[endIndex] == null)
            throw new RuntimeException("???");

        MPQASentence s1 = nodeToSen.get(starts[startIndex]);
        MPQASentence s2 = nodeToSen.get(ends[endIndex]);

        if(s1 != s2) {
            int n1 = s1.dg.nodes.length - starts[startIndex].position;
            int n2 = ends[endIndex].position;
                    
            if(n1 >= n2)
                endIndex = ((EPEDepNode) s1.dg.nodes[s1.dg.nodes.length-1]).end;
            else {
                startIndex = ((EPEDepNode) s2.dg.nodes[1]).start;
                s1 = s2;
            }
        }
        
        if(starts[startIndex] == null) {
            //System.err.println(startIndex + " / " + endIndex);
            //throw new RuntimeException("starts[startIndex] == null");
            System.err.println("starts[startIndex] == null");
            return null;
        }
        if(ends[endIndex] == null) {
            //System.out.println(expr);
            //System.out.println(startIndex + " / " + endIndex);            
            ///throw new RuntimeException("ends[endIndex] == null");            
            System.err.println("ends[endIndex] == null");
            return null;
        }
        
        ArrayList<DepNode> out = new ArrayList();
        DepGraph dg = s1.dg;
        for(int i = starts[startIndex].position; i <= ends[endIndex].position; i++)
            out.add(dg.nodes[i]);
        
        return out;
    }
    
    private String createLabel(String label) {
        if(label.equals("GATE_expressive-subjectivity"))
            return "es";
        if(label.equals("GATE_direct-subjective"))
            return "ds";
        if(label.equals("GATE_objective-speech-event"))
            return "os";
        throw new RuntimeException("unknown label: " + label);
    }
    
    private String convertPolarity(String label) {
        if(label == null)
            return null;
        label = label.trim();
        if(label.matches("(uncertain-)?positive"))
            return "positive";
        if(label.matches("(uncertain-)?negative"))
            return "negative";
        if(label.matches("(uncertain-)?neutral"))
            return "neutral";
        if(label.matches("(uncertain-)?both"))
            return "both";
        
        if(label.isEmpty()) {
            //System.err.println("Warning: polarity = |" + label + "|");
            return "neutral";
        }
        
        throw new RuntimeException("unknown label: " + label);
    }
    
    private String convertIntensity(String label) {
        if(label == null)
            return null;
        label = label.trim();
        if(label.equals("extreme"))
            return "extreme";
        if(label.equals("high"))
            return "high";
        if(label.matches("medium|neutral") || label.isEmpty())
            return "medium";
        if(label.equals("low"))
            return "low";
        throw new RuntimeException("unknown label: " + label);
    }
    
    private void createExpressions(HashMap<DepNode, MPQASentence> nodeToSen,
            EPEDepNode[] starts, EPEDepNode[] ends,
            AnnotationLayer l, String text, HashMap<String, MPQAAgent> agentMap) {
        
        for(Object e: l.spans) {
            Span expr = (Span) e;

            if(expr.start == expr.end || expr.start >= ends.length)
                continue;
            if(expr.hasProperty("implicit"))
                continue;
            if(text.substring(expr.start, expr.end).trim().isEmpty())
                continue;

            MPQAExpression mexpr = new MPQAExpression();
            mexpr.type = createLabel(expr.label);            
            
            mexpr.span = createSpan(expr, starts, ends, nodeToSen, text);
            if(mexpr.span != null) {
                mexpr.start = expr.start;
                mexpr.end = expr.end;
                mexpr.containingSentence = nodeToSen.get(mexpr.span.get(0));            
                mexpr.containingSentence.exprs.add(mexpr);                
            }
            
            String ns = (String) expr.getProperty("nested-source");
            if(ns != null)
                ns = ns.replaceAll("\\s", "");

            mexpr.polarity = convertPolarity((String) expr.getProperty("polarity"));
            mexpr.intensity = convertIntensity((String) expr.getProperty("intensity"));
            
            if(ns != null && !ns.isEmpty()) {
                mexpr.nestedSource = new ArrayList();
                for(String src: ns.split(",")) // what if reference doesn't exist?
                    mexpr.nestedSource.add(agentMap.get(src)); 
            }
            
        }
    }
    
    private Pair<ArrayList<MPQAAgent>, HashMap<String, MPQAAgent>> createAgents(HashMap<DepNode, MPQASentence> nodeToSen,
            EPEDepNode[] starts, EPEDepNode[] ends, 
            AnnotationLayer l, String text) {

        ArrayList<MPQAAgent> out = new ArrayList();
        HashMap<String, MPQAAgent> agentMap = new HashMap();

        for(Object s: l.spans) {
            Span as = (Span) s;
            
            String id = (String) as.getProperty("id");
            if(id == null)
                continue;
            
            id = id.trim();
            if(id.isEmpty())
                continue;

            MPQAAgent ag = new MPQAAgent();
            ag.mpqaID = id;
            
            if(id.equals("w"))
                ag.isWriter = true;
            else if(id.equals("implicit"))
                ag.isImplicit = true;
            
            agentMap.put(id, ag);
            out.add(ag);
        }
        
        //System.out.println("out = " + out);
        
        for(Object s: l.spans) {
            MPQAAgentInstance ai = new MPQAAgentInstance();
            Span as = (Span) s;
            
            String ns = (String) as.getProperty("nested-source");
            if(ns != null)
                ns = ns.replaceAll("\\s", "");
            
            //System.out.println(as + " / " + ns);
            
            MPQAAgent ag;
            if(ns == null || ns.isEmpty()) {
                /* Why would this be needed? */
                /*ag = new MPQAAgent();
                out.add(ag);*/
                String id = (String) as.getProperty("id");
                if(id == null)
                    continue;
                id = id.trim();
                if(id.isEmpty())
                    continue;
                ag = agentMap.get(id);
            } else {
                ai.nestedSource = new ArrayList();
                for(String src: ns.split(",")) {
                    MPQAAgent srcAgent = agentMap.get(src);
                    if(srcAgent == null) {
                        srcAgent = new MPQAAgent();
                        out.add(srcAgent);
                        agentMap.put(src, srcAgent);
                    }
                    ai.nestedSource.add(srcAgent);
                }
                ag = ai.nestedSource.get(ai.nestedSource.size() - 1);
            }
            ai.mpqaNbr = as.id;
            ai.agent = ag;
            ag.instances.add(ai);
            
            if(as.start == as.end || as.start >= ends.length || text.substring(as.start, as.end).trim().isEmpty()) {
                // empty
            } else {                
                ai.span = createSpan(as, starts, ends, nodeToSen, text);
                if(ai.span != null) {
                    ai.start = as.start;
                    ai.end = as.end;
                    ai.containingSentence = nodeToSen.get(ai.span.get(0));                    
                }
            }

        }
        
        /*System.out.println("out = " + out);
        System.out.println(agentMap);*/
        
        return new Pair(out, agentMap);
    }
    
    private Pair<EPEDepNode[], EPEDepNode[]> makeCharMaps(ArrayList<DepGraph> dgs) {
        //DepGraph lastSen = dgs.get(dgs.size()-1);

        /*EPEDepNode lastNode = (EPEDepNode) lastSen.nodes[lastSen.nodes.length-1];
        EPEDepNode[] starts = new EPEDepNode[lastNode.end+1];
        EPEDepNode[] ends = new EPEDepNode[lastNode.end+1];*/

        int maxEnd = 0;
        for(DepGraph dg: dgs)
            for(DepNode n: dg.nodes) {
                EPEDepNode en = (EPEDepNode) n;
                if(en.end > maxEnd)
                    maxEnd = en.end;
            }
        EPEDepNode[] starts = new EPEDepNode[maxEnd+1];
        EPEDepNode[] ends = new EPEDepNode[maxEnd+1];
        
        for(DepGraph dg: dgs) {
            for(DepNode n: dg.nodes) {
                if(n.position == 0)
                    continue;
                
                EPEDepNode en = (EPEDepNode) n;

                // hmm...
                if(en.start == en.end)
                    continue;
                
                if(en.start >= starts.length) {
                    System.err.println(dg);
                    System.err.println("---");
                    System.err.println(en);
                    System.err.println("---");
                    System.err.println(starts.length);
                    System.err.println("---");
                    //System.err.println(lastNode);
                    throw new RuntimeException("start outside text");
                }
                
                if(starts[en.start] != null) {
                    System.err.println(dg);
                    throw new RuntimeException("two tokens start at same index");
                }
                starts[en.start] = en;
                if(ends[en.end] != null)
                    throw new RuntimeException("two tokens end at same index: " + ends[en.end] + " / " + en);
                ends[en.end] = en;
            }
        }
        return new Pair(starts, ends);
    }
    
    private MPQAText getNext() {
        AnnotatedText text = iter.next();
        
        MPQAText out = new MPQAText();
        
        out.source = (String) text.getProperty("mpqa_file");
        out.original = text;

        /*if(!selectedFiles.contains(out.source)) {
            out.isCorrupt = true;
            return out;
        }*/

        /*for(AnnotationLayer l: text.layers)
            for(Object expr: l.spans)
                System.out.println(expr);*/
        
        String bn = out.source.replaceFirst(".*/", "");
        String epeFileName = epeDir + "/" + bn + ".epe"; 
        ArrayList<DepGraph> dgs = readEPE(epeFileName);

        //if(true)
        //    return out;
        
        if(dgs == null) {
            out.isCorrupt = true;
            return out;
        }

        if(false)
            checkText(text.text, dgs);
        
        Pair<EPEDepNode[], EPEDepNode[]> ms = null;
        try {
            ms = makeCharMaps(dgs);
        } catch(Exception e) {
            System.err.println(out.source);
            e.printStackTrace();
            System.exit(1);
        }
        
        Pair<ArrayList<MPQASentence>, HashMap<DepNode, MPQASentence>> p = createSentences(dgs);
        out.sentences = p.left;

        text.splitLayers();

        Pair<ArrayList<MPQAAgent>, HashMap<String, MPQAAgent>> ap = null;
        try {
            for(AnnotationLayer l: text.layers)
                if(l.label.equals("GATE_agent"))
                    ap = createAgents(p.right, ms.left, ms.right, l, text.text);
        } catch(Exception ee) {
            System.out.println(out.source);

            for(int i = 0; i < Math.min(20, text.text.length()); i++)
                System.out.println(i + ": " + text.text.charAt(i));
            
            for(DepGraph dg: dgs)
                System.out.println(dg);

            for(int i = 0; i < ms.left.length; i++) {
                System.out.println(i + ": " + ms.left[i]);
            }
            
            for(AnnotationLayer<Span> l: text.layers) {
                for(Span s: l) {
                    if(s.label.equals("GATE_agent"))
                        System.out.println(s);  
                }
            }
            
            ee.printStackTrace();
            System.exit(1);
        }
        out.agents = ap.left;
                
        for(AnnotationLayer l: text.layers)
            if(l.label.matches("GATE_direct-subjective|GATE_expressive-subjectivity|GATE_objective-speech-event")) {
                createExpressions(p.right, ms.left, ms.right, l, text.text, ap.right);
            }
        
        /*
        for(MPQASentence sen: out.sentences) {
            for(MPQAExpression expr: sen.exprs) {
                if(!expr.span.isEmpty()) {
                    //System.out.println(expr);
                    //System.out.println(expr.start + "," + expr.end);
                    //System.exit(0);
                    if(expr.start == Integer.MIN_VALUE || expr.end == Integer.MIN_VALUE)
                        throw new RuntimeException("!");
                }
            }            
        }
        for(MPQAAgent ag: out.agents)
            for(MPQAAgentInstance inst: ag.instances) {
                if(inst.span != null && !inst.span.isEmpty()) {
                    //System.out.println(expr);
                    //System.out.println(expr.start + "," + expr.end);
                    //System.exit(0);
                    if(inst.start == Integer.MIN_VALUE || inst.end == Integer.MIN_VALUE)
                        throw new RuntimeException("!");
                }
            }
        */
        
        return out;
    }

    private void checkText(String text, ArrayList<DepGraph> dgs) {
        for(DepGraph dg: dgs) {
            //System.out.println(dg);
            for(DepNode n: dg.nodes) {
                if(n.position == 0)
                    continue;
                //if(n.parents.length != 1)
                //    throw new RuntimeException("!!");
                EPEDepNode en = (EPEDepNode) n;
 
                String w = text.substring(en.start, en.end);
                if(!Util.isPunc(w) && !en.word.equals(text.substring(en.start, en.end))) {
                    System.out.println(en);
                    System.out.println("|" + text.substring(en.start, en.end) + "|");
                    throw new RuntimeException("!");
                }
            }
        }
        //System.exit(0);
    }


}
