package mpqa_structlearn.io;

import java.util.*;

import se.lth.cs.nlp.depsrl.format.*;

import se.lth.cs.nlp.nlputils.annotations.Span;
import se.lth.cs.nlp.nlputils.core.PRFStats;
import se.lth.cs.nlp.nlputils.depgraph.*;
import se.lth.cs.nlp.nlputils.pstree.*;

public class MPQASentence implements java.io.Serializable {

	private static final long serialVersionUID = 0L;
	
	public DepGraph dg;
	public PhraseStructureTree pstree;
	public ArrayList<Span> discConns;
	
	public ArrayList<String> slClues;
	public ArrayList<PAStructure> pas = new ArrayList();
	public ArrayList<MPQAExpression> exprs = new ArrayList();

	private static final double ONE = 1.0 - 1e-6;
	
	public PRFStats compareExpressions(MPQASentence sen) {
		PRFStats out = new PRFStats();

		out.nInGold += exprs.size();
		out.nGuesses += sen.exprs.size();
		
		/* I think this only works when expressions in one layer don't overlap. */
		for(MPQAExpression e1: this.exprs) {
			boolean foundOverlap = false;
			boolean foundExact = false;
			for(MPQAExpression e2: sen.exprs) {
				if(!e1.type.equals(e2.type))
					continue;

				double pfc = e1.getSpanCoverage(e2);
				double pgc = e2.getSpanCoverage(e1);
				
				out.propFound += pfc;
				out.propGuessed += pgc;
				
				if(pfc > 0)
					foundOverlap = true;
				if(pfc > ONE && pgc > ONE)
					foundExact = true;				
			}
			if(foundOverlap)
				out.nOverlap++;
			if(foundExact)
				out.nCorrect++;
		}
				
		return out;
	}
	
	public PRFStats compareExpressionsClustered(MPQASentence sen, HashMap<String, String> clusters) {
		PRFStats out = new PRFStats();

		for(MPQAExpression e: exprs)
			if(clusters.containsKey(e.type))
				out.nInGold++;
		for(MPQAExpression e: sen.exprs)
			if(clusters.containsKey(e.type))
				out.nInGold++;

		for(MPQAExpression e1: this.exprs) {
			String cl1 = clusters.get(e1.type);  
			if(cl1 == null)
				continue;
			boolean foundOverlap = false;
			boolean foundExact = false;
			for(MPQAExpression e2: sen.exprs) {
				if(clusters.get(e2.type) != cl1)
					continue;

				double pfc = e1.getSpanCoverage(e2);
				double pgc = e2.getSpanCoverage(e1);
				
				out.propFound += pfc;
				out.propGuessed += pgc;
				
				if(pfc > 0)
					foundOverlap = true;
				if(pfc > ONE && pgc > ONE)
					foundExact = true;				
			}
			if(foundOverlap)
				out.nOverlap++;
			if(foundExact)
				out.nCorrect++;
		}
				
		return out;
	}
	
	public PRFStats compareExpressionsUnique(MPQASentence sen) {
		PRFStats out = new PRFStats();

		out.nInGold += exprs.size();
		out.nGuesses += sen.exprs.size();
		
		HashSet<MPQAExpression> done2 = new HashSet();
		
		for(MPQAExpression e1: this.exprs) {
			boolean foundOverlap = false;
			boolean foundExact = false;

			MPQAExpression max = null;
			double maxScore = Double.NEGATIVE_INFINITY;
			
			for(MPQAExpression e2: sen.exprs) {
				if(done2.contains(e2))
					continue;
				if(!e1.type.equals(e2.type))
					continue;

				double pfc = e1.getSpanCoverage(e2);
				double pgc = e2.getSpanCoverage(e1);
				
				if(pfc + pgc > maxScore) {
					max = e2;
					maxScore = pfc + pgc;
				}
				
				if(pfc > 0)
					foundOverlap = true;
				if(pfc > ONE && pgc > ONE)
					foundExact = true;				
			}
			if(foundOverlap)
				out.nOverlap++;
			if(foundExact)
				out.nCorrect++;
			if(max != null) {
				double pfc = e1.getSpanCoverage(max);
				double pgc = max.getSpanCoverage(e1);
				out.propFound += pfc;
				out.propGuessed += pgc;
				done2.add(max);
			}
			
		}
				
		return out;
	}
	
	public PRFStats compareExpressionsSimple(MPQASentence sen) {
		PRFStats out = new PRFStats();

		for(MPQAExpression e1: this.exprs)
			if(!e1.type.equals("os"))
				out.nInGold++;
		for(MPQAExpression e2: sen.exprs)
			if(!e2.type.equals("os"))
				out.nGuesses++;
		
		/* I think this only works when expressions in one layer don't overlap. */
		for(MPQAExpression e1: this.exprs) {
			if(e1.type.equals("os"))
				continue;
			boolean foundOverlap = false;
			boolean foundExact = false;
			for(MPQAExpression e2: sen.exprs) {
				if(e2.type.equals("os"))
					continue;

				double pfc = e1.getSpanCoverage(e2);
				double pgc = e2.getSpanCoverage(e1);
				
				out.propFound += pfc;
				out.propGuessed += pgc;
				
				if(pfc > 0)
					foundOverlap = true;
				if(pfc > ONE && pgc > ONE)
					foundExact = true;				
			}
			if(foundOverlap)
				out.nOverlap++;
			if(foundExact)
				out.nCorrect++;
		}
				
		return out;
	}

	
	public boolean equalsAnnotation(MPQASentence sen) {
		int n = exprs.size();
		if(n != sen.exprs.size())
			return false;
		for(int i = 0; i < n; i++) {
			MPQAExpression e1 = exprs.get(i);
			MPQAExpression e2 = sen.exprs.get(i);
			if(!e1.equalsAnnotation(e2))
				return false;
		}
		return true;
	}
	
	/* Does not take expression labels into account. */
	public PRFStats compareLinksOld(MPQASentence sen) {
		PRFStats out = new PRFStats();

		out.nInGold += getNLinks();
		out.nGuesses += sen.getNLinks();
		
		for(MPQAExpression e1: this.exprs) {
			if(e1.nestedSource == null)
				continue;
			MPQAAgent a1 = e1.nestedSource.get(e1.nestedSource.size() - 1);
			if(a1 == null)
				continue;
			
			//System.out.println("a1.instance.span = " + a1.instances.get(0).span);
						
			for(MPQAExpression e2: sen.exprs)
				if(e1.overlaps(e2)) {
					if(e2.nestedSource == null)
						continue;
					MPQAAgent a2 = e2.nestedSource.get(e2.nestedSource.size() - 1);
					if(a2 == null)
						continue;
					
					// restriction
					//if(!a1.isImplicit || !a2.isImplicit)
					//continue;
					
					//System.out.println("a2.instance.span = " + a2.instances.get(0).span);
					
					if(a1.isWriter && a2.isWriter || a1.isImplicit && a2.isImplicit) {
						out.nCorrect++;
						out.nOverlap++;
						out.propFound += 1.0;
						out.propGuessed += 1.0;
						break;
					}

					MPQAAgentInstance bestI1 = null, bestI2 = null;
					double bestOverlap = Double.MIN_VALUE;
					for(MPQAAgentInstance i1: a1.instances)
						for(MPQAAgentInstance i2: a2.instances) {
							if(i2 == null)
								continue;
							if(i1.containingSentence == null) // varför?
								continue;
							if(i2.containingSentence == null) // varför?
								continue;
							if(i1.containingSentence.dg != i2.containingSentence.dg)
								continue;
							double overlap = i1.getSpanCoverage(i2);
							//System.out.println("overlap = " + overlap);
							if(overlap > bestOverlap) {
								bestOverlap = overlap;
								bestI1 = i1;
								bestI2 = i2;
							}
						}
					if(bestI1 != null) {
						out.nOverlap++;
						out.propFound += bestOverlap;
						out.propGuessed += bestI2.getSpanCoverage(bestI1);
						if(bestI1.span.size() == bestI2.span.size() && bestOverlap > ONE)
							out.nCorrect++;
					}
					
				}
		}
		
		return out;		
	}

	private int getNLinks() {
		int count = 0;
		for(MPQAExpression e1: this.exprs) {
			if(e1.nestedSource == null || e1.nestedSource.isEmpty())
				continue;
			MPQAAgent a1 = e1.nestedSource.get(e1.nestedSource.size() - 1);
			if(a1 == null)
				continue;

			// restriction
			//if(!a1.isImplicit)
			//	continue;
			
			count++;
		}
		return count;
	}
	
	private int getNLinks(String etype, 
			Boolean nonlocal, Boolean writer, Boolean implicit) {
		int count = 0;
		for(MPQAExpression e1: this.exprs) {
			if(relevantLink(e1, etype, nonlocal, writer, implicit))
				count++;
		}
		return count;
	}
	
	private static boolean relevantLink(MPQAExpression e1, String etype, 
			Boolean nonlocal, Boolean writer, Boolean implicit) {
		if(e1.nestedSource == null || e1.nestedSource.isEmpty())
			return false;
		MPQAAgent a1 = e1.nestedSource.get(e1.nestedSource.size() - 1);
		if(a1 == null)
			return false;
		
		/*
		if(a1.isWriter && a1.instances.size() == 0) {
			System.out.println("instances.size == 0");
			System.exit(0);
		}
		Local => instances.size > 0
		Nonlocal => instances.size == 0
		*/
		
		if(nonlocal != null) {
			if(nonlocal && a1.instances.size() > 0)
				return false;
			if(!nonlocal && a1.instances.size() == 0)
				return false;
		}
		if(writer != null) {
			if(writer && !a1.isWriter)
				return false;
			if(!writer && a1.isWriter)
				return false;
		}
		if(implicit != null) {
			if(implicit && !a1.isImplicit)
				return false;
			if(!implicit && a1.isImplicit)
				return false;
		}
		if(etype != null && !e1.type.equals(etype))
			return false;
		return true;
	}
	

	
	public PRFStats compareLinks(MPQASentence sen) {
		return compareLinks(sen, null, null, null, null);
	}
	
	public PRFStats compareLinks(MPQASentence otherSen, String etype,
			Boolean nonlocal, Boolean writer, Boolean implicit) {
		PRFStats out = new PRFStats();

		out.nInGold += getNLinks(etype, nonlocal, writer, implicit);
		out.nGuesses += otherSen.getNLinks(etype, nonlocal, writer, implicit);
		
		for(MPQAExpression e1: this.exprs) {

			if(!relevantLink(e1, etype, nonlocal, writer, implicit))
				continue;
			MPQAAgent a1 = e1.nestedSource.get(e1.nestedSource.size() - 1);
						
			for(MPQAExpression e2: otherSen.exprs)
				if(e1.overlaps(e2)) {
					if(!relevantLink(e2, etype, nonlocal, writer, implicit))
						continue;
					MPQAAgent a2 = e2.nestedSource.get(e2.nestedSource.size() - 1);
					
					// restriction
					//if(!a1.isImplicit || !a2.isImplicit)
					//continue;
					
					//System.out.println("a2.instance.span = " + a2.instances.get(0).span);
					
					if(a1.isWriter && a2.isWriter || a1.isImplicit && a2.isImplicit) {
						out.nCorrect++;
						out.nOverlap++;
						out.propFound += 1.0;
						out.propGuessed += 1.0;
						break;
					}

					MPQAAgentInstance bestI1 = null, bestI2 = null;
					double bestOverlap = Double.MIN_VALUE;
					for(MPQAAgentInstance i1: a1.instances)
						for(MPQAAgentInstance i2: a2.instances) {
							if(i2 == null)
								continue;
							if(i1.containingSentence == null) // varför?
								continue;
							if(i2.containingSentence == null) // varför?
								continue;
							if(i1.containingSentence.dg != i2.containingSentence.dg)
								continue;
							double overlap = i1.getSpanCoverage(i2);
							//System.out.println("overlap = " + overlap);
							if(overlap > bestOverlap) {
								bestOverlap = overlap;
								bestI1 = i1;
								bestI2 = i2;
							}
						}
					if(bestI1 != null) {
						out.nOverlap++;
						out.propFound += bestOverlap;
						out.propGuessed += bestI2.getSpanCoverage(bestI1);
						if(bestI1.span.size() == bestI2.span.size() && bestOverlap > ONE)
							out.nCorrect++;
					}
					
				}
		}
		
		return out;		
	}
	
	private static final boolean polMatch(String p1, String p2) {
		if(p1 != null)
			p1 = p1.trim();
		if(p2 != null)
			p2 = p2.trim();
		if(p1 == null || p1.equals(""))
			p1 = "neutral";
		if(p2 == null || p2.equals(""))
			p2 = "neutral";
		if(p1.charAt(0) == 'u')
			p1 = p1.substring(10);
		if(p2.charAt(0) == 'u')
			p2 = p2.substring(10);

		if(p1.equals("both"))
			p1 = "neutral";
		if(p2.equals("both"))
			p2 = "neutral";
				
		return p1.equals(p2);		
	}
	
	private static final boolean intMatch(String i1, String i2) {
		if(i1 == null || i1.equals(""))
			i1 = "medium";
		if(i2 == null || i2.equals(""))
			i2 = "medium";
		
		if(i1.equals("extreme"))
			i1 = "high";
		if(i2.equals("extreme"))
			i2 = "high";
		
		return i1.equals(i2);
	}
	
	public PRFStats comparePolInt(MPQASentence sen, boolean forPol, String value) {
		PRFStats out = new PRFStats();
		
		//System.out.println("s1: " + this.exprs);
		//System.out.println("s2: " + sen.exprs);		

		/*
		for(MPQAExpression e1: this.exprs)
			if(e1.intensity != null && e1.intensity.equals("extreme"))
					e1.intensity = "high";
		for(MPQAExpression e2: sen.exprs)
			if(e2.intensity != null && e2.intensity.equals("extreme"))
					e2.intensity = "high";
		
		for(MPQAExpression e1: this.exprs) {
			if(e1.polarity != null && e1.polarity.startsWith("u"))
				e1.polarity = e1.polarity.substring("uncertain-".length());
			if(e1.polarity != null && e1.polarity.equals("both"))
					e1.polarity = "neutral";
		}
		for(MPQAExpression e2: sen.exprs) {
			if(e2.polarity != null && e2.polarity.startsWith("u"))
				e2.polarity = e2.polarity.substring("uncertain-".length());
			if(e2.polarity != null && e2.polarity.equals("both"))
					e2.polarity = "neutral";
		}*/
		
		for(MPQAExpression e1: this.exprs)
			if(!e1.type.equals("os") 
					   && (forPol && polMatch(e1.polarity, value) || !forPol && intMatch(e1.intensity, value)))
				out.nInGold++;
		for(MPQAExpression e2: sen.exprs)
			if(!e2.type.equals("os")
					   && (forPol && polMatch(e2.polarity, value) || !forPol && intMatch(e2.intensity, value)))
				out.nGuesses++;
		
		/* I think this only works when expressions in one layer don't overlap. */
		for(MPQAExpression e1: this.exprs) {
			if(e1.type.equals("os"))
				continue;
			if(forPol && !polMatch(e1.polarity, value))
				continue;
			if(!forPol && !intMatch(e1.intensity, value))
				continue;
			
			boolean foundOverlap = false;
			boolean foundExact = false;
			for(MPQAExpression e2: sen.exprs) {
				if(e2.type.equals("os"))
					continue;
				if(forPol && !polMatch(e2.polarity, value))
					continue;
				if(!forPol && !intMatch(e2.intensity, value))
					continue;

				double pfc = e1.getSpanCoverage(e2);
				double pgc = e2.getSpanCoverage(e1);
				
				out.propFound += pfc;
				out.propGuessed += pgc;
				
				if(pfc > 0)
					foundOverlap = true;
				if(pfc > ONE && pgc > ONE)
					foundExact = true;				
			}
			if(foundOverlap)
				out.nOverlap++;
			if(foundExact)
				out.nCorrect++;
		}
		
		//System.out.println("comp: " + out);
		
		return out;
	}
	
	public PRFStats comparePolInt(MPQASentence sen, boolean forPol) {
		PRFStats out = new PRFStats();
		
		//System.out.println("s1: " + this.exprs);
		//System.out.println("s2: " + sen.exprs);		

		/*for(MPQAExpression e1: this.exprs)
			if(e1.intensity != null && e1.intensity.equals("extreme"))
					e1.intensity = "high";
		for(MPQAExpression e2: sen.exprs)
			if(e2.intensity != null && e2.intensity.equals("extreme"))
					e2.intensity = "high";
		
		for(MPQAExpression e1: this.exprs)
			if(e1.polarity != null && e1.polarity.startsWith("unc"))
				throw new RuntimeException("uncertain");
		for(MPQAExpression e2: this.exprs)
			if(e2.polarity != null && e2.polarity.startsWith("unc"))
				throw new RuntimeException("uncertain");
		
		for(MPQAExpression e1: this.exprs)
			if(e1.polarity != null && e1.polarity.equals("both"))
					e1.polarity = "neutral";
		for(MPQAExpression e2: sen.exprs)
			if(e2.polarity != null && e2.polarity.equals("both"))
					e2.polarity = "neutral";*/
		
		for(MPQAExpression e1: this.exprs)
			if(!e1.type.equals("os"))
				out.nInGold++;
		for(MPQAExpression e2: sen.exprs)
			if(!e2.type.equals("os"))
				out.nGuesses++;
		
		/* I think this only works when expressions in one layer don't overlap. */
		for(MPQAExpression e1: this.exprs) {
			if(e1.type.equals("os"))
				continue;

			/*if(forPol && (e1.polarity == null || e1.polarity.equals("")))
				continue;
			if(!forPol && (e1.intensity == null || e1.intensity.equals("")))
				continue;*/
			
			boolean foundOverlap = false;
			boolean foundExact = false;
			for(MPQAExpression e2: sen.exprs) {
				if(e2.type.equals("os"))
					continue;

				/*if(forPol && (e2.polarity == null || e2.polarity.equals("")))
					continue;
				if(!forPol && (e2.intensity == null || e2.intensity.equals("")))
					continue;*/
				
				if(forPol && polMatch(e1.polarity, e2.polarity)// e1.polarity.equals(e2.polarity) 
						|| !forPol && intMatch(e1.intensity, e2.intensity)) {
					double pfc = e1.getSpanCoverage(e2);
					out.propFound += pfc;

					double pgc = e2.getSpanCoverage(e1);
					out.propGuessed += pgc;

					if(pfc > 0)
						foundOverlap = true;
					if(pfc > ONE && pgc > ONE)
						foundExact = true;
				}
			}
			if(foundOverlap)
				out.nOverlap++;
			if(foundExact)
				out.nCorrect++;
		}
		
		//System.out.println("comp: " + out);
		
		return out;
	}

	public String toString() {
		return dg.toString() + "\n" + exprs.toString();
	}
	
	public String toBrackets() {
		StringBuilder sb = new StringBuilder("[");
		boolean first = true;
		for(MPQAExpression e: exprs) {
			if(first)
				first = false;
			else
				sb.append(";");
			sb.append(e.type);
			sb.append("/");
			sb.append(e.span.get(0).position);
			sb.append("/");
			sb.append(e.span.get(e.span.size()-1).position + 1);
		}
		sb.append("]");
		return sb.toString();
	}
	
	public void pruneIdentical() {
		HashSet<MPQAExpression> toRemove = new HashSet();
		for(MPQAExpression e1: exprs)
			for(MPQAExpression e2: exprs) {
				if(e1 == e2)
					continue;
				if(e1.type.equals(e2.type) && e1.span.equals(e2.span))
					toRemove.add(e2);
			}
		exprs.removeAll(toRemove);
	}
	
	public void pruneOverlapping() {
		HashSet<MPQAExpression> toRemove = new HashSet();
		for(MPQAExpression e1: exprs)
			for(MPQAExpression e2: exprs) {
				if(e1 == e2)
					continue;
				if(e1.overlaps(e2)) {
					if(e1.span.size() >= e2.span.size())
						toRemove.add(e2);
					else
						toRemove.add(e1);
				}
			}
		exprs.removeAll(toRemove);
		
	}

	// does NOT clone the parse trees
	public MPQASentence clone() {
		MPQASentence out = new MPQASentence();
		out.dg = dg;
		out.pstree = pstree;
		out.pas = pas;
		if(slClues != null)
			out.slClues = (ArrayList) slClues.clone();		
		if(discConns != null)
			out.discConns = (ArrayList) discConns.clone();
		if(exprs != null) {
			out.exprs = new ArrayList();
			for(MPQAExpression ex: exprs) {
				MPQAExpression cex = ex.clone();
				cex.containingSentence = out;
				out.exprs.add(cex);
			}
		}
		return out;
	}
	
	public void convertToUnlabeledOpinion() {
		for(Iterator<MPQAExpression> it = exprs.iterator(); it.hasNext(); ) {
			MPQAExpression ex = it.next();
			if(ex.type.equals("os"))
				it.remove();
			else
				ex.type = "OP";
		}
	}

	public void replaceDepTree(DepGraph newDg) {
	    //if(newDg.nodes.length != dg.nodes.length) {
	    String dgs0 = dg.toTokenString().replaceAll("\\s", "");
	    String dgs1 = newDg.toTokenString().replaceAll("\\s", "");
	    if(!dgs0.equals(dgs1)) {
			System.err.println("Warning: parse tree mismatch");

			System.err.println("old:");
			System.err.println(dg);
			System.err.println("new:");
			System.err.println(newDg);

			throw new RuntimeException("!!!");
			
			//System.exit(1);
			
			//return;
		}
		
		HashSet<MPQAAgent> ags = new HashSet();
		
		for(MPQAExpression expr: exprs) {
			int n = expr.span.size();
			for(int i = 0; i < n; i++) {
				int pos = expr.span.get(i).position;
				expr.span.set(i, newDg.nodes[pos]);
			}
			if(expr.nestedSource != null)
				ags.addAll(expr.nestedSource);

			if(n > 0 && expr.span.get(0) instanceof EPEDepNode) {
			    if(expr.start != Integer.MIN_VALUE) {
			        /*if(expr.start != ((EPEDepNode) expr.span.get(0)).start) {
			            System.out.println(expr);
			            System.out.println(expr.start + "/" + expr.end);
	                    System.out.println(expr.span);
			            throw new RuntimeException("incompatible spans!");
			        }*/
			    }
                expr.start = ((EPEDepNode) expr.span.get(0)).start; 
                expr.end = ((EPEDepNode) expr.span.get(n-1)).end;
			}
			
		}

		for(MPQAAgent ag: ags) {
			if(ag == null || ag.instances == null) // hur kan ag vara null?
				continue;
			for(MPQAAgentInstance inst: ag.instances) {
				if(inst.containingSentence != this)
					continue;
				int n = inst.span.size();
				for(int i = 0; i < n; i++) {
					int pos = inst.span.get(i).position;
					inst.span.set(i, newDg.nodes[pos]);
				}
				if(n > 0 && inst.span.get(0) instanceof EPEDepNode) {
	                //if(inst.start != Integer.MIN_VALUE)
	                //    throw new RuntimeException("inst start and end already set");
	                inst.start = ((EPEDepNode) inst.span.get(0)).start; 
	                inst.end = ((EPEDepNode) inst.span.get(n-1)).end; 
	            }
			}
		}

		if(pas != null)
			for(PAStructure pa: pas) {
				pa.pred = newDg.nodes[pa.pred.position];
				for(int i = 0; i < pa.args.size(); i++) {
					int pos = pa.args.get(i).position;
					pa.args.set(i, newDg.nodes[pos]);
				}			
			}

		dg = newDg;
	}



}
