/* Configuration.java

   Created on Mar 18, 2008 by Richard Johansson (richard@cs.lth.se).

   $Log: Configuration.java,v $
   Revision 1.1  2010-06-07 11:33:30  johansson
   Added.

   Revision 1.1  2009/01/16 14:50:55  johansson
   Added to brenta repository.

   Revision 1.8  2008/05/16 11:47:54  richard
   Bug with empty set.

   Revision 1.7  2008/04/28 14:36:19  richard
   Includes global features.

   Revision 1.6  2008/04/08 11:47:51  richard
   Added the PRED classifiers.

   Revision 1.5  2008/04/07 12:14:18  richard
   Pairs.

   Revision 1.4  2008/04/07 07:34:16  richard
   Added SUPP and separated NN and VB for WSD.

   Revision 1.3  2008/04/03 13:38:06  richard
   WSD feature extractor.

   Revision 1.2  2008/04/02 11:29:59  richard
   Separated parameters for different classifiers.

   Revision 1.1  2008/03/18 15:18:29  richard
   Imported from fnlabeler.


 */

package mpqa_structlearn.locallink;

import org.w3c.dom.*;

import java.io.*;
import java.util.*;

import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.ml_long.*;

/**
 * 
 * @author richard
 *
 */
public class Configuration {

    public static enum TaskName { TRAIN, PRINT, PRINT_HTML,
        GOLD_TARGETS, AUTOMATIC_TARGETS,
        GOLD_FRAMES, AUTOMATIC_FRAMES,
        GOLD_ARGUMENTS, AUTOMATIC_ARGUMENTS,
        GOLD_ROLES, AUTOMATIC_ROLES,
    };
    
    private Document doc;
    
    private Set<TaskName> tasks;    

    private String modelName;
    
    private String[] methods = new String[2];
    private String[] args = new String[2];
    private String[][] features = new String[2][];
    private HashSet<Pair<String, String>>[] pairs = new HashSet[2];

    public static enum ClassifierID { LOCAL, NONLOCAL };
    
    public Configuration(InputStream is) throws IOException {
        doc = DOMUtils.readDocument(is, false);
        if(doc == null)
            throw new IOException("Couldn't read configuration.");

        if(!doc.getDocumentElement().getNodeName().equals("configuration")) {
            throw new IOException("This is not a configuration file.");
        }

        NodeList nl = doc.getDocumentElement().getElementsByTagName("model");
        if(nl.getLength() > 0)
            modelName = nl.item(0).getTextContent().trim();

        nl = doc.getDocumentElement().getElementsByTagName("tasks");

        if(nl.getLength() > 0) {
            //task = TaskName.valueOf(nl.item(0).getTextContent().trim());
            String ts = nl.item(0).getTextContent().trim();
            if(ts.endsWith(","))
                ts = ts.substring(0, ts.length() - 1);
            ts = ts.trim();
            tasks = EnumSet.noneOf(TaskName.class);
            if(ts.length() > 0)
                for(String t: ts.split("\\,\\s*"))
                    tasks.add(TaskName.valueOf(t));

            if(tasks.contains(TaskName.TRAIN) && tasks.size() > 1)
                throw new IOException("Should not add more tasks than TRAIN");
            
            if((tasks.contains(TaskName.PRINT) || tasks.contains(TaskName.PRINT_HTML)) && tasks.size() > 1)
                throw new IOException("Should not add more tasks than PRINT");
            
            if(tasks.contains(TaskName.AUTOMATIC_TARGETS)
                    && (tasks.contains(TaskName.GOLD_FRAMES)
                        || tasks.contains(TaskName.GOLD_ARGUMENTS)
                        || tasks.contains(TaskName.GOLD_ROLES)))
                     throw new IOException("Automatically detected targets -> should not use gold frames, arguments, or roles");             
            
            if(tasks.contains(TaskName.AUTOMATIC_ARGUMENTS)
               && tasks.contains(TaskName.GOLD_ROLES))
                throw new IOException("Automatically detected arguments -> should not use gold roles");             

            if(tasks.contains(TaskName.GOLD_TARGETS) && tasks.contains(TaskName.AUTOMATIC_TARGETS))
                throw new IOException("Should not specify both automatic and gold targets");
            
            if(tasks.contains(TaskName.GOLD_FRAMES) && tasks.contains(TaskName.AUTOMATIC_FRAMES))
                throw new IOException("Should not specify both automatic and gold frames");
            
            if(tasks.contains(TaskName.GOLD_ARGUMENTS) && tasks.contains(TaskName.AUTOMATIC_ARGUMENTS))
                throw new IOException("Should not specify both automatic and gold arguments");
            
            if(tasks.contains(TaskName.GOLD_ROLES) && tasks.contains(TaskName.AUTOMATIC_ROLES))
                throw new IOException("Should not specify both automatic and gold roles");
            
            if((tasks.contains(TaskName.GOLD_FRAMES) || tasks.contains(TaskName.AUTOMATIC_FRAMES))
                    && !tasks.contains(TaskName.GOLD_TARGETS) && !tasks.contains(TaskName.AUTOMATIC_TARGETS))
                     throw new IOException("Assign frames -> should specify how to find the targets");
            
            if((tasks.contains(TaskName.GOLD_ARGUMENTS) || tasks.contains(TaskName.AUTOMATIC_ARGUMENTS))
                    && !tasks.contains(TaskName.GOLD_FRAMES) && !tasks.contains(TaskName.AUTOMATIC_FRAMES))
                     throw new IOException("Find arguments -> should specify how to assign the frames");
            
            if((tasks.contains(TaskName.GOLD_ROLES) || tasks.contains(TaskName.AUTOMATIC_ROLES))
                    && !tasks.contains(TaskName.GOLD_ARGUMENTS) && !tasks.contains(TaskName.AUTOMATIC_ARGUMENTS))
                     throw new IOException("Assign roles -> should specify how to find the arguments");
            
        } else
            throw new IOException("Must declare tasks");
        
        nl = doc.getDocumentElement().getElementsByTagName("classifier");
        if(nl.getLength() > 0 && !tasks.contains(TaskName.TRAIN))
            throw new IOException("Should not specify classifiers in testing mode!");       
        for(int i = 0; i < nl.getLength(); i++) {
            Element el = (Element) nl.item(i);

            String funs = el.getAttribute("funcs").trim();
            if(funs.equals(""))
                throw new IOException("Must declare functions of classifier");
            
            for(String fun: funs.split("\\,\\s*")) {        
                int f = ClassifierID.valueOf(fun).ordinal();
                methods[f] = el.getAttribute("method");
                args[f] = el.getAttribute("args");

                NodeList nl2 = el.getElementsByTagName("features");
                for(int j = 0; j < nl2.getLength(); j++) {
                    Element fs = (Element) nl2.item(j);
                    String fstr = fs.getTextContent().trim();
                    if(fstr.endsWith(","))
                        fstr = fstr.substring(0, fstr.length() - 1);
                    fstr = fstr.trim();
                    Pair<String[], HashSet<Pair<String, String>>> p
                    = createFeaturesAndPairs(fstr);
                    //createFeaturesAndPairs(fstr, f);
                    features[f] = p.left;
                    pairs[f] = p.right;
                }
            }            
        }
        
    }

    private Pair<String[], HashSet<Pair<String, String>>> createFeaturesAndPairs(String s) throws IOException {
        s = s.trim();
        String[] ss = s.equals("")? new String[0]: s.split("\\,\\s*");
        ArrayList<String> fs = new ArrayList();
        HashSet<Pair<String, String>> ps = new HashSet();
        for(String f: ss) {
            int ix = f.indexOf('+');
            if(ix != -1) {
                String s1 = f.substring(0, ix);
                String s2 = f.substring(ix + 1);
                if(s2.contains("+"))
                    throw new IOException("Only combinations of 2 allowed");
                ps.add(new Pair(s1, s2));
            } else
                fs.add(f);
        }
        return new Pair(fs.toArray(new String[0]), ps);
    }
    
    public void write(OutputStream out) throws IOException {
        DOMUtils.writeDocument(doc, out);
    }

    public String toXMLString() {
    	ByteArrayOutputStream bos = new ByteArrayOutputStream();
    	try {
    		write(bos);
        	return bos.toString("UTF-8");
    	} catch(Exception e) {
    		throw new RuntimeException(e);
    	}
    }
    
    public String getModelName() {
        return modelName;
    }
    
    public Set<TaskName> getTasks() {
        return Collections.unmodifiableSet(tasks);
    }    
    
    public LearningAlgorithm getAlgorithm(ClassifierID c) {
        return new AlgorithmFactory().create(getClassifierMethod(c),
                                             getClassifierArguments(c));
    }
    
    private String getClassifierMethod(ClassifierID c) {
        return methods[c.ordinal()];
    }

    private String getClassifierArguments(ClassifierID c) {
        return args[c.ordinal()];
    }

    public HashSet<Pair<String, String>> getClassifierPairs(ClassifierID c) {
        return pairs[c.ordinal()];
    }
    
    public FeatureDeclaration getClassifierFeatures(ClassifierID c) {
        return subset(LocalLinkClassifier.FEATURE_DECL, features[c.ordinal()]);
    }

    private FeatureDeclaration subset(FeatureDeclaration decl, String[] fs) {
        if(fs == null)
            return decl;
        else
            return decl.subset(new HashSet<String>(Arrays.asList(fs)));
    }

}