/*
 EnglishFeatures.java
  
 Created on Nov 15, 2006 by Richard Johansson (richard@cs.lth.se).

 $Log: EnglishFeatures.java,v $
 Revision 1.1  2010-06-07 11:33:30  johansson
 Added.

 Revision 1.1  2009/01/16 14:50:55  johansson
 Added to brenta repository.

 Revision 1.8  2008/05/16 11:48:21  richard
 POS paths, siblings.

 Revision 1.7  2008/04/23 09:06:07  richard
 Handles multiple roots.

 Revision 1.6  2008/04/17 14:51:11  richard
 Replaces numbers.

 Revision 1.5  2008/04/10 07:09:31  richard
 New features.

 Revision 1.4  2008/04/07 07:34:47  richard
 Support features.

 Revision 1.3  2008/04/02 11:30:21  richard
 Added getRelativePosition.

 Revision 1.2  2008/03/19 15:54:12  richard
 Serializable.

 Revision 1.1  2008/03/18 15:18:48  richard
 Imported from fnlabeler.

 Revision 1.7  2007/10/12 06:34:44  richard
 Some more functions.

 Revision 1.6  2007/10/08 12:33:02  richard
 Refactored again to remove atoms.

 Revision 1.5  2007/10/04 12:28:02  richard
 Refactored to atom-based.

 Revision 1.4  2007/08/21 10:19:00  richard
 Changed collapsing in ps and dep paths.

 Revision 1.3  2007/08/20 15:03:41  richard
 Added phrases.

 Revision 1.2  2007/08/17 10:54:22  richard
 Cleaned up. Loads feature declarations.

 Revision 1.1  2007/08/17 10:15:10  richard
 Refactored.

 Revision 1.1  2007/04/05 13:26:42  richard
 Version used in SemEval.

 Revision 1.2  2006/11/20 16:19:13  richard
 Labels a treebank.

 Revision 1.1  2006/11/15 14:21:46  richard
 Added the file.

   
 */
package mpqa_structlearn.locallink;

import java.util.*;
import java.io.Serializable;
import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.depgraph.*;
import se.lth.cs.nlp.depsrl.format.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class EnglishFeatures implements Serializable {

    private static final long serialVersionUID = 0L;
    
    /* Whether the preposition should be attached to PP for prepositional
     * phrases, i.e. a PP headed by "at" gets the phrase type PP-at.
     */
    private static final boolean PP_ATTACH_PREP = false; //true;

    private static final String IDENTICAL = "<IDENTICAL>";
    private static final String EMPTY = "<EMPTY>";
    private static final String NONE = "<NONE>";
    private static final String ACTIVE = "active";
    private static final String PASSIVE = "passive";
    
    private static final String LEFT = "left";
    private static final String RIGHT = "right";
    private static final String ON = "on";

    
    public Pair<String, DepNode> getRelAndHead(DepNode node, DepNode target) {
        if(node == target)
            return new Pair(IDENTICAL, node);
        if(node.isSameOrAncestorOf(target))
            //for(DepNode c: node.children)
            //if(target == c)
            return new Pair("UP->" + target.relations[0], node);

        if(!PP_ATTACH_PREP)
            return new Pair(node.relations[0], node);

        /* TODO prep som head eller inte */        

        DepNode pchild = null;
        for(DepNode c: node.children)
            if(c.relations[0].equals("PMOD")) {
                pchild = c;
                break;
            }
        
        if(pchild == null)
            return new Pair(node.relations[0], node);
        
        String rel = node.relations[0] + "-" + node.word.toLowerCase();
        
        return new Pair(rel, pchild);
    }
    
    private DepNode findCommonAncestor(DepNode d1, DepNode d2) {
        HashSet<DepNode> s = new HashSet<DepNode>();
        while(true) {
            s.add(d1);
            if(d1.position == 0)
                break;
            d1 = d1.parents[0];
        }
        while(d2 != null) {
            if(s.contains(d2))
                return d2;
            if(d2.position == 0)
                break;
            d2 = d2.parents[0];
        }
        return null;
    }

    public String findDepPOSPath(DepNode d1, DepNode d2) {
        if(d1 == d2)
            return IDENTICAL;
        StringBuilder sb = new StringBuilder();         
        DepNode a = findCommonAncestor(d1, d2);
        if(a.position == 0)
            return NONE;
        while(d1 != a) {
            sb.append(d1.pos + "_UP_");
            d1 = d1.parents[0];
        }
        sb.append(a.pos);
        int ix = sb.length();        
        while(d2 != a) {
            sb.insert(ix, "_DOWN_" + d2.pos);
            d2 = d2.parents[0];
        }
        return sb.toString();
    }
    
    private String findDepPathString(DepNode d1, DepNode d2) {
        if(d1 == d2)
            return null;
        StringBuilder sb = new StringBuilder();        
        DepNode a = findCommonAncestor(d1, d2);
        while(d1 != a) {
            sb.append(d1.relations[0] + "_UP_");
            d1 = d1.parents[0];
        }
        int ix = sb.length();        
        while(d2 != a) {
            sb.insert(ix, d2.relations[0] + "_DOWN_");
            d2 = d2.parents[0];
        }
        return sb.toString().substring(0, sb.length()-1);
    }
    
    public String findDepPath(DepNode d1, DepNode d2) {
        if(d1 == d2)
            return IDENTICAL;
        return findDepPathString(d1, d2);
    }
    
    public String findGeneralPath(DepNode d1, DepNode d2,
            boolean removeCoord, boolean removeVC, boolean removeOPRD) {
        if(d1 == d2)
            return IDENTICAL;
        String p = findDepPathString(d1, d2);
        if(removeCoord)
            p = p.replaceAll("COORD_UP|CONJ_UP", "");
        if(removeVC)
            p = p.replaceAll("VC_UP", "");
        if(removeOPRD)
            p = p.replaceAll("OPRD_UP", "");
        
        p = p.replaceAll("_+", "_");
        p = p.replaceAll("(^_|_$)", "");
        
        if(p.equals(""))
            return EMPTY;

        return p;        
    }
    
    public String findSimplifiedDepPath(DepNode d1, DepNode d2) {
        if(d1 == d2)
            return IDENTICAL;
        String p = findDepPathString(d1, d2);
        //p = p.replaceAll("COORD_UP|VC_UP", "");
        p = p.replaceAll("VC_UP", "");

        //p = p.replaceAll("VCOORD_UP", "");
        
        p = p.replaceAll("_+", "_");
        p = p.replaceAll("(^_|_$)", "");
        
        if(p.equals(""))
            return EMPTY;
        return p;
    }
    
    public String findPathToSupports(PAStructure pa, DepNode d2) {
        HashSet<DepNode> supports = new HashSet();
        for(int i = 0; i < pa.args.size(); i++)
            if(pa.argLabels.get(i).startsWith("SU"))
                supports.add(pa.args.get(i));
        String best = null;
        int bestSteps = Integer.MAX_VALUE;
        if(supports.isEmpty())
            return EMPTY;
        for(DepNode su: supports) {
            String path = findDepPath(su, d2);
            int steps = Strings.count(path, '_');
            if(steps < bestSteps) {
                best = path;
                bestSteps = steps;
            }
        }
        return best;
    }
    
    public String findPathToVerb(DepNode npred, DepNode d2) {
        if(npred.parents[0] == null)
            return EMPTY;
        DepNode anc = findCommonAncestor(npred, d2);
        DepNode n = d2;
        while(n.pos != null && !n.pos.startsWith("V")) {
            if(n == anc)
                break;
            n = n.parents[0];
        }
        if(n.pos == null || !n.pos.startsWith("V"))
            return EMPTY;
        else
            return findSimplifiedDepPath(n, d2);
    }
    
    public void findDeps(DepNode target, boolean includeSubject, HashSet<String> deps) {
        deps.clear();
        for(DepNode c: target.children)
            if(!c.relations[0].matches("P|COORD|PRN|VMOD")
               && (includeSubject || !c.relations[0].equals("SBJ")))
                deps.add(getRelAndHead(c, target).left);
    }
    
    public String findDepSubCat(DepNode target) {
        if(!target.pos.startsWith("V"))
            return EMPTY;
        List<String> deps = new ArrayList<String>();
        for(DepNode c: target.children)
            if(!c.relations[0].matches("(SBJ|P|COORD|CONJ|PRN|DEP)|(.*GAP.*)"))
                deps.add(getRelAndHead(c, target).left.toString());
        if(deps.size() == 0)
            return EMPTY;
        return Strings.join(deps, "_");
    }

    public void findWordContext(DepNode n, DepGraph dg, int wsize, HashSet<String> context,
            boolean replaceNumbers) {
        context.clear();
        int count = 0;
        for(int i = n.position - 1; i > 0 && count < wsize; i--)
            if(!dg.nodes[i].pos.matches(("CC|HYPH|:|\\.|\\,|``|''|SYM|UH|W.*|POS|LS|DT|#|\\$|\\(|\\)"))) {
                count++;
                if(replaceNumbers)
                    context.add(dg.nodes[i].word.toLowerCase().replaceAll("[0-9]", "0"));
                else
                    context.add(dg.nodes[i].word.toLowerCase());
            }
        count = 0;
        for(int i = n.position + 1; i < dg.nodes.length && count < wsize; i++)
            if(!dg.nodes[i].pos.matches(("CC|HYPH|:|\\.|\\,|``|''|SYM|UH|W.*|POS|LS|DT|#|\\$|\\(|\\)"))) {
                count++;
                if(replaceNumbers)
                    context.add(dg.nodes[i].word.toLowerCase().replaceAll("[0-9]", "0"));
                else
                    context.add(dg.nodes[i].word.toLowerCase());
            }
    }
    
    public boolean hasChildRelation(DepNode n, String rel) {
        for(DepNode c: n.children)
            if(c.relations[0].equals(rel))
                return true;
        return false;
    }

    public boolean hasObjectChild(DepNode n) {
        return hasChildRelation(n, "OBJ");
    }
    
    public boolean hasSubjectChild(DepNode n) {
        return hasChildRelation(n, "SBJ");
    }
    
    public boolean hasSubjectInVC(DepNode n) {
        if(hasChildRelation(n, "SBJ"))
            return true;
        if(n.relations[0].equals("VC"))
            return hasSubjectInVC(n.parents[0]);
        return false;
    }
    
    public String oprdParentHasObject(DepNode n) {
        if(n.relations[0].equals("OPRD"))
            return hasObjectChild(n.parents[0])? "HAS_OBJ": "NO_OBJ";
        if(n.relations[0].matches("IM|VC|COORD|CONJ"))
            return oprdParentHasObject(n.parents[0]);
        return NONE;
    }
    
    public String findVoice(DepNode tn) {
        if(!tn.pos.startsWith("V"))
            return NONE;
        if(tn.word.toLowerCase().equals("been"))
            return ACTIVE;
        if(!tn.pos.equals("VBN"))
            return ACTIVE;

        return findPassiveInParents(tn.parents[0]);
    }
    
    private String findPassiveInParents(DepNode n) {
        if(n == null || n.position == 0)
            return PASSIVE;
        String word = n.word.toLowerCase(), pos = n.pos;
        
        if(pos.startsWith("NN"))
            return PASSIVE;
        
        if(word.matches("am|are|is|was|were|be|been|being"))
            return PASSIVE;
        
        if(word.matches("ha(ve|s|d|ving)"))
            return ACTIVE;
        
        if(pos.matches("VBZ|VBD|VBP|MD"))
            return PASSIVE;
        
        return findPassiveInParents(n.parents[0]);
    }
    
    public boolean isVerb(DepNode n) {
        return n.pos != null 
            && (n.pos.startsWith("V"));
    }
    
    public String getCoarsePos(DepNode n) {
        return n.pos.substring(0, 2);
    }
    
    public DepNode getLeftmostChild(DepNode node) {
        ArrayList<DepNode> l = new ArrayList<DepNode>();
        for(DepNode c: node.children)
            if(c.position < node.position)
                l.add(c);
        int min = Integer.MAX_VALUE;
        DepNode minNode = null;
        for(DepNode d: l)
            if(d.position < min) {
                min = d.position;
                minNode = d;
            }
        return minNode;
    }
    
    public DepNode getRightmostChild(DepNode node) {
        ArrayList<DepNode> l = new ArrayList<DepNode>();
        for(DepNode c: node.children)
            if(c.position > node.position)
                l.add(c);
        int max = -1;
        DepNode maxNode = null;
        for(DepNode d: l)
            if(d.position > max) {
                max = d.position;
                maxNode = d;
            }
        return maxNode;
    }
    
    public String getRelativePosition(DepNode node, DepNode target) {
        if(node.position < target.position)
            return LEFT;
        if(node.position == target.position)
            return ON;
        return RIGHT;
    }
    
    public DepNode getLeftSibling(DepNode node) {
        DepNode p = node.parents[0];
        int max = -1;
        DepNode maxn = null;
        for(DepNode n: p.children)
            if(n.position < node.position && n.position > max) {
                max = n.position;
                maxn = n;
            }
        return maxn;
    }
    
    public DepNode getRightSibling(DepNode node) {
        DepNode p = node.parents[0];
        int min = Integer.MAX_VALUE;
        DepNode minn = null;
        for(DepNode n: p.children)
            if(n.position > node.position && n.position < min) {
                min = n.position;
                minn = n;
            }
        return minn;
    }
 
    public int getDistance(DepNode from, DepNode to) {
    	if(from == to)
    		return 0;
    	DepNode a = findCommonAncestor(from, to);
    	int out = 0;
    	while(from != a) {
    		out++;
    		from = from.parents[0];
    	}
    	while(to != a) {
    		out++;
    		to = to.parents[0];
    	}    	
    	return out;
    }
    
    public DepNode getClosestInSpan(DepNode from, ArrayList<DepNode> span) {
    	int size = span.size();
    	if(size == 0)
    		throw new IllegalArgumentException("span is empty");
    	if(size == 1)
    		return span.get(0);
    	int minDist = Integer.MAX_VALUE;
    	DepNode minNode = null;
    	for(DepNode n: span) {
    		int dist = getDistance(from, n);
    		if(dist < minDist) {
    			minDist = dist;
    			minNode = n;
    		}
    	}
    	return minNode;	
    }
    
    public DepNode getHighest(ArrayList<DepNode> span) {
    	int size = span.size();
    	if(size == 0)
    		throw new IllegalArgumentException("span is empty");
    	if(size == 1)
    		return span.get(0);
    	HashSet<DepNode> removed = new HashSet();
    	for(DepNode n1: span)
    		for(DepNode n2: span)
    			if(n1 != n2 && n1.isSameOrAncestorOf(n2))
    				removed.add(n2);
    	for(DepNode n: span)
    		if(!removed.contains(n))
    			return n;
    	throw new RuntimeException("no highest node found");
    }
    
}
