/*
 NodeFeatureExtractor.java
 
 Created on Nov 15, 2006 by Richard Johansson (richard@cs.lth.se).
 
 $Log: NodeFeatureExtractor.java,v $
 Revision 1.1  2010-06-07 11:33:30  johansson
 Added.

 Revision 1.1  2009/01/16 14:50:55  johansson
 Added to brenta repository.

 Revision 1.6  2008/05/16 11:49:33  richard
 POS paths, siblings.

 Revision 1.5  2008/04/17 14:51:29  richard
 Replaces numbers.

 Revision 1.4  2008/04/10 07:12:05  richard
 Full path instead of VC-less path.

 Revision 1.3  2008/04/07 08:02:13  richard
 Support.

 Revision 1.2  2008/04/02 11:30:49  richard
 Moved getRelativePosition to EnglishFeatures.

 Revision 1.1  2008/03/18 15:18:48  richard
 Imported from fnlabeler.

 Revision 1.7  2007/10/08 12:33:02  richard
 Refactored again to remove atoms.

 Revision 1.6  2007/10/04 12:28:02  richard
 Refactored to atom-based.

 Revision 1.5  2007/08/21 10:20:48  richard
 Deprel.

 Revision 1.4  2007/08/20 15:03:41  richard
 Added phrases.

 Revision 1.3  2007/08/20 12:05:15  richard
 Restructured task configuration. Gold standard steps.

 Revision 1.2  2007/08/17 10:54:22  richard
 Cleaned up. Loads feature declarations.

 Revision 1.1  2007/08/17 10:15:10  richard
 Refactored.

 Revision 1.1  2007/04/05 13:26:42  richard
 Version used in SemEval.

 Revision 1.2  2006/11/20 16:19:13  richard
 Labels a treebank.

 Revision 1.1  2006/11/15 14:21:46  richard
 Added the file.


 */
package mpqa_structlearn.locallink;

import se.lth.cs.nlp.depsrl.format.PAStructure;
import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.ml_long.*;
import se.lth.cs.nlp.nlputils.depgraph.*;
import java.util.*;
//import se.lth.cs.nlp.depsrl.format.*;
import mpqa_structlearn.io.*;

/**
 * Feature extractor that finds features for a node in a parse tree with
 * respect to a predicate node.
 * 
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class NodeFeatureExtractor extends FeatureExtractor<Triple<MPQAExpression, DepNode, MPQASentence>> {
    
    private static final long serialVersionUID = 0L;
    
    private EnglishFeatures lang;
    
    /* Feature indices. */
    int N_POSITION;
    int N_HOLDER_HEAD;
    int N_HOLDER_HEAD_POS;
    int N_DEPREL;

    int N_NEAREST_EXPR_WORD;
    
    int N_LEFT_WORD;
    int N_LEFT_POS;
    int N_RIGHT_WORD;
    int N_RIGHT_POS;
    
    int N_PATH;
    
    int N_LEFTSIBLING_WORD, N_LEFTSIBLING_POS;
    int N_RIGHTSIBLING_WORD, N_RIGHTSIBLING_POS;
    
    int N_POS_PATH;
    
    int N_PB_REL;
    
    private static final String NONE = "<NONE>";
    private static final String IDENTICAL = "<IDENTICAL>";
    
    private static final boolean REPLACE_NUMBERS = true;
    
    public NodeFeatureExtractor(EnglishFeatures lang, FeatureDeclaration decl){
        this.lang = lang;

        N_POSITION = decl.getFeatureIndex("N_POSITION");
        N_HOLDER_HEAD = decl.getFeatureIndex("N_HOLDER_HEAD");
        N_HOLDER_HEAD_POS = decl.getFeatureIndex("N_HOLDER_HEAD_POS");
        N_DEPREL = decl.getFeatureIndex("N_DEPREL");
        
        N_NEAREST_EXPR_WORD = decl.getFeatureIndex("N_NEAREST_EXPR_WORD");
        
        N_LEFT_WORD = decl.getFeatureIndex("N_LEFT_WORD");
        N_LEFT_POS = decl.getFeatureIndex("N_LEFT_POS");
        N_RIGHT_WORD = decl.getFeatureIndex("N_RIGHT_WORD");
        N_RIGHT_POS = decl.getFeatureIndex("N_RIGHT_POS");
        
        N_PATH = decl.getFeatureIndex("N_PATH");
        
        N_LEFTSIBLING_WORD = decl.getFeatureIndex("N_LEFTSIBLING_WORD");
        N_RIGHTSIBLING_WORD = decl.getFeatureIndex("N_RIGHTSIBLING_WORD");
        N_LEFTSIBLING_POS = decl.getFeatureIndex("N_LEFTSIBLING_POS");
        N_RIGHTSIBLING_POS = decl.getFeatureIndex("N_RIGHTSIBLING_POS");    

        N_POS_PATH = decl.getFeatureIndex("N_POS_PATH");
        N_PB_REL = decl.getFeatureIndex("N_PB_REL");
    }

    /**
     * Extracts features to a feature list.
     * 
     * @param p a pair consisting of two DepNodes
     * @param featureList the feature list.
     */
    public void extractFeatures(Triple<MPQAExpression, DepNode, MPQASentence> tr,
            FeatureList featureList) {
    	MPQAExpression expr = tr.first;
    	DepNode holdernode = tr.second;
    	MPQASentence sen = tr.third;
    	
        if(holdernode == null)
            throw new RuntimeException("depnode == null");
        
        DepNode exprnode = lang.getClosestInSpan(holdernode, expr.span);
        
        featureList.setFeature(N_POSITION, lang.getRelativePosition(holdernode, exprnode));

        // from predicate to argument
        featureList.setFeature(N_PATH, lang.findDepPath(exprnode, holdernode));
        featureList.setFeature(N_POS_PATH, lang.findDepPOSPath(exprnode, holdernode));
        if(holdernode != exprnode)
            featureList.setFeature(N_DEPREL, holdernode.relations[0]);
        else
            featureList.setFeature(N_DEPREL, IDENTICAL);
        
        featureList.setFeature(N_HOLDER_HEAD, normalize(holdernode.word));
        featureList.setFeature(N_HOLDER_HEAD_POS, holdernode.pos);

        featureList.setFeature(N_NEAREST_EXPR_WORD, normalize(exprnode.word));
        
        DepNode l = lang.getLeftmostChild(holdernode);
        if(l == null) {
            featureList.setFeature(N_LEFT_WORD, NONE);
            featureList.setFeature(N_LEFT_POS, NONE);
        } else {
            featureList.setFeature(N_LEFT_WORD, normalize(l.word));
            featureList.setFeature(N_LEFT_POS, l.pos);
        }
        DepNode r = lang.getRightmostChild(holdernode);
        if(r == null) {
            featureList.setFeature(N_RIGHT_WORD, NONE);
            featureList.setFeature(N_RIGHT_POS, NONE);
        } else {
            featureList.setFeature(N_RIGHT_WORD, normalize(r.word));
            featureList.setFeature(N_RIGHT_POS, r.pos);
        }

        DepNode ls = lang.getLeftSibling(holdernode);
        if(ls == null) {
            featureList.setFeature(N_LEFTSIBLING_WORD, NONE);
            featureList.setFeature(N_LEFTSIBLING_POS, NONE);
        } else {
            featureList.setFeature(N_LEFTSIBLING_WORD, normalize(ls.word));
            featureList.setFeature(N_LEFTSIBLING_POS, ls.pos);
        }
        
        DepNode rs = lang.getRightSibling(holdernode);
        if(rs == null) {
            featureList.setFeature(N_RIGHTSIBLING_WORD, NONE);
            featureList.setFeature(N_RIGHTSIBLING_POS, NONE);
        } else {
            featureList.setFeature(N_RIGHTSIBLING_WORD, normalize(rs.word));
            featureList.setFeature(N_RIGHTSIBLING_POS, rs.pos);
        }
        
        featureList.setFeature(N_PB_REL, getPBRel(holdernode, expr, sen));        
    }

    private DepNode getLeftmostDescendant(DepNode node) {
        ArrayList<DepNode> l = new ArrayList<DepNode>();
        node.collectSubtree(l);
        int min = Integer.MAX_VALUE;
        DepNode minNode = null;
        for(DepNode d: l)
            if(d.position < min) {
                min = d.position;
                minNode = d;
            }
        return minNode;
    }
    
    private DepNode getRightmostDescendant(DepNode node) {
        ArrayList<DepNode> l = new ArrayList<DepNode>();
        node.collectSubtree(l);
        int max = -1;
        DepNode maxNode = null;
        for(DepNode d: l)
            if(d.position > max) {
                max = d.position;
                maxNode = d;
            }
        return maxNode;
    }
    
    private String normalize(String word) {
        word = word.toLowerCase();
        if(REPLACE_NUMBERS)
            word = word.replaceAll("[0-9]", "0");
        return word;
    }
    
    private String getPBRel(DepNode cand, MPQAExpression expr, MPQASentence sen) {
    	for(PAStructure pa: sen.pas)
    		if(expr.span.contains(pa.pred)) {
    			for(int i = 0; i < pa.args.size(); i++)
    				if(pa.args.get(i) == cand) {
    					String label = pa.argLabels.get(i);
    					if(!label.startsWith("SU"))
    						return label;
    				}
    		}
    	return NONE;
    }
    
}
