package mpqa_structlearn.sequences;

import mpqa_structlearn.io.*;

import java.io.*;
import java.util.*;

import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.depgraph.*;

public class MPQASeqExtractor {

	public static SetMap<DepNode, MPQAExpression> createEndMap(MPQASentence sen) {
		SetMap<DepNode, MPQAExpression> out = new SetMap();
		for(MPQAExpression e: sen.exprs)
			out.put(e.span.get(e.span.size() - 1), e);
		return out;
	}
	
	private static String createIOBLabel(boolean starts, MPQAExpression e, boolean usePolarity) {
		if(e == null)
			return "O";
		//if(usePolarity && e.type.equals("os"))
		//	return "O";
		//return (starts? "B-": "I-") + (usePolarity? getPolString(e): e.type);

		String out = starts? "B-": "I-";
		if(!usePolarity)
		    out += e.type;
		else if(false)
		    out += getPolString(e);
		else
		    out += e.type + "-" + getPolString(e);
		return out;
	}
	
	private static String getPolString(MPQAExpression e) {
		if(e.type.equals("os"))
			return "obj";
		if(e.polarity == null)
			return "neu";
		String pol = e.polarity.trim();
		if(pol.equals(""))
			return "neu";
		if(pol.contains("neutral"))
			return "neu";
		if(pol.contains("positive"))
			return "pos";
		if(pol.contains("negative"))
			return "neg";
		if(pol.contains("both"))
			return "neu";
		throw new RuntimeException("Unhandled polarity label: |" + e.polarity + "|");
	}
	
	private static HashMap<DepNode, MPQAExpression> createStartMap(MPQASentence sen) {
		HashMap<DepNode, MPQAExpression> out = new HashMap();
		for(MPQAExpression expr: sen.exprs) {
			DepNode n = expr.span.get(0);
			if(out.containsKey(n)) {
				MPQAExpression other = out.get(n);
				if(expr.span.size() > other.span.size())
					out.put(n, expr);				
			} else
				out.put(n, expr);
		}
		return out;
	}
	
	/**
	 * Extracts the sequential training set for the expression detector,
	 * polarity classifier, and holder extractor. 
	 */
	public static void extractSequences(MPQAText text, PrintWriter pw,
			boolean usePolarity, boolean extractHolders, boolean printFeatures) throws IOException {

	    //usePolarity = true;

		for(MPQASentence sen: text.sentences) {
			HashMap<DepNode, MPQAExpression> startMap = createStartMap(sen);
			SetMap<DepNode, MPQAExpression> endMap = createEndMap(sen);
			MPQAExpression currentExpr = null;
			
			StringBuilder[] sbs = new StringBuilder[sen.dg.nodes.length];
			
			for(int i = 1; i < sen.dg.nodes.length; i++) {
				DepNode n = sen.dg.nodes[i];
				if(n.pos == null) {
					//if(!n.word.startsWith("<"))
					throw new RuntimeException("pos = null: " + n);
					//continue;
				}
				if(n.word.equals(".") && !n.pos.equals(".")) {
				    //System.out.println(sen.dg);
				    //System.err.println(n.word + " / " + n.pos);
				    //throw new RuntimeException("!!!");
				}

				boolean starts = false;
				
				if(currentExpr == null) {
					MPQAExpression e = startMap.get(n);
					if(e != null) {
						currentExpr = e;
						starts = true;
					}
				}
				String label = createIOBLabel(starts, currentExpr, usePolarity);
				if(currentExpr != null) {
					HashSet<MPQAExpression> es = endMap.get(n);
					if(es != null && es.contains(currentExpr))
						currentExpr = null;
				}
			
				String subjClue = sen.slClues.get(i);
				if(subjClue == null)
					subjClue = "_";
				String isDom = "_";
				String dom = "_";
				
				//pw.println(n.word + "\t" + n.pos + "\t" + n.lemma + "\t" + subjClue
				//		+ "\t" + isDom + "\t" + dom + "\t" + label);
				//pw.flush();
				
				sbs[i] = new StringBuilder();
				sbs[i].append(n.word);
				sbs[i].append("\t");
				if(printFeatures) {
					sbs[i].append(n.pos);
					sbs[i].append("\t");
					sbs[i].append(n.lemma);
					sbs[i].append("\t");
					sbs[i].append(subjClue);
					sbs[i].append("\t");
					sbs[i].append(isDom);
					sbs[i].append("\t");
					sbs[i].append(dom);
					sbs[i].append("\t");					
				}
				sbs[i].append(label);
			}
			/*pw.println();
			pw.flush();*/

			if(extractHolders) {
				MPQAExpression prev = null;
				for(MPQAExpression expr: sen.exprs) {

					if(prev != null && expr.span.get(0).position < prev.span.get(0).position)
						throw new RuntimeException("illegal expression order: " + sen);

					if(expr.nestedSource == null || expr.nestedSource.size() == 0) {
						for(int i = 1; i < sbs.length; i++)
							sbs[i].append("\t_");
					} else {
						MPQAAgent holder = expr.nestedSource.get(expr.nestedSource.size() - 1);
						if(holder.isImplicit) {
							sbs[1].append("\t<I>");
							for(int i = 2; i < sbs.length; i++)
								sbs[i].append("\t_");
						} else if(holder.isWriter) {
							sbs[1].append("\t<W>");
							for(int i = 2; i < sbs.length; i++)
								sbs[i].append("\t_");
						} else if(holder.instances == null || holder.instances.size() == 0) {
							for(int i = 1; i < sbs.length; i++)
								sbs[i].append("\t_");
						} else {
						    
						    /*MPQAAgentInstance fi = holder.instances.get(0);
							int start = fi.span.get(0).position;
							int end = fi.span.get(fi.span.size()-1).position;
							for(int i = 1; i < start; i++)
								sbs[i].append("\t_");
							sbs[start].append("\tB");
							for(int i = start + 1; i <= end; i++)
								sbs[i].append("\tI");
							for(int i = end + 1; i < sbs.length; i++)
								sbs[i].append("\t_");
								*/

						    boolean[] coveredTokens = new boolean[sbs.length];
						    for(MPQAAgentInstance ai: holder.instances) { 
						        if(ai.containingSentence != sen)
						            continue;
	                            for(DepNode n: ai.span)
	                                coveredTokens[n.position] = true;
						    }

						    boolean inside = false;
						    
						    for(int i = 1; i < coveredTokens.length; i++) {
						        if(coveredTokens[i]) {
						            if(inside)
						                sbs[i].append("\tI");
						            else
						                sbs[i].append("\tB");
						            inside = true;
						        } else {
						            sbs[i].append("\t_");
						            inside = false;
						        }

						    }
						}
					}					
					prev = expr;
				}
			}
			
			for(int i = 1; i < sbs.length; i++)
				pw.println(sbs[i]);
			pw.println();
			pw.flush();
			
		}

	}
	
	public static void main(String[] argv) {
        String mode = argv[0];
	    String dir = argv[1];
		String subjLexFile = argv[2];
		String docListFile = argv[3];
		String outFile = argv[4];
		
		try {
			ArrayList<String> selectedFiles = new ArrayList();
			
			BufferedReader br = Util.openFileReader(docListFile);
			String fn = br.readLine();
			while(fn != null) {
				fn = fn.replaceAll(".*docs/", "");
			
				if(fn.startsWith("wsj"))
				    fn = "xbank_" + fn;

				selectedFiles.add(fn);
				fn = br.readLine();
			}
			
			//Ax.readLines(docListFile, selectedFiles);

			System.out.println(selectedFiles);
			
			selectedFiles.remove("20020203/20.46.36-9539");

			Iterator<MPQAText> r;
			if(mode.equals("lk"))
			    r = new LKReader(dir, subjLexFile, selectedFiles);
			else if(mode.equals("ese"))
			    r = new MPQAReaderEPE(dir, subjLexFile, selectedFiles, null, null);
			else
			    throw new RuntimeException("illegal reader mode: must be lk or ese");
			
			PrintWriter pw = new PrintWriter(new OutputStreamWriter(new FileOutputStream(outFile), "UTF-8"));
			while(r.hasNext()) {
				MPQAText text = r.next();
				String shortSrc = text.source.substring(23);
				//if(!selectedFiles.contains(text.source) && !selectedFiles.contains(shortSrc))
				//	throw new RuntimeException("source = " + text.source);
				System.out.println(text.source);
				pw.println("<MPQA TEXT ID=\"" + shortSrc + "\">");
				extractSequences(text, pw, false, false, true);
			}
			pw.close();
			
		} catch(Exception e) {
			e.printStackTrace();
		}
	}
    
}
