/*
 AnnotatedText.java
 
 Created on Nov 30, 2005 by Richard Johansson (richard@cs.lth.se).
 
 $Log: AnnotatedText.java,v $
 Revision 1.3  2011-07-11 11:54:15  johansson
 Trim punctuation.

 Revision 1.2  2010-06-07 10:54:51  johansson
 Major update.

 Revision 1.1  2009/01/16 10:04:53  johansson
 Added to brenta repository.

 Revision 1.11  2007/08/10 13:29:32  richard
 Tests.

 Revision 1.10  2006/12/22 15:43:29  richard
 Added replaceAllContext.

 Revision 1.9  2006/12/20 09:23:42  richard
 Text replacing methods.

 Revision 1.8  2006/12/15 10:30:31  richard
 Inserts, deletes, and replaces text and modifies the corresponding spans.

 Revision 1.7  2006/09/07 17:51:19  richard
 New revision.

 Revision 1.6  2006/04/11 08:10:13  richard
 Added documentation.
 
 Revision 1.5  2006/04/07 13:58:14  richard
 Added a constructor.
 
 Revision 1.4  2006/04/05 12:09:13  richard
 Added get/setProperty.
 
 Revision 1.3  2006/03/29 13:54:51  richard
 Added colored HTML output.
 
 Revision 1.2  2005/12/02 13:31:32  richard
 Added tokens and lemmas.
 
 Revision 1.1  2005/11/30 13:29:43  richard
 Added the file.
 
 
 */
package se.lth.cs.nlp.nlputils.annotations;

import java.util.*;
import java.util.regex.*;
import se.lth.cs.nlp.nlputils.core.*;

/**
 * A text with a list of annotation layers.
 * 
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class AnnotatedText {
    
    public String text;
    public HashMap<String, Object> properties;
    public List<AnnotationLayer> layers;
    
    public String[] tokens;
    public String[] poss;
    public String[] lemmas;
    
    public int[] tokenStarts;
    public int[] tokenEnds;
    
    /**
     * Creates a new AnnotatedText.
     */
    public AnnotatedText() {
        properties = new HashMap<String, Object>();
        layers = new ArrayList<AnnotationLayer>();
    }
    
    public AnnotatedText clone() {
        AnnotatedText out = new AnnotatedText();
        out.text = text;
        if(tokens != null)
            out.tokens = tokens.clone();
        if(poss != null)
            out.poss = poss.clone();
        if(lemmas != null)
            out.lemmas = lemmas.clone();
        if(tokenStarts != null)
            out.tokenStarts = tokenStarts.clone();
        if(tokenEnds != null)
            out.tokenEnds = tokenEnds.clone();        
        out.properties = (HashMap) properties.clone();
        out.layers = new ArrayList();
        for(AnnotationLayer l: layers)
            out.layers.add(l.clone());
        return out;
    }
    
    public String toString() {
        return "[AnnotatedText:\nText:\n" + text + "\nProperties:\n"
        + properties + "\nLayers:\n" + layers + "]";
    }
    
    public void splitLayers() {
    	ArrayList<AnnotationLayer> newLayers = new ArrayList();
    	for(AnnotationLayer l: layers)
    		newLayers.addAll(l.splitBySpanType());
    	layers = newLayers;
    }
    
    public AnnotationLayer getLayer(String name) {
    	for(AnnotationLayer l: layers)
    		if(name.equals(l.label))
    			return l;
    	return null;
    }
    
    /**
     * Gets a property of the annotated text.
     * 
     * @param name the name of the property to check.
     * @return the property, or null if not found.
     */
    public Object getProperty(String name) {
        return properties.get(name);
    }
    
    /**
     * Sets a property of the annotated text.
     * 
     * @param name the name of the property to set.
     * @param value the new value of the property.
     */
    public void setProperty(String name, Object value) {
        properties.put(name, value);
    }

    public void tokenize() {
        tokenize(new StringTokenizer(text));
    }
    
    public void tokenize(StringTokenizer tokenizer) {
        ArrayList<String> ts = new ArrayList<String>();
        while(tokenizer.hasMoreTokens())
            ts.add(tokenizer.nextToken());
        tokens = ts.toArray(new String[0]);
    }
    
    private void findStartEndIndices() {
    	if(tokens == null)
            throw new IllegalStateException("Text must be tokenized!");
    	tokenStarts = new int[tokens.length];
    	tokenEnds = new int[tokens.length];
    	
    	int position = 0;
    	
    	for(int i = 0; i < tokens.length; i++) {    		
    		tokenStarts[i] = getNext(position);
    		tokenEnds[i] = getEnd(tokens[i], tokenStarts[i]);
    		
    		/*if(tokenEnds[i] == -1) { // special case
    			tokenStarts[i] = text.indexOf(tokens[i]);
    			tokenEnds[i] = tokenStarts[i] + tokens[i].length();
    		}*/
    		
    		position = tokenEnds[i];
    	}
    	
    	//System.out.println(Arrays.toString(tokenStarts));
    	//System.out.println(Arrays.toString(tokenEnds));    	
    }
    
    private int getNext(int position) {
    	int l = text.length();
    	while(position < l && Character.isWhitespace(text.charAt(position)))
			position++;
		return position;
	}
    
    private int getEnd(String token, int position) {
		int len = Math.min(text.length() - position, token.length());
		String t2 = text.substring(position, position + len);
		
		if(!token.equals(t2)) {
			
			/*if(token.matches("``|''") && t2.startsWith("\"")) {
				len = 1;
				t2 = "\"";
			} else if(token.equals("`") && t2.startsWith("'")) {
				len = 1;
				t2 = "'";
			} else if(token.replaceAll("`", "'").equals(t2.replaceAll("`", "'"))) {
				len = token.length();
			} else {*/
				//System.out.println(line);
				System.out.println(getContext(position));
				System.out.println("position = " + position);
				System.out.println("ss = " + text.substring(position));
				throw new RuntimeException(token + " != " + t2);
		    //}
		}
		//System.out.println(token);        
		position += len;
		return position;
	}

	private String getContext(int pos) {
		int start = Math.max(0, pos - 100);
		int end = Math.min(text.length(), pos + 100);
		return text.substring(start, end);
	}
    
    public void findTokenIndices() {
        if(tokens == null)
            throw new IllegalStateException("Text must be tokenized!");
        findStartEndIndices();
        for(AnnotationLayer l: layers)
        	//l.findTokenIndices(tokens);
        	l.findTokenIndicesNew(tokens, text, tokenStarts, tokenEnds);
    }
    
    public void findCharIndices() {
        if(tokens == null)
            throw new IllegalStateException("Text must be tokenized!");
        for(AnnotationLayer l: layers)
            l.findCharIndices(tokens);
    }
    
    public void insertTextAt(String inserted, int index) {
        if(tokens != null || poss != null || lemmas != null)
            throw new IllegalStateException("Cannot modify a tokenized text!");
        if(index < 0 || index > text.length())
            throw new IllegalArgumentException("Cannot insert outside text!");
        StringBuilder sb = new StringBuilder(text);
        sb.insert(index, inserted);
        text = sb.toString();
        for(AnnotationLayer l: layers)
            l.insertAt(inserted.length(), index);
    }

    public void deleteTextAt(int start, int end) {
        if(tokens != null || poss != null || lemmas != null)
            throw new IllegalStateException("Cannot modify a tokenized text!");        
        if(start > end)
            throw new IllegalArgumentException("start > end!");
        if(start < 0 || end > text.length())
            throw new IllegalArgumentException("Cannot delete outside text!");
        StringBuilder sb = new StringBuilder(text);
        sb.delete(start, end);
        text = sb.toString();
        for(AnnotationLayer l: layers)
            l.deleteAt(start, end);
    }

    public void replaceTextAt(String replacement, int start, int end) {
        if(tokens != null || poss != null || lemmas != null)
            throw new IllegalStateException("Cannot modify a tokenized text!");        
        if(start > end)
            throw new IllegalArgumentException("start > end!");
        if(start < 0 || end > text.length())
            throw new IllegalArgumentException("Cannot replace outside text!");
        StringBuilder sb = new StringBuilder(text);
        sb.replace(start, end, replacement);
        text = sb.toString();
        for(AnnotationLayer l: layers) {
            l.deleteAt(start, end);
            l.insertAt(replacement.length(), start);
        }
    }

    private Pair<TreeMap<Integer, ArrayList<Integer>>, StringBuilder> createReplacement(String repl) {
        StringBuilder sb = new StringBuilder();

        TreeMap<Integer, ArrayList<Integer>> tm
            = new TreeMap<Integer, ArrayList<Integer>>();
        
        for(int i = 0; i < repl.length(); i++) {
            if(repl.charAt(i) == '\\') {
                sb.append(repl.charAt(i + 1));
                i++;
            } else if(repl.charAt(i) == '$') {
                int ref = repl.charAt(i + 1) - '0';
                if(ref < 0 || ref > 9)
                    throw new IllegalArgumentException("Ref id must be 0-9!");
                ArrayList<Integer> is = tm.get(-sb.length());
                if(is == null) {
                    is = new ArrayList<Integer>(1);
                    tm.put(-sb.length(), is);
                }
                is.add(0, ref);
                i++;
            } else
                sb.append(repl.charAt(i));
        }

        return new Pair<TreeMap<Integer, ArrayList<Integer>>, StringBuilder>(tm, sb);
    }

    private int getNBrackets(String regex) {
        int n = 0;
        for(int i = 0; i < regex.length(); i++) {
            if(regex.charAt(i) == '\\')
                i++;
            else if(regex.charAt(i) == '(')
                n++;
        }
        return n;
    }
    
    public void replaceAll(String regex, String replacement) {
        Pattern pattern = Pattern.compile("()(" + regex + ")()");
        replaceInternal(pattern, 2, getNBrackets(regex)+3,
                replacement, true);
    }
    
    public void replaceAllContext(String regex, String lc, String rc, String replacement) {
        Pattern pattern = Pattern.compile("(" + lc + ")(" + regex + ")(" + rc + ")");
        int n1 = getNBrackets(lc);
        int n2 = getNBrackets(regex);
        replaceInternal(pattern, n1 + 2, n1 + n2 + 3,
                replacement, true);
    }
    
    public void replaceFirst(String regex, String replacement) {
        Pattern pattern = Pattern.compile("()(" + regex + ")()");
        replaceInternal(pattern, 2, getNBrackets(regex)+3, 
                replacement, false);
    }
    
    private void replaceInternal(Pattern pattern, int middleBracketStart,
            int rightBracketStart,
            String replacement, boolean all) {
        Matcher matcher = pattern.matcher(text);
        Pair<TreeMap<Integer, ArrayList<Integer>>, StringBuilder> p
            = createReplacement(replacement);
        
        StringBuffer ntext = new StringBuffer();
        int offset = 0;        
        while(matcher.find()) {
            
            StringBuilder sb = new StringBuilder();
            
            sb.append(matcher.group(1));
            int len_start = sb.length();
            
            sb.append(p.right);
            
            for(Map.Entry<Integer, ArrayList<Integer>> e: p.left.entrySet()) {
                for(Integer g: e.getValue())
                    sb.insert(len_start-e.getKey(), matcher.group(g + middleBracketStart));
            }

            sb.append(matcher.group(rightBracketStart));
            
            //System.out.println("match = " + matcher.group());
            //System.out.println("sb = " + sb);

            int len = sb.length();
            int len_end = matcher.group(rightBracketStart).length();
            int len_sub = len - len_start - len_end;
            
            //System.out.println("len_start = " + len_start);
            //System.out.println("len = " + len);
            //System.out.println("len_end = " + len_end);
            
            int startpos = matcher.start() + offset + len_start;
            int endpos = matcher.end() + offset - len_end; 
            
            for(AnnotationLayer l: layers) {
                l.insertAt(len_sub, startpos);
                l.deleteAt(startpos + len_sub,
                           endpos + len_sub);
            }
            offset += len - (matcher.end() - matcher.start());
            matcher.appendReplacement(ntext, Matcher.quoteReplacement(sb.toString()));
            if(!all)
                break;
        }
        matcher.appendTail(ntext);        
        text = ntext.toString();
    }

    public void trimPunctuationTokens() {
    	for(AnnotationLayer l: layers)
    		l.trimPunctuationTokens(this);
    }
    
    public void trimSpaces() {
        for(AnnotationLayer l: layers)
            l.trimSpaces(this);
    }

    public void removeWordInternalAnnotations() {
        for(AnnotationLayer l: layers)
            l.removeWordInternalAnnotations(this);
    }
    
    public void nestedLayers() {
        for(int i = 0; i < layers.size(); i++) {
            AnnotationLayer l = layers.get(i);
            Collections.sort(l.spans, Span.NestingOrder.instance());
            AnnotationLayer l2 = nextLevelLayer(l);
            if(l2 != null)
                layers.add(i+1, l2);
        }
    }
    
    private AnnotationLayer nextLevelLayer(AnnotationLayer<Span> l) {
        AnnotationLayer l2 = new AnnotationLayer();
        for(int i = 0; i < l.spans.size() - 1; i++) {
            Span s = l.spans.get(i);
            for(int j = i + 1; j < l.spans.size(); j++) {
                Span s2 = l.spans.get(j);
                if(s2.start >= s.end)
                    break;
                if(!l2.spans.contains(s2))
                    l2.add(s2);                
            }
        }
        l.spans.removeAll(l2.spans);
        if(l2.spans.size() == 0)
            return null;
        else
            return l2;
    }
    
//    public void replaceAll(Pattern pattern, String replacement) {
//        replaceInternal(pattern, 0, replacement, true);
//    }
//    
//    public void replaceFirst(Pattern pattern, String replacement) {
//        replaceInternal(pattern, 0, replacement, false);
//    }
    
    /* Tests deleting. */

    private static void testIns(String ins, int index, int start, int end) {
        AnnotatedText testtext = new AnnotatedText();
        testtext.text = "01234567890123456789";
        AnnotationLayer testlayer = new AnnotationLayer();        
        testtext.layers.add(testlayer);
        Span s = new Span(2, 6, "s1");
        testlayer.spans.add(s);
        testtext.insertTextAt(ins, index);
        if(s.start != start)
            throw new RuntimeException("when inserting " + ins + " at " + index + ": start is incorrect: should be " + start + ", was " + s.start);
        if(s.end != end)
            throw new RuntimeException("when inserting " + ins + " at " + index + ": end is incorrect: should be " + end + ", was " + s.end);        
    }
    
    private static void testIns2(String ins, int index, int start, int end) {
        AnnotatedText testtext = new AnnotatedText();
        testtext.text = "0123";
        AnnotationLayer testlayer = new AnnotationLayer();        
        testtext.layers.add(testlayer);
        Span s = new Span(1, 2, "s1");
        testlayer.spans.add(s);
        testtext.insertTextAt(ins, index);
        if(s.start != start)
            throw new RuntimeException("when inserting " + ins + " at " + index + ": start is incorrect: should be " + start + ", was " + s.start);
        if(s.end != end)
            throw new RuntimeException("when inserting " + ins + " at " + index + ": end is incorrect: should be " + end + ", was " + s.end);        
    }
    
    private static void testDel(int rstart, int rend, int start, int end) {
        AnnotatedText testtext = new AnnotatedText();
        testtext.text = "01234567890123456789";
        AnnotationLayer testlayer = new AnnotationLayer();        
        testtext.layers.add(testlayer);
        Span s = new Span(2, 6, "s1");
        testlayer.spans.add(s);
        testtext.deleteTextAt(rstart, rend);
        if(s.start != start)
            throw new RuntimeException("when deleting from " + rstart + " to " + rend + ": start is incorrect: should be " + start + ", was " + s.start);
        if(s.end != end)
            throw new RuntimeException("when deleting from " + rstart + " to " + rend + ": end is incorrect: should be " + end + ", was " + s.end);
    }

    private static void testRepl(String string, int[][] prespans, String regex, String repl, int[][] postspans, boolean all) {
        AnnotatedText testtext = new AnnotatedText();
        testtext.text = string;
        AnnotationLayer<Span> testlayer = new AnnotationLayer();        
        testtext.layers.add(testlayer);

        for(int i = 0; i < prespans.length; i++)
            testlayer.add(new Span(prespans[i][0], prespans[i][1], "s" + i));

        System.out.println("*** At start of testRepl(" + regex + ", " + repl + ")");
        RawXMLAnnotation.instance().printLayers(testtext);
        
        String xx = null;
        if(all) {
            xx = testtext.text.replaceAll(regex, repl);
            testtext.replaceAll(regex, repl);
        } else {
            xx = testtext.text.replaceFirst(regex, repl);
            testtext.replaceFirst(regex, repl);
        }

        if(!xx.equals(testtext.text))
            throw new RuntimeException("when replacing " + regex + " with " + repl + ": replacement went wrong: " + testtext.text + " != " + xx);        

        //System.out.println(testtext);
        
        RawXMLAnnotation.instance().printLayers(testtext);

        for(int i = 0; i < postspans.length; i++) {
            Span s = testlayer.spans.get(i);
            if(s.start != postspans[i][0])
                throw new RuntimeException("when replacing " + regex + " with " + repl + ": start is incorrect: should be " + postspans[i][0] + ", was " + s.start);                
            if(s.end != postspans[i][1])
                throw new RuntimeException("when replacing " + regex + " with " + repl + ": end is incorrect: should be " + postspans[i][1] + ", was " + s.end);                
        }
    }
    
    private static void testReplC(String string, int[][] prespans, 
            String lc,
            String regex, 
            String rc,
            String repl, int[][] postspans, boolean all) {
        AnnotatedText testtext = new AnnotatedText();
        testtext.text = string;
        AnnotationLayer<Span> testlayer = new AnnotationLayer();        
        testtext.layers.add(testlayer);

        for(int i = 0; i < prespans.length; i++)
            testlayer.add(new Span(prespans[i][0], prespans[i][1], "s" + i));

        System.out.println("*** At start of testReplC(" + regex + ", " + repl + ")");
        RawXMLAnnotation.instance().printLayers(testtext);
        
        //String xx = null;
        if(all) {
            //xx = testtext.text.replaceAll(regex, repl);
            testtext.replaceAllContext(regex, lc, rc, repl);
        } else {
            //xx = testtext.text.replaceFirst(regex, repl);
            testtext.replaceFirst(regex, repl);
        }

        //if(!xx.equals(testtext.text))
        //    throw new RuntimeException("when replacing " + regex + " with " + repl + ": replacement went wrong: " + testtext.text + " != " + xx);        

        System.out.println(testtext);
        
        RawXMLAnnotation.instance().printLayers(testtext);

        for(int i = 0; i < postspans.length; i++) {
            Span s = testlayer.spans.get(i);
            if(s.start != postspans[i][0])
                throw new RuntimeException("when replacing " + regex + " with " + repl + ": start is incorrect: should be " + postspans[i][0] + ", was " + s.start);                
            if(s.end != postspans[i][1])
                throw new RuntimeException("when replacing " + regex + " with " + repl + ": end is incorrect: should be " + postspans[i][1] + ", was " + s.end);                
        }
    }
    
    static void testHTTP() {
        AnnotatedText testtext = new AnnotatedText();
        testtext.text = "testar/testar http://www.df.lth.se/~richardj/test abc/def / ghi";
        AnnotationLayer testlayer = new AnnotationLayer();        
        testtext.layers.add(testlayer);
        //testtext.replaceAllContext("\\/", "http:[^ ]*?", "", "__SLASH__");
        testtext.replaceAll("(http:[^ ]*?)\\/", "$1__SLASH__");
        System.out.println(testtext.text);
    }
    
    public static void main(String[] argv) {

        /*testIns("xx", 0, 4, 8);        
        testIns("xx", 1, 4, 8);
        testIns("xx", 2, 2, 8);
        testIns("xx", 3, 2, 8);
        testIns("xx", 5, 2, 8);
        testIns("xx", 6, 2, 6);
        testIns("xx", 7, 2, 6);
        
        testDel(0, 1, 1, 5);
        testDel(1, 2, 1, 5);
        testDel(1, 3, 1, 4);
        testDel(2, 3, 2, 5);
        testDel(5, 6, 2, 5);
        testDel(5, 7, 2, 5);
        testDel(1, 6, 1, 1);
        testDel(6, 7, 2, 6);
        testDel(7, 8, 2, 6);
        testDel(2, 6, 2, 2);
        testDel(2, 7, 2, 2);
        testDel(1, 7, 1, 1);*/
        
        //testRepl("23", "xx", 2, 6);
        //testRepl("45", "xx", 2, 6);        

    	/*testHTTP();

        if(true)
            System.exit(0);*/

    	/*
    	
        String test = "0123456789012345789";
        int[][] s0 = new int[][] {{2,6},{10,11}};
        
        testRepl(test, s0, "0", "X", s0, true);        
        testRepl(test, s0, "1", "X", s0, true);       
        testRepl(test, s0, "2", "X", s0, true);
        testRepl(test, s0, "5", "X", s0, true);
        testRepl(test, s0, "6", "X", s0, true);
        testRepl(test, s0, "7", "X", s0, true);

        testIns2("XX", 1, 1, 4);
        
        testRepl("90123", new int[][] {{1,2}}, "0", "XX", new int[][] {{1,3}}, true);      
        testRepl(test, new int[][] {{2,6},{10,11}}, "0", "XX", new int[][] {{3,7},{11,12}}, false);
        testRepl(test, new int[][] {{2,6},{10,11},{14,18}}, "0", "XX", new int[][] {{3,7},{11,13},{16,20}}, true);
        //        testRepl("1", "XX", 3, 7, 11, 12, true);        
//        testRepl("2", "XX", 2, 7, 11, 12, true);
//        testRepl("5", "XX", 2, 7, 11, 12, true);
//        testRepl("6", "XX", 2, 6, 11, 12, true);
//        testRepl("7", "XX", 2, 6, 11, 12, true);
        
        testRepl(test, new int[][] {{2,6},{10,11},{11,15}}, "(2)(.)(4)", "y$3xz$2$1y", new int[][]{{2,10},{14,15},{15,23}}, true);

        testReplC("90123", new int[][] {{1,2}}, "9", "0", "1", "X", new int[][] {{1,2}}, true);      
        testReplC("90123", new int[][] {{2,4}}, "((9)(0))", "((((1)))((2)))", "3", "$5X$2", new int[][] {{2,5}}, true);      
        
        //testReplC(test, s0, "0", "1", "2", "X", s0, true);

         */

        String test = "0123456789012345789";
        int[][] s0 = new int[][] {{2,6}};
        int[][] s1 = new int[][] {{2,4}};
        
        testRepl(test, s0, "12", "X", s1, true);        
    	
        
        
    }
    
}
