/*
 AnnotationLayer.java
 
 Copyright (C) 2005, Richard Johansson (richard@cs.lth.se).
 
 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.
 
 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.
 
 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
 Created on Nov 25, 2005 by Richard Johansson (richard@cs.lth.se).
 
 $Log: AnnotationLayer.java,v $
 Revision 1.3  2011-07-11 11:53:40  johansson
 Uses the new PRFStats.

 Revision 1.2  2010-06-07 10:54:51  johansson
 Major update.

 Revision 1.1  2009/01/16 10:04:53  johansson
 Added to brenta repository.

 Revision 1.16  2007/10/30 15:14:42  richard
 Clear.

 Revision 1.15  2007/09/05 14:14:32  richard
 Rewrote compare.

 Revision 1.14  2007/08/22 10:19:54  richard
 Added overlaps.

 Revision 1.13  2007/08/17 10:16:48  richard
 RemoveSpansByLabel. Patched findTokenIndices partly.

 Revision 1.12  2007/08/13 12:24:44  richard
 Properties. Name constructor.

 Revision 1.11  2006/12/20 09:21:58  richard
 Changed semantics of insertAt and deleteAt.

 Revision 1.10  2006/12/15 10:30:31  richard
 Inserts, deletes, and replaces text and modifies the corresponding spans.

 Revision 1.9  2006/11/13 10:06:04  richard
 Iterable over spans.

 Revision 1.8  2006/09/07 17:51:19  richard
 New revision.

 Revision 1.7  2006/04/11 08:10:13  richard
 Added documentation.
 
 Revision 1.6  2006/04/05 12:09:44  richard
 Added findTokenIndices and findCharIndices.
 
 Revision 1.5  2006/03/29 13:55:45  richard
 Added colored HTML output.
 
 Revision 1.4  2005/11/30 12:51:58  richard
 Added removeEmptySpans.
 
 Revision 1.3  2005/11/30 12:42:34  richard
 Added the toString and printDetailed methods.
 
 Revision 1.2  2005/11/29 13:51:10  richard
 Implemented layer comparison.
 
 Revision 1.1  2005/11/28 15:41:00  richard
 Added the file.
 
 
 */
package se.lth.cs.nlp.nlputils.annotations;

import java.util.*;
import java.util.regex.*;

import se.lth.cs.nlp.nlputils.core.PRFStats;

/**
 * Annotation layer.
 * 
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class AnnotationLayer<T extends Span> implements Iterable<T> { 
    
    public String label;
    public ArrayList<T> spans = new ArrayList<T>();
    
    /**
     * Properties.
     */
    public HashMap<String, Object> properties;
    
    public AnnotationLayer() {        
    }
    
    public AnnotationLayer(String label) {        
        this();
        this.label = label;
    }
    
    public AnnotationLayer clone() {
    	AnnotationLayer out = new AnnotationLayer();
    	out.label = label;
    	out.spans = new ArrayList();
    	for(Span s: spans)
    		out.spans.add(s.clone());
    	if(properties != null)
    		out.properties = (HashMap) properties.clone();
    	return out;
    }
    
    public List<AnnotationLayer> splitBySpanType() {
    	HashMap<String, AnnotationLayer> types = new HashMap();
    	for(Span s: spans) {
    		AnnotationLayer l = types.get(s.label);
    		if(l == null) {
    			l = new AnnotationLayer(s.label);
    			types.put(s.label, l);
    		}
    		l.spans.add(s);
    	}    	
    	TreeSet<String> keys = new TreeSet(types.keySet());
    	ArrayList<AnnotationLayer> out = new ArrayList();
    	for(String k: keys)
    		out.add(types.get(k));    	
    	return out;
    }
    
    /**
     * Compares this layer to another layer.
     * 
     * TODO: this method has quadratic complexity.
     * 
     * @param other the layer to compare with.
     * @return the result of the comparison.
     */
    public CompareResult compare(AnnotationLayer other) {
        
        CompareResult out = new CompareResult();
        out.nInGold = spans.size();
        out.nPredicted = other.spans.size();
        
        ArrayList<Span> copy1 = new ArrayList<Span>(spans);
        ArrayList<Span> copy2 = new ArrayList<Span>(other.spans);

        Collections.sort(copy1, Span.LengthOrder.instance());
        Collections.sort(copy2, Span.LengthOrder.instance());
        
        //System.out.println("*** COMPARE: ***");
        //System.out.println("l1 = " + this);
        //System.out.println("l2 = " + other);
        
        for(Iterator<Span> it = copy1.iterator(); it.hasNext(); ) {
            Span s = it.next();
            if(removeMatch(copy2, s)) {
                it.remove();
                out.nCorrect++;
                out.weightedPScore++;
                out.weightedRScore++;
                //System.out.println("Exact match: " + s);
            }
        }
        
        for(Iterator<Span> it = copy2.iterator(); it.hasNext(); ) {
            Span s = it.next();
            
            double bestf = 0;
            Span bestMatch = null;
            for(Span s2: copy1) {
                int intersect = s.intersectionLength(s2);
                double p = (double) intersect / s.length();
                double r = (double) intersect / s2.length();
                double f = p+r == 0? 0: 2*p*r/(p+r);
                if(f > bestf) {
                    bestf = f;
                    bestMatch = s2;
                }
            }
            if(bestMatch != null) {
                copy1.remove(bestMatch);
                int intersect = s.intersectionLength(bestMatch);
                out.weightedPScore += (double) intersect / s.length();
                out.weightedRScore += (double) intersect / bestMatch.length();
                out.nPartialMatches++;
                it.remove();
                //System.out.println("Partial match: " + s + " / " + bestMatch);
            }

            //if(findPartialMatch(copy1, s))
            //    out.nPartialMatches++;
        }

        //System.out.println("Unmatched gold spans: " + copy1);
        //System.out.println("Unmatched output spans: " + copy2);
        
        return out;
    }
    
    public double[] compareApprox(AnnotationLayer other) {
    	double[] out = new double[9];

    	// TODO assumes sequential
    	
        ArrayList<Span> copy1 = new ArrayList<Span>(spans);
        ArrayList<Span> copy2 = new ArrayList<Span>(other.spans);

        Collections.sort(copy1, Span.ByLeftOrder.instance());
        Collections.sort(copy2, Span.ByLeftOrder.instance());

        int nInGold = copy1.size();
        int nGuesses = copy2.size();
        int nExact = 0;

        int nOverlap = 0;
        
        if(nInGold != spans.size())
        	throw new RuntimeException("e1");
        if(nGuesses != other.spans.size())
        	throw new RuntimeException("e1");
        
        double propFoundCorrect = 0;
        double propGuessCorrect = 0;
        
        ListIterator<Span> i2 = copy2.listIterator();

        if(i2.hasNext()) {
        	Span s2 = i2.next();
        	for(Span s1: copy1) {

        		while(s2.tokenEnd <= s1.tokenStart) {
        			if(!i2.hasNext()) {
        				s2 = null;
        				break;
        			}
        			s2 = i2.next();
        		}

        		if(s2 == null)
        			break;

        		while(s2.tokenStart < s1.tokenEnd) {

        			if(s1.label.equals(s2.label) 
        			   && s1.tokenStart == s2.tokenStart && s1.tokenEnd == s2.tokenEnd)
        				nExact++;        				

        			double overlap = s1.intersectionLengthTokens(s2);
        			if(overlap > 0) {
        				propFoundCorrect += overlap / s1.lengthTokens();
        				propGuessCorrect += overlap / s2.lengthTokens();
        				nOverlap++;
        			}
        			
        			if(!i2.hasNext()) {
        				s2 = null;
        				break;
        			}
        			s2 = i2.next();        		
        		}

        		if(s2 == null)
        			break;

        	}
        }
        //System.out.println(propGuessCorrect);
        //System.out.println(propFoundCorrect);
        
        double p = propGuessCorrect / nGuesses;
        double r = propFoundCorrect / nInGold;
        double f = 2*p*r/(p+r);
        out[0] = p;
        out[1] = r;
        out[2] = f;
        
        out[3] = nExact;
        out[4] = nGuesses;
        out[5] = nInGold;
        out[6] = propGuessCorrect;
        out[7] = propFoundCorrect;

        out[8] = nOverlap;
        
    	return out;
    }

    public PRFStats compareApproxPRF(AnnotationLayer other) {
    	// TODO assumes sequential
    	
        ArrayList<Span> copy1 = new ArrayList<Span>(spans);
        ArrayList<Span> copy2 = new ArrayList<Span>(other.spans);

        Collections.sort(copy1, Span.ByLeftOrder.instance());
        Collections.sort(copy2, Span.ByLeftOrder.instance());

        PRFStats out = new PRFStats();
        out.nInGold = copy1.size();
        out.nGuesses = copy2.size();

        // TODO this one is simplistic quadratic-time...
        for(Span s1: copy1) {
        	Span maxMatch = null;
        	int maxOverlap = 0;
        	for(Span s2: copy2) {        	
        		int overlap = s1.intersectionLengthTokens(s2);
    			if(overlap > maxOverlap) {
    				maxMatch = s2;
    				maxOverlap = overlap;
    			}
        	}        	
        	if(maxOverlap > 0) {
        		copy2.remove(maxMatch);        		
        		if(s1.lengthTokens() == maxOverlap && s1.lengthTokens() == maxMatch.lengthTokens())
        			out.nCorrect++;
        		out.propFound += (double) maxOverlap / s1.lengthTokens();
        		out.propGuessed += (double) maxOverlap / maxMatch.lengthTokens();
        		out.nOverlap++;
        	}
        }
        
    	return out;
    }

    
    private boolean removeMatch(ArrayList<Span> l, Span s) {
        for(Iterator<Span> it = l.iterator(); it.hasNext(); ) {
            Span s2 = it.next();
            if(s2.matches(s)) {
                it.remove();
                return true;
            }
        }
        return false;
    }
    
    private boolean findPartialMatch(ArrayList<Span> l, Span s) {
        for(Span s2: l)
            if(s.matchesPartially(s2))
                return true;
        return false;
    }
    
    public String toString() {
        String out = "(AnnotationLayer(" + label + ")";
        if(properties != null)
            out += ", properties = " + properties;
        out += ": " + spans + ")";
        return out;
    }
    
    /**
     * Detailed printout.
     * 
     * @param text
     */
    public void printDetailed(String text) {
        for(int i = 0; i < text.length(); i++) {
            System.out.print(i + ": " + text.charAt(i));
            for(Span sp: spans)
                if(sp.start == i)
                    System.out.print(" " + sp);
            System.out.println();
        }
    }
    
    /**
     * Removes all empty spans from the layer.
     * 
     * @return the removed spans.
     */
    public ArrayList<T> removeEmptySpans() {
        ArrayList<T> out = new ArrayList<T>();
        for(Iterator<T> it = spans.iterator(); it.hasNext(); ) {
            T s = it.next();
            if(s.start < 0 || s.end < 0 || s.start == s.end) {
                out.add(s);
                it.remove();
            }
        }
        return out;
    }
    
    public ArrayList<T> removeSpansByLabel(String regex) {
        Pattern p = Pattern.compile(regex);
        ArrayList<T> out = new ArrayList<T>();
        for(Iterator<T> it = spans.iterator(); it.hasNext(); ) {
            T s = it.next();
            if(s.label != null && p.matcher(s.label).matches()) {
                out.add(s);
                it.remove();
            }
        }
        return out;
    }
    
    /**
     * Sets the token indices for all spans, assuming that the character
     * indices are already set.
     * 
     * @param tokens the tokens the indices refer to.
     */
    public void findTokenIndices(String[] tokens) {
        /* TODO effektivare */

    	/* TODO denna är buggig om ett spann börjar inne i ett ord. */

        for(Span s: spans)
            s.tokenEnd = s.tokenStart = -1;
        
        int position = 0;
        
        for(int i = 0; i < tokens.length; i++) {
            
            int endPos = position + tokens[i].length();
            
            for(int j = 0; j < spans.size(); j++) {
                Span s = spans.get(j);
                
                if(s.start == s.end || (s.tokenStart != -1 && s.tokenEnd != -1)) // tidigare tokenEnd
                    continue;
                
                if(s.tokenStart == -1) {
                    if(s.start < position) {
                        s.tokenStart = i - 1;
                    } else if(s.start == position) {
                        s.tokenStart = i;
                    }
                }
                
                if(s.end <= endPos)
                    s.tokenEnd = i + 1;
                
            }
            
            position = endPos + 1;
        }
        
    }
    
    public void findTokenIndicesNew(String[] tokens, String text, int[] starts, int[] ends) {
    	for(Span s: this) {				
			int st = s.start;
			while(st < s.end && st < text.length() && Character.isWhitespace(text.charAt(st)))
				st++;
			int en = s.end;
			while(en > st && en < text.length() && Character.isWhitespace(text.charAt(en - 1)))
				en--;

			s.tokenStart = findIndexBSearch(starts, st, true);

			//s.tokenEnd = findIndexBSearch(ends, s.end) + 1;
			if(st < en && s.tokenStart < starts.length) {
				s.tokenEnd = findIndexBSearch(ends, en, false) + 1;
				if(s.tokenEnd > tokens.length)
					s.tokenEnd = tokens.length;
			}
			else
				s.tokenEnd = s.tokenStart;
		}
    }
    
    
    private static int findIndexBSearch(int[] arr, int c, boolean isStart) {

		if(arr.length == 0)
			return -1;
		if(c < arr[0]) {
			if(isStart)
				return -1;
			else
				return 0;
		}

		if(c > arr[arr.length - 1]) {
			if(isStart)
				return arr.length - 1;
			else
				return -1;
		}
		//return arr.length - 1; // ???
		
		int low = 0, high = arr.length - 1, mid = (low + high) / 2;				
		
		while(low < mid && mid < high) {
			if(arr[mid] == c)
				return mid;
			if(arr[low] == c)
				return low;
			if(arr[high] == c)
				return high;
			if(c < arr[mid])
				high = mid;
			else
				low = mid;
			mid = (low + high) / 2;
		}

		//System.out.println("2. low = " + low + ", mid = " + mid + ", high = " + high);
		
		if(low < high) {
			if(arr[low] == c)
				return low;
			if(arr[high] == c)
				return high;

			// TODO testa
			if(isStart)
				return low;
			else
				return high;
		}

		return mid;
	}
    
    
    /**
     * Sets the character indices for all spans, assuming that the token
     * indices are already set.
     * 
     * @param tokens the tokens the indices refer to.
     */
    public void findCharIndices(String[] tokens) {
        /* TODO effektivare */
        
        int[] starts = new int[tokens.length];
        int[] ends = new int[tokens.length];
        
        int position = 0;
        for(int i = 0; i < tokens.length; i++) {
            starts[i] = position;
            ends[i] = position + tokens[i].length();
            position = ends[i] + 1;
        }
        
        for(Span s: spans) {
            s.start = starts[s.tokenStart];
            s.end = ends[s.tokenEnd - 1];
        }
        
    }
    
    /**
     * Adds all spans of another annotation layer to the list of spans.
     * Does not change the label.
     * 
     * @param otherLayer the other annotation layer.
     */
    public void merge(AnnotationLayer otherLayer) {
        spans.addAll(otherLayer.spans);
    }
    
    
    /**
     * Returns an HTML representation of the layer using colors.
     * 
     * @param text the text.
     * @return the HTML string.
     */
    public String toColoredHTML(String text) {
        StringBuffer out = new StringBuffer();
        
        ArrayList<Span> coloredSpans = new ArrayList<Span>();
        
        for(Span s: spans)
            if(s.properties != null 
                    && (s.properties.containsKey("fgColor") || s.properties.containsKey("bgColor")))
                coloredSpans.add(s);
        
        Collections.sort(coloredSpans, Span.NestingOrder.instance());
        
        int position = 0;
        
        for(Span s: coloredSpans) {
            if(s.start < position)
                continue;
            
            String fg = (String) s.properties.get("fgColor");
            String bg = (String) s.properties.get("bgColor");
            String font = (String) s.properties.get("font");
            
            out.append(text.substring(position, s.start));
            if(font != null && font.equals("it"))
                out.append("<i>");
            out.append("<font style='");
            
            if(fg != null)
                out.append("color: " + fg);
            if(fg != null && bg != null)
                out.append(";");
            if(bg != null)
                out.append("background-color: " + bg);
            
            out.append("'>");
            if(font != null && font.equals("it"))
                out.append("</i>");
            
            out.append(text.substring(s.start, s.end) + "</font>");
            
            position = s.end;
        }
        
        out.append(text.substring(position));
        
        return out.toString();
    }
    
    /**
     * Returns an HTML representation of the layer using brackets.
     * 
     * @param text the text.
     * @return the HTML string.
     */
    public String toBracketedHTML(String text) {
        StringBuffer out = new StringBuffer();
        
        ArrayList<Span> tmpSpans = new ArrayList<Span>();
        
        tmpSpans.addAll(spans);
        
        Collections.sort(tmpSpans, Span.NestingOrder.instance());
        
        System.out.println("tmpSpans = " + tmpSpans);
        
        int position = 0;
        
        for(Span s: tmpSpans) {
            if(s.start < position)
                continue;
            
            String font = s.properties == null? null: (String) s.properties.get("font");
            
            out.append(text.substring(position, s.start));
            
            out.append("<b>[</b>");
            
            if(font != null && font.equals("it"))
                out.append("<i>");			
            out.append(text.substring(s.start, s.end));
            if(font != null && font.equals("it"))
                out.append("</i>");
            
            out.append("<b>]<sub>" + s.label + "</sub></b>");
            
            position = s.end;
        }
        
        out.append(text.substring(position));
        
        return out.toString();
    }
    
    public Iterator<T> iterator() {
        return spans.iterator();
    }
    
    public void add(T s) {
        spans.add(s);
    }
    
    public void clear() {
        spans.clear();
    }
    
    public void insertAt(int length, int index) {
        //String m = "I(" + length + "," + index + "): ";
        for(Span s: spans)
            if(s.start <= index && s.end > index) {
                //System.out.println(m+"Changing end of " + s.label + " from " + s.end + " to " + (s.end + length));
                s.end += length;
            } else if(s.start > index) {
                //System.out.println(m+"Changing start of " + s.label + " from " + s.start + " to " + (s.start + length));
                //System.out.println(m+"Changing end of " + s.label + " from " + s.end + " to " + (s.end + length));
                s.start += length;
                s.end += length;
            }
    }
    
    public void deleteAt(int start, int end) {
        //String m = "D(" + start + "," + end + "): ";

        int length = end - start;

        for(Span s: spans) {
            
            if(start < s.start && s.start < end) {
                //System.out.println(m+"Changing start of " + s.label + " from " + s.start + " to " + start);
                s.start = start;
            } else if(s.start >= end) {
                //System.out.println(m+"Changing start of " + s.label + " from " + s.start + " to " + (s.start - length));
                s.start -= length;
            }

            if(end > s.end && s.end > start) {
                //System.out.println(m+"Changing end of " + s.label + " from " + s.end + " to " + start);
                s.end = start;
            } else if(s.end > start) {
                //System.out.println(m+"Changing end of " + s.label + " from " + s.end + " to " + (s.end - length));
                s.end -= length;
            }
        }
    }

    public void setProperty(String property, Object value) {
        if(properties == null)
            properties = new HashMap<String, Object>(1);
        properties.put(property, value);
    }

    public Object getProperty(String property) {
        if(properties == null)
            return null;
        return properties.get(property);
    }
 
    /**
     * Returns true if any span of this layer overlaps with the other span.
     * 
     */
    public boolean overlaps(Span s) {
        for(Span sp: this)
            if(sp.overlaps(s))
                return true;
        return false;
    }
    
    public void trimPunctuationTokens(AnnotatedText text) {
    	for(Span sp: this)
    		sp.trimPunctuationTokens(text);
    }
    
    public void trimSpaces(AnnotatedText text) {
        for(Span sp: this)
            sp.trimSpaces(text);
    }

    public void removeWordInternalAnnotations(AnnotatedText text) {
        for(Iterator<T> it = spans.iterator(); it.hasNext(); ) {
            Span s = it.next();
            if(s.isWordInternal(text))
                it.remove();
        }
    }
    
    public void removeCoveredSpans() {
    	// TODO linear-time algorithm
    	HashSet<Span> toRemove = new HashSet();
    	
    	for(Span s1: this)
    		for(Span s2: this)
    			if(s1 != s2 && s1.includes(s2)) {
    				toRemove.add(s2);
    			}
    				
    	spans.removeAll(toRemove);    				
    }
    
    public void removeByLabel(String label) {
    	for(Iterator<T> it = spans.iterator(); it.hasNext(); )
    		if(it.next().label.equals(label))
    			it.remove();    		
    }
    
    public static void main(String[] argv) {
    	
    	AnnotationLayer l = new AnnotationLayer();
    	
        Span s = new Span();
        s.start = 5;
        s.end = 9;
        s.label = "S1";
        l.add(s);

        s = new Span();
        s.start = 5;
        s.end = 10;
        s.label = "S2";
        l.add(s);

        Collections.sort(l.spans, Span.NestingOrder.instance());
        
    	System.out.println(l);
    }


    
}
