/*
 CompareResult.java
 
 Copyright (C) 2005, Richard Johansson (richard@cs.lth.se).
 
 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.
 
 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.
 
 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
 Created on Nov 25, 2005 by Richard Johansson (richard@cs.lth.se).
 
 $Log: CompareResult.java,v $
 Revision 1.1  2009/01/16 10:04:53  johansson
 Added to brenta repository.

 Revision 1.7  2007/09/05 14:15:02  richard
 Partial precision and recall.

 Revision 1.6  2006/09/07 17:51:19  richard
 New revision.

 Revision 1.5  2006/04/11 08:21:04  richard
 Added documentation.
 
 Revision 1.4  2006/04/05 12:07:44  richard
 Changed the printout. Adds results.
 
 Revision 1.3  2005/11/30 12:41:44  richard
 Added the toString method and the F-measures.
 
 Revision 1.2  2005/11/29 13:51:10  richard
 Implemented layer comparison.
 
 Revision 1.1  2005/11/28 15:41:00  richard
 Added the file.
 
 
 */
package se.lth.cs.nlp.nlputils.annotations;

/**
 * The result of a comparison of two layers.
 * 
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class CompareResult {
    
    public int nCorrect;
    public int nPartialMatches;
    public int nPredicted;
    public int nInGold;
    
    public double weightedPScore;
    public double weightedRScore;
    
    /**
     * Returns the precision, that is the probability that a guess is correct.
     * 
     * @return the precision.
     */
    public double getPrecision() {
        return (double) nCorrect / nPredicted;
    }
    
    /**
     * Returns the recall, that is the probability that an item from the gold
     * standard is retrieved.
     * 
     * @return the recall.
     */
    public double getRecall() {
        return (double) nCorrect / nInGold;
    }
    
    /**
     * Returns the F-measure with equal weights, that
     * is the harmonic mean of the precision and the recall.
     * 
     * @return the unweighted F-measure.
     */
    public double getF1() {
        double p = getPrecision();
        double r = getRecall();
        return 2*p*r / (p + r);
    }
    
    /**
     * Returns a weighted F-measure. A <code>beta</code> value below 1
     * pulls the F-measure towards the precision and a value above 1 
     * towards the recall.
     * 
     * @param beta the parameter <code>beta</code>
     * @return the weighted F-measure.
     */
    public double getF(double beta) {
        double beta2 = beta * beta;
        double p = getPrecision();
        double r = getRecall();
        return (beta2 + 1)*p*r / (beta2 * p + r);
    }
    
    /**
     * Returns the precision when partial matches are counted as correct.
     * 
     * @return the lenient precision.
     */
    public double getLenientPrecision() {
        return (double) (nCorrect + nPartialMatches) / nPredicted;
    }
    
    /* TODO: lenient recall. */
    
    public String toString() {
        return "[Result: nCorrect = " + nCorrect + ", nPartialMatches = "
        + nPartialMatches + ", nPredicted = " + nPredicted + ", nInGold = "
        + nInGold 
        + ", precision = " + getPrecision() + ", recall = " + getRecall()
        + ", F1 = " + getF1() + "]";
    }
    
    /**
     * Adds another CompareResult to this one.
     * 
     * @param other the other CompareResult.
     */
    public void add(CompareResult other) {
        nCorrect += other.nCorrect;
        nPartialMatches += other.nPartialMatches;
        nPredicted += other.nPredicted;
        nInGold += other.nInGold;
        weightedPScore += other.weightedPScore;
        weightedRScore += other.weightedRScore;
    }
    
}
