/*
 FrameNetAnnotationLayers.java
  
 Created on Aug 13, 2007 by Richard Johansson (richard@cs.lth.se).

 $Log: FrameNetAnnotationLayers.java,v $
 Revision 1.1  2009/01/16 10:04:53  johansson
 Added to brenta repository.

 Revision 1.2  2007/08/17 10:17:18  richard
 Public methods.

 Revision 1.1  2007/08/13 12:23:49  richard
 Added.

   
 */
package se.lth.cs.nlp.nlputils.annotations;

import gnu.trove.TObjectIntHashMap;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;

import se.lth.cs.nlp.nlputils.core.CollectionUtils;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class FrameNetAnnotationLayers {

    protected static final String[] ORDER = {
        "Target",
        "FE",
        "GF",
        "PT",
        "Other",
        "Sent",
        
        "Adj",
        "Adv",
        "Noun",
        "Prep",
        "Verb"
    };
    
    protected static final TObjectIntHashMap<String> ORDER_MAP = CollectionUtils.toOrderMap(ORDER);
    
    protected static final Comparator<AnnotationLayer> LAYER_ORDER = new Comparator<AnnotationLayer>() {
        public int compare(AnnotationLayer l1, AnnotationLayer l2) {
            int i1 = ORDER_MAP.get(l1.label);
            if(i1 == 0)
                throw new RuntimeException("Layer type " + l1.label + " unknown");
            int i2 = ORDER_MAP.get(l2.label);
            if(i2 == 0)
                throw new RuntimeException("Layer type " + l2.label + " unknown");
            return Double.compare(i1, i2);
        }        
    };
    
    public static String posLayerName(String lu) {
        if(lu == null)
            return null;
        if(lu.endsWith(".n"))
            return "Noun";
        if(lu.endsWith(".v"))
            return "Verb";
        if(lu.endsWith(".a"))
            return "Adj";
        if(lu.endsWith(".prep"))
            return "Prep";
        if(lu.endsWith(".adv"))
            return "Adv";
        if(lu.endsWith(".num")) // ska vara med?
            return null;
        if(lu.endsWith(".c")) // ska vara med?
            return null;
        if(lu.endsWith(".art")) // ska vara med?
            return null;
        throw new RuntimeException("Unknown pos: " + lu);
    }
    
    public static boolean isPosLayerLabel(String l) {
        return l.matches("Noun|Verb|Adj|Prep|Adv");
    }
    
    public static void checkLayers(AnnotatedText text) {
        if(text.layers.size() == 0) {
            AnnotationLayer nes = new AnnotationLayer();
            nes.label = "NER";
            text.layers.add(nes);
        } else {
            for(int i = 1; i < text.layers.size(); i++) {
                AnnotationLayer l = text.layers.get(i);
                if(l.label.equals("NER")) {
                    System.err.println("Warning: moved NER layer to position 0");
                    text.layers.remove(i);
                    text.layers.add(0, l);
                    i--;
                }
            }
            
            if(!text.layers.get(0).label.equals("NER")) {
                AnnotationLayer nes = new AnnotationLayer();
                nes.label = "NER";
                text.layers.add(0, nes);
            }
            if(text.layers.size() > 1) {

                AnnotationLayer[] ls = text.layers.toArray(new AnnotationLayer[0]);
                
                int prev = 1;
                for(int i = 2; i <= ls.length; i++)
                    if(i == ls.length
                       || ls[i].label.equals("Target")) {
                        Arrays.sort(ls, prev, i, LAYER_ORDER);
                        prev = i;
                    }
                
                ArrayList<AnnotationLayer> copy = new ArrayList<AnnotationLayer>();
                copy.add(ls[0]);
                int i = 1;
                while(i < ls.length) {
                    AnnotationLayer tl = ls[i++];
                    if(!tl.label.equals("Target")) {
                        printLayers(ls, tl);
                        throw new RuntimeException("First layer must be Target. text = " + text);
                    }
                    copy.add(tl);
                    if(i < ls.length && ls[i].label.equals("FE"))
                        copy.add(ls[i++]);
                    else
                        copy.add(new AnnotationLayer("FE"));
                    if(i < ls.length && ls[i].label.equals("GF"))
                        copy.add(ls[i++]);
                    else
                        copy.add(new AnnotationLayer("GF"));
                    if(i < ls.length && ls[i].label.equals("PT"))
                        copy.add(ls[i++]);
                    else
                        copy.add(new AnnotationLayer("PT"));
                    if(i < ls.length && ls[i].label.equals("Other"))
                        copy.add(ls[i++]);
                    else
                        copy.add(new AnnotationLayer("Other"));
                    if(i < ls.length && ls[i].label.equals("Sent"))
                        copy.add(ls[i++]);
                    else
                        copy.add(new AnnotationLayer("Sent"));

                    String lu = (String) tl.getProperty("lu");
                    String pl = posLayerName(lu);
                    if(i < ls.length && lu == null && ls[i].label.matches("Adj|Adv|Noun|Prep|Verb"))
                        throw new RuntimeException("Unspecified LU, but POS-specific layer exists.");
                    if(i < ls.length && ls[i].label.matches("Adj|Adv|Noun|Prep|Verb")) {
                        if(!ls[i].label.equals(pl))
                            System.err.println("Warning: lu = " + lu + ", pos layer label = " + ls[i].label);
                        copy.add(ls[i++]);
                    } else
                        copy.add(new AnnotationLayer(pl));
                }
                text.layers = copy;
                
                if(text.layers.size() % 7 != 1) {
                    printLayers(text.layers.toArray(new AnnotationLayer[0]), null);
                    throw new RuntimeException("Bug: number of layers should be 1 % 7, was " + text.layers.size());
                }
                
                for(int j = 1; j < text.layers.size(); j += 7)
                    if(!text.layers.get(j).label.equals("Target")) {
                        printLayers(text.layers.toArray(new AnnotationLayer[0]), null);
                        throw new RuntimeException("Bug: layer at position " + j + " should be Target");
                    }
            }
        }
    }
    
    protected static void printLayers(AnnotationLayer[] ls, AnnotationLayer tl) {
        int i = 0;
        for(AnnotationLayer l: ls) {
            System.out.print(i++ + ". " + l.label);
            if(l.label.equals("Target"))
                System.out.print(" (" + l.getProperty("lu") + ")");
            if(l == tl)
                System.out.println(" <--");
            else
                System.out.println();
        }
    }
    
}
