/*
 RawXMLAnnotation.java
 
 Created on Nov 29, 2005 by Richard Johansson (richard@cs.lth.se).
 
 $Log: RawXMLAnnotation.java,v $
 Revision 1.1  2009/01/16 10:04:53  johansson
 Added to brenta repository.

 Revision 1.9  2006/12/20 09:23:12  richard
 Exception handling.

 Revision 1.8  2006/11/15 11:45:28  richard
 Null check.

 Revision 1.7  2006/09/07 17:51:19  richard
 New revision.

 Revision 1.6  2006/04/11 08:35:19  richard
 Added documentation.
 
 Revision 1.5  2006/04/05 12:10:27  richard
 Prints properties.
 
 Revision 1.4  2006/03/29 13:58:00  richard
 Implemented printLayers.
 
 Revision 1.3  2005/11/30 13:32:19  richard
 Changed the interface.
 
 Revision 1.2  2005/11/30 12:44:02  richard
 Implemented the raw XML import.
 
 Revision 1.1  2005/11/30 09:52:25  richard
 Renamed the annotation types.
 
 Revision 1.2  2005/11/30 09:49:09  richard
 Changed the interface.
 
 Revision 1.1  2005/11/30 09:33:42  richard
 Added the file.
 
 
 */
package se.lth.cs.nlp.nlputils.annotations;

import java.util.*;
import java.io.*;

import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.*;
import javax.xml.parsers.*;

/**
 * This AnnotationType implements functionality to handle classical nested
 * XML annotations.
 * 
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class RawXMLAnnotation extends AnnotationType {
    
    private RawXMLAnnotation() {}
    
    private static final RawXMLAnnotation instance = new RawXMLAnnotation();
    
    /**
     * Returns the instance of this AnnotationType.
     * 
     * @return the instance of this AnnotationType.
     */
    public static RawXMLAnnotation instance() {
        return instance;
    }
    
    private final class XMLCallback extends DefaultHandler implements ContentHandler {
        
        private int index = 0;
        private int count = 0;
        private LinkedList<Span> stack = new LinkedList<Span>();
        ArrayList<Span> finished = new ArrayList<Span>();
        StringBuffer text = new StringBuffer();
        
        public void startElement(String namespace, String localname,  
                String type, Attributes attributes) {
            Span s = new Span();
            s.label = type;
            s.start = index;
            s.id = "s" + count;
            count++;
            
            int n = attributes.getLength();
            if(n > 0) {
                s.properties = new HashMap<String, Object>();
                for(int i = 0; i < n; i++) {
                    s.properties.put(attributes.getQName(i),
                            attributes.getValue(i));
                }
            }
            stack.add(s);
        }
        
        public void endElement(String namespace, String localname,  
                String type) {
            Span s = stack.removeLast();
            if(!s.label.equals(type))
                throw new RuntimeException("Bad type!");
            s.end = index;
            finished.add(s);
        }
        
        public void characters(char[] ch, int start, int len) {
            text.append(new String(ch, start, len));
            index += len;
        }
        
    }
    
    private XMLReader makeXMLReader() throws SAXException, ParserConfigurationException { 
        javax.xml.parsers.SAXParserFactory saxParserFactory 
        = javax.xml.parsers.SAXParserFactory.newInstance(); 
        javax.xml.parsers.SAXParser saxParser = saxParserFactory.newSAXParser(); 
        org.xml.sax.XMLReader parser = saxParser.getXMLReader(); 
        return parser; 
    }
    
    /**
     * Creates an annotated text with a single layer.
     * 
     * @param in the input stream from which the annotation is read.
     * @return the resulting AnnotatedText.
     */
    public AnnotatedText createLayers(InputStream in) throws IOException {
        try {
            XMLReader reader = makeXMLReader(); 
            XMLCallback c = new XMLCallback();
            reader.setContentHandler(c);
            reader.parse(new org.xml.sax.InputSource(in));
            
            AnnotationLayer l = new AnnotationLayer();
            l.spans = c.finished;
            
            ArrayList<AnnotationLayer> out = new ArrayList<AnnotationLayer>();
            out.add(l);
            
            AnnotatedText t = new AnnotatedText();
            t.layers = out;
            t.text = c.text.toString();
            return t;
        } catch(IOException e) {
            throw e;
        } catch(Exception e) {
            e.printStackTrace();            
            throw new IOException("Couldn't parse!");
        }
    }
    
    /**
     * This method is unimplemented since raw XML text should never be put 
     * into Java strings (because this may break the encoding).
     */
    public AnnotatedText createLayers(String s) {
        throw new RuntimeException("NEVER use raw strings for XML!!!");
    }
    
    /**
     * Prints a list of annotation layers on a text to an output stream. The 
     * spans in the layers must be nestable.
     
     * @param out the output stream.
     * @param text the text.
     * @param l the list of annotation layers.
     */
    public void printLayers(OutputStream out, String text, 
            List<AnnotationLayer> l) throws IOException {
        
        /* TODO escaping */
        
        //try {
            
            ArrayList<Span> spans = new ArrayList<Span>();
            
            for(AnnotationLayer layer: l)
                if(layer != null)
                    spans.addAll(layer.spans);
            
            Collections.sort(spans, Span.NestingOrder.instance());
            
            Span first = spans.isEmpty()? null: spans.get(0);
            
            if(first == null || first.start != 0 || first.end != text.length()) {
                Span dummy = new Span();
                dummy.start = 0;
                dummy.end = text.length();
                dummy.label = "text";
                spans.add(0, dummy);
            }
            
            PrintWriter pw = new PrintWriter(new OutputStreamWriter(out, "UTF-8"));
            
            pw.println("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
            
            ArrayList<Span> stack = new ArrayList<Span>();
            
            int position = 0;
            int nextSpan = 0;
            //while(position < text.length()) {
            while(position <= text.length()) {
                int nextPos = position;
                while(!stack.isEmpty()) {
                    Span top = stack.get(stack.size() - 1);
                    if(position == top.end) {
                        pw.print("</" + top.label + ">");
                        stack.remove(stack.size() - 1);
                        
                        //System.err.println("popping" + top);
                        
                    } else
                        break;
                }
                
                if(position == text.length()) {
                    if(!stack.isEmpty())
                        throw new RuntimeException("Stack is not empty!");
                    break;
                }
                
                if(!stack.isEmpty()) {
                    nextPos = stack.get(stack.size() - 1).end;
                }
                while(nextSpan < spans.size()) {
                    Span next = spans.get(nextSpan);
                    if(position == next.start) {						
                        
                        pw.print("<" + next.label);
                        if(next.properties != null)
                            for(Map.Entry<String, Object> e: next.properties.entrySet()) {
                                pw.print(" " + e.getKey() + "=");
                                String v = e.getValue().toString().replaceAll("\"", "\\\"");
                                pw.print("\"" + v + "\"");
                            }
                        
                        if(position == next.end) {
                            pw.print("/>");
                        } else {
                            pw.print(">");
                            stack.add(next);
                            nextPos = next.end;
                        }
                        nextSpan++;
                    } else {
                        if(nextPos > next.start)
                            nextPos = next.start;
                        break;
                    }
                }
                
                pw.print(text.substring(position, nextPos));
                position = nextPos;
            }
            
            pw.println();
            
            pw.flush();
            
        //} catch(Exception e) {
        //    e.printStackTrace();
        //    throw new IOException("Could not print layer.");
        //}
    }
    
}
