/*
 Span.java
 
 Copyright (C) 2005, Richard Johansson (richard@cs.lth.se).
 
 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.
 
 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.
 
 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
 Created on Nov 25, 2005 by Richard Johansson (richard@cs.lth.se).
 
 $Log: Span.java,v $
 Revision 1.3  2011-07-11 11:52:52  johansson
 Trim punctuation.

 Revision 1.2  2010-06-07 10:54:51  johansson
 Major update.

 Revision 1.1  2009/01/16 10:04:53  johansson
 Added to brenta repository.

 Revision 1.19  2008/12/03 15:07:39  richard
 Added doIntersect.

 Revision 1.18  2008/06/02 12:26:49  richard
 Overlap bug.

 Revision 1.17  2007/10/04 13:58:46  richard
 Empty spans non-overlapping.

 Revision 1.16  2007/09/07 09:15:25  richard
 Added hasProperty.

 Revision 1.15  2007/09/05 14:15:37  richard
 Length and intersection length.

 Revision 1.14  2007/08/22 10:20:09  richard
 Added overlaps.

 Revision 1.13  2007/08/13 12:24:08  richard
 Changed toString.

 Revision 1.12  2006/12/15 10:30:53  richard
 New constructors.

 Revision 1.11  2006/11/13 10:06:15  richard
 Set/get properties.

 Revision 1.10  2006/09/07 17:51:19  richard
 New revision.

 Revision 1.9  2006/04/11 08:26:58  richard
 Added documentation.
 
 Revision 1.8  2006/04/05 11:35:24  richard
 Changed the printout.
 
 Revision 1.7  2006/03/29 13:57:26  richard
 Added the nesting comparator.
 
 Revision 1.6  2005/12/02 13:31:32  richard
 Added tokens and lemmas.
 
 Revision 1.5  2005/11/30 13:30:33  richard
 Properties can store any object. Changed toString.
 
 Revision 1.4  2005/11/30 12:41:25  richard
 Added the toString method.
 
 Revision 1.3  2005/11/30 09:34:06  richard
 Added the properties map.
 
 Revision 1.2  2005/11/29 13:51:10  richard
 Implemented layer comparison.
 
 Revision 1.1  2005/11/28 15:41:00  richard
 Added the file.
 
 
 */
package se.lth.cs.nlp.nlputils.annotations;

import java.util.*;

import se.lth.cs.nlp.nlputils.core.Util;

/**
 * A span, that is a contiguous piece of text.
 * 
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class Span {
    
    public String label;
    public String id;
    
    public Span() {        
    }
    
    public Span(int start, int end) {
        this.start = start;
        this.end = end;
    }
    
    public Span(int start, int end, String label) {
        this(start, end);
        this.label = label;
    }
    
    public Span clone() {
    	Span out = new Span();
    	out.start = start;
    	out.end = end;
    	out.tokenStart = tokenStart;
    	out.tokenEnd = tokenEnd;
    	out.label = label;
    	out.id = id;
    	if(properties != null)
    		out.properties = new HashMap(properties);
    	return out;
    }
    
    /**
     * The start of the span, inclusive.
     */
    public int start; // inclusive
    
    /**
     * The end of the span, exclusive.
     */
    public int end;   // exclusive
    
    /**
     * The start of the span as counted in tokens, inclusive.
     */
    public int tokenStart = -1; // inclusive
    
    /**
     * The end of the span as counted in tokens, exclusive.
     */
    public int tokenEnd = -1;   // exclusive
    
    /**
     * Properties.
     */
    public HashMap<String, Object> properties;
    
    /**
     * Returns true if this span has the same start and end and the same
     * label as the other span.
     * 
     * @param other the span to compare this span to.
     * @return true if they match.
     */
    public boolean matches(Span other) {
        return start == other.start && end == other.end 
        && label.equals(other.label);
    }
    
    public boolean matchesUnlabeled(Span other) {
        return start == other.start && end == other.end;
    }
        
    public boolean includes(Span other) {
    	return start <= other.start && end >= other.end;
    }
    
    public int length() {
        return Math.max(end - start, 0);
    }
    
    public int lengthTokens() {
        return Math.max(tokenEnd - tokenStart, 0);    	
    }
    
    /**
     * Returns true if this span overlaps and has the same label as the
     * other span.
     * 
     * @param other the span to compare this span to.
     * @return true if they overlap and have the same label.
     */
    public boolean matchesPartially(Span other) {
        if(!label.equals(other.label))
            return false;
        if(start >= other.start && start < other.end)
            return true;
        if(end > other.start && end <= other.end)
            return true;
        return false;
    }
    
    public int intersectionLength(Span other) {
        if(!label.equals(other.label))
            return 0;
        if(start >= end)
            return 0;
        if(other.start >= other.end)
            return 0;
        if(!overlaps(other))
            return 0;
        int maxstart = Math.max(start, other.start);
        int minend = Math.min(end, other.end);
        return Math.max(minend - maxstart, 0);
    }

    public int intersectionLengthTokens(Span other) {
        if(!label.equals(other.label))
            return 0;
        if(tokenStart >= tokenEnd)
            return 0;
        if(other.tokenStart >= other.tokenEnd)
            return 0;
        if(!overlapsTokensUnlabeled(other))
            return 0;
        int maxstart = Math.max(tokenStart, other.tokenStart);
        int minend = Math.min(tokenEnd, other.tokenEnd);
        return Math.max(minend - maxstart, 0);
    }

    public boolean overlapsTokens(Span other) {
	if(!label.equals(other.label))
	    return false;
	return overlapsTokensUnlabeled(other);
    }    

    public boolean overlapsTokensUnlabeled(Span other) {
        if(tokenStart >= tokenEnd || other.tokenStart >= other.tokenEnd)
            return false;
        if(tokenStart >= other.tokenStart && tokenStart < other.tokenEnd)
            return true;
        if(other.tokenStart >= tokenStart && other.tokenStart < tokenEnd)
            return true;
        return false;
    }
    
    public boolean overlapsUnlabeled(Span other) {
        if(start >= end || other.start >= other.end)
            return false;
        if(start >= other.start && start < other.end)
            return true;
        if(other.start >= start && other.start < end)
            return true;
        return false;
    }
    
    
    public void doIntersect(Span other) {
        if(!label.equals(other.label))
            throw new IllegalArgumentException("Labels must match");
        if(start >= end)
            return;
        if(other.start >= other.end || !overlaps(other)) {
            end = start;
            return;
        }
        int maxstart = Math.max(start, other.start);
        int minend = Math.min(end, other.end);
        start = maxstart;
        end = minend;
    }
    
    /**
     * Returns true if this span overlaps with the other span.
     * Empty spans never overlap.
     * 
     * @param other the span to compare this span to.
     * @return true if they overlap.
     */
    public boolean overlaps(Span other) {
        if(start >= end || other.start >= other.end)
            return false;
        if(start >= other.start && start < other.end)
            return true;
        if(other.start >= start && other.start < end)
            return true;
        return false;
    }
    
    private static boolean isPunc(String t) {
    	if(t.length() == 0)
    		return false;
    	for(int i = t.length()-1; i >= 0; i--)
    		if(!Util.isPunc(t.charAt(i)))
    			return false;
    	return true;
    }
    
    public boolean isWordInternal(AnnotatedText text) {
        if(start < 0 || end < 0 || end < start)
            return false;
        if(start > 0) {
            char sc = text.text.charAt(start-1);
            if(!Util.isPunc(sc) && !Util.isSpace(sc))
                return true;
        }
        if(end < text.text.length()) {
            char ec = text.text.charAt(end);
            if(!Util.isPunc(ec) && !Util.isSpace(ec))
                return true;
        }
        return false;
    }
    
    public void trimPunctuationTokens(AnnotatedText text) {
    	if(tokenStart == -1 || tokenEnd == -1)
    		throw new RuntimeException("not tokenized");
    	while(tokenStart < tokenEnd && isPunc(text.tokens[tokenStart])) {
            System.out.println("START PUNC");
    		tokenStart++;
    		start = text.tokenStarts[tokenStart];
    	}    	
    	while(tokenStart < tokenEnd && isPunc(text.tokens[tokenEnd-1])) {
            System.out.println("END PUNC");
    	    tokenEnd--;
    		end = text.tokenEnds[tokenEnd];
    	}
    }
    
    public void trimSpaces(AnnotatedText text) {
        while(start >= end && Character.isSpace(text.text.charAt(start)))
            start += 1;
        while(Character.isSpace(text.text.charAt(end - 1)))
            end -= 1;
    }
    
    public String toString() {
        String s = "(";
        if(id != null)
            s += "[" + id + "] ";
        s += label + ", chars=[" + start + ", " + end + "]";
        if(properties != null)
            s += ", properties: " + properties;
        if(tokenStart > -1 || tokenEnd > -1)
            s += ", tokens=[" + tokenStart + ", " + tokenEnd + "]";
        return s + ")";
    }
    
    public static class ByLeftOrder implements Comparator<Span> {

        private ByLeftOrder() {}
        private static ByLeftOrder instance = new ByLeftOrder();

        /**
         * Returns the single instance of this class.
         * 
         * @return the instance.
         */
        public static ByLeftOrder instance() {
            return instance;
        }
        
        public int compare(Span left, Span right) {
        	int c = left.start - right.start;
        	if(c != 0)
        		return c;
        	int c2 = left.tokenStart - right.tokenEnd;
        	if(c2 != 0)
        		return c2;
        	int c3 = left.end - right.end;
        	if(c3 != 0)
        		return c3;
        	return left.tokenEnd - right.tokenEnd;
        }
        
    };
    
    /**
     * A comparator that can be used to sort spans according to nesting order.
     * Throws a RuntimeException if the spans cannot be nested.
     */
    public static class NestingOrder implements Comparator<Span> {

        private NestingOrder() {}
        private static NestingOrder instance = new NestingOrder();

        /**
         * Returns the single instance of this class.
         * 
         * @return the instance.
         */
        public static NestingOrder instance() {
            return instance;
        }
        
        public int compare(Span left, Span right) {
            if(left.start == right.start && left.end == right.end)
                //return left.id.compareTo(right.id);
                return 0;
            
            if(left.end <= right.start)
                return -1;
            
            if(right.end <= left.start)
                return 1;
            
            if(left.start <= right.start && left.end >= right.end)
                return -1;
            
            if(left.start >= right.start && left.end <= right.end)
                return 1;
            
            throw new RuntimeException("Not nestable!");
        }
        
    };
    
    /**
     * A comparator that can be used to sort spans according to length.
     */
    public static class LengthOrder implements Comparator<Span> {

        private LengthOrder() {}
        private static LengthOrder instance = new LengthOrder();

        /**
         * Returns the single instance of this class.
         * 
         * @return the instance.
         */
        public static LengthOrder instance() {
            return instance;
        }
        
        public int compare(Span left, Span right) {
            int l1 = left.end - left.start;
            int l2 = right.end - right.start;
            
            if(l1 < l2)
                return 1;
            if(l1 > l2)
                return -1;

            return Double.compare(left.start, right.start);
        }
        
    };    
    
    public void setProperty(String property, Object value) {
        if(properties == null)
            properties = new HashMap<String, Object>(1);
        properties.put(property, value);
    }

    public boolean hasProperty(String property) {
        if(properties == null)
            return false;
        return properties.containsKey(property);
    }
    
    public Object getProperty(String property) {
        if(properties == null)
            return null;
        return properties.get(property);
    }

    public int dist(Span s2) {
        if(overlapsUnlabeled(s2))
            return 0;
        if(end <= s2.start)
            return s2.start - end;
        else
            return start - s2.end;
    }




    
}
