/*
 BigDoubleArray.java
  
 Created on Apr 8, 2008 by Richard Johansson (richard@cs.lth.se).

 $Log: BigDoubleArray.java,v $
 Revision 1.2  2009/11/16 11:36:04  johansson
 Renamed for compatibility with Windows systems.

 Revision 1.1  2009/01/16 10:05:11  johansson
 Added to brenta repository.

 Revision 1.4  2008/04/08 11:43:43  richard
 More methods.

 Revision 1.3  2008/04/08 09:17:04  richard
 Serializable.

 Revision 1.2  2008/04/08 09:00:07  richard
 Timing tests.

 Revision 1.1  2008/04/08 08:41:24  richard
 Added.

   
 */
package se.lth.cs.nlp.nlputils.core;

import java.io.Serializable;
import java.util.Arrays;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class BigDoubleArray implements Serializable {

    private static final int DEFAULT_CHUNK = 1 << 24;   
    
    private static final long serialVersionUID = 0L;
    
    private double[][] arrs;
    private int size;

    private int chunkSizeBits;
    private int chunkIndexMask;
    
    public BigDoubleArray(int size, int chunkSize) {
        this.size = size;
        int bit = Util.msb(chunkSize) - 1;
        if((chunkSize & ~(1 << bit)) != 0)
            throw new IllegalArgumentException("Chunk size must be a power of 2.");                
        chunkSizeBits = bit;        
        chunkIndexMask = (1 << bit) - 1;        
        int nChunks = (int) Math.ceil((double) size / chunkSize);
        arrs = new double[nChunks][];
        int lastSize = size - chunkSize*(nChunks - 1);
        for(int i = 0; i < nChunks - 1; i++)
            arrs[i] = new double[chunkSize];        
        arrs[nChunks-1] = new double[lastSize];
    }    
    
    public BigDoubleArray(int size) {
        this(size, DEFAULT_CHUNK);
    }
    
    public int length() {
        return size;
    }
    
    public void set(int index, double value) {
        if(index < 0 || index >= size)
            throw new ArrayIndexOutOfBoundsException("illegal index: " + index);
        int index1 = index >>> chunkSizeBits;
        int index2 = index & chunkIndexMask;
        arrs[index1][index2] = value;
    }

    public void fill(double value) {
        for(int i = 0; i < arrs.length; i++)
            Arrays.fill(arrs[i], value);
    }       
    
    public double modify(int index, double value) {
        if(index < 0 || index >= size)
            throw new ArrayIndexOutOfBoundsException("illegal index: " + index);
        int index1 = index >>> chunkSizeBits;
        int index2 = index & chunkIndexMask;
        arrs[index1][index2] += value;
        return arrs[index1][index2];
    }

    public double get(int index) {
        if(index < 0 || index >= size)
            throw new ArrayIndexOutOfBoundsException("illegal index: " + index);
        int index1 = index >>> chunkSizeBits;
        int index2 = index & chunkIndexMask;
        return arrs[index1][index2];
    }
    
    /* Test. */
    
    public static void main(String[] argv) {
        int size = 100000000;
        int chunkSize = 1 << 5;
        double[] arr = new double[size];
        BigDoubleArray barr = new BigDoubleArray(size, chunkSize);

        long l0 = System.currentTimeMillis();
        double tmp = 0;
        for(int i = 0; i < size; i++)
            tmp = i / 10.0;
        long l1 = System.currentTimeMillis();
        for(int i = 0; i < size; i++)
            arr[i] = i / 10.0;
        long l2 = System.currentTimeMillis();        
        for(int i = 0; i < size; i++)
            barr.set(i, i / 10.0);
        long l3 = System.currentTimeMillis();

        long dt1 = l1 - l0;
        long dt2 = l2 - l1 - dt1;
        long dt3 = l3 - l2 - dt1;

        System.out.println("dt1 = " + dt1);
        System.out.println("dt2 = " + dt2);
        System.out.println("dt3 = " + dt3);
        
        for(int i = 0; i < size; i++) {
            if(Math.abs(arr[i] - barr.get(i)) > 1e-9)
                throw new RuntimeException("Error");
        }
    }
    
}
