package se.lth.cs.nlp.nlputils.core;

import java.util.Arrays;

public class BloomFilter {

	private static final int LOG_INT_SIZE = 5;
	private static final int INT_LOWER_BITS = (1 << LOG_INT_SIZE) - 1; 
	private int[][] stored; 

	private int nhashes, mod;
	
	public BloomFilter(int size, int nhashes) {
		int logsize = Util.msb(Math.max(0, size - 1));
		if(size > 1 << logsize)
			logsize++;
		int len = 1 << Math.max(0, logsize - LOG_INT_SIZE);		
		stored = new int[len][nhashes];		
		mod = len - 1;
		this.nhashes = nhashes;
	}

	public void put(String k) {
		int h = 0;
		for(int i = 0; i < nhashes; i++) {
			h = i == 0? ihash(k, i): ihashNext(h);			
			int ix = (h >>> LOG_INT_SIZE) & mod;
			int flag = 1 << (h & INT_LOWER_BITS);			
			stored[ix][i] |= flag;
		}
	}
	
	public void putOld(String k) {
		for(int i = 0; i < nhashes; i++) {
			int h = ihash(k, i);			
			int ix = (h >>> LOG_INT_SIZE) & mod;
			int flag = 1 << (h & INT_LOWER_BITS);			
			stored[ix][i] |= flag;
		}
	}
	
	public boolean contains(String k) {
		int h = 0;
		for(int i = 0; i < nhashes; i++) {
			h = i == 0? ihash(k, i): ihashNext(h);	
			int ix = (h >>> LOG_INT_SIZE) & mod;
			int flag = 1 << (h & INT_LOWER_BITS);			
			if((stored[ix][i] & flag) == 0)
				return false;
		}
		return true;
	}
	
	public boolean containsOld(String k) {
		for(int i = 0; i < nhashes; i++) {
			int h = ihash(k, i);
			int ix = (h >>> LOG_INT_SIZE) & mod;
			int flag = 1 << (h & INT_LOWER_BITS);			
			if((stored[ix][i] & flag) == 0)
				return false;
		}
		return true;
	}
	
	private static int ihash(String s, int start) {
		int hash = start;
		int l = s.length();
		for(int i = 0; i < l; i++) {			
			hash += s.charAt(i);
			hash += (hash << 10L);
			hash ^= (hash >> 6L);
		}
		hash += (hash << 3L);
		hash ^= (hash >> 11L);
		hash += (hash << 15L);
		return hash;
	}
	
	private static int ihashNext(int hash) {
		hash += (hash << 3L);
		hash ^= (hash >> 11L);
		hash += (hash << 15L);
		return hash;
	}
	
	private static long lhash(String s, long start) {
		long hash = start;
		int l = s.length();
		for(int i = 0; i < l; i++) {			
			hash += s.charAt(i);
			hash += (hash << 10L);
			hash ^= (hash >> 6L);
		}
		hash += (hash << 3L);
		hash ^= (hash >> 11L);
		hash += (hash << 15L);
		return hash;
	}
	
	public static void main(String[] argv) {
		
		BloomFilter bf = new BloomFilter((1 << 16)*1, 8);

		System.out.println(bf.stored.length);
		
		int n = 100000;
		int ncoll = 0;
		
		for(int i = 0; i < n; i++) {
			String s = "" + i;
			if(bf.contains(s))
				ncoll++;
			bf.put(s);
		}
		
		for(int i = 0; i < n; i++) {
			String s = "" + i;
			if(!bf.contains(s))
				throw new RuntimeException("error: " + s);
		}		
		
		System.out.println(ncoll + "/" + n + " = " + ((double) ncoll / n));
		
	}
	
}
