/*
 ClosedIntIntMap.java
  
 Created on Nov 6, 2007 by Richard Johansson (richard@cs.lth.se).

 $Log: ClosedIntIntMap.java,v $
 Revision 1.1  2009/01/16 10:05:11  johansson
 Added to brenta repository.

 Revision 1.4  2008/03/03 12:22:16  richard
 Handles negative keys.

 Revision 1.3  2008/02/22 13:16:59  richard
 Added getKey.

 Revision 1.2  2007/12/18 15:15:08  richard
 Added toString.

 Revision 1.1  2007/11/07 15:37:30  richard
 Added.

   
 */
package se.lth.cs.nlp.nlputils.core;

import java.io.Serializable;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class ClosedIntIntMap implements Serializable {

    private static final long serialVersionUID = 0L;
    
    private int[] keys;
    private int[] values;
    private int size;

    private int threshold;    
    private double growFactor;
    
    public ClosedIntIntMap(int size, double loadFactor, double growFactor) {
        this(size);
        this.growFactor = growFactor;
        this.threshold = (int) (size*loadFactor);
    }
    
    public ClosedIntIntMap(int size) {
        keys = new int[size];
        values = new int[size];
    }

    public int getOrSet(int key, int newValue) {
        if(size == values.length)
            throw new RuntimeException("Map is full!");
        int index = key % keys.length;

        //
        if(index < 0)
            index = -index;
        
        int offset = 0;
        while(true) {
            if(keys[index] == 0) {
                keys[index] = key;
                values[index] = newValue;
                size++;
                if(size == threshold)
                    resize();
                return newValue;                
            }
            if(keys[index] == key)
                return values[index];
            offset++;
            index += offset * offset;
            if(index < 0)
                index = -index;
            index %= keys.length;
        }        
    }
    
    public int increment(int key) {
        if(size == values.length)
            throw new RuntimeException("Map is full!");
        int index = (int) (key % keys.length);

        //
        if(index < 0)
            index = -index;
        
        int offset = 0;
        while(true) {
            if(keys[index] == 0) {
                keys[index] = key;
                values[index] = 1;
                size++;
                if(size == threshold)
                    resize();
                return 0;
            } else if(keys[index] == key) {
                return values[index]++;
            }
            offset++;
            index += offset * offset;
            if(index < 0)
                index = -index;
            index %= keys.length;
        }        
    }
    
    public int put(int key, int value) {
        if(size == values.length)
            throw new RuntimeException("Map is full!");
        int index = (int) (key % keys.length);

        //
        if(index < 0)
            index = -index;
        
        int offset = 0;
        while(true) {
            if(keys[index] == 0) {
                keys[index] = key;
                values[index] = value;
                size++;
                if(size == threshold)
                    resize();
                return 0;
            } else if(keys[index] == key) {
                int oldValue = values[index];
                values[index] = value;
                return oldValue;
            }
            offset++;
            index += offset * offset;
            if(index < 0)
                index = -index;
            index %= keys.length;
        }        
    }
    
    public int get(int key) {
        int index = key % keys.length;

        //
        if(index < 0)
            index = -index;
        
        int offset = 0;
        while(true) {
            if(keys[index] == 0)
                return 0;
            if(keys[index] == key)
                return values[index];
            offset++;
            index += offset * offset;
            index %= keys.length;

            //
            if(index < 0)
                index = -index;
        }        
    }
    
    public boolean containsKey(int key) {
        if(key == 0)
            return false;
        int index = key % keys.length;

        //
        if(index < 0)
            index = -index;
        
        int offset = 0;
        while(true) {
            if(keys[index] == 0)
                return false;
            if(keys[index] == key)
                return true;
            offset++;
            index += offset * offset;
            index %= keys.length;
            
            //
            if(index < 0)
                index = -index;            
        }        
    }
    
    private void resize() {
        int newSize = (int) (values.length * growFactor);
        threshold = (int) (threshold * growFactor);
        int[] oldKeys = keys;
        int[] oldValues = values;
        size = 0;
        keys = new int[newSize];
        values = new int[newSize];
        for(int i = 0; i < oldValues.length; i++) {
            if(oldKeys[i] != 0)
                put(oldKeys[i], oldValues[i]);
        }
    }
    
    public int[] getKeys() {
        return keys;
    }

    public int[] getValues() {
        return values;
    }
 
    public int size() {
        return size;
    }
    
    public String toString() {
        StringBuffer sb = new StringBuffer();
        for(int i = 0; i < keys.length; i++)
            if(values[i] != 0)
                sb.append(keys[i] + "->" + values[i] + ", ");
        if(sb.length() > 0)
            sb.delete(sb.length() - 2, sb.length());
        return "{ " + sb + " }";
    }
    
    public int getKey(int value) {
        for(int i = 0; i < keys.length; i++)
            if(values[i] == value)
                return keys[i];
        return 0;
    }
    
}
