/*
 ClosedLongDoubleMap.java
  
 Created on Apr 28, 2008 by Richard Johansson (richard@cs.lth.se).

 $Log: ClosedLongDoubleMap.java,v $
 Revision 1.1  2009/01/16 10:05:11  johansson
 Added to brenta repository.

 Revision 1.2  2008/04/28 14:43:16  richard
 Load reports.

 Revision 1.1  2008/04/28 07:01:16  richard
 Added.

   
 */
package se.lth.cs.nlp.nlputils.core;

import java.io.Serializable;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class ClosedLongDoubleMap implements Serializable {

    private static final long serialVersionUID = 0L;
    
    private BigLongArray keys;
    private BigDoubleArray values;
    private int size;

    private int threshold;    
    private double growFactor;

    int nreported = 0;
    
    public ClosedLongDoubleMap(int size, double loadFactor, double growFactor) {
        this.growFactor = growFactor;
        this.threshold = (int) (size*loadFactor);
        keys = new BigLongArray(size);
        values = new BigDoubleArray(size);
    }
    
    public ClosedLongDoubleMap(int size) {
        keys = new BigLongArray(size);
        values = new BigDoubleArray(size);
    }

    private int hash(long l) {
        int hash = 0;
        for(int i = 0; i < 8; i++) {
            hash += (l & 255);
            l >>= 8;
            hash += (hash << 10);
            hash ^= (hash >> 6);
        }
        hash += (hash << 3);
        hash ^= (hash >> 11);
        hash += (hash << 15);
        return hash;
    }
    
    private int maxOffset = 0;
    
    public double getOrSet(long key, double newValue) {
        if(size == values.length())
            throw new RuntimeException("Map is full!");       
        int index = hash(key) % keys.length();

        //
        if(index < 0)
            index = -index;
        
        int offset = 0;
        while(true) {
            long k = keys.get(index);
            if(k == 0) {
                keys.set(index, key);
                values.set(index, newValue);
                size++;
                if(size == threshold)
                    resize();

                /* Temporary. */
                int v = (int) ((double) size * 100.0d / (double) keys.length());
                if(v > nreported) {
                    nreported++;
                    System.err.println(" Load: " + nreported + "%");
                }
                /* End. */
                
                return newValue;                
            }
            if(k == key)
                return values.get(index);
            offset++;
            if(offset > maxOffset) {
                System.err.println(" Offset: " + offset);
                maxOffset = offset;
            }
            index += offset * offset;
            if(index < 0)
                index = -index;
            index %= keys.length();
        }        
    }   
    
	// Returns the old value.
    public double add(long key, double d) {
        if(size == values.length())
            throw new RuntimeException("Map is full!");
        int index = hash(key) % keys.length();

        //
        if(index < 0)
            index = -index;
        
        int offset = 0;
        while(true) {
            long k = keys.get(index);
            if(k == 0) {
                keys.set(index, key);
                values.set(index, d);
                size++;
                if(size == threshold)
                    resize();
                
                /* Temporary. */
                int v = (int) ((double) size * 100.0d / (double) keys.length());
                if(v > nreported) {
                    nreported++;
                    System.err.println(" Load: " + nreported + "%");
                }
                /* End. */                
                
                return 0.0;
            } else if(k == key) {
		    double oldval = values.modify(index, d) - d; // TODO check!
		    return oldval; //values.modify(index, d);
                //return;
            }
            offset++;
            index += offset * offset;
            if(index < 0)
                index = -index;
            index %= keys.length();
        }        
    }
    
    public double put(long key, double value) {
        if(size == values.length())
            throw new RuntimeException("Map is full!");
        int index = hash(key) % keys.length();

        //
        if(index < 0)
            index = -index;
        
        int offset = 0;
        while(true) {
            long k = keys.get(index);
            if(k == 0) {
                keys.set(index, key);
                values.set(index, value);
                size++;
                if(size == threshold)
                    resize();
                return 0;
            } else if(k == key) {
                double oldValue = values.get(index);
                values.set(index, value);
                return oldValue;
            }
            offset++;
            index += offset * offset;
            if(index < 0)
                index = -index;
            index %= keys.length();
        }        
    }
    
    public double get(long key) {
        int index = hash(key) % keys.length();

        //
        if(index < 0)
            index = -index;
        
        int offset = 0;
        while(true) {
            long k = keys.get(index);
            if(k == 0)
                return 0;
            if(k == key)
                return values.get(index);
            offset++;
            index += offset * offset;
            index %= keys.length();
            //
            if(index < 0)
                index = -index;

        }        
    }
    
    public boolean containsKey(long key) {
        if(key == 0)
            return false;
        int index = hash(key) % keys.length();

        //
        if(index < 0)
            index = -index;
        
        int offset = 0;
        while(true) {
            long k = keys.get(index);
            if(k == 0)
                return false;
            if(k == key)
                return true;
            offset++;
            index += offset * offset;
            index %= keys.length();
            //
            if(index < 0)
                index = -index;

        }        
    }
    
    private void resize() {
        int newSize = (int) (values.length() * growFactor);
        threshold = (int) (threshold * growFactor);
        BigLongArray oldKeys = keys;
        BigDoubleArray oldValues = values;
        size = 0;
        keys = new BigLongArray(newSize);
        values = new BigDoubleArray(newSize);
        for(int i = 0; i < oldValues.length(); i++) {
            if(oldKeys.get(i) != 0)
                put(oldKeys.get(i), oldValues.get(i));
        }
    }
    
    public long[] getKeys() {
        //return keys;
        throw new RuntimeException("Unimplemented!");
    }

    public double[] getValues() {
        //return values;
        throw new RuntimeException("Unimplemented!");
    }
 
    public int size() {
        return size;
    }
    
    public long getKey(double value) {
        for(int i = 0; i < keys.length(); i++)
            if(values.get(i) == value)
                return keys.get(i);
        return 0;
    }
    
    private static final int N = 100000000;
    
    public static void main(String[] argv) {
        ClosedLongDoubleMap map = new ClosedLongDoubleMap(2*N);
        java.util.Random rand = new java.util.Random(0);
        for(int i = 0; i < N; i++) {
            long l = rand.nextLong();
            int v = rand.nextInt();
            map.getOrSet(l, v);
        }
        System.err.println("Done.");
        
        while(true);
        
    }
    
}
