/*
 ClosedLongIntMap.java
  
 Created on Nov 6, 2007 by Richard Johansson (richard@cs.lth.se).

 $Log: ClosedLongIntMap.java,v $
 Revision 1.1  2009/01/16 10:05:11  johansson
 Added to brenta repository.

 Revision 1.15  2008/04/08 11:44:04  richard
 Uses big arrays.

 Revision 1.14  2008/04/07 12:12:33  richard
 Test method.

 Revision 1.13  2008/03/03 14:16:15  richard
 Jenkins hash.

 Revision 1.11  2008/03/03 13:51:11  richard
 New hash.

 Revision 1.10  2008/03/03 13:49:11  richard
 New hash (bug fix).

 Revision 1.9  2008/03/03 13:48:01  richard
 New hash.

 Revision 1.8  2008/03/03 13:43:07  richard
 Temporary report.

 Revision 1.7  2008/03/03 13:41:38  richard
 Temporary report.

 Revision 1.6  2008/03/03 13:11:15  richard
 Temporary report.

 Revision 1.5  2008/03/03 13:09:43  richard
 Temporary report.

 Revision 1.4  2008/03/03 12:54:34  richard
 Temporary report.

 Revision 1.3  2008/03/03 12:22:16  richard
 Handles negative keys.

 Revision 1.2  2008/02/22 13:16:59  richard
 Added getKey.

 Revision 1.1  2007/11/07 15:37:30  richard
 Added.

   
 */
package se.lth.cs.nlp.nlputils.core;

import java.io.Serializable;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class ClosedLongIntMap implements Serializable {

    private static final long serialVersionUID = 0L;
    
    private BigLongArray keys;
    private BigIntArray values;
    private int size;

    private int threshold;    
    private double growFactor;

    int nreported = 0;
    
    public ClosedLongIntMap(int size, double loadFactor, double growFactor) {
        this.growFactor = growFactor;
        this.threshold = (int) (size*loadFactor);
        keys = new BigLongArray(size);
        values = new BigIntArray(size);
    }
    
    public ClosedLongIntMap(int size) {
        keys = new BigLongArray(size);
        values = new BigIntArray(size);
    }

    private int hash(long l) {
        int hash = 0;
        for(int i = 0; i < 8; i++) {
            hash += (l & 255);
            l >>= 8;
            hash += (hash << 10);
            hash ^= (hash >> 6);
        }
        hash += (hash << 3);
        hash ^= (hash >> 11);
        hash += (hash << 15);
        return hash;
    }
    
    private int maxOffset = 0;
    
    public int getOrSet(long key, int newValue) {
        if(size == values.length())
            throw new RuntimeException("Map is full!");       
        int index = hash(key) % keys.length();

        //
        if(index < 0)
            index = -index;
        
        int offset = 0;
        while(true) {
            long k = keys.get(index);
            if(k == 0) {
                keys.set(index, key);
                values.set(index, newValue);
                size++;
                if(size == threshold)
                    resize();

                /* Temporary. */
                int v = (int) ((double) size * 100.0d / (double) keys.length());
                if(v > nreported) {
                    nreported++;
                    System.err.println(" Load: " + nreported + "%");
                }
                /* End. */
                
                return newValue;                
            }
            if(k == key)
                return values.get(index);
            offset++;
            if(offset > maxOffset) {
                System.err.println(" Offset: " + offset);
                maxOffset = offset;
            }
            index += offset * offset;
            if(index < 0)
                index = -index;
            index %= keys.length();
        }        
    }
    
    public int increment(long key) {
        if(size == values.length())
            throw new RuntimeException("Map is full!");
        int index = hash(key) % keys.length();

        //
        if(index < 0)
            index = -index;
        
        int offset = 0;
        while(true) {
            long k = keys.get(index);
            if(k == 0) {
                keys.set(index, key);
                values.set(index, 1);
                size++;
                if(size == threshold)
                    resize();
                return 0;
            } else if(k == key) {
                return values.modify(index, 1); //values[index]++;
            }
            offset++;
            index += offset * offset;
            if(index < 0)
                index = -index;
            index %= keys.length();
        }        
    }
    
    public int put(long key, int value) {
        if(size == values.length())
            throw new RuntimeException("Map is full!");
        int index = hash(key) % keys.length();

        //
        if(index < 0)
            index = -index;
        
        int offset = 0;
        while(true) {
            long k = keys.get(index);
            if(k == 0) {
                keys.set(index, key);
                values.set(index, value);
                size++;
                if(size == threshold)
                    resize();
                return 0;
            } else if(k == key) {
                int oldValue = values.get(index);
                values.set(index, value);
                return oldValue;
            }
            offset++;
            index += offset * offset;
            if(index < 0)
                index = -index;
            index %= keys.length();
        }        
    }
    
    public int get(long key) {
        int index = hash(key) % keys.length();

        //
        if(index < 0)
            index = -index;
        
        int offset = 0;
        while(true) {
            long k = keys.get(index);
            if(k == 0)
                return 0;
            if(k == key)
                return values.get(index);
            offset++;
            index += offset * offset;
            index %= keys.length();
            //
            if(index < 0)
                index = -index;

        }        
    }
    
    public boolean containsKey(long key) {
        if(key == 0)
            return false;
        int index = hash(key) % keys.length();

        //
        if(index < 0)
            index = -index;
        
        int offset = 0;
        while(true) {
            long k = keys.get(index);
            if(k == 0)
                return false;
            if(k == key)
                return true;
            offset++;
            index += offset * offset;
            index %= keys.length();
            //
            if(index < 0)
                index = -index;

        }        
    }
    
    private void resize() {
        int newSize = (int) (values.length() * growFactor);
        threshold = (int) (threshold * growFactor);
        BigLongArray oldKeys = keys;
        BigIntArray oldValues = values;
        size = 0;
        keys = new BigLongArray(newSize);
        values = new BigIntArray(newSize);
        for(int i = 0; i < oldValues.length(); i++) {
            if(oldKeys.get(i) != 0)
                put(oldKeys.get(i), oldValues.get(i));
        }
    }
    
    public long[] getKeys() {
        //return keys;
        throw new RuntimeException("Unimplemented!");
    }

    public int[] getValues() {
        //return values;
        throw new RuntimeException("Unimplemented!");
    }
 
    public int size() {
        return size;
    }
    
    public long getKey(int value) {
        for(int i = 0; i < keys.length(); i++)
            if(values.get(i) == value)
                return keys.get(i);
        return 0;
    }
    
    private static final int N = 100000000;
    
    public static void main(String[] argv) {
        ClosedLongIntMap map = new ClosedLongIntMap(2*N);
        java.util.Random rand = new java.util.Random(0);
        for(int i = 0; i < N; i++) {
            long l = rand.nextLong();
            int v = rand.nextInt();
            map.getOrSet(l, v);
        }
        System.err.println("Done.");
        
        while(true);
        
    }
    
}
