/*
 ClosedLongIntMap.java
  
 Created on Nov 6, 2007 by Richard Johansson (richard@cs.lth.se).

 $Log: ClosedObjectIntMap.java,v $
 Revision 1.1  2010-06-07 11:02:59  johansson
 Added.

 Revision 1.1  2009/01/16 10:05:11  johansson
 Added to brenta repository.

 Revision 1.15  2008/04/08 11:44:04  richard
 Uses big arrays.

 Revision 1.14  2008/04/07 12:12:33  richard
 Test method.

 Revision 1.13  2008/03/03 14:16:15  richard
 Jenkins hash.

 Revision 1.11  2008/03/03 13:51:11  richard
 New hash.

 Revision 1.10  2008/03/03 13:49:11  richard
 New hash (bug fix).

 Revision 1.9  2008/03/03 13:48:01  richard
 New hash.

 Revision 1.8  2008/03/03 13:43:07  richard
 Temporary report.

 Revision 1.7  2008/03/03 13:41:38  richard
 Temporary report.

 Revision 1.6  2008/03/03 13:11:15  richard
 Temporary report.

 Revision 1.5  2008/03/03 13:09:43  richard
 Temporary report.

 Revision 1.4  2008/03/03 12:54:34  richard
 Temporary report.

 Revision 1.3  2008/03/03 12:22:16  richard
 Handles negative keys.

 Revision 1.2  2008/02/22 13:16:59  richard
 Added getKey.

 Revision 1.1  2007/11/07 15:37:30  richard
 Added.

   
 */
package se.lth.cs.nlp.nlputils.core;

import java.io.Serializable;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class ClosedObjectIntMap<V> implements Serializable {

    private static final long serialVersionUID = 0L;
    
    private BigArray<V> keys;
    private BigIntArray values;
    private int size;

    int nreported = 0;
    
    public ClosedObjectIntMap(int size) {
        keys = new BigArray(size);
        values = new BigIntArray(size);
    }

    private int maxOffset = 0;
    
    protected int hash(V key) {
    	return key.hashCode();
    }
    
    public int getOrSet(V key, int newValue) {
        if(size == values.length())
            throw new RuntimeException("Map is full!");       
        int index = hash(key) % keys.length();

        //
        if(index < 0)
            index = -index;
        
        int offset = 0;
        while(true) {
            V k = keys.get(index);
            if(k == null) {
                keys.set(index, key);
                values.set(index, newValue);
                size++;

                /* Temporary. */
                int v = (int) ((double) size * 100.0d / (double) keys.length());
                if(v > nreported) {
                    nreported++;
                    System.err.println(" Load: " + nreported + "%");
                }
                /* End. */
                
                return newValue;                
            }
            if(k.equals(key))
                return values.get(index);
            offset++;
            if(offset > maxOffset) {
                System.err.println(" Offset: " + offset);
                maxOffset = offset;
            }
            index += offset * offset;
            if(index < 0)
                index = -index;
            index %= keys.length();
        }        
    }
    
    public int increment(V key) {
        if(size == values.length())
            throw new RuntimeException("Map is full!");
        int index = hash(key) % keys.length();

        //
        if(index < 0)
            index = -index;
        
        int offset = 0;
        while(true) {
            V k = keys.get(index);
            if(k == null) {
                keys.set(index, key);
                values.set(index, 1);
                size++;
                return 0;
            } else if(k.equals(key)) {
                return values.modify(index, 1); //values[index]++;
            }
            offset++;
            index += offset * offset;
            if(index < 0)
                index = -index;
            index %= keys.length();
        }        
    }
    
    public int put(V key, int value) {
        if(size == values.length())
            throw new RuntimeException("Map is full!");
        int index = hash(key) % keys.length();

        //
        if(index < 0)
            index = -index;
        
        int offset = 0;
        while(true) {
            V k = keys.get(index);
            if(k == null) {
                keys.set(index, key);
                values.set(index, value);
                size++;
                return 0;
            } else if(k.equals(key)) {
                int oldValue = values.get(index);
                values.set(index, value);
                return oldValue;
            }
            offset++;
            index += offset * offset;
            if(index < 0)
                index = -index;
            index %= keys.length();
        }        
    }
    
    public int get(V key) {
        int index = hash(key) % keys.length();

        //
        if(index < 0)
            index = -index;
        
        int offset = 0;
        while(true) {
            V k = keys.get(index);
            if(k == null)
                return 0;
            if(k.equals(key))
                return values.get(index);
            offset++;
            index += offset * offset;
            index %= keys.length();
            //
            if(index < 0)
                index = -index;

        }        
    }
    
    public boolean containsKey(V key) {
        if(key == null)
            return false;
        int index = hash(key) % keys.length();

        //
        if(index < 0)
            index = -index;
        
        int offset = 0;
        while(true) {
            V k = keys.get(index);
            if(k == null)
                return false;
            if(k.equals(key))
                return true;
            offset++;
            index += offset * offset;
            index %= keys.length();
            //
            if(index < 0)
                index = -index;

        }        
    }
    
    public String[] getKeys() {
        //return keys;
        throw new RuntimeException("Unimplemented!");
    }

    public int[] getValues() {
        //return values;
        throw new RuntimeException("Unimplemented!");
    }
 
    public int size() {
        return size;
    }
    
}
