
/*
 DOMUtils.java
 
 Copyright (C) 2005, Richard Johansson (richard@cs.lth.se).
 
 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.
 
 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.
 
 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
 Created in 2003 by Richard Johansson (richard@cs.lth.se).
 
 $Log: DOMUtils.java,v $
 Revision 1.1  2009/01/16 10:05:11  johansson
 Added to brenta repository.

 Revision 1.6  2007/08/17 10:19:52  richard
 Added removeDTDString.

 Revision 1.5  2006/09/26 13:10:18  richard
 Ignore document type if null.

 Revision 1.4  2006/09/26 13:07:39  richard
 May skip validation.

 Revision 1.3  2006/09/22 15:47:33  richard
 Throws exception.

 Revision 1.2  2006/09/07 17:51:19  richard
 New revision.

 Revision 1.1  2006/05/16 13:38:13  richard
 Re-added because of CVS problems.
 
 Revision 1.2  2006/04/07 14:46:08  richard
 Added documentation.
 
 Revision 1.1  2005/11/28 15:44:29  richard
 Added the file.
 
 Revision 1.10  2004/09/03 08:48:48  richard
 Turned on validation.
 
 Revision 1.9  2004/08/30 14:16:40  richard
 Added some more utility methods.
 
 Revision 1.8  2004/01/16 16:04:45  richard
 Replaced the input stream with an input source.
 
 Revision 1.7  2004/01/14 13:39:32  richard
 Allowed documents without declared document type.
 
 Revision 1.6  2003/12/18 09:22:16  richard
 Faster id->node lookup by means of a hashtable.
 
 Revision 1.5  2003/10/14 09:02:06  xd99rj
 Fixed a bug in the string->tree conversion.
 
 Revision 1.4  2003/09/15 14:10:42  xd99rj
 The output now takes locale into account.
 
 Revision 1.3  2003/06/25 14:35:57  xd99rj
 Added support for the printout and document reading.
 
 Revision 1.2  2003/06/25 11:46:14  xd99rj
 Created utility methods to make the traversing easier.
 
 Revision 1.1  2003/05/22 08:33:45  xd99rj
 Added the source files.
 
 */

package se.lth.cs.nlp.nlputils.core;

import java.util.Hashtable;

import java.util.regex.*;

import javax.xml.parsers.*;

import org.w3c.dom.*;
import org.xml.sax.InputSource;
import java.io.*;

import javax.xml.transform.*;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
//import javax.xml.transform.stream.StreamSource;
import javax.xml.transform.OutputKeys;

/**
 * A utility class which contains static methods to simplify some chores
 * common when working with DOM trees.
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class DOMUtils {
    
    private static final Pattern encodingPattern 
    = Pattern.compile("\\<\\?xml.*?encoding=\"(.*?)\".*?\\?\\>");
       
    private DOMUtils() {}
    
    /**
     * Replaces a child node.
     *
     * This method replaces a certain child node when the children must appear
     * in a specific order.
     *
     * @param n The parent node.
     * @param childName The node name of the child to replace.
     * @param newChild The new child node.
     * @param predecessors The names of those children which may appear before
     *                     the child to replace.
     */
    public static void replaceChildByName(Node n,
            String childName, 
            Node newChild,
            String[] predecessors) {
        int predNbr = 0;
        Node child = n.getFirstChild();
        
        while(predNbr < predecessors.length && child != null) {
            if(child.getNodeName().equals(predecessors[predNbr]))
                child = child.getNextSibling();
            predNbr++;
        }
        
        if(child == null)
            n.appendChild(newChild);
        else if(child.getNodeName().equals(childName))
            n.replaceChild(newChild, child);
        else
            n.insertBefore(newChild, child);
        
    }
    
    /**
     * Finds the first child node with a certain name if one exists.
     * @param n The parent node.
     * @param childName The name of the child node.
     * @return The child node.
     */
    public static Node findChildByName(Node n, String childName) {
        Node child = n.getFirstChild();
        while(child != null) {
            if(child.getNodeName().equals(childName))
                return child;
            child = child.getNextSibling();
        }
        return null;
    } 
    
    /**
     * Creates a new document.
     */
    public static Document createDocument(String docName, String dtd) {
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        
        factory.setValidating(true);
        //factory.setNamespaceAware(true);
        
        try {
            DocumentBuilder builder = factory.newDocumentBuilder();
            DOMImplementation domImpl = builder.getDOMImplementation();
            
            DocumentType docType = null;
            
            if(dtd != null)
                docType = domImpl.createDocumentType(docName, null, dtd);
            return domImpl.createDocument(null, docName, docType);
            
        } catch (ParserConfigurationException e) {
            e.printStackTrace();
            return null;
        }
    }
    
    /**
     * Reads a document.
     */
    public static Document readDocument(InputSource in) {
        
        DocumentBuilderFactory factory =
            DocumentBuilderFactory.newInstance();
        
        factory.setValidating(true);   
        //factory.setNamespaceAware(true);
        
        try {
            
            in.setSystemId("file:.");
            
            DocumentBuilder builder = factory.newDocumentBuilder();
            //Document document = builder.parse(in, "file:.");
            Document document = builder.parse(in);
            
            return document;
            
        } catch (Exception e) {
            e.printStackTrace();
            return null;
        }
        
    }
    
    public static Document readDocument(InputStream in) {
        return readDocument(in, true);
    }
    
    /**
     * Reads a document.
     */
    public static Document readDocument(InputStream in, boolean validate) {
        //return readDocument(new InputSource(in));
        DocumentBuilderFactory factory =
            DocumentBuilderFactory.newInstance();
        
        factory.setValidating(validate);   
        //factory.setNamespaceAware(true);
        
        try {
            
            //in.setSystemId("file:.");
            
            DocumentBuilder builder = factory.newDocumentBuilder();
            //Document document = builder.parse(in, "file:.");
            Document document = builder.parse(in);
            
            return document;
            
        } catch (Exception e) {
            e.printStackTrace();
            return null;
        }
    }	
    
    /**
     *
     */
    public static void writeDocument(Document document, StreamResult out) throws IOException {
    	try {
            StringWriter writer = new StringWriter();
            TransformerFactory tFactory = TransformerFactory.newInstance();
            Transformer transformer = tFactory.newTransformer();
            DOMSource source = new DOMSource(document);

            if(document.getDoctype() != null
               && document.getDoctype().getSystemId() != null)
                transformer.setOutputProperty(OutputKeys.DOCTYPE_SYSTEM, 
                        document.getDoctype().getSystemId());

            //if(charset != null)
            //transformer.setOutputProperty(OutputKeys.ENCODING, charset);
            
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");
            transformer.setOutputProperty("{http://xml.apache.org/xslt}"
                    + "indent-amount", "4");
            transformer.transform(source, out);
        } catch(Exception e) {
            e.printStackTrace();
            throw new IOException(e.getMessage());
        }
    }
    
    /**
     *
     */
    public static void writeDocument(Document document, OutputStream out) throws IOException {
    	StreamResult sr = new StreamResult(out);
    	writeDocument(document, sr);
    }
    
    /**
     * Returns a string containing the XML printout of a document.
     */
    public static String treeToString(Document document, String charset) {
        try {
            StringWriter writer = new StringWriter();
            TransformerFactory tFactory = TransformerFactory.newInstance();
            Transformer transformer = tFactory.newTransformer();
            DOMSource source = new DOMSource(document);
            StreamResult result = new StreamResult(writer);
            
            if(document.getDoctype() != null 
                    && document.getDoctype().getSystemId() != null)
                transformer.setOutputProperty(OutputKeys.DOCTYPE_SYSTEM, 
                        document.getDoctype().getSystemId());
            if(charset != null)
                transformer.setOutputProperty(OutputKeys.ENCODING, charset);
            
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");
            transformer.setOutputProperty("{http://xml.apache.org/xslt}"
                    + "indent-amount", "4");
            transformer.transform(source, result);
            return writer.toString();
        } catch(Exception e) {
            e.printStackTrace();
            return null;
        }
    }
    
    public static String treeToString(Document document) {
        String charset = System.getProperty("file.encoding");
        return treeToString(document, charset);
    }
    
    
    /**
     * Converts the byte array to a string, taking the declared XML character
     * encoding into account.
     */
//  public static String XMLBytesToString(byte[] bytes) {
//  String tmp = new String(bytes);
    
//  Matcher m = encodingPattern.matcher(tmp);
    
//  if(m.find()) {
//  String encoding = xmlEncodingNameToJavaEncodindName(m.group(1));
//  if(encoding != null)
//  try {
//  return new String(bytes, encoding);
//  } catch(Exception e) {
//  e.printStackTrace();
//  }
//  } 
//  return tmp;
//  }
    
    /**
     * Converts the byte array to a Document, taking the declared XML character
     * encoding into account.
     */
    public static Document XMLBytesToDocument(byte[] bytes) {
        System.out.print("\n|");
        try { System.out.write(bytes); } catch(Exception e) {e.printStackTrace();}
        System.out.println("|");
        //InputSource is = new InputSource(new ByteArrayInputStream(bytes));
        //return readDocument(is);
        return readDocument(new ByteArrayInputStream(bytes));
    }
    
    /**
     * Converts the string to a byte array, taking the declared XML character
     * encoding into account.
     */
//  public static byte[] stringToXMLBytes(String s) {
//  Matcher m = encodingPattern.matcher(s);
//  if(m.find()) {
//  String encoding = //xmlEncodingNameToJavaEncodindName(m.group(1));
//  m.group(1);
//  if(encoding != null)
//  try {
//  //return s.getBytes(encoding);
    
//  Charset charset = Charset.forName(encoding);
//  //CharsetDecoder decoder = charset.newDecoder();
//  CharsetEncoder encoder = charset.newEncoder();
    
//  ByteBuffer bbuf = encoder.encode(CharBuffer.wrap(s));
    
//  } catch(Exception e) {
//  e.printStackTrace();
//  }
//  }
//  return s.getBytes();
//  }
    
    /**
     * Converts the Document to a byte array, taking the declared XML character
     * encoding into account.
     */
    public static byte[] documentToXMLBytes(Document document, 
            String encoding) {
        try {
            ByteArrayOutputStream os = new ByteArrayOutputStream();
            TransformerFactory tFactory = TransformerFactory.newInstance();
            Transformer transformer = tFactory.newTransformer();
            DOMSource source = new DOMSource(document);
            StreamResult result = new StreamResult(os);
            
            if(document.getDoctype() != null)
                transformer.setOutputProperty(OutputKeys.DOCTYPE_SYSTEM, 
                        document.getDoctype().getSystemId());
            if(encoding != null)
                transformer.setOutputProperty(OutputKeys.ENCODING, encoding);
            
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");
            transformer.setOutputProperty("{http://xml.apache.org/xslt}"
                    + "indent-amount", "4");
            
            transformer.transform(source, result);
            
            return os.toByteArray();
        } catch(Exception e) {
            e.printStackTrace();
            return null;
        }
        
    }
    
    
//  private static String xmlEncodingNameToJavaEncodindName(String xmlName) {
//  if(xmlName.matches("ISO-8859-1"))
//  return "ISO8859-1";
//  if(xmlName.matches("UTF-8"))
//  return "UTF-8";
//  System.err.println("Unknown encoding: " + xmlName);
//  return null;
//  }
    
    private static Hashtable nodesToIds = new Hashtable();
    
    private static class Pair {
        public Object left, right;
        public Pair(Object left, Object right) {
            this.left = left;
            this.right = right;
        }
        public int hashCode() {
            return 31 * left.hashCode() + right.hashCode();
        }
        public boolean equals(Object o) {
            if(!(o instanceof Pair))
                return false;
            Pair p = (Pair) o;
            return left.equals(p.left) && right.equals(p.right);
        }
    }
    
    private static final boolean USE_TABLE = true;
    private static final int MAX_TABLE_SIZE = 10000;
    
    private static long total = 0;
    private static int ncalls = 0;
    
    private static int nhits = 0;
    
    /**
     * This is needed because Document.getElementById does not seems to work
     * well in our DOM implementation.
     * We assume that every attribute which is called "id" is an id.
     * Will only find nodes which are reachable from the first node.
     */
    public static Node findNodeById(Node node, String id) {
        long t0 = System.currentTimeMillis();
        
        Node n = null;
        if(USE_TABLE) {
            Pair p = new Pair(node, id);
            n = (Node) nodesToIds.get(p);
            if(n == null) {
                n = findNodeByIdInternal(node, id);
                if(n != null) {
                    if(nodesToIds.size() > MAX_TABLE_SIZE)
                        nodesToIds = new Hashtable();
                    nodesToIds.put(p, n);
                }		    
            } else
                nhits++;
        } else {
            n = findNodeByIdInternal(node, id);
        }
        
        total += (System.currentTimeMillis() - t0);
        ncalls += 1;
        return n;
    }
    
    public static void cacheNode(Node fromNode, String id, Node toNode) {
        if(nodesToIds.size() > MAX_TABLE_SIZE)
            nodesToIds = new Hashtable();
        nodesToIds.put(new Pair(fromNode, id), toNode);
    }
    
    public static void printStats() {
        System.out.println("Total time in findNodeById: " + 
                (double)total/1000.0);
        System.out.println("Total number of calls: " + ncalls); 
        System.out.println("Number of hits: " + nhits);
        System.out.println("Average time per call: "
                + ((double)total/(double)ncalls)/1000.0);
        ncalls = 0;
        nhits = 0;
        total = 0;
    }
    
    private static Node findNodeByIdInternal(Node node, String id) {
        if(node instanceof Element) {
            if(((Element) node).getAttribute("id").equals(id)) {
                return node;
            }
        }
        Node child = node.getFirstChild();
        while(child != null) {
            Node result = findNodeByIdInternal(child, id);
            if(result != null)
                return result;
            child = child.getNextSibling();
        }
        return null;
    }
    
    /**
     * Validates a document against its DTD.
     */
    public static void validate(Document document) {
        
        String text = treeToString(document);
        InputSource source = new InputSource(new StringReader(text));
        
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        
        factory.setValidating(true);   
        
        try {
            
            DocumentBuilder builder = factory.newDocumentBuilder();
            builder.parse(source);
            
        } catch (Exception e) {
            e.printStackTrace();
        }
    }
    
    /**
     * Returns the first non-text child of the node.
     */
    public static Node firstChild(Node node) {
        Node child = node.getFirstChild();
        while(child != null && child instanceof Text)
            child = child.getNextSibling();
        return child;
    }
    
    /**
     * Returns the next non-text sibling of the node.
     */
    public static Node nextSibling(Node node) {
        do {
            node = node.getNextSibling();
        } while(node != null && node instanceof Text);
        return node;
    }
    
    /**
     * Returns the last non-text child of the node.
     */
    public static Node lastChild(Node node) {
        Node child = node.getLastChild();
        while(child != null && child instanceof Text)
            child = child.getPreviousSibling();
        return child;
    }
    
    /**
     * Returns the previous non-text sibling of the node.
     */
    public static Node previousSibling(Node node) {
        do {
            node = node.getPreviousSibling();
        } while(node != null && node instanceof Text);
        return node;
    }
   
    static final int BUF_SIZE = 1 << 12;
    public static InputStream removeDTDString(InputStream is) throws IOException {
        ByteArrayOutputStream os = new ByteArrayOutputStream();
        byte[] buf = new byte[BUF_SIZE];
        while(true) {
            int len = is.read(buf);
            if(len == -1)
                break;
            os.write(buf, 0, len);
        }
        String s = os.toString("UTF-8");
        s = s.replaceFirst("<!\\s*DOCTYPE.*?>", "\n");
        return new ByteArrayInputStream(s.getBytes("UTF-8"));
    }
    
}
