/*
 Multimap.java
 
 Copyright (C) 2006, Richard Johansson (richard@cs.lth.se).
 
 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.
 
 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.
 
 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
 Created on Nov 7, 2006 by Richard Johansson (richard@cs.lth.se).
 
 $Log: Multimap.java,v $
 Revision 1.1  2009/01/16 10:05:11  johansson
 Added to brenta repository.

 Revision 1.7  2006/11/20 10:21:18  richard
 Added keySet.

 Revision 1.6  2006/11/16 21:18:03  richard
 New revision.

 Revision 1.5  2006/11/16 09:45:19  richard
 Added putAll.

 Revision 1.4  2006/11/08 10:29:09  richard
 Serializable.

 Revision 1.3  2006/11/07 10:37:03  richard
 Added.

 
 */
package se.lth.cs.nlp.nlputils.core;

import java.util.*;
import java.lang.reflect.*;
import java.io.Serializable;

/**
 * A map from a key type to a collection of the value type.
 * 
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class Multimap<K, V> implements Serializable {
    
    private static final long serialVersionUID = 0;
    
    private HashMap<K, Collection<V>> map = new HashMap<K, Collection<V>>();
    private Constructor con;
    
    public Multimap(Class cl) {
        if(!Collection.class.isAssignableFrom(cl))
            throw new IllegalArgumentException("Must specify a Collection implementation!");
        try {
            con = cl.getConstructor(new Class[0]);
        } catch(Exception e) {
            throw new RuntimeException(e);
        }
    }

    private Collection<V> create() {
        try {
            return (Collection<V>) con.newInstance(null);
        } catch(Exception e) {
            throw new RuntimeException(e);
        }
    }
    
    /**
     * Adds <code>value</code> to the association list of <code>key</code>.
     * @param key the key.
     * @param value the value.
     */
    public void put(K key, V value) {
        Collection<V> l = map.get(key);
        if(l == null) {
            l = create();
            map.put(key, l);
        }
        l.add(value);
    }

    public void putAll(K key, Collection<V> values) {
        Collection<V> l = map.get(key);        
        if(l == null) {
            l = create();
            map.put(key, l);
        }
        l.addAll(values);
    }
    
    public void putAll(K key, V[] values) {
        putAll(key, Arrays.asList(values));
    }
    
    /**
     * Returns the association list for <code>key</code>.
     * @param key the key.
     * @return the list.
     */
    public Collection<V> get(K key) {
        return map.get(key);
    }
    
    /**
     * Checks whether the map contains a given key.
     * @param key the key.
     */
    public boolean containsKey(K key) {
        return map.containsKey(key);
    }

    public String toString() {
        return map.toString();
    }

    public int size() {
        return map.size();
    }
    
    public Set<K> keySet() {
        return map.keySet();
    }
    
}

