/*
 MultipleHashIntMap.java
  
 Created on Apr 10, 2008 by Richard Johansson (richard@cs.lth.se).

 $Log: MultipleHashIntMap.java,v $
 Revision 1.1  2009/01/16 10:05:11  johansson
 Added to brenta repository.

 Revision 1.1  2008/04/14 10:45:43  richard
 Added.

   
 */
package se.lth.cs.nlp.nlputils.core;

import java.io.Serializable;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class MultipleHashIntMap implements Serializable {

    private static final long serialVersionUID = 0L;
    
    private int[] data;    
    private int mask;
    private int size;
    private int limit;
    
    public MultipleHashIntMap(int nbits, double maxLoad) {
        this.mask = (1 << nbits) - 1;
        data = new int[1 << (nbits + 1)];
        limit = (int) Math.ceil(data.length/2 * maxLoad);
    }
    
    public int get(int hash1, int hash2) {
        int index = hash1 & mask;
        int offset = 0;
        while(true) {
            int i1 = index << 1;
            if(data[i1] == 0 || data[i1] == hash2)
                return data[i1 + 1];
            offset++;
            index = (index + offset) & mask;
        }
    }
    
    public int put(int hash1, int hash2, int value) {
        int index = hash1 & mask;
        int offset = 0;
        while(true) {
            int i1 = index << 1;
            if(data[i1] == 0) {
                data[i1] = hash2;
                data[i1 + 1] = value;
                size++;
                if(size > limit)
                    throw new RuntimeException("Map is full!");
                return 0;                
            } else if(data[i1] == hash2) {
                int old = data[i1 + 1];
                data[i1 + 1] = value;
                return old;
            }
            offset++;
            index = (index + offset) & mask;
        }
    }
    
    public int getOrPut(int hash1, int hash2, int value) {
        int index = hash1 & mask;
        int offset = 0;
        while(true) {
            int i1 = index << 1;
            if(data[i1] == 0) {
                data[i1] = hash2;
                data[i1 + 1] = value;
                size++;
                if(size > limit)
                    throw new RuntimeException("Map is full!");
                return 0;                
            } else if(data[i1] == hash2)
                return data[i1 + 1];
            offset++;
            index = (index + offset) & mask;
        }
    }
    
    public int size() {
        return size;
    }
    
    /* Tests. */
    
    private static int hash1(long l) {
        int hash = 0;
        for(int i = 0; i < 8; i++) {
            hash += (l & 255);
            l >>>= 8;
            hash += (hash << 10);
            hash ^= (hash >>> 6);

        }
        hash += (hash << 3);
        hash ^= (hash >>> 11);
        hash += (hash << 15);

        if(hash == 0)
            throw new RuntimeException("hash1 = 0");
        return hash;
    }
    
    private static int hash3(long l) {
        int hash = 0;
        for(int i = 0; i < 4; i++) {
            hash += (l & 65535);
            l >>>= 16;
            hash += (hash << 10);
            hash ^= (hash >>> 6);
        }
        hash += (hash << 3);
        hash ^= (hash >>> 11);
        hash += (hash << 15);
        if(hash == 0)
            throw new RuntimeException("hash1 = 0");
        return hash;
    }
    
    private static final long HM = (1L << 32L) - 1L;
    
    private static int hash4(long l) {
        int hash = 0;
        for(int i = 0; i < 2; i++) {
            hash += (l & HM);
            l >>>= 32;
            hash += (hash << 10);
            hash ^= (hash >>> 6);            
        }
        hash += (hash << 3);        
        hash ^= (hash >>> 11);
        hash += (hash << 15);
        if(hash == 0)
            throw new RuntimeException("hash1 = 0");
        return hash;
    }
    
    static long MASK = 255L << 56L; 
    
    private static int hash2(long l) {
        int hash = 0;
        for(int i = 0; i < 8; i++) {
            hash += (l & MASK) >> 56L;
            l <<= 8;
            hash += (hash << 10);
            hash ^= (hash >>> 6);
        }
        hash += (hash << 3);
        hash ^= (hash >>> 11);
        hash += (hash << 15);
        if(hash == 0)
            throw new RuntimeException("hash2 = 0");
        return hash;
    }

    private static int nDifferingBits(int l1, int l2) {
        int xor = l1 ^ l2;
        int count = 0;
        while(xor != 0) {
            count += xor & 1;
            xor >>>= 1;
        }
        return count;
    }
    
    private static void testHash() {
        java.util.Random rand = new java.util.Random(1);
        long l = rand.nextLong();
        int N = 20000000;
        int sum = 0;
        for(int i = 0; i < N; i++) {
            long bit = rand.nextInt(64);
            long l2 = l ^ (1L << bit);
            sum += nDifferingBits(hash1(l), hash1(l2));
            l = l2;
        }
        double d = (double) sum / N;
        System.out.println("Average number of bit flips in hash: " + d);
    }
    
    public static void main(String[] argv) {
        //testHash();
        
        /*
        java.util.Random rand = new java.util.Random(0);
        int N = 10;
        int nbits = 5;
        
        java.util.HashMap<Long, Integer> checkm = new java.util.HashMap(); 
        int ncollisions = 0;
        
        MultipleHashIntMap m = new MultipleHashIntMap(nbits, 0.7);
        int idCounter = 1;
        for(int i = 0; i < N; i++) {
            long l = rand.nextLong() % (N/2);
            while(l == 0)
                l = rand.nextLong() % (N/2);
            int h1 = hash1(l);
            int h2 = hash2(l);
            int result = m.getOrPut(h1, h2, idCounter);
            if(result == 0) {
                if(checkm.containsKey(l))
                    throw new RuntimeException("Shouldn't contain l");
                checkm.put(l, idCounter);
                idCounter++;
            } else {
                Integer checkvalue = checkm.get(l);
                if(checkvalue == null || !checkvalue.equals(result))
                    ncollisions++;
            }
        }
        
        System.out.println(checkm.size());
        System.out.println(m.size());
        System.out.println("ncollisions = " + ncollisions);

        */

        java.util.Random rand = new java.util.Random(0);
        int nbits = 5;
        MultipleHashIntMap m = new MultipleHashIntMap(nbits, 0.7);
        long l = rand.nextLong();
        int h1 = hash1(l);
        int h2 = hash2(l);
        
        System.out.println(m.getOrPut(h1, h2, 12));
        System.out.println(m.getOrPut(h1, h2, 13));
        System.out.println(m.getOrPut(h1, h2, 14));

        long l2 = rand.nextLong();
        h2 = hash2(l2);
        
        System.out.println(m.getOrPut(h1, h2, 22));
        System.out.println(m.getOrPut(h1, h2, 23));
        System.out.println(m.getOrPut(h1, h2, 24));
        
        long l3 = rand.nextLong();
        h1 = hash1(l3);
        
        System.out.println(m.getOrPut(h1, h2, 32));
        System.out.println(m.getOrPut(h1, h2, 33));
        System.out.println(m.getOrPut(h1, h2, 34));

        h1 = hash1(l);
        h2 = hash2(l);
        
        System.out.println(m.getOrPut(h1, h2, 32));
        System.out.println(m.getOrPut(h1, h2, 33));
        System.out.println(m.getOrPut(h1, h2, 34));
        
    }    
    
}
