package se.lth.cs.nlp.nlputils.core;

import java.lang.reflect.*;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.TreeMap;

public class OptionParser {

    private String prefix;
    private boolean allowShortBools;
    
    public OptionParser(String prefix, boolean allowShortBools) {
        this.prefix = prefix;
        this.allowShortBools = allowShortBools;
    }
    
    public <T extends CommandLineOptions> T parse(Class<T> cl, String[] argv, int start, int end) {
        try {
            Constructor con;
            try {
                con = cl.getConstructor();
            } catch(Exception e) {
                con = cl.getDeclaredConstructor();
            }
            T out = (T) con.newInstance();

            for(int i = start; i < end; i++) {
                String opt = argv[i];
                if(!opt.startsWith(prefix))
                    throw new IllegalArgumentException("Illegal option format: " + opt);
                String[] ts = opt.substring(prefix.length()).split("=", 2);
                String optName = ts[0];
                if(ts.length > 2)
                    throw new IllegalArgumentException("Illegal option format: " + opt);
                
                Field f = null;
                try { 
                    f = cl.getField(optName);
                } catch(NoSuchFieldException e) {
                    try { 
                        f = cl.getDeclaredField(optName);
                    } catch(NoSuchFieldException e2) {
                        throw new IllegalArgumentException("Illegal option: " + prefix + optName);
                    }
                }
                
                Object val;
                Class ftype = f.getType();
                if(allowShortBools && (ftype == boolean.class || ftype == Boolean.class)) {
                    if(ts.length < 2) {
                        if(allowShortBools)
                            val = true;
                        else
                            throw new IllegalArgumentException("Option value missing: " + opt);
                    } else
                        val = Boolean.parseBoolean(ts[1]);
                } else {
                    if(ts.length < 2)
                        throw new IllegalArgumentException("Option value missing: " + opt);
                    val = parseValue(ftype, optName, ts[1]);
                }
                f.set(out, val);
            }
            
            return out;
        } catch(Exception e) {
            throw new RuntimeException(e);
        }
    }
    
    private Object parseValue(Class ftype, String optName, String s) {
        Class ctype = ftype.getComponentType();
        if(ctype != null)
            return parseArray(ftype, ctype, optName, s);
        else if(ftype == String.class) {
            if(s.startsWith("\"") && s.endsWith("\""))
                s = s.substring(1, s.length()-1);
            return s;
        } else if(ftype == int.class || ftype == Integer.class)
            return Integer.parseInt(s);
        else if(ftype == double.class || ftype == Double.class)
            return Double.parseDouble(s);
        else if(ftype == byte.class || ftype == Byte.class)
            return Byte.parseByte(s);
        else if(Enum.class.isAssignableFrom(ftype))
            return Enum.valueOf(ftype, s);
        else 
            throw new IllegalArgumentException("Illegal class of option: " + optName);
    }
    
    private Object parseArray(Class ftype, Class ctype, String optName, String s) {
        Object[] os = ListParser.parse(s);
        String[] ss = flatten(os);
        Object arr = Array.newInstance(ctype, ss.length);
        for(int i = 0; i< ss.length; i++)
            Array.set(arr, i, parseValue(ctype, optName, ss[i]));
        return arr;
    }
    
    private String[] flatten(Object[] os) {
        String[] out = new String[os.length];
        for(int i = 0; i < os.length; i++) {
            if(os[i] instanceof Object[])
                out[i] = Arrays.deepToString((Object[]) os[i]);
            else
                out[i] = (String) os[i];
        }
        return out;
    }
    
    public <T extends CommandLineOptions> T parse(Class<T> cl, String[] argv) {
        return parse(cl, argv, 0, argv.length);
    }
    
    public void execute(Class cl, String[] argv, int start) {
        if(argv.length < start + 1)
            throw new IllegalArgumentException("Illegal option list: " + Arrays.toString(argv));
        try {
            String opt = argv[start];
            if(!opt.startsWith(prefix))
                throw new IllegalArgumentException("Illegal option: " + opt);
            opt = opt.substring(prefix.length());
            Method method = cl.getMethod(opt, String[].class);
            method.invoke(null, (Object) argv);
        } catch(Exception e) {
            throw new RuntimeException(e);
        }
    }

    public static void print(CommandLineOptions opts) {
        Class cl = opts.getClass();
        TreeMap<String, Object> values = new TreeMap();        
        int maxLen = 0;
        for(Field f: cl.getFields()) {
            try {
                String k = f.getName();
                if(k.length() > maxLen)
                    maxLen = k.length();
                //System.out.println(f.getName() + " = " + f.get(opts));
                values.put(k, f.get(opts));
            } catch(IllegalAccessException e) {
                // 
            }
        }
        for(Field f: cl.getDeclaredFields()) {
            try {
                String k = f.getName();
                if(k.length() > maxLen)
                    maxLen = k.length();
                //System.out.println(f.getName() + " = " + f.get(opts));
                values.put(k, f.get(opts));
            } catch(IllegalAccessException e) {
                // 
            }            
        }
        for(String k: values.keySet()) {
            System.out.print(k);
            for(int i = k.length(); i < maxLen; i++)
                System.out.print(" ");
            System.out.println(" = " + values.get(k));
        }
    }
    
    static enum TestEnum {
        VAL1,
        VAL2;
    }
    
    static class TestOptions implements CommandLineOptions {
        
        int intOpt1;
        int intOpt2;
        
        boolean boolOpt;
        
        String stringOpt;
        
        double doubleOpt;

        double[] darrOpt;
        String[] strArrOpt;
        
        double[][] deepOpt;
        
        TestEnum enumOpt;
        
        TestEnum[][][] deepOpt2;
        
    }
    
    public static void main(String[] argv) {
        OptionParser op = new OptionParser("-", true);
        TestOptions opts = op.parse(TestOptions.class, argv);
        
        //System.out.println(opts.enumOpt);
        //System.out.println(Arrays.toString(opts.enumOpt));
        //System.out.println(Arrays.toString(opts.strArrOpt));
        System.out.println(Arrays.deepToString(opts.deepOpt2));

        //Class cl = double[].class;
        //Class cl2 = cl.getComponentType();
        //System.out.println(cl2.getName());
        
        //System.out.println(Object[].class.isAssignableFrom(Object[][].class));
        
    }
    
}
