package se.lth.cs.nlp.nlputils.core;

import java.io.*;
import java.util.*;

import se.lth.cs.nlp.nlputils.math.PermutationTest;
import se.lth.cs.nlp.nlputils.math.Resampler;
import se.lth.cs.nlp.nlputils.math.Statistics;

public class PRFStats implements java.io.Serializable {

	private static final long serialVersionUID = 1L;
	
	public double nCorrect, nInGold, nGuesses, propFound, propGuessed, nOverlap;

	private ArrayList<PRFStats> samples;
	
	public PRFStats() {		
	}
	
	public PRFStats(boolean allSamples) {		
		if(allSamples)
			samples = new ArrayList();
	}
	
	public PRFStats clone() {
		PRFStats out = new PRFStats();
		out.nCorrect = nCorrect;
		out.nInGold = nInGold;
		out.nGuesses = nGuesses;
		out.propFound = propFound;
		out.propGuessed = propGuessed;
		out.nOverlap = nOverlap;
		if(samples != null) {
			out.samples = new ArrayList();
			for(PRFStats sample: samples)
				out.samples.add(sample.clone());
		}
		return out;
	}
	
	public PRFStats(String[] ss) {		
		if(ss.length != 6)
			throw new RuntimeException("illegal number of tokens, should be 6");			
		nCorrect = Double.parseDouble(ss[0]);
		nGuesses = Double.parseDouble(ss[1]);
		nInGold = Double.parseDouble(ss[2]);

		// pg:3, pf: 4
		propGuessed = Double.parseDouble(ss[3]);
		propFound = Double.parseDouble(ss[4]);
		nOverlap = Double.parseDouble(ss[5]);
	}

	public int size() {
		if(samples == null)
			return -1;
		else
			return samples.size();
	}
	
	public void add(PRFStats other) {
		nCorrect += other.nCorrect;
        nGuesses += other.nGuesses;
        nInGold += other.nInGold;
        propFound += other.propFound;
        propGuessed += other.propGuessed;
        nOverlap += other.nOverlap;
        if(samples != null) {
        	if(other.samples != null && other.samples.size() > 0)
        		samples.addAll(other.samples);
        	else {
        		other.samples = null;
        		samples.add(other);
        	}
        }
	}

	public double getPrecision() {
		if(nGuesses == 0) // should return 1.0?
			return nInGold == 0? 1: 0;
		else
			return (double) nCorrect / nGuesses;
	}

	public double getRecall() {
		if(nInGold == 0)
			return nGuesses == 0? 1: 0;
		else
			return (double) nCorrect / nInGold;
	}
	
	public double getF1() {
		double p = getPrecision();
		double r = getRecall();
		if(p == 0 || r == 0)
			return 0;
		else
			return 2*p*r/(p + r);
	}
	
	public double getOverlapPrecision() {
		if(nGuesses == 0)
			return nInGold == 0? 1: 0;
		else
			return (double) nOverlap / nGuesses;
	}

	public double getOverlapRecall() {
		if(nInGold == 0)
			return nGuesses == 0? 1: 0;
		else
			return (double) nOverlap / nInGold;
	}
	
	public double getOverlapF1() {
		double p = getOverlapPrecision();
		double r = getOverlapRecall();
		if(p == 0 || r == 0)
			return 0;
		else
			return 2*p*r/(p + r);
	}
	
	public double getPartialPrecision() {
		if(nGuesses == 0)
			return nInGold == 0? 1: 0;
		else
			return propGuessed / nGuesses;
	}

	public double getPartialRecall() {
		if(nInGold == 0)
			return nGuesses == 0? 1: 0;
		else
			return propFound / nInGold;
	}
	
	public double getPartialF1() {
		double p = getPartialPrecision();
		double r = getPartialRecall();
		if(p == 0 || r == 0)
			return 0;
		else
			return 2*p*r/(p + r);
	}
	
	// 0..1 -> precision-weighted
	// 1    -> equal weight
	// > 1  -> recall-weighted
	public double getPartialF(double alpha) {
		double p = getPartialPrecision();
		double r = getPartialRecall();
		if(p == 0 || r == 0)
			return 0;
		else
			return (1+alpha)*p*r/(alpha*p + r);
	}
	
	public String toString() {
		return "{ncorr = " + nCorrect + ", nguess = " + nGuesses + ", ningold = " + nInGold
		+ ", pfound = " + propFound + ", pguess = " + propGuessed + ", nover = " + nOverlap + "}";
	}
	
	public void printRaw() {
		// pg:3, pf: 4
		// TODO THIS MAY HAVE CAUSED ERRORS!!!!
		System.out.format("%f %f %f %f %f %f\n", nCorrect, nGuesses, nInGold, propFound, propGuessed, nOverlap);
	}
	
	public String toRawString() {
		// pg:3, pf: 4
		return nCorrect + " " + nGuesses + " " + nInGold
		+ " " + propGuessed + " " + propFound + " " + nOverlap;
	}
	
	public double getPropError() {
		System.out.println("getPropError: " + this);
		
		return (nGuesses - propGuessed) + (nInGold - propFound);
	}

	public void print() {
		if(nCorrect > 0 || propFound == 0 && propGuessed == 0 && nOverlap == 0)
			System.out.format("Exact:   p = %f, r = %f, f1 = %f\n", 
					getPrecision(), getRecall(), getF1());
		if(propFound > 0 || propGuessed > 0)
			System.out.format("Partial: p = %f, r = %f, f1 = %f\n", 
					getPartialPrecision(), getPartialRecall(), getPartialF1());
		if(nOverlap > 0)
			System.out.format("Overlap: p = %f, r = %f, f1 = %f\n", 
					getOverlapPrecision(), getOverlapRecall(), getOverlapF1());
	}
	
	public void printSamples(PrintWriter pw) throws IOException {
		// pg:3, pf: 4
		if(samples == null)
			throw new IllegalStateException("not in sampling mode");
		for(PRFStats s: samples)
			pw.format("%f %f %f %f %f %f\n", s.nCorrect, s.nGuesses, s.nInGold, 
					s.propGuessed, s.propFound, s.nOverlap);
	}

	public void printSamplesToFile(String filename) throws IOException {
		PrintWriter out = new PrintWriter(new FileWriter(filename));
		printSamples(out);
		out.close();
	}
	
	public static PRFStats readSamples(BufferedReader br) throws IOException {
		PRFStats out = new PRFStats(true);
		String line = br.readLine();
		while(line != null) {
			String[] ss = line.split(" ");
			out.add(new PRFStats(ss));
			line = br.readLine();
		}			
		return out;
	}

	/* Functors for resampling. */
	
	public DoubleFunction<int[]> getPrecisionEvaluator() {
		return new DoubleFunction<int[]>() {
			public double apply(int[] selected) {
				double nCorrect = 0, nGuesses = 0, nInGold = 0;
				for(int index: selected) {
					PRFStats s = samples.get(index);
					nCorrect += s.nCorrect;
					nGuesses += s.nGuesses;
					nInGold += s.nInGold;
				}
				if(nGuesses == 0)
					return nInGold == 0? 1: 0;
				else
					return nCorrect / nGuesses;
			}
		};
	}
	
	public DoubleFunction<int[]> getRecallEvaluator() {
		return new DoubleFunction<int[]>() {
			public double apply(int[] selected) {
				double nCorrect = 0, nGuesses = 0, nInGold = 0;
				for(int index: selected) {
					PRFStats s = samples.get(index);
					nCorrect += s.nCorrect;
					nGuesses += s.nGuesses;
					nInGold += s.nInGold;
				}
				if(nInGold == 0)
					return nGuesses == 0? 1: 0;
				else
					return nCorrect / nInGold;
			}
		};
	}
	
	public DoubleFunction<int[]> getF1Evaluator() {
		return new DoubleFunction<int[]>() {
			public double apply(int[] selected) {
				double nCorrect = 0, nGuesses = 0, nInGold = 0;
				for(int index: selected) {
					PRFStats s = samples.get(index);
					nCorrect += s.nCorrect;
					nGuesses += s.nGuesses;
					nInGold += s.nInGold;
				}
				double p;
				if(nGuesses == 0)
					p = nInGold == 0? 1: 0;
				else
					p = nCorrect / nGuesses;
				double r;
				if(nInGold == 0)
					r = nGuesses == 0? 1: 0;
				else
					r = nCorrect / nInGold;
				if(p == 0 || r == 0)
					return 0;
				else
					return 2*p*r/(p + r);			
			}
		};
	}

	public DoubleFunction<int[]> getPartialPrecisionEvaluator() {
		return new DoubleFunction<int[]>() {
			public double apply(int[] selected) {
				double propGuessed = 0, nGuesses = 0, nInGold = 0;
				for(int index: selected) {
					PRFStats s = samples.get(index);
					propGuessed += s.propGuessed;
					nGuesses += s.nGuesses;
					nInGold += s.nInGold;
				}
				if(nGuesses == 0)
					return nInGold == 0? 1: 0;
				else
					return propGuessed / nGuesses;
			}
		};
	}
	
	public DoubleFunction<int[]> getPartialRecallEvaluator() {
		return new DoubleFunction<int[]>() {
			public double apply(int[] selected) {
				double propFound = 0, nGuesses = 0, nInGold = 0;
				for(int index: selected) {
					PRFStats s = samples.get(index);
					propFound += s.propFound;
					nGuesses += s.nGuesses;
					nInGold += s.nInGold;
				}
				if(nInGold == 0)
					return nGuesses == 0? 1: 0;
				else
					return propFound / nInGold;
			}
		};
	}
	
	public DoubleFunction<int[]> getPartialF1Evaluator() {
		return new DoubleFunction<int[]>() {
			public double apply(int[] selected) {
				double propGuessed = 0, propFound = 0, nGuesses = 0, nInGold = 0;
				for(int index: selected) {
					PRFStats s = samples.get(index);
					propGuessed += s.propGuessed;
					propFound += s.propFound;
					nGuesses += s.nGuesses;
					nInGold += s.nInGold;
				}
				double p;
				if(nGuesses == 0)
					p = nInGold == 0? 1: 0;
				else
					p = propGuessed / nGuesses;		
				double r;
				if(nInGold == 0)
					r = nGuesses == 0? 1: 0;
				else
					r = propFound / nInGold;				
				if(p == 0 || r == 0)
					return 0;
				else
					return 2*p*r/(p + r);			
			}
		};
	}
	
	public DoubleFunction<int[]> getOverlapPrecisionEvaluator() {
		return new DoubleFunction<int[]>() {
			public double apply(int[] selected) {
				double nOverlap = 0, nGuesses = 0, nInGold = 0;
				for(int index: selected) {
					PRFStats s = samples.get(index);
					nOverlap += s.nOverlap;
					nGuesses += s.nGuesses;
					nInGold += s.nInGold;
				}
				if(nGuesses == 0)
					return nInGold == 0? 1: 0;
				else
					return nOverlap / nGuesses;
			}
		};
	}
	
	public DoubleFunction<int[]> getOverlapRecallEvaluator() {
		return new DoubleFunction<int[]>() {
			public double apply(int[] selected) {
				double nOverlap = 0, nGuesses = 0, nInGold = 0;
				for(int index: selected) {
					PRFStats s = samples.get(index);
					nOverlap += s.nOverlap;
					nGuesses += s.nGuesses;
					nInGold += s.nInGold;
				}
				if(nInGold == 0)
					return nGuesses == 0? 1: 0;
				else
					return nOverlap / nInGold;
			}
		};
	}
	
	public DoubleFunction<int[]> getOverlapF1Evaluator() {
		return new DoubleFunction<int[]>() {
			public double apply(int[] selected) {
				double nOverlap = 0, nGuesses = 0, nInGold = 0;
				for(int index: selected) {
					PRFStats s = samples.get(index);
					nOverlap += s.nOverlap;
					nGuesses += s.nGuesses;
					nInGold += s.nInGold;
				}
				double p;
				if(nGuesses == 0)
					p = nInGold == 0? 1: 0;
				else
					p = nOverlap / nGuesses;
				double r;
				if(nInGold == 0)
					r = nGuesses == 0? 1: 0;
				else
					r = nOverlap / nInGold;
				if(p == 0 || r == 0)
					return 0;
				else
					return 2*p*r/(p + r);			
			}
		};
	}

	/* Difference evaluator for resampling tests. */	
	/* Should not be here? */
	
	public static DoubleFunction<int[]> createDifferenceEvaluator(final DoubleFunction<int[]> e1,
			final DoubleFunction<int[]> e2) {
		return new DoubleFunction<int[]>() {
			public double apply(int[] selected) {
				return e1.apply(selected) - e2.apply(selected);
			}
		};
	}
	
	/* Converts a pair of evaluators for resamplers into a mixer for permutation tests for differences. */

	public static DoubleFunction<int[]> createPermTestMixer(final PRFStats s1,
			final PRFStats s2, final DoubleFunction<int[]> eval1, 
			final DoubleFunction<int[]> eval2) {
		if(s1.samples == null)
			throw new IllegalArgumentException("Must evaluate in detailed mode");
		if(s1.nInGold != s2.nInGold)
			throw new IllegalArgumentException("Must evaluate on same test set!");
		final int[] iter = CollectionUtils.range(0, s1.samples.size());
		
		return new DoubleFunction<int[]>() {
			public double apply(int[] flips) {
				for(int i = 0; i < flips.length && flips[i] >= 0; i++) {
					int index = flips[i];
					PRFStats sample1 = s1.samples.get(index);
					PRFStats sample2 = s2.samples.get(index);
					s1.samples.set(index, sample2);
					s2.samples.set(index, sample1);
				}
				double out = eval1.apply(iter) - eval2.apply(iter);
				for(int i = 0; i < flips.length && flips[i] >= 0; i++) {
					int index = flips[i];
					PRFStats sample1 = s1.samples.get(index);
					PRFStats sample2 = s2.samples.get(index);
					s1.samples.set(index, sample2);
					s2.samples.set(index, sample1);
				}
				return out;
			}
		};
	}
	
	private DoubleFunction<int[]> getEvaluatorByString(String id) {
		if(id.equals("p"))
			return getPrecisionEvaluator();
		if(id.equals("r"))
			return getRecallEvaluator();
		if(id.equals("f1"))
			return getF1Evaluator();
		if(id.equals("ep"))
			return getPrecisionEvaluator();
		if(id.equals("er"))
			return getRecallEvaluator();
		if(id.equals("ef1"))
			return getF1Evaluator();
		if(id.equals("pp"))
			return getPartialPrecisionEvaluator();
		if(id.equals("pr"))
			return getPartialRecallEvaluator();
		if(id.equals("pf1"))
			return getPartialF1Evaluator();
		if(id.equals("op"))
			return getOverlapPrecisionEvaluator();
		if(id.equals("or"))
			return getOverlapRecallEvaluator();
		if(id.equals("of1"))
			return getOverlapF1Evaluator();
		System.err.println("Illegal evaluator name.");
		System.exit(1);
		return null;
	}
	
	private static int findCrossoverIndex(double[] d) {
		/* Requires the array to be sorted. */
		for(int i = 0; i < d.length - 1; i++)
			if(d[i] <= 0 && d[i+1] > 0)
				return i;
		return -1;
	}
	
	public static void testSignificance1(PRFStats stats, String evalType) {		
		DoubleFunction<int[]> eval = stats.getEvaluatorByString(evalType);
		Random rand = new Random(0);
		System.out.println("Evaluation value: " + eval.apply(CollectionUtils.range(0, stats.samples.size())));
		Statistics result = Resampler.resamplingTest(eval, stats.samples.size(), 
				stats.samples.size(), rand, 0.0001, 100000, 200000);
		System.out.println(result);
		double i95len = result.q975 - result.q025;
		System.out.format("95%% interval: mean ± %f\n", i95len/2);
	}
	
	public static void testSignificance2(PRFStats stats0, PRFStats stats1,
			String evalType, boolean usePermTest) {
		DoubleFunction<int[]> eval0 = stats0.getEvaluatorByString(evalType);
		DoubleFunction<int[]> eval1 = stats1.getEvaluatorByString(evalType);
		if(stats0.samples.size() != stats1.samples.size())
			throw new IllegalArgumentException("must be same size");			
		DoubleFunction<int[]> eval;
		if(usePermTest)
			eval = createPermTestMixer(stats0, stats1, eval0, eval1);
		else 					
			eval = createDifferenceEvaluator(eval0, eval1);								
		Random rand = new Random(0);				
		Statistics result;

		System.out.println("Evaluation value: " + eval.apply(CollectionUtils.range(0, stats0.samples.size())));

		// TODO improve convergence tests!
		if(usePermTest)
			result = PermutationTest.permutationTest(eval, stats0.samples.size(),
					rand, 0.0001, 100000, 200000);
		else
			result = Resampler.resamplingTest(eval, stats0.samples.size(), 
					stats0.samples.size(), rand, 0.0001, 100000, 200000);				
		System.out.println(result);
		int crossover = findCrossoverIndex(result.data);
		if(crossover == -1) {
			System.out.println("No crossover observed, difference is significant.");
		} else {
			if(crossover > result.data.length / 2) {
				crossover = result.data.length - crossover - 1;
				double pco = (double) crossover / result.data.length;
				System.out.println("Crossover quantile: 1 - " + pco);
			} else {
				double pco = (double) crossover / result.data.length;
				System.out.println("Crossover quantile: " + pco);
			}

		}				
	}
	
	public static void main(String[] argv) {		

		if(argv.length < 3) {
			System.err.println("Must specify evaluator, one or two input files and an output");
			System.exit(1);
		}
		try {
			if(argv.length == 3) {
				BufferedReader br = Util.openFileReader(argv[1]);
				PRFStats stats = readSamples(br);
				DoubleFunction<int[]> eval = stats.getEvaluatorByString(argv[0]);
				Random rand = new Random(0);
				System.out.println("Evaluation value: " + eval.apply(CollectionUtils.range(0, stats.samples.size())));
				Statistics result = Resampler.resamplingTest(eval, stats.samples.size(), 
						stats.samples.size(), rand, 0.0001, 100000, 200000);
				System.out.println(result);
				double i95len = result.q975 - result.q025;
				System.out.format("95%% interval: mean ± %f\n", i95len/2);
				if(argv[2].matches("NULL|EMPTY|NONE|/dev/null"))
					System.exit(0);
				if(!argv[2].endsWith(".m")) {
					System.err.println("Warning: skipping non-matlab file");
					System.exit(0);
				}
				PrintWriter pw = new PrintWriter(new FileWriter(argv[2]));
				result.printMatlabVector(pw, argv[2].substring(0, argv[2].length()-2));
				pw.close();				
			} else if(argv.length == 4) {
				String evalName = argv[0];
				boolean usePermTest = evalName.startsWith("perm_");
				if(usePermTest)
					evalName = evalName.substring(5);
				BufferedReader br0 = Util.openFileReader(argv[1]);
				PRFStats stats0 = readSamples(br0);
				DoubleFunction<int[]> eval0 = stats0.getEvaluatorByString(evalName);
				BufferedReader br1 = Util.openFileReader(argv[2]);
				PRFStats stats1 = readSamples(br1);
				DoubleFunction<int[]> eval1 = stats1.getEvaluatorByString(evalName);
				if(stats0.samples.size() != stats1.samples.size())
					throw new IllegalArgumentException("must be same size");			
				DoubleFunction<int[]> eval;
				if(usePermTest)
					eval = createPermTestMixer(stats0, stats1, eval0, eval1);
				else 					
					eval = createDifferenceEvaluator(eval0, eval1);								
				Random rand = new Random(0);				
				Statistics result;

				System.out.println("Evaluation value: " + eval.apply(CollectionUtils.range(0, stats0.samples.size())));
				
				// TODO improve convergence tests!
				if(usePermTest)
					result = PermutationTest.permutationTest(eval, stats0.samples.size(),
							rand, 0.0001, 100000, 200000);
				else
					result = Resampler.resamplingTest(eval, stats0.samples.size(), 
							stats0.samples.size(), rand, 0.0001, 100000, 200000);				
				System.out.println(result);
				int crossover = findCrossoverIndex(result.data);
				if(crossover == -1) {
					System.out.println("No crossover observed.");
				} else {
					if(crossover > result.data.length / 2) {
						crossover = result.data.length - crossover - 1;
						double pco = (double) crossover / result.data.length;
						System.out.println("Crossover quantile: 1 - " + pco);
					} else {
						double pco = (double) crossover / result.data.length;
						System.out.println("Crossover quantile: " + pco);
					}

				}				
				if(argv[3].matches("NULL|EMPTY|NONE|/dev/null"))
					System.exit(0);
				if(!argv[3].endsWith(".m")) {
					System.err.println("Warning: skipping non-matlab file");
					System.exit(0);
				}
				PrintWriter pw = new PrintWriter(new FileWriter(argv[3]));
				result.printMatlabVector(pw, argv[3].substring(0, argv[3].length()-2));
				pw.close();
			} else {
				System.out.println("3 or 4 arguments");
				System.exit(1);
			}
		} catch(Exception e) {
			e.printStackTrace();
			System.exit(1);
		}
	}
	
}