/*
 Copyright (2006) Richard Johansson (richard.johansson@gu.se)
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at 
 
 http://www.apache.org/licenses/LICENSE-2.0
 
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License. 
*/

/*
 Util.java
 
 Created on Jan 11, 2006 by Richard Johansson (richard@cs.lth.se).
 
 $Log: Ax.java,v $
 Revision 1.5  2011-07-11 11:19:00  johansson
 More methods.

 Revision 1.4  2011-02-10 11:42:31  johansson
 Looze bzip2 library coupling and equals with tolerance.

 Revision 1.3  2010-07-05 13:28:59  johansson
 BZip2 in openStream.

 Revision 1.2  2010-06-07 11:02:38  johansson
 Major update.

 Revision 1.1  2009/11/12 08:26:54  johansson
 Renamed for compatibility with Windows systems.

 Revision 1.1  2009/01/16 10:05:11  johansson
 Added to brenta repository.

 Revision 1.11  2008/05/25 14:06:44  richard
 Updated.

 Revision 1.10  2007/11/08 14:23:26  richard
 Added msb.

 Revision 1.9  2007/03/09 13:40:57  richard
 Changed retval of fromSystem.

 Revision 1.8  2007/03/02 09:00:54  richard
 Added fromSystem.

 Revision 1.7  2007/03/01 12:48:23  richard
 Flushes writer.

 Revision 1.6  2007/03/01 12:20:13  richard
 Added readFile.

 Revision 1.5  2007/02/08 15:47:23  richard
 Added system, writeLines, and getTempFileName.

 Revision 1.4  2007/01/08 12:56:05  richard
 Added readLines.

 Revision 1.3  2006/09/17 17:52:46  richard
 UniqueSymbol and openTempFile.

 Revision 1.2  2006/09/07 17:51:19  richard
 New revision.

 Revision 1.1  2006/05/16 13:38:13  richard
 Re-added because of CVS problems.
 
 Revision 1.2  2006/04/07 14:45:49  richard
 Added documentation.
 
 Revision 1.1  2006/02/23 09:46:33  richard
 Added the file.
 
 
 */
package se.lth.cs.nlp.nlputils.core;

import java.util.*;
import java.util.zip.GZIPInputStream;
import java.io.*;
import java.lang.reflect.*;

/**
 * Contains a set of auxiliary methods used by all libraries.
 * 
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class Util {
    
    private Util() {}
    
    /**
     * Asserts that <code>condition</code> holds,
     * otherwise throws a <code>RuntimeException</code>.
     * @param condition the condition to check.
     * @param message a message for the exception.
     */
    public static void check(boolean condition, String message) {
        if(!condition)
            throw new RuntimeException("Check failed: " + message);
    }
    
    public static String uniqueSymbol() {
    	return "" + new Random().nextInt() + "_" + System.nanoTime();
    }

    public static String getTempFileName(String prefix) {
        return System.getProperty("java.io.tmpdir") 
        + System.getProperty("file.separator") 
        + prefix + uniqueSymbol();
    }
    
    public static String readFile(String filename) throws IOException {
        Scanner sc = new Scanner(new File(filename));
        StringBuilder sb = new StringBuilder();
        while(sc.hasNextLine()) {
            String l = sc.nextLine();
            sb.append(l + "\n");
        }
        return sb.toString();
    }
    
    public static PrintWriter openTempFile(String prefix) throws IOException {
    	return new PrintWriter(new FileWriter(getTempFileName(prefix)));
    }

    public static void readLines(BufferedReader r, Collection<String> coll) throws IOException {
    	String line = r.readLine();
    	while(line != null) {
            coll.add(line);
            line = r.readLine();
    	}
    }
    
    public static void readLines(String filename, Collection<String> coll) throws IOException {
    	BufferedReader br = new BufferedReader(new FileReader(filename));
    	readLines(br, coll);
    }
    
    public static void writeLines(PrintWriter w, Collection<String> coll) {
        for(String s: coll)
            w.println(s);
        w.flush();
    }
    
    public static Iterable<String> readLines(final String filename) throws IOException {
        return readLines(Util.openFileReader(filename));
    }
    
    public static Iterable<String> readLines(final String filename, final String encoding) throws IOException {
        return readLines(Util.openFileReader(filename, encoding));
    }
    
    public static Iterable<String> readLines(final BufferedReader br) throws IOException {
        return new Iterable<String>() {
            public Iterator<String> iterator() {
                try {
                    return lineIterator(br);
                } catch(IOException e) {
                    throw new RuntimeException(e);
                }
            }
        };
    }
    
    public static Iterator<String> lineIterator(final BufferedReader br) throws IOException {
        return new Iterator<String>() {
            private String current;
            public boolean hasNext() {
                if(current != null)
                    return true;
                try {
                    current = br.readLine();
                } catch(Exception e) {
                    throw new RuntimeException(e);
                }
                if(current == null) {
                    try {
                        br.close();
                    } catch(Exception e) { /* ignored */ }
                    return false;
                }
                return true;
                //return current != null;
            }

            public String next() {
                if(!hasNext())
                    throw new IllegalStateException("end of stream");
                String out = current;
                current = null;
                return out;
            }

            public void remove() {
                throw new UnsupportedOperationException("remove unsupported");
            }
            
        };
    }
    
    public static void writeLines(String filename, Collection<String> coll) throws IOException {
        PrintWriter pw = new PrintWriter(filename);
        writeLines(pw, coll);
        pw.close();
    }

    public static BufferedReader makeReader(InputStream is, String encoding) throws IOException {
        InputStreamReader ir;
        if(encoding == null)
            ir = new InputStreamReader(is);
        else
            ir = new InputStreamReader(is, encoding);
        return new BufferedReader(ir);
    }
    
    public static PrintWriter makeWriter(OutputStream os, String encoding) throws IOException {
        OutputStreamWriter osw;
        if(encoding == null)
            osw = new OutputStreamWriter(os);
        else
            osw = new OutputStreamWriter(os, encoding);
        return new PrintWriter(osw);
    }
    
    public static BufferedReader openStdinReader(String encoding) throws IOException {
        return makeReader(System.in, encoding);
    }
    
    public static PrintWriter openStdoutWriter(String encoding) throws IOException {
        return makeWriter(System.out, encoding);
    }
    
    public static BufferedReader openFileReader(String fileName) throws IOException {
    	return openFileReader(fileName, null);
    }
    
    public static BufferedReader openFileReader(String fileName, String encoding) throws IOException {
    	InputStream is = openFileStream(fileName);
    	return makeReader(is, encoding);
    }
    
    public static PrintWriter openFileWriter(String fileName, String encoding) throws IOException {
        OutputStream os = openFileOutStream(fileName);
        return makeWriter(os, encoding);
    }
    
    public static InputStream openFileStream(String fileName) throws IOException {
    	InputStream is = new FileInputStream(fileName);
    	if(fileName.endsWith(".gz"))
    		is = new GZIPInputStream(is);
    	else if(fileName.endsWith(".bz2")) {
    		//is = new org.apache.tools.bzip2.CBZip2InputStream(is);
    		is = openBZip2InputStream(is);
    	}
    	return is;
    }

    public static OutputStream openFileOutStream(String fileName) throws IOException {
        OutputStream os = new FileOutputStream(fileName);
        /* TODO */
        /*if(fileName.endsWith(".gz"))
            is = new GZIPInputStream(is);
        else if(fileName.endsWith(".bz2")) {
            //is = new org.apache.tools.bzip2.CBZip2InputStream(is);
            is = openBZip2InputStream(is);
        }*/
        return os;
    }
    
    public static ObjectInputStream openFileObjectStream(String filename) throws IOException {
        InputStream is = openFileStream(filename);
        return new ObjectInputStream(is);
    }
    
    public static void saveObject(String filename, Object o) throws IOException {
        OutputStream os = openFileOutStream(filename);
        ObjectOutputStream oos = new ObjectOutputStream(os);
        oos.writeObject(o);
        oos.close();        
    }
    
    public static Object loadObject(String filename) throws IOException, ClassNotFoundException {
        ObjectInputStream ois = openFileObjectStream(filename);
        Object out = ois.readObject();
        ois.close();
        return out;
    }
    
    private static InputStream openBZip2InputStream(InputStream in) throws IOException {
		int b1 = in.read();
		int b2 = in.read();
		if(b1 != 'B' || b2 != 'Z')
			throw new IllegalArgumentException("Should start with BZ");
    	try {
    		Class cl = Class.forName("org.apache.tools.bzip2.CBZip2InputStream");
    		Constructor con = cl.getConstructor(InputStream.class);
    		return (InputStream) con.newInstance(in);
    	} catch(Exception e) {
    		e.printStackTrace();
    		throw new RuntimeException(e);
    	}
    }
    
    private static class StreamThread extends Thread {
        private InputStream in;
        private OutputStream out;
        StreamThread(InputStream in, OutputStream out) {
            super();
            this.in = in;
            this.out = out;
        }
        public void run() {
            try {
                while(true) {
                    int c = in.read();
                    if(c == -1)
                        break;
                    if(out != null)
                        out.write(c);
                }
            } catch(IOException e) {
                e.printStackTrace();
            }            
        }
    }

    public static int system(String command, OutputStream out, OutputStream err) throws IOException {
        Process process = Runtime.getRuntime().exec(command);
        Thread t1 = new StreamThread(process.getInputStream(), out);
        Thread t2 = new StreamThread(process.getErrorStream(), err);
        t1.start();
        t2.start();
        int status = -1;
        try {
            status = process.waitFor();
            t1.join();
            t2.join();
        } catch(InterruptedException e) {
            throw new IOException(e.getMessage());
        }
        return status;
    }
    
    public static int system(String command) throws IOException {
        return system(command, System.out, System.err);
    }
    
    public static Pair<String, Integer> fromSystem(String command) throws IOException {
        ByteArrayOutputStream bs = new ByteArrayOutputStream();
        int status = system(command, bs, System.err);
        return new Pair<String, Integer>(new String(bs.toByteArray()), status);
    }    
    
    public static int msb(int n) {
        int nbits = 0;
        while(n != 0) {
            nbits++;
            n >>>= 1;
        }
        return nbits;
    }
    
    public static final int DAYS = 1;
    public static final int HOURS = 2;
    public static final int MINUTES = 4;
    public static final int SECONDS = 8;
    public static final int MILLISECONDS = 16;
    
    public static final int ALL_TIME_FIELDS = 31;
    
    public static String toTimeString(long time, int format) {
        long msecs = time % 1000L;
        long totalSecs = time / 1000L;
        long secs = totalSecs % 60L;
        long totalMins = totalSecs / 60L;
        long mins = totalMins % 60L;
        long totalHours = totalMins / 60L;
        long hours = totalHours % 24L;
        long days = totalHours / 24L;

        if((format & DAYS) == 0) {
            hours += days * 24L;
            days = 0;
        }
        if((format & HOURS) == 0) {
            mins += hours * 60L;
            hours = 0;
        }
        if((format & MINUTES) == 0) {
            secs += mins * 60L;
            mins = 0;
        }
        if((format & SECONDS) == 0) {
            msecs += secs * 1000L;
            secs = 0;
        }
        if((format & MILLISECONDS) == 0) {
            msecs = 0;
        }
        
        StringBuilder sb = new StringBuilder();
        if(days > 0) {
            sb.append(days + " day");
            if(days > 1)
                sb.append("s");
        }
        if(hours > 0) {
            if(sb.length() > 0)
                sb.append(", ");
            sb.append(hours + " hour");
            if(hours > 1)
                sb.append("s");
        }
        if(mins > 0) {
            if(sb.length() > 0)
                sb.append(", ");
            sb.append(mins + " minute");
            if(mins > 1)
                sb.append("s");
        }
        if(secs > 0) {
            if(sb.length() > 0)
                sb.append(", ");
            sb.append(secs + " second");
            if(secs > 1)
                sb.append("s");
        }
        if(msecs > 0) {
            if(sb.length() > 0)
                sb.append(", ");
            sb.append(msecs + " millisecond");
            if(msecs > 1)
                sb.append("s");
        }
        return sb.toString();
    }

    public static void main(String[] argv) {
        long t0 = System.currentTimeMillis();
        /*try {
            Thread.sleep(1500);
        } catch(Exception e) {};        
        long t1 = System.currentTimeMillis();*/
        
        long t1 = t0 + 1000*3 + 1000*60*71 + 1000*60*60*25 + 3456;
        
        int f = MINUTES | MILLISECONDS;
        
        System.out.println(toTimeString(t1 - t0, f));
    }
    
    public static String charToString(char c) {
    	switch(c) {
    	case 13: return "\\r";
    	case 10: return "\\n";
    	default: {
    			if(c < ' ')
    				return "(" + (int) c + ")";
    			else
    				return "" + c;
    		}
    	}
    }
    
    public static boolean isPunc(char c) {
    	int type = Character.getType(c);
    	switch(type) {
    	case Character.COMBINING_SPACING_MARK:
    	case Character.CONNECTOR_PUNCTUATION:
    	case Character.CONTROL:
    	case Character.DASH_PUNCTUATION:
    	case Character.ENCLOSING_MARK:
    	case Character.END_PUNCTUATION:
    	case Character.FINAL_QUOTE_PUNCTUATION:
    	case Character.INITIAL_QUOTE_PUNCTUATION:
    	case Character.OTHER_PUNCTUATION:
    	case Character.START_PUNCTUATION:
    		return true;
    	default:
    		return false;
    	}
    }
    
    public static boolean isPunc(String s) {
        int l = s.length();
        if(l == 0)
            return false;
        for(int i = 0; i < l; i++)
            if(!isPunc(s.charAt(i)))
                return false;
        return true;
    }
    
    public static boolean isSpace(char c) {
        if(c == ' ' || c == '\t' || c == '\n' || c == '\r')
            return true;
        int type = Character.getType(c);
        switch(type) {
        case Character.PARAGRAPH_SEPARATOR:
        case Character.SPACE_SEPARATOR:
            return true;
        default:
            return false;
        }
    }
    
    public static void printFileBytes(String file) {
    	try {
    		BufferedInputStream is = new BufferedInputStream(new FileInputStream(file));
    		byte[] buf = new byte[10000];

    		int ix = 0;
    		
    		int n = is.read(buf);
    		while(n != -1) {
    			for(int i = 0; i < n; i++)    				    				
    				System.out.format("%d\t%s\n", ix + i, charToString((char) buf[i]));
    			ix += n;
    			
    			n = is.read(buf);
    		}
    		
    	} catch(Exception e) {
    		throw new RuntimeException(e);
    	}
    }

    private static double TOLERANCE = 1e-12;
    
	public static boolean equal(double x, double y) {
		double diff = x - y;
		return diff < TOLERANCE && diff > -TOLERANCE;
	}
	
	public static boolean equal(double x, double y, double tol) {
		double diff = x - y;
		return diff < tol && diff > -tol;
	}
    
	public <T> T[] makeArray(T t, int n) {
		return (T[]) Array.newInstance(t.getClass(), n);
	}

	public static boolean isNullFile(String file) {
		return file == null || file.toLowerCase().matches("null|empty|none|/dev/null|/dev/zero");
	}
	
	// could be made in log n if we used cumulative distribution
	public static int drawRandom(Random rand, double[] dist) {
		double r = rand.nextDouble();
		double sum = 0.0;		
		for(int i = 0; i < dist.length; i++) {
			sum += dist[i];
			if(r <= sum)
				return i;
		}
		return -1;
	}

    public static void printToFile(String filename, String text, String encoding) throws IOException {
        PrintWriter pw = openFileWriter(filename, encoding);
        pw.print(text);
        pw.close();
    }
	
}
