package se.lth.cs.nlp.nlputils.math;

import java.io.*;
import java.util.*;

import se.lth.cs.nlp.nlputils.core.*;

public class AUCEstimator extends DoubleFunction<int[]> { 

	private double[] scores;
	private boolean[] values;
	
	public AUCEstimator(double[] scores, boolean[] values) {
		this.scores = scores;
		this.values = values;
	}
	
	public double apply(int[] selected) {
		int ncorr = 0, npos = 0, nneg = 0;
		for(int i = 0; i < selected.length; i++) {
			if(!values[selected[i]]) {
				nneg++;
				continue;
			} else
				npos++;
			for(int j = 0; j < selected.length; j++) {
				if(values[selected[j]])
					continue;
				if(scores[selected[i]] > scores[selected[j]])
					ncorr++;
			}
		}
		if(npos == 0 || nneg == 0)
			return 1.0;
		return (double) ncorr / (npos * nneg);
	}

	private static DoubleFunction<int[]> createDifferenceEvaluator(final DoubleFunction<int[]> e1,
			final DoubleFunction<int[]> e2) {
		return new DoubleFunction<int[]>() {
			public double apply(int[] selected) {
				return e1.apply(selected) - e2.apply(selected);
			}
		};
	}
	
	/* Converts a pair of evaluators for resamplers into a mixer for permutation tests for differences. */

	private static DoubleFunction<int[]> createPermTestMixer(final AUCEstimator eval1, 
			final AUCEstimator eval2) {
		if(!Arrays.equals(eval1.values, eval2.values))
			throw new IllegalArgumentException("Must evaluate on same set!");
		int n = eval1.values.length;
		final int[] iter = CollectionUtils.range(0, n);
		
		return new DoubleFunction<int[]>() {
			public double apply(int[] flips) {
				for(int i = 0; i < flips.length; i++) {
					int index = flips[i];
					double swap = eval1.scores[index];
					eval1.scores[index] = eval2.scores[index];
					eval2.scores[index] = swap;
				}
				double out = eval1.apply(iter) - eval2.apply(iter);
				for(int i = 0; i < flips.length; i++) {
					int index = flips[i];
					double swap = eval1.scores[index];
					eval1.scores[index] = eval2.scores[index];
					eval2.scores[index] = swap;
				}
				return out;
			}
		};
	}

	private static final int LABEL_COL = 1;
	private static final int SCORE_COL = 2;
	
	private static AUCEstimator readFile(String file) throws IOException {
		BufferedReader br = Util.openFileReader(file);
		ArrayList<Boolean> vals = new ArrayList();
		ArrayList<Double> scores = new ArrayList();
		String line = br.readLine();
		while(line != null) {
			String[] ss = line.split("\\s+");
			vals.add(ss[LABEL_COL].matches("1|\\+1"));
			scores.add(Double.parseDouble(ss[SCORE_COL]));
			line = br.readLine();
		}
		return new AUCEstimator(CollectionUtils.toDoubleArray(scores),
				CollectionUtils.toBooleanArray(vals));	
	}
	
	public static void main(String[] argv) {
		try {

			if(argv.length < 2) {
				System.err.println("Must specify res/perm and one or two input files");
				System.exit(1);
			}			
			
			boolean res = false;
			if(argv[0].equals("res"))
				res = true;
			else if(!argv[0].equals("perm")) {
				System.err.println("Must specify res/perm");
				System.exit(1);
			}
			if(!res && argv.length < 3) {
				System.err.println("Must specify two input files");
				System.exit(1);
			}

			AUCEstimator est1 = readFile(argv[1]);
			AUCEstimator est2 = null;
			if(argv.length >= 3)
				est2 = readFile(argv[2]);			
			int n = est1.values.length;
			Random rand = new Random(0);
						
			if(argv.length == 2) {
				System.out.println("Estimated AUC: " + est1.apply(CollectionUtils.range(0, n)));
				Statistics result = Resampler.resamplingTest(est1, n, n, rand, 1e-12, 2000, 2000);
				System.out.println(result);
				double i95len = result.q975 - result.q025;
				System.out.format("95%% interval: mean ± %f\n", i95len/2);
			} else {
				System.out.println("Estimated AUC 1: " + est1.apply(CollectionUtils.range(0, n)));
				System.out.println("Estimated AUC 2: " + est2.apply(CollectionUtils.range(0, n)));
				Statistics result;
				if(res) {
					DoubleFunction<int[]> eval = createDifferenceEvaluator(est1, est2);
					result = Resampler.resamplingTest(eval, n, n, rand, 1e-12, 1000, 1000);
				} else {
					DoubleFunction<int[]> eval = createPermTestMixer(est1, est2);
					result = PermutationTest.permutationTest(eval, n, rand, 1e-12, 1000, 1000);
				}

				System.out.println(result);
				int crossover = CollectionUtils.findCrossoverIndex(result.data);
				if(crossover == -1) {
					System.out.println("No crossover observed.");
				} else {
					if(crossover > result.data.length / 2) {
						crossover = result.data.length - crossover - 1;
						double pco = (double) crossover / result.data.length;
						System.out.println("Crossover quantile: 1 - " + pco);
					} else {
						double pco = (double) crossover / result.data.length;
						System.out.println("Crossover quantile: " + pco);
					}
				}
			}
			
		} catch(Exception e) {
			e.printStackTrace();
		}
	}
	
}
