/*
 Copyright (2007) Richard Johansson (richard.johansson@gu.se)
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at 
 
 http://www.apache.org/licenses/LICENSE-2.0
 
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License. 
*/

/*
 BLAS.java
  
 Created on May 10, 2007 by Richard Johansson (richard@cs.lth.se).

 $Log: BLAS.java,v $
 Revision 1.1  2009/01/16 10:06:31  johansson
 Added to brenta repository.

 Revision 1.6  2007/06/01 11:00:37  richard
 Added apply_exp.

 Revision 1.5  2007/05/22 14:13:54  richard
 Renamed library.

 Revision 1.4  2007/05/22 08:54:03  richard
 Index bugs.

 Revision 1.3  2007/05/11 14:56:30  richard
 Added dgemv.

 Revision 1.2  2007/05/11 14:12:18  richard
 Added native methods.

 Revision 1.1  2007/05/11 13:27:11  richard
 Renamed package. Added dswap.

 Revision 1.1  2007/05/10 08:52:23  richard
 Added.

   
 */
package se.lth.cs.nlp.nlputils.math;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class BLAS {

    private static boolean useNative = false;
    private static boolean loadedLibrary = false;

    public static void setNative(boolean b) {
        if(b && !loadedLibrary) {
            try {
                System.loadLibrary("jniblas");
                loadedLibrary = true;
                useNative = true;
            } catch(Throwable t) {
                t.printStackTrace();
            }
        } else
            useNative = b;
    }
    
    /* Level 1 native routines. */
    
    public static native void jni_daxpy(int n, double alpha,
            double[] x, int incx, double[] y, int incy,
            int xoff, int yoff);
    
    public static native void jni_dcopy(int n,
            double[] x, int incx, double[] y, int incy,
            int xoff, int yoff);
    
    public static native void jni_dscal(int n, double alpha,
            double[] x, int incx, int xoff);

    public static native double jni_ddot(int n, 
            double[] x, int incx, double[] y, int incy,
            int xoff, int yoff);
    
    public static native float jni_sdot(int n, 
            float[] x, int incx, float[] y, int incy,
            int xoff, int yoff);
    
    public static native void jni_dswap(int n,
            double[] x, int incx, double[] y, int incy,
            int xoff, int yoff);

    /* Level 1 Java routines. */
    
    /* The C Blas does not seem to permit negative step lengths. */
    
    public static void jblas_daxpy(int n, double alpha,
            double[] x, int incx, double[] y, int incy,
            int xoff, int yoff) {
        for(int i = 0; i < n; i++)
            y[i*incy+yoff] += alpha*x[i*incx+xoff];
    }
    
    public static void jblas_saxpy(int n, float alpha,
            float[] x, int incx, float[] y, int incy,
            int xoff, int yoff) {
        for(int i = 0; i < n; i++)
            y[i*incy+yoff] += alpha*x[i*incx+xoff];
    }
    
    public static void jblas_dcopy(int n,
            double[] x, int incx, double[] y, int incy,
            int xoff, int yoff) {
        for(int i = 0; i < n; i++)
            y[i*incy+yoff] = x[i*incx+xoff];
    }
    
    public static void jblas_dscal(int n, double alpha,
            double[] x, int incx, int xoff) {
        for(int i = 0; i < n; i++)
            x[i*incx+xoff] *= alpha;
    }

    public static void jblas_sscal(int n, float alpha,
            float[] x, int incx, int xoff) {
        for(int i = 0; i < n; i++)
            x[i*incx+xoff] *= alpha;
    }
    
    public static double jblas_ddot(int n, 
            double[] x, int incx, double[] y, int incy,
            int xoff, int yoff) {
        double sum = 0.0;
        for(int i = 0; i < n; i++)
            sum += y[i*incy+yoff]*x[i*incx+xoff];
        return sum;
    }
    
    public static float jblas_sdot(int n, 
            float[] x, int incx, float[] y, int incy,
            int xoff, int yoff) {
        float sum = 0.0f;
        for(int i = 0; i < n; i++)
            sum += y[i*incy+yoff]*x[i*incx+xoff];
        return sum;
    }
    
    public static void jblas_dswap(int n,
            double[] x, int incx, double[] y, int incy,
            int xoff, int yoff) {
        double tmp;
        for(int i = 0; i < n; i++) {
            tmp = x[i*incx + xoff];
            x[i*incx + xoff] = y[i*incy + yoff];
            y[i*incy + yoff] = tmp;
        }
    }
    
    /* Level 1 dispatching routines. */

    public static void daxpy(int n, double alpha,
            double[] x, int incx, double[] y, int incy,
            int xoff, int yoff) {
        if(useNative)
            jni_daxpy(n, alpha, x, incx, y, incy, xoff, yoff);
        else
            jblas_daxpy(n, alpha, x, incx, y, incy, xoff, yoff);
    }

    public static void saxpy(int n, float alpha,
			     float[] x, int incx, float[] y, int incy,
			     int xoff, int yoff) {
	    if(useNative)
		    throw new RuntimeException("unimplemented!");
	    else
		    jblas_saxpy(n, alpha, x, incx, y, incy, xoff, yoff);
    }
    
    public static void dcopy(int n,
            double[] x, int incx, double[] y, int incy,
            int xoff, int yoff) {
        if(useNative)
            jni_dcopy(n, x, incx, y, incy, xoff, yoff);
        else
            jblas_dcopy(n, x, incx, y, incy, xoff, yoff);
    }
    
    public static void sscal(int n, float alpha,
            float[] x, int incx, int xoff) {
        if(useNative)
		throw new RuntimeException("unimplemented!");
        else
		jblas_sscal(n, alpha, x, incx, xoff);
    }

    public static void dscal(int n, double alpha,
            double[] x, int incx, int xoff) {
        if(useNative)
            jni_dscal(n, alpha, x, incx, xoff);
        else
            jblas_dscal(n, alpha, x, incx, xoff);
    }
    
    public static double ddot(int n, 
            double[] x, int incx, double[] y, int incy,
            int xoff, int yoff) {
        if(useNative)
            return jni_ddot(n, x, incx, y, incy, xoff, yoff);
        else
            return jblas_ddot(n, x, incx, y, incy, xoff, yoff);
    }
    
    public static float sdot(int n, 
            float[] x, int incx, float[] y, int incy,
            int xoff, int yoff) {
        if(useNative)
            return jni_sdot(n, x, incx, y, incy, xoff, yoff);
        else
            return jblas_sdot(n, x, incx, y, incy, xoff, yoff);
    }
    
    public static void dswap(int n,
            double[] x, int incx, double[] y, int incy,
            int xoff, int yoff) {
        if(useNative)
            jni_dswap(n, x, incx, y, incy, xoff, yoff);
        else
            jblas_dswap(n, x, incx, y, incy, xoff, yoff);
    }
    
    /* Level 2 native routines. */
    
    public static native void jni_dgemv(boolean trans,
            int m, int n, double alpha, double[] a, int lda,
            double[] x, int incx, double beta, double[] y, int incy,
            int aroff, int acoff, int xoff, int yoff);
    
    /* Level 2 Java routines. */
    
    public static void jblas_dgemv(boolean trans,
            int m, int n, double alpha, double[] a, int lda,
            double[] x, int incx, double beta, double[] y, int incy,
            int aroff, int acoff, int xoff, int yoff) {
        if(!trans)
            for(int i = 0; i < m; i++)
                y[i*incy+yoff] = beta * y[i*incy+yoff]
                                          + jblas_ddot(n, a, 1, x, incx,
                                                       lda*(aroff+i) + acoff,
                                                       xoff) * alpha;

            
        else
            for(int i = 0; i < n; i++)
                y[i*incy+yoff] = beta * y[i*incy+yoff]
                                          + jblas_ddot(m, a, lda, x, incx,
                                                       lda*aroff + (acoff+i),
                                                       xoff) * alpha;
    }
    
    /* Level 2 dispatching routines. */
    
    public static void dgemv(boolean trans,
            int m, int n, double alpha, double[] a, int lda,
            double[] x, int incx, double beta, double[] y, int incy,
            int aroff, int acoff, int xoff, int yoff) {
        if(useNative)
            jni_dgemv(trans, m, n, alpha, a, lda, x, incx, beta, y, incy,
                    aroff, acoff, xoff, yoff);
        else
            jblas_dgemv(trans, m, n, alpha, a, lda, x, incx, beta, y, incy,
                        aroff, acoff, xoff, yoff);
    }
    
    
    /* Other auxiliary routines. */

    public static native void jni_apply_exp(double[] x, int n);
    
    public static void jblas_apply_exp(double[] x, int n) {
        for(int i = 0; i < n; i++)
            x[i] = Math.exp(x[i]);
    }
    
    public static void apply_exp(double[] x, int n) {
        if(useNative)
            jni_apply_exp(x, n);
        else
            jblas_apply_exp(x, n);
    }
    
}
