/*
 DMatrix.java
  
 Created on May 11, 2007 by Richard Johansson (richard@cs.lth.se).

 $Log: DMatrix.java,v $
 Revision 1.2  2009/04/16 12:05:20  johansson
 Added scaleRow and rowDaxpy.

 Revision 1.1  2009/01/16 10:06:31  johansson
 Added to brenta repository.

 Revision 1.6  2007/06/14 12:11:53  richard
 Changed increment. Removed factor.

 Revision 1.5  2007/05/30 09:04:05  richard
 ToString.

 Revision 1.4  2007/05/29 07:06:53  richard
 Zero lda, transpose, toString.

 Revision 1.3  2007/05/22 09:09:52  richard
 Lda bug.

 Revision 1.2  2007/05/16 14:42:52  richard
 Enlarges stepwise.

 Revision 1.1  2007/05/11 14:56:55  richard
 Added.

   
 */
package se.lth.cs.nlp.nlputils.math;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class DMatrix {

    private static final int INCREMENT = 100;
    
    public double[] data;
    public int m, n;
    public int lda;
     
    public boolean trans;
    
    public DMatrix(int m, int n) {
        this.data = new double[m*n];
        this.m = m;
        this.lda = n;
        this.n = n;
    }

    public double get(int i, int j) {
        int index = trans? j*lda + i: i*lda + j;
        return data[index];
    }

    public void set(int i, int j, double aij) {
        int index = trans? j*lda + i: i*lda + j;
        data[index] = aij;
    }
    
    private void addRowsInt(int nrows) {
        if(lda > 0) {
            int ntotal = data.length / lda;
            if(m + nrows > ntotal) {
                double[] copy = new double[lda * (m + nrows)];
                System.arraycopy(data, 0, copy, 0, m*lda);
                data = copy;
            }
        }
        m += nrows;
    }

    private void addColsInt(int ncols) {
        if(n + ncols > lda) {
            int nlda = n + ncols + INCREMENT;
            int nr = lda > 0? data.length / lda: m;
            double[] copy = new double[nr * nlda];
            for(int j = 0; j < n; j++)
                BLAS.dcopy(m, data, lda, copy, nlda, j, j);
            data = copy;
            lda = nlda;
        }
        n += ncols;
    }
    
    public void addRows(int nrows) {
        if(trans)
            addColsInt(nrows);
        else
            addRowsInt(nrows);
    }
    
    public void addCols(int ncols) {
        if(trans)
            addRowsInt(ncols);
        else
            addColsInt(ncols);
    }
    
    public void removeLastCols(int ncols) {
        if(trans)
            m -= ncols;
        else
            n -= ncols;
    }

    public void removeLastRows(int nrows) {
        if(trans)
            n -= nrows;
        else
            m -= nrows;
    }
    
    public void swapRows(int i1, int i2) {
        if(trans)
            BLAS.dswap(n, data, lda, data, lda, i1, i2);
        else
            BLAS.dswap(n, data, 1, data, 1, i1*lda, i2*lda);
    }
    
    public void swapCols(int i1, int i2) {
        if(trans)
            BLAS.dswap(n, data, 1, data, 1, i1*lda, i2*lda);
        else
            BLAS.dswap(n, data, lda, data, lda, i1, i2);
    }
    
    public void scale(double d) {
        for(int i = 0; i < m; i++)
            BLAS.dscal(n, d, data, 1, i*lda);
    }
    
    public void transpose() {
        trans = !trans;
    }
    
    public String toString() {
        StringBuilder sb = new StringBuilder();
        for(int i = 0; i < m; i++) {
            for(int j = 0; j < n; j++) {
                sb.append(get(i, j) + " ");
            }
            sb.append("\n");
        }
        return sb.toString();
    }

    public void scaleRow(int row, double scale) {
        if(trans)
            BLAS.dscal(m, scale, data, lda, row);
        else
            BLAS.dscal(n, scale, data, 1, row*lda);
    }
    
    /** 
     * 
     * 
     * @param row1
     * @param row2
     * @param scale
     */
    public void rowDaxpy(int row1, int row2, double scale) {
        if(row1 == row2)
            throw new IllegalArgumentException("row1 == row2");
        if(trans)
            BLAS.daxpy(m, scale, data, lda, data, lda, row1, row2);            
        else
            BLAS.daxpy(n, scale, data, 1, data, 1, row1*lda, row2*lda);
    }
    
}
