/*
 EncodedMulticlassDefinition.java

 Created on Mar 19, 2008 by Richard Johansson (richard@cs.lth.se).

 $Log: EncodedMulticlassDefinition.java,v $
 Revision 1.4  2010-06-08 09:49:10  johansson
 Added freeze. Fixed encoding bug -- forgot to encode on output.

 Revision 1.3  2010-06-07 11:29:59  johansson
 Added.

 Revision 1.2  2009/04/16 12:04:24  johansson
 More methods.

 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:04  richard
 Branched.

 Revision 1.1  2008/03/19 15:51:32  richard
 Added.


 */
package se.lth.cs.nlp.nlputils.ml_long;

import gnu.trove.TIntArrayList;
import gnu.trove.TIntHashSet;

import java.io.*;
import se.lth.cs.nlp.nlputils.core.*;

import java.util.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class EncodedMulticlassDefinition extends MulticlassDefinition<SparseVector, Integer> implements Serializable {    

    private static final long serialVersionUID = 1L;

    public EncodedMulticlassDefinition() {
        super();
    }

    private SparseVectorClassEncoder enc;
    private int[] range;
    
    public EncodedMulticlassDefinition(int[] range, boolean compress) {
        super();
        this.range = range;
        enc = new SparseVectorClassEncoder(compress);
    }

    public EncodedMulticlassDefinition(int[] range) {
	this(range, true);
    }
    
    public double cost(SparseVector x, Integer y, Integer ybar) {
    	return y.equals(ybar)? 0: 1;
    }

    protected boolean frozen = false;
    
    public void freeze() {
    	if(frozen)
        	System.err.println("Warning: MulticlassDefinition already frozen");
        else {
        	if(listEncoder != null)
        		listEncoder.freeze();
        	if(svEncoder != null)
        		svEncoder.freeze();
        	if(enc != null)
        		enc.freeze();
        }
    }
    
    public int encodeClass(Integer y) {
        return y;
    }

    public Integer decodeClass(int cl) {
        return cl;
    }

    public void encodeInstance(SparseVector x, SparseVector sv) {
        if(sv != x) {
            if(x.keys.length <= sv.keys.length) {
                sv.index = x.index;
                System.arraycopy(x.keys, 0, sv.keys, 0, sv.index);
                System.arraycopy(x.values, 0, sv.values, 0, sv.index);
            } else {
                sv.clear();
                sv.putAll(x);
            }
        }
    }

    public void encode(SparseVector x, Integer y, SparseVector sv) {
    	if(enc == null)
            throw new RuntimeException("Must specify classes!");

    	//System.out.println("Entering EMD.encode:\nx = " + x);
    	//System.out.println("y = " + y);
    	
    	sv.clear();        
        enc.encode(x, y, sv);
        
        //System.out.println("Leaving EMD.encode:\nsv = " + sv);
    }

    public Integer argmax(SparseVector x, Model model, SparseVector sv, double[] result) {
    	double max = Double.NEGATIVE_INFINITY;
        int maxc = Integer.MIN_VALUE;
        for(int i = 0; i < range.length; i++) {
            sv.clear();
            enc.encode(x, range[i], sv);
            double score = model.score(sv);
            if(score > max) {
                max = score;
                maxc = range[i];
            }
        }
        result[0] = max;
        
        enc.encode(x, maxc, sv); // bugfix 100608
        
        return maxc;
    }

    private static final class Item implements Comparable<Item> {
    	int id;
    	double score;
    	SparseVector rep;
    	public int compareTo(Item i) {
    		return Double.compare(i.score, score);
    	}
    	Item(int id, double score, SparseVector rep) {
    		this.id = id;
    		this.score = score;
    		this.rep = rep;
    	}    	
    }
    
    public int findNBest(SparseVector x, Model model, int n, Integer[] ybars, SparseVector[] svs, double[] scores) {
    	ArrayList<Item> items = new ArrayList();
    	for(int i = 0; i < range.length; i++) {
    		SparseVector sv = new SparseVector();
            enc.encode(x, range[i], sv);
            double score = model.score(sv);
            items.add(new Item(range[i], score, sv));
    	}
    	PriorityQueue<Item> q = new PriorityQueue(items);
    	int nout = Math.min(n, q.size());
    	for(int i = 0; i < nout; i++) {
    		Item item = q.poll();
    		ybars[i] = item.id;
    		svs[i] = item.rep;
    		scores[i] = item.score;
    	}
    	return nout;
    }    
    
    public Integer maxLoss(SparseVector x, Integer y, Model model, SparseVector sv, double[] result) {
        double max = Double.NEGATIVE_INFINITY;
        int maxc = Integer.MIN_VALUE;
        for(int i = 0; i < range.length; i++) {
            sv.clear();
            enc.encode(x, range[i], sv);
            double score = model.score(sv) + Math.sqrt(cost(x, y, range[i]));
            //double score = model.score(sv) + cost(x, y, range[i]);
            if(score > max) {
                max = score;
                maxc = range[i];
            }
        }
        sv.clear();
        enc.encode(x, maxc, sv);        
        result[0] = model.score(sv);       
        result[1] = cost(x, y, maxc);
        return maxc;
    }

    public Decomposition decompose(SparseVector x, Integer y) {
        throw new RuntimeException("Unimplemented");
    }
    
    private static ArrayList<Pair<SparseVector, Integer>> encodeSet(ArrayList<Pair<FeatureList, String>> set,
    		FeatureListEncoder lenc, SymbolEncoder clEnc) {    	
    	ArrayList<Pair<SparseVector, Integer>> tsEnc = new ArrayList();
        for(Pair<FeatureList, String> p: set) {
        	SparseVector sv = new SparseVector(); 	
        	lenc.encode(p.left, sv);
        	sv.sortIndices();
        	sv.trim();
        	int ce = clEnc.encode(p.right);
        	tsEnc.add(new Pair(sv, ce));
        }
        return tsEnc;
    }
    
    private static int[] getEncRange(SymbolEncoder enc) {
    	TIntArrayList l = new TIntArrayList();    	
    	for(String s: enc.inverse())
    		if(s != null)
    			l.add(enc.encode(s));
    	int[] out = l.toNativeArray();
    	Arrays.sort(out);
    	return out;    		
    }

    private static int[] getEncRange(ArrayList<Pair<SparseVector, Integer>> ts) {
    	TIntHashSet s = new TIntHashSet();
    	for(Pair<SparseVector, Integer> p: ts)
    		s.add(p.right);
    	int[] out = s.toArray();
    	Arrays.sort(out);
    	return out;
    }
    
    // haven't found out how to format to this effect
    private static void printDouble(double d) {
    	long l = (long) d;
    	if(d - l < 1e-6)
    		System.out.print(l);
    	else
    		System.out.format("%.6g", d);
    }
    
    public static void printSet(ArrayList<Pair<SparseVector, Integer>> ts) {    	
    	for(Pair<SparseVector, Integer> p: ts) {
    		SparseVector sv = p.left;
    		sv.sortIndices();
    		System.out.print(p.right);
    		for(int i = 0; i < sv.index; i++) {
    			System.out.format(" %d:", sv.keys[i]);
    			printDouble(sv.values[i]);
    		}
    		System.out.println();
    	}    		
    }
    
    public static ArrayList<Pair<SparseVector, Integer>> readEncodedSet(String file) throws IOException {
    	BufferedReader br = Util.openFileReader(file);
    	ArrayList<Pair<SparseVector, Integer>> out = new ArrayList();
    	String line = br.readLine();
    	while(line != null) {
    		line = line.replaceAll("#.*", "");
    		line = line.trim();
    		String[] ts = line.split("\\s+");
    		SparseVector sv = new SparseVector();
    		for(int i = 1; i < ts.length; i++) {
    			int ix = ts[i].indexOf(':');
    			if(ix == -1)
    				throw new RuntimeException("format error: " + line);
    			long k = Long.parseLong(ts[i].substring(0, ix));
    			double v = Double.parseDouble(ts[i].substring(ix+1));
    			sv.put(k, v);
    		}
    		sv.sortIndices();
    		sv.trim();
    		out.add(new Pair(sv, Integer.parseInt(ts[0])));    		
    		line = br.readLine();    		
    	}
    	return out;
    }
    
    // TODO varför ger PA olika resultat med denna jfrt med multiclassdef?
    
    private static final boolean ENCODED = false;
    
    public static void main(String[] argv) {
        try {
        	
        	ArrayList<Pair<SparseVector, Integer>> tsEnc;
    		
        	FeatureDeclaration decl = null;
        	FeatureListEncoder lenc = null;
        	SymbolEncoder clEnc = null;
        	if(!ENCODED) {
        		ArrayList<Pair<FeatureList, String>> l = MulticlassDefinition.readDataSet(argv[0], null);
        		decl = l.get(0).left.decl;

        		ArrayList<Pair<String, String>> pairings = new ArrayList();
        		lenc = new PairingFeatureListEncoder(decl, pairings);
        		clEnc = new SymbolEncoder();

        		tsEnc = encodeSet(l, lenc, clEnc);
        		l = null;
        		clEnc.freeze();
        		lenc.freeze();
        	} else
        		tsEnc = readEncodedSet(argv[0]);

            String methodString = argv[2];
            String argString = argv[3];
            
            LearningAlgorithm method = new AlgorithmFactory().create(methodString, argString);
            EncodedMulticlassDefinition def = new EncodedMulticlassDefinition(getEncRange(tsEnc));

            long t0 = System.currentTimeMillis();
            Classifier<SparseVector, Integer> cl = method.train(def, tsEnc);
            System.out.println("Training time: " + (System.currentTimeMillis() - t0));

            tsEnc = null;

            //sc = new Scanner(new FileInputStream(argv[1]));
            int n = 0, ncorrect = 0;            

            long cltime = 0;

            HashMap<Object, int[]> perClassResults = new HashMap();

            if(!ENCODED) {
            	ArrayList<Pair<FeatureList, String>> l2 = argv[1].equals("NULL")? null: readDataSet(argv[1], decl);

            	if(l2 != null) {
            		System.out.println("Loaded test set.");

            		ArrayList<String> clDec = clEnc.inverse();

            		for(Pair<FeatureList, String> p: l2) {
            			n++;
            			if(n % 1000 == 0) {
            				System.out.print(".");
            				System.out.flush();
            			}

            			t0 = System.currentTimeMillis();

            			SparseVector sv = new SparseVector();
            			lenc.encode(p.left, sv);
            			sv.sortIndices();

            			Integer aenc = cl.classify(sv);
            			cltime += System.currentTimeMillis() - t0;

            			String answer = clDec.get(aenc);

            			if(answer.equals(p.right))
            				ncorrect++;                

            			int[] c1 = null;
            			if(!answer.equals(MulticlassDefinition.NULL)) {
            				c1 = perClassResults.get(answer);
            				if(c1 == null) {
            					c1 = new int[3];
            					perClassResults.put(answer, c1);
            				}
            				c1[0]++;
            				if(answer.equals(p.right))
            					c1[2]++;
            			}
            			if(!p.right.equals(NULL)) {
            				int[] c2 = perClassResults.get(p.right);
            				if(c2 == null) {
            					c2 = new int[3];
            					perClassResults.put(p.right, c2);
            				}
            				c2[1]++;
            			}

            		}
            	}
            } else {
            	ArrayList<Pair<SparseVector, Integer>> testSet 
            		= readEncodedSet(argv[1]);
        		System.out.println("Loaded test set.");
            	for(Pair<SparseVector, Integer> p: testSet) {
            		n++;
        			if(n % 1000 == 0) {
        				System.out.print(".");
        				System.out.flush();
        			}
        			Integer aenc = cl.classify(p.left);
        			if(aenc.equals(p.right))
        				ncorrect++;
            	}
            }

            double acc = (double) ncorrect / n;

            System.out.println();           
            System.out.println(ncorrect + "/" + n + " = " + acc);
            System.out.println("Classification time: " + cltime);

            if(false)
            	for(Object c: perClassResults.keySet()) {
            		int[] res = perClassResults.get(c);
            		double pc = (double) res[2] / res[0];
            		double rc = (double) res[2] / res[1];
            		double fc = 2*pc*rc/(pc + rc);
            		pc = Math.round(10000.0*pc)/10000.0;
            		rc = Math.round(10000.0*rc)/10000.0;
            		fc = Math.round(10000.0*fc)/10000.0;
            		System.out.println(c + ": p = " + pc + ", r = " + rc + ", f1 = " + fc);
            	}         
            
            if(argv.length > 4) {
            	System.out.println("Writing model to " + argv[4]);
            	ObjectOutputStream oos = new ObjectOutputStream(new FileOutputStream(argv[4]));
            	oos.writeObject(cl);
            	oos.close();
            }
            
        } catch(Exception e) {
            e.printStackTrace();
        }
    }
    
}