/*
 FeatureDeclaration.java
 
 Created on Dec 22, 2005 by Richard Johansson (richard@cs.lth.se).
 
 $Log: FeatureDeclaration.java,v $
 Revision 1.3  2010-06-07 11:29:59  johansson
 Added.

 Revision 1.2  2009/03/06 13:21:40  johansson
 Fixed bug in subset.

 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:04  richard
 Branched.

 Revision 1.10  2008/04/02 11:22:43  richard
 New constructors.

 Revision 1.9  2007/10/08 12:31:34  richard
 Refactored again to remove atoms.

 Revision 1.8.2.1  2007/10/08 12:19:04  richard
 Refactored again to remove atoms.

 Revision 1.8  2007/10/04 12:33:08  richard
 Subset.

 Revision 1.7  2007/09/28 14:47:53  richard
 Refactored to atom-based.

 Revision 1.6  2007/02/20 12:26:22  richard
 Added getFeatures.

 Revision 1.5  2007/01/26 09:46:18  richard
 Uses trove hashmap.

 Revision 1.4  2006/09/07 17:51:19  richard
 New revision.

 Revision 1.3  2006/04/11 08:47:43  richard
 Added documentation.
 
 Revision 1.2  2006/04/05 11:41:09  richard
 Implements serializable. Uses subsets.
 
 Revision 1.1  2006/01/13 13:19:37  richard
 Added the file.
 
 
 */
package se.lth.cs.nlp.nlputils.ml_long;

import se.lth.cs.nlp.nlputils.core.*;
import java.util.*;
import java.io.Serializable;
import gnu.trove.*;     

/**
 * Declaration of a feature set.
 * 
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class FeatureDeclaration implements Serializable {
    
    private static final long serialVersionUID = 0L;
    
    private TObjectIntHashMap<String> nameToIndex = new TObjectIntHashMap<String>();
    
    private int index = 0;
    
    private int nUsed;
    
    /**
     * Creates a new, empty, feature set.
     *
     */
    public FeatureDeclaration() {
    }
    
    /**
     * Creates a new feature set from the given feature names.
     *
     * @param names the feature names.
     */
    public FeatureDeclaration(String[] names) {
        this();
        for(int i = 0; i < names.length; i++)
            addFeature(names[i]);
    }

    public FeatureDeclaration(TObjectIntHashMap<String> nameToIndex) {
        this();
        this.nameToIndex = nameToIndex;
        TObjectIntIterator iterator = nameToIndex.iterator();
        nUsed = 0;
        for(int i = nameToIndex.size(); i-- > 0;) {
            iterator.advance();
            int val = iterator.value();
            if(val != MLConstants.EXCLUDED_MAGIC_NUMBER) {
                nUsed++;
                if(val + 1 > index)
                    index = val + 1;
            }
        }
    }
    
    public FeatureDeclaration(String fromString) {
        this();
        nameToIndex = new TObjectIntHashMap();
        String[] ss = fromString.substring(1, fromString.length() - 1).split("\\,\\s*");
        for(String s: ss) {
            if(s.startsWith("("))
                nameToIndex.put(s.substring(1, s.length() - 1), MLConstants.EXCLUDED_MAGIC_NUMBER);
            else {
                String[] p = s.split(" --> ");
                int ix = Integer.parseInt(p[1]);
                nameToIndex.put(p[0], ix);
                nUsed++;
                if(ix + 1 > index)
                    index = ix + 1;
            }
        }
    }
    
    /**
     * Adds a new feature to the feature set.
     * 
     * @param name the name of the new feature.
     */
    public void addFeature(String name) {
        Util.check(!nameToIndex.containsKey(name), "Feature name " + name + " already in use.");
        int n = index++;
        nameToIndex.put(name, n);
        nUsed++;
    }
    

    
//    /**
//     * Adds a new feature to the feature set, and assigns it the given number.
//     * 
//     * @param name the name of the new feature.
//     * @param n the number to assign it.
//     */
//    public void addFeature(String name, int n) {
//        Aux.check(!nameToIndex.containsKey(name), "Feature name " + name + " already in use.");
//        index = n + 1;
//        nameToIndex.put(name, n);
//    }
    
    /**
     * Returns the feature number of the given feature.
     * 
     * @param name the feature name.
     * @return the number.
     */
    public int getFeatureIndex(String name) {
        Util.check(nameToIndex.containsKey(name), "Unknown feature " + name);
        int n = nameToIndex.get(name);
        return n;
    }
    
    public String getFeatureName(int index) {
        for(Object k: nameToIndex.keys()) {
            String key = (String) k;
            if(nameToIndex.get(key) == index)
                return key;
        }
        return null;
    }
    
    /**
     * Returns the number of active features in this feature set.
     * 
     * @return the number of active features.
     */
    public int size() {
        return nUsed;
    }
    
    /**
     * Returns a FeatureDeclaration that is a subset of this declaration.
     * 
     * @param used the names of the features that should be included in the 
     * subset.
     * @return the subset.
     */
    public FeatureDeclaration subset(Set<String> used) {
        for(String f: used)
            if(!nameToIndex.containsKey(f))
                throw new IllegalArgumentException("Unknown feature " + f);
        
        TObjectIntHashMap<String> n2 = nameToIndex.clone();
        int count = 0;

        TObjectIntIterator iterator = n2.iterator();
        for (int i = n2.size(); i-- > 0;) {
            iterator.advance();
            
            if(iterator.value() != MLConstants.EXCLUDED_MAGIC_NUMBER)            
                if(!used.contains(iterator.key()))
                    iterator.setValue(MLConstants.EXCLUDED_MAGIC_NUMBER);
                else
                    iterator.setValue(count++);
        }
        FeatureDeclaration out = new FeatureDeclaration();
        out.nameToIndex = n2;
        out.nUsed = count;
        return out;
    }

    public int[] getSubsetIndices(FeatureDeclaration subset) {
    	int[] out = new int[nUsed];
    	for(Object k: nameToIndex.keys()) {
    		String f = (String) k;
    		int i1 = nameToIndex.get(f);
    		int i2 = subset.nameToIndex.get(f);
    		if(i1 != MLConstants.EXCLUDED_MAGIC_NUMBER)
    			out[i1] = i2;
    	}
    	return out;
    }
    
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append("[");
        TObjectIntIterator iterator = nameToIndex.iterator();
        for (int i = nameToIndex.size(); i-- > 0;) {
            iterator.advance();
            if(iterator.value() != MLConstants.EXCLUDED_MAGIC_NUMBER)
                sb.append(iterator.key() + " --> " + iterator.value());
            else
                sb.append("(" + iterator.key() + ")");
            if(i > 0)
                sb.append(", ");
        }     
        sb.append("]");
        return sb.toString();
    }
    
    public Set<String> getFeatures() {
        TreeSet<String> out = new TreeSet<String>();
        TObjectIntIterator<String> iterator = nameToIndex.iterator();
        for (int i = nameToIndex.size(); i-- > 0;) {
            iterator.advance();
            if(iterator.value() != MLConstants.EXCLUDED_MAGIC_NUMBER)
                out.add(iterator.key());
        }     
        return out;
    }
    

}