/*
 FeatureList.java
 
 Created on Dec 22, 2005 by Richard Johansson (richard@cs.lth.se).
 
 $Log: FeatureList.java,v $
 Revision 1.2  2010-06-07 11:29:59  johansson
 Added.

 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:04  richard
 Branched.

 Revision 1.13  2008/04/02 11:23:09  richard
 Better set output.

 Revision 1.12  2007/10/08 12:31:34  richard
 Refactored again to remove atoms.

 Revision 1.11.2.1  2007/10/08 12:19:04  richard
 Refactored again to remove atoms.

 Revision 1.11  2007/10/04 12:33:22  richard
 Clone.

 Revision 1.10  2007/09/28 14:47:53  richard
 Refactored to atom-based.

 Revision 1.9  2007/08/14 07:05:40  richard
 New toString.

 Revision 1.8  2007/08/13 14:41:12  richard
 Clone.

 Revision 1.7  2006/09/18 15:17:17  richard
 Serializable.

 Revision 1.6  2006/09/08 15:31:51  richard
 Parametrized FeatureList.

 Revision 1.5  2006/09/07 17:51:19  richard
 New revision.

 Revision 1.4  2006/04/11 08:58:31  richard
 Added documentation.
 
 Revision 1.3  2006/04/05 11:42:16  richard
 Adapted for subsets. Prints training string.
 
 Revision 1.2  2006/01/13 13:35:22  richard
 Made the array public for convenience.
 
 Revision 1.1  2006/01/13 13:19:37  richard
 Added the file.
 
 
 */
package se.lth.cs.nlp.nlputils.ml_long;

import java.util.*;

import se.lth.cs.nlp.nlputils.core.*;

import java.io.Serializable;

/**
 * A list of feature values.
 *
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class FeatureList implements Serializable {
    private static final long serialVersionUID = 2L;

    /**
     * The feature declaration that is used by this feature list.
     */
    public FeatureDeclaration decl;
    
    /**
     * The array of feature values.
     */
    public Object[] values;
    
    /**
     * Creates a new feature list given a declaration.
     * 
     * @param decl the declaration to use.
     */
    public FeatureList(FeatureDeclaration decl) {
        values = new Object[decl.size()];
        this.decl = decl;
    }
    
    /**
     * Creates a new feature list given a declaration, using the given array
     * as storage.
     * 
     * @param decl the declaration to use.
     * @param features the array to use as storage.
     */
    public FeatureList(FeatureDeclaration decl, Object[] values) {
        Util.check(values.length == decl.size(), "Wrong feature vector size.");
        this.values = values;
        this.decl = decl;
    }
    
    /**
     * Sets the value of a feature given by name. If the feature is declared
     * inactive, then this method returns silently.
     * 
     * @param featureName the name of the feature.
     * @param value the value of the feature.
     */
    public void setFeature(String featureName, String value) {
        int index = decl.getFeatureIndex(featureName);
        if(index != MLConstants.EXCLUDED_MAGIC_NUMBER)
            values[index] = value;
    }
    
    public void setFeature(String featureName, HashSet<String> value) {
        int index = decl.getFeatureIndex(featureName);
        if(index != MLConstants.EXCLUDED_MAGIC_NUMBER)
            values[index] = value;
    }

    public void setFeature(String featureName, double[] value) {
        int index = decl.getFeatureIndex(featureName);
        if(index != MLConstants.EXCLUDED_MAGIC_NUMBER)
            values[index] = value;
    }
    
    public void setFeature(String featureName, double value) {
        int index = decl.getFeatureIndex(featureName);
        if(index != MLConstants.EXCLUDED_MAGIC_NUMBER)
            values[index] = value;
    }
    
    /**
     * Sets the value of a feature given by index. If the feature is declared
     * inactive, then this method returns silently.
     * 
     * @param featureIndex the index of the feature.
     * @param value the value of the feature.
     */
    public void setFeature(int featureIndex, String value) {
        if(featureIndex != MLConstants.EXCLUDED_MAGIC_NUMBER)
            values[featureIndex] = value;
    }
    
    public void setFeature(int featureIndex, HashSet<String> value) {
        if(featureIndex != MLConstants.EXCLUDED_MAGIC_NUMBER)
            values[featureIndex] = value;
    }
    
    public void setFeature(int featureIndex, double[] value) {
        if(featureIndex != MLConstants.EXCLUDED_MAGIC_NUMBER)
            values[featureIndex] = value;
    }
    
    public void setFeature(int featureIndex, double value) {
        if(featureIndex != MLConstants.EXCLUDED_MAGIC_NUMBER)
            values[featureIndex] = value;
    }
    
    /**
     * Gets the value of a feature given by name. If the feature is declared
     * inactive, then this method returns null.
     * 
     * @param featureName the name of the feature.
     * @return the value of the feature, or null if the feature is inactive.
     */
    public Object getFeature(String featureName) {
        int index = decl.getFeatureIndex(featureName);
        if(index != MLConstants.EXCLUDED_MAGIC_NUMBER)
            return values[index];
        else
            return null;
    }
    
    /**
     * Gets the value of a feature given by index. If the feature is declared
     * inactive, then this method returns null.
     * 
     * @param index the index of the feature.
     * @return the value of the feature, or null if the feature is inactive.
     */
    public Object getFeature(int index) {
        if(index != MLConstants.EXCLUDED_MAGIC_NUMBER)
            return values[index];
        else
            return null;
    }
    
    public int hashCode() {
        int out = decl.hashCode();
        for(int i = 0; i < values.length; i++)
            if(values[i] != null)
                out += values[i].hashCode();
        return out;
    }
    
    public boolean equals(Object o) {
        FeatureList l = (FeatureList) o;
        return Arrays.deepEquals(values, l.values);
    }
    
    public String toString() {
        //return "[decl:" + decl.hashCode() + " " + Arrays.toString(features) + "]";
        StringBuilder sb = new StringBuilder("{");        
        for(String fname: decl.getFeatures()) {
            sb.append(fname);
            sb.append("=");
            sb.append(getFeature(fname) + ", ");
        }
        sb.setLength(sb.length() - 2);
        sb.append("}");
        return sb.toString();
    }
    
    /**
     * 
     * @return the training string.
     */
    public String toTrainingString() {
        StringBuilder out = new StringBuilder();
        for(int i = 0; i < values.length; i++) {
            if(values[i] instanceof String)
                out.append(escapeNS(values[i]));
            else
                out.append(join(values[i]));
            out.append(MLConstants.FEATURE_SEPARATOR);
        }
        return out.substring(0, out.length() - 1);
    }

    private String escapeNS(Object o) {
        String s = (String) o;
        s = s.replaceAll("\\" + MLConstants.FEATURE_SEPARATOR, MLConstants.ESCAPE_FEATURE_SEPARATOR);
        s = s.replaceAll("" + MLConstants.SET_SEPARATOR, MLConstants.ESCAPE_SET_SEPARATOR);
        return s;
    }

    private String join(Object o) {
        HashSet<String> ss = (HashSet) o;
        if(ss.size() == 0)
            return MLConstants.EMPTY_SET;
        StringBuilder sb = new StringBuilder();
        boolean start = true;
        for(String s: ss) {
            if(!start)
                sb.append(MLConstants.SET_SEPARATOR);
            sb.append(escapeNS(s));
            start = false;
        }
        return sb.toString();
    }
    
    public FeatureList clone() {
        FeatureList cl = new FeatureList(decl);
        for(int i = 0; i < values.length; i++) {
            Object v = values[i];
            if(v instanceof String || v instanceof Double)
                cl.values[i] = v;
            else if(v instanceof HashSet)
                cl.values[i] = ((HashSet) v).clone();
            else if(v instanceof double[])
                cl.values[i] = ((double[]) v).clone();
        }
        return cl;
    }
 
}