/*
 FeatureSearch.java
  
 Created on Apr 2, 2008 by Richard Johansson (richard@cs.lth.se).

 $Log: FeatureSearch.java,v $
 Revision 1.2  2010-06-07 11:29:59  johansson
 Added.

 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:04  richard
 Branched.

 Revision 1.2  2008/04/03 08:38:23  richard
 Specify a start set.

 Revision 1.1  2008/04/02 11:23:45  richard
 Added.

   
 */
package se.lth.cs.nlp.nlputils.ml_long;

import java.util.*;

import se.lth.cs.nlp.nlputils.core.Pair;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public abstract class FeatureSearch {

    public static abstract class Evaluator {
        public abstract double evaluate(Classifier<FeatureList, String> cl,
                List<Pair<FeatureList, String>> testset);
    }

    public static class AccuracyEvaluator extends Evaluator {
        public double evaluate(Classifier<FeatureList, String> cl,
                List<Pair<FeatureList, String>> testset) {
            int ncorrect = 0;
            int n = testset.size();
            for(Pair<FeatureList, String> instance: testset) {
                String guess = cl.classify(instance.left);
                if(guess.equals(instance.right))
                    ncorrect++;
            }
            return (double) ncorrect / n;
        }
    }
    
    public static class FMeasureEvaluator extends Evaluator {
        private String zero;
        public FMeasureEvaluator(String zero) {
            this.zero = zero;
        }
        public double evaluate(Classifier<FeatureList, String> cl,
                List<Pair<FeatureList, String>> testset) {
            int ncorrect = 0;
            int ngold = 0;
            int nguesses = 0;
            for(Pair<FeatureList, String> instance: testset) {
                String guess = cl.classify(instance.left);
                if(!guess.equals(zero)) {
                    nguesses++;
                    if(guess.equals(instance.right)) {
                        ncorrect++;
                    }
                }
                if(!instance.right.equals(zero))
                    ngold++;
            }
            double p = (double) ncorrect / nguesses;
            double r = (double) ncorrect / ngold;
            if(p == 0 || r == 0)
                return 0;
            else
                return 2*p*r/(p+r);
        }
    }
    
    public abstract List<Pair<Object, Double>> search(ArrayList<Pair<FeatureList, String>> trainingSet,
            ArrayList<Pair<FeatureList, String>> testSet,
            String alg, String algArgs, Evaluator eval);

    public abstract void setStartSets(Collection<String> features,
            Collection<Pair<String, String>> featurePairs);
    
    private static Pair<HashSet<String>, HashSet<Pair<String, String>>> 
        parseStartSet(String arg) {

        arg = arg.trim();

        if(!arg.startsWith("{") || !arg.endsWith("}"))
            throw new IllegalArgumentException("Illegal set format");

        arg = arg.substring(1, arg.length() - 1).trim();

        String[] ss = arg.split(",\\s*");

        HashSet<String> fs = new HashSet();
        HashSet<Pair<String, String>> ps = new HashSet();

        for(int i = 0; i < ss.length; i++) {        
            String s = ss[i];
            if(s.startsWith("(")) {
                if(i == ss.length - 1)
                    throw new IllegalArgumentException("Illegal pair format " + s);
                String s2 = ss[++i];
                if(!s2.endsWith(")"))
                    throw new IllegalArgumentException("Illegal pair format " + s2);
                s = s.substring(1);
                s2 = s2.substring(0, s2.length() - 1).trim();
                if(s.equals("") || s2.equals(""))
                    throw new IllegalArgumentException("Illegal pair format (empty)");
                ps.add(new Pair(s, s2));
            } else
                fs.add(s);
        }
        
        return new Pair(fs, ps);
    }
    
    public static void main(String[] argv) {
        try {
            if(argv.length < 4) {
                System.err.println("Must specify training set, test set, algorithm and arguments.");
                System.exit(1);
            }
            
            ArrayList<Pair<FeatureList, String>> trs = MulticlassDefinition.readDataSet(argv[0], null);
            FeatureDeclaration decl = trs.get(0).left.decl;
            
            ArrayList<Pair<FeatureList, String>> tes = MulticlassDefinition.readDataSet(argv[1], decl);
            
            FeatureSearch fs = new GreedyForwardFeatureSearch();
            Evaluator eval = new FMeasureEvaluator("False"); //<NONE>");

            if(argv.length >= 5) {
                Pair<HashSet<String>, HashSet<Pair<String, String>>> p
                    = parseStartSet(argv[4]);
                fs.setStartSets(p.left, p.right);
            }
            
            List out = fs.search(trs, tes, argv[2], argv[3], eval);
            System.out.println(out);
        } catch(Exception e) {
            e.printStackTrace();
        }
    }
    
}
