/*
 GreedyForwardFeatureSearch.java
  
 Created on Apr 2, 2008 by Richard Johansson (richard@cs.lth.se).

 $Log: GreedyForwardFeatureSearch.java,v $
 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:04  richard
 Branched.

 Revision 1.6  2008/04/10 07:00:07  richard
 Cross-validates.

 Revision 1.5  2008/04/07 12:11:42  richard
 Update.

 Revision 1.4  2008/04/03 08:38:31  richard
 Specify a start set.

 Revision 1.3  2008/04/03 07:48:18  richard
 Bug in pair set.

 Revision 1.2  2008/04/02 13:26:25  richard
 Skips inverse pairs.

 Revision 1.1  2008/04/02 11:23:52  richard
 Added.

   
 */
package se.lth.cs.nlp.nlputils.ml_long;

import java.util.*;

import se.lth.cs.nlp.nlputils.core.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class GreedyForwardFeatureSearch extends FeatureSearch {

    private static final double MIN_IMPROVEMENT = 1e-6;
    
    static final boolean CROSS_VALIDATION = true;
    static final int NFOLDS = 5;
    
    private static final class RemappingFE extends FeatureExtractor<FeatureList> {
        private static final long serialVersionUID = 1L;
        private int[] mapping;
        RemappingFE(FeatureDeclaration fullDecl, FeatureDeclaration subset) {
            mapping = new int[fullDecl.size()];
            for(int i = 0; i < mapping.length; i++) {
                String fname = fullDecl.getFeatureName(i);
                mapping[i] = subset.getFeatureIndex(fname);
            }
        }
        
        public void extractFeatures(FeatureList l, FeatureList l2) {
            for(int i = 0; i < l.values.length; i++) {
                int ix = mapping[i];
                if(ix != MLConstants.EXCLUDED_MAGIC_NUMBER)
                    l2.values[ix] = l.values[i];
            }                
        }
    };   

    private HashSet<String> startFs;
    private HashSet<Pair<String, String>> startPs;
    
    public void setStartSets(Collection<String> features,
            Collection<Pair<String, String>> featurePairs) {
        startFs = new HashSet(features);
        startPs = new HashSet(featurePairs);
    }
    
    private double run(FeatureDeclaration fullDecl,
            HashSet<String> currentFeatures,
            HashSet<Pair<String, String>> currentPairs,
            ArrayList<Pair<FeatureList, String>> trainingSet,
            ArrayList<Pair<FeatureList, String>> testSet,
            String algString, String algArgs,
            Evaluator eval) {

        FeatureDeclaration decl = fullDecl.subset(currentFeatures);
        RemappingFE fe = new RemappingFE(fullDecl, decl);
        
        if(!CROSS_VALIDATION) {
            FeatureListEncoder lenc = new PairingFeatureListEncoder(decl, currentPairs);

            MulticlassDefinition<FeatureList, String> spec
            = new MulticlassDefinition(trainingSet, decl, fe, lenc);        
            LearningAlgorithm alg = new AlgorithmFactory().create(algString, algArgs);
            Classifier<FeatureList, String> cl = alg.train(spec, trainingSet);

            return eval.evaluate(cl, testSet);
        } else {
            double sum = 0;            
            ArrayList<Pair<FeatureList, String>> fullSet = new ArrayList();
            fullSet.addAll(trainingSet);
            fullSet.addAll(testSet);

            int foldSize = fullSet.size() / NFOLDS; 
            for(int i = 0; i < NFOLDS; i++) {

                ArrayList<Pair<FeatureList, String>> trs = new ArrayList();
                ArrayList<Pair<FeatureList, String>> tes = new ArrayList();
                
                int index = 0;
                for(int j = 0; j < i; j++)
                    for(int k = 0; k < foldSize; k++)
                        trs.add(fullSet.get(index++));
                for(int k = 0; k < foldSize; k++)
                    tes.add(fullSet.get(index++));
                while(index < fullSet.size())
                    trs.add(fullSet.get(index++));                    
                
                FeatureListEncoder lenc = new PairingFeatureListEncoder(decl, currentPairs);
                MulticlassDefinition<FeatureList, String> spec
                = new MulticlassDefinition(trs, decl, fe, lenc);
                LearningAlgorithm alg = new AlgorithmFactory().create(algString, algArgs);
                Classifier<FeatureList, String> cl = alg.train(spec, trs);
                sum += eval.evaluate(cl, tes);
            }
            return sum / NFOLDS;
        }        
    }
    
    public List<Pair<Object, Double>> search(ArrayList<Pair<FeatureList, String>> trainingSet,
            ArrayList<Pair<FeatureList, String>> testSet,
            String algString, String algArgs, Evaluator eval) {
        FeatureDeclaration decl = trainingSet.get(0).left.decl;
        List<Pair<Object, Double>> out = new ArrayList();
        
        HashSet<String> usedFeatures = startFs == null? new HashSet(): startFs;
        HashSet<Pair<String, String>> usedPairs = startPs == null? new HashSet(): startPs;

        for(String s: usedFeatures)
            out.add(new Pair(s, Double.NaN));
        for(Pair p: usedPairs)
            out.add(new Pair(p, Double.NaN));
        
        AlgorithmFactory.setVerbosity(0);
        
        double bestResult = run(decl, usedFeatures, usedPairs, trainingSet, testSet,
                algString, algArgs, eval);
        
        System.out.println("* Baseline value: " + bestResult);
        
        /* 1. Add single features. */
        while(true) {
            double previousResult = bestResult;
            
            HashSet<String> bestFeatures = usedFeatures;
            Pair<Object, Double> histItem = null;
            
            for(String feature: decl.getFeatures()) {
                if(usedFeatures.contains(feature))
                    continue;
                System.out.println("* Trying " + feature);
                HashSet<String> newSet = new HashSet(usedFeatures);
                newSet.add(feature);
                
                double result = run(decl, newSet, usedPairs, trainingSet, testSet,
                        algString, algArgs, eval);
                
                System.out.println("* Result: " + result);
                if(result >= bestResult + MIN_IMPROVEMENT) {
                    bestFeatures = newSet;
                    bestResult = result;
                    histItem = new Pair(feature, result - previousResult);
                    System.out.println("* Improvement: " + histItem);
                    /*try {
                        System.in.read();
                    } catch(Exception e) {}*/
                }
            }
            
            if(histItem == null)
                break;
            else {
                out.add(histItem);
                usedFeatures = bestFeatures;
                System.out.println("*** Best result: " + histItem);
                System.out.println("*** Current set: " + out);
                previousResult = bestResult;
            }
        }
        
        while(true) {
            double previousResult = bestResult;
            Pair<Object, Double> histItem = null;

            HashSet<String> bestSet = usedFeatures;
            HashSet<Pair<String, String>> bestPairs = usedPairs;

            HashSet<Pair<String, String>> done = new HashSet();
            
            for(String f1: usedFeatures) {
                for(String f2: decl.getFeatures()) {
                    if(f2.equals(f1))
                        continue;
                    Pair<String, String> p = new Pair(f1, f2);
                    Pair<String, String> pinv = new Pair(f2, f1);
                    if(usedPairs.contains(p) || usedPairs.contains(pinv))
                        continue;
                    if(done.contains(p) || done.contains(pinv))
                        continue;
                    done.add(p);
                    done.add(pinv);
                    
                    System.out.println("* Trying " + p);

                    HashSet<String> newSet = new HashSet(usedFeatures);
                    HashSet<Pair<String, String>> newPairs = new HashSet(usedPairs);
                    newSet.add(f2);
                    newPairs.add(p);

                    double result = run(decl, newSet, newPairs, trainingSet, testSet,
                            algString, algArgs, eval);
                    
                    System.out.println("* Result: " + result);
                    if(result >= bestResult + MIN_IMPROVEMENT) {
                        bestPairs = newPairs;
                        bestSet = newSet;
                        bestResult = result;
                        histItem = new Pair(p, result - previousResult);
                        System.out.println("* Improvement: " + histItem);
                        /*try {
                            System.in.read();
                        } catch(Exception e) {}*/
                    }
                    
                }
            }
            
            if(histItem == null)
                break;
            else {
                out.add(histItem);
                usedPairs = bestPairs;
                usedFeatures = bestSet;
                System.out.println("*** Best result: " + histItem);
                System.out.println("*** Current set: " + out);
                previousResult = bestResult;
            }
        }
        
        System.out.println("**** Best result: " + bestResult);
        System.out.println("**** Final set: " + out);
        
        return out;
    }

}
