/*
 IntPairEncoder.java
  
 Created on Oct 5, 2007 by Richard Johansson (richard@cs.lth.se).

 $Log: IntPairEncoder.java,v $
 Revision 1.2  2010-06-07 11:29:59  johansson
 Added.

 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:04  richard
 Branched.

 Revision 1.7  2008/04/02 11:24:06  richard
 Limits.

 Revision 1.6  2008/02/22 13:17:33  richard
 Added decode.

 Revision 1.5  2007/11/06 09:57:23  richard
 Changed remapLeft.

 Revision 1.4  2007/10/22 07:47:00  richard
 Uses IntHistogram.

 Revision 1.3  2007/10/22 07:31:35  richard
 Renamed IntPairEncoder2 to IntPairEncoder.

 Revision 1.3  2007/10/22 07:27:37  richard
 Added remapLeft.

 Revision 1.2  2007/10/20 14:46:01  richard
 Added remap and freeze.

 Revision 1.1  2007/10/19 13:50:31  richard
 Added.

 Revision 1.2  2007/10/08 12:31:34  richard
 Refactored again to remove atoms.

 Revision 1.1.2.1  2007/10/08 12:19:04  richard
 Refactored again to remove atoms.

   
 */
package se.lth.cs.nlp.nlputils.ml_long;

import gnu.trove.*;

import java.io.Serializable;
import se.lth.cs.nlp.nlputils.core.*;


/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class IntPairEncoder implements Serializable {

    private static final long serialVersionUID = 0L;

    // 100510
    private TLongIntHashMap encoding2 = new TLongIntHashMap();
    private TIntIntHashMap encoding = new TIntIntHashMap();

    private int size = 0;

    private int N_F_LOW_BITS;
    private int N_C_LOW_BITS;

    private int LOW_PART_F;
    private int LOW_PART_C;

    private static final int DEFAULT = 12;
    
    private IntHistogram hist;
    
    private boolean frozen = false;
    
    public IntPairEncoder(int halfbits, boolean collectStats) {
        if(halfbits > 32 || halfbits < 0)
            throw new IllegalArgumentException("0 < halfbits < 32");
        N_F_LOW_BITS = halfbits;
        N_C_LOW_BITS = 32 - halfbits;
        LOW_PART_F = 1 << N_F_LOW_BITS;
        LOW_PART_C = 1 << N_C_LOW_BITS;        
        if(collectStats)
            hist = new IntHistogram();
    }

    public IntPairEncoder(int halfbits) {
        this(halfbits, false);
    }
    
    public IntPairEncoder() {
        this(DEFAULT, false);
    }
    
    public IntPairEncoder clone() {
    	IntPairEncoder out = new IntPairEncoder();
    	out.encoding = (TIntIntHashMap) encoding.clone();
    	out.encoding2 = (TLongIntHashMap) encoding2.clone();
    	out.frozen = frozen;
    	out.size = size;
    	out.N_C_LOW_BITS = N_C_LOW_BITS;
    	out.N_F_LOW_BITS = N_F_LOW_BITS;
    	out.LOW_PART_C = LOW_PART_C;
    	out.LOW_PART_F = LOW_PART_F;
    	return out;
    }
    
    public int encode(int c, int a) {
        int encoded;
        if(a < LOW_PART_F && c < LOW_PART_C) {
            int l = a + (c << N_F_LOW_BITS);
            encoded = encoding.get(l);
            if(encoded == 0 && !frozen) {
                encoded = ++size;
                encoding.put(l, encoded);
            }
        } else {
            long l = a + ((long) c << 32L);
            encoded = encoding2.get(l);
            if(encoded == 0 && !frozen) {
                encoded = ++size;
                encoding2.put(l, encoded);
            }
        }
        if(encoded != 0 && hist != null)
            hist.add(encoded);

        if(encoded == 0) {
            //System.out.println("(" + c + ", " + a + ") -> " + encoded);
            //System.exit(1);
        }
        
        return encoded;
    }
    
    public void remapLeft(int ntop) {
        
        TIntIntHashMap remapping = new TIntIntHashMap();
        int n = 1;

        remapping.put(1, 1);
        
        for(IntPair p: hist.asSortedList()) {
            if(p.right != 1)
                remapping.put(p.right, ++n);
            if(n > ntop)
                break;
        }

        TIntIntHashMap newEncoding = new TIntIntHashMap();
        TLongIntHashMap newEncoding2 = new TLongIntHashMap();

        IntHistogram newHist = new IntHistogram();
        //hist = null;
        
        TIntIntIterator it = encoding.iterator();
        for(int i = encoding.size(); i-- > 0; ) {
            it.advance();
            int oldValue = it.value();
            
            int firstKey = it.key() >> N_F_LOW_BITS;
            int secondKey = it.key() & ((1 << N_F_LOW_BITS) - 1);

            int remappedFirstKey = remapping.get(firstKey);
            int remappedTo = remapping.get(oldValue);

            if (remappedTo != 0 && remappedFirstKey != 0) {
                if(remappedFirstKey < LOW_PART_C) {
                    int l = secondKey + (remappedFirstKey << N_F_LOW_BITS);
                    newEncoding.put(l, remappedTo);
                } else {
                    long l = secondKey + ((long) remappedFirstKey << 32L);
                    newEncoding2.put(l, remappedTo);
                }
                newHist.setFrequency(remappedTo, hist.getFrequency(oldValue));
            }
        }
        
        TLongIntIterator it2 = encoding2.iterator();
        for(int i = encoding2.size(); i-- > 0; ) {
            it2.advance();
            int oldValue = it2.value();
            
            int firstKey = (int) (it2.key() >> 32L);
            int secondKey = (int) (it2.key() & ((1L << 32L) - 1L));

            int remappedFirstKey = remapping.get(firstKey);
            int remappedTo = remapping.get(oldValue);
            
            if (remappedTo != 0 && remappedFirstKey != 0) {
                if(remappedFirstKey < LOW_PART_C) {
                    int l = secondKey + (remappedFirstKey << N_F_LOW_BITS);
                    newEncoding.put(l, remappedTo);
                } else {
                    long l = secondKey + ((long) remappedFirstKey << 32L);
                    newEncoding2.put(l, remappedTo);
                }
                newHist.setFrequency(remappedTo, hist.getFrequency(oldValue));
            }
        }

        int oldSize = encoding.size() + encoding2.size();
        int newSize = newEncoding.size() + newEncoding2.size();
        
        encoding = newEncoding;
        encoding2 = newEncoding2;
        hist = newHist;
        //hist = new IntHistogram();
        
        System.out.println("Reduction: " + oldSize + " to " + newSize);
        
    }
    
    public void freeze() {
        frozen = true;
        hist = null;
    }
    
    public int size() {
        return encoding.size() + encoding2.size();
    }
    
    public IntPair decode(int encoded) {
        int[] keys = encoding.keys();
        int[] values = encoding.getValues();
        for(int i = 0; i < keys.length; i++)
            if(values[i] == encoded) {
                int upper = keys[i] >> N_F_LOW_BITS;
                int lower = keys[i] & ((1 << N_F_LOW_BITS) - 1);
                return new IntPair(upper, lower);
            }
        long[] keys2 = encoding2.keys();
        int[] values2 = encoding2.getValues();
        for(int i = 0; i < keys2.length; i++)
            if(values2[i] == encoded) {
                int upper = (int) (keys2[i] >> 32L);
                int lower = (int) (keys2[i] & ((1L << 32L) - 1L));
                return new IntPair(upper, lower);
            }
        return null;
    }
 
    public void remap(TIntIntHashMap m) {
    	if(hist != null)
    		throw new IllegalStateException("unsupported");
    		
    	TIntIntHashMap newEncoding = new TIntIntHashMap();
        TLongIntHashMap newEncoding2 = new TLongIntHashMap();

        {
        	TIntIntIterator it = encoding.iterator();
        	for(int i = encoding.size(); i-- > 0; ) {
        		it.advance();
        		int pair = it.key();
        		int id = it.value();            
        		int remapped = m.get(id);
        		if(remapped != 0)
        			newEncoding.put(pair, remapped);            
        	}
        }

        {
        	TLongIntIterator it2 = encoding2.iterator();
        	for(int i = encoding2.size(); i-- > 0; ) {
        		it2.advance();
        		long pair = it2.key();
        		int id = it2.value();            
        		int remapped = m.get(id);
        		if(remapped != 0)
        			newEncoding2.put(pair, remapped);            
        	}
        }
        
        encoding = newEncoding;
        encoding2 = newEncoding2;
        size = encoding.size() + encoding2.size();
    }
    
}
