/*
 LinearModel.java
  
 Created on Sep 4, 2006 by Richard Johansson (richard@cs.lth.se).

 $Log: LinearModel2.java,v $
 Revision 1.3  2010-07-05 14:04:22  johansson
 Added set.

 Revision 1.2  2009/04/16 11:59:36  johansson
 Added update on index.

 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:04  richard
 Branched.

 Revision 1.10  2007/02/26 08:04:54  richard
 Added factor.

 Revision 1.9  2006/10/25 07:07:39  richard
 SparseVector as a sequence.

 Revision 1.8  2006/10/05 06:59:01  richard
 New serial id.

 Revision 1.7  2006/09/26 14:48:14  richard
 Added standard vector operations to SparseVector.

 Revision 1.6  2006/09/18 15:17:17  richard
 Serializable.

 Revision 1.5  2006/09/17 16:10:04  richard
 May repeat indices.

 Revision 1.4  2006/09/09 16:18:05  richard
 Uses the new SparseVector and GNU Trove.

 Revision 1.3  2006/09/08 15:32:51  richard
 Parametrized FeatureList.

 Revision 1.2  2006/09/08 11:54:39  richard
 Quadratic kernel experiments.

 Revision 1.1  2006/09/07 17:54:18  richard
 Added files.

   
 */
package se.lth.cs.nlp.nlputils.ml_long;

import se.lth.cs.nlp.nlputils.core.*;
import java.io.Serializable;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class LinearModel2 extends Model implements Serializable {

    private static final long serialVersionUID = 0L;
    
	/* TODO b�ttre nollkontroll �VERALLT */
	
    private ClosedLongDoubleMap table;
    private int capacity;
    
    public LinearModel2(int size) {
        this.capacity = size;
        table = new ClosedLongDoubleMap(size);
    }
    
    public void update(double d, SparseVector rep) {
        while(rep != null) {
            for(int i = 0; i < rep.index; i++) {
                double d2 = d*rep.values[i]*rep.factor;
                table.add(rep.keys[i], d2);
            }
            rep = rep.next;
        }
    }

    public void update(long index, double delta) {
    	table.add(index, delta);
    }
    
    public void set(long index, double value) {
    	table.put(index, value);
    }
    
    private static final double TOLERANCE = 1e-15;
    
    public void updateDiff(double d, SparseVector r1, SparseVector r2, boolean sorted) {
        if(Math.abs(d) < TOLERANCE)
            return;
        if(!sorted) {
            r1.sortIndices();
            r2.sortIndices();
        }
        
        if(r1.next != null || r2.next != null)
            throw new RuntimeException("Stacked vectors unimplemented...");
        int ix1 = 0, ix2 = 0;
        while(ix1 < r1.index || ix2 < r2.index) {
            long k1 = ix1 == r1.index? 0: r1.keys[ix1];
            long k2 = ix2 == r2.index? 0: r2.keys[ix2];
            
            if(k2 == 0 || (k1 != 0 && k1 < k2)) {
                double dd = d*r1.values[ix1]*r1.factor;
                if(dd > TOLERANCE || dd < -TOLERANCE)
                    table.add(k1, dd);
                ix1++;
            } else if(k1 == 0 || (k2 != 0 && k2 < k1)) {
                double dd = -d*r2.values[ix2]*r2.factor;
                if(dd > TOLERANCE || dd < -TOLERANCE)
                    table.add(k2, dd);
                ix2++;                
            } else if(k1 == k2) {
                double dd = d*(r1.values[ix1]*r1.factor - r2.values[ix2]*r2.factor);
                if(dd > TOLERANCE || dd < -TOLERANCE)
                    table.add(k1, dd);
                ix1++;
                ix2++;
            } else {
                throw new RuntimeException("bug, shouldn't be here...");
            }            
        }
        
    }
    
    public void clear() {
        table = new ClosedLongDoubleMap(capacity);
    }
        
    public int size() {
        return table.size();
    }
    
    public double score(SparseVector rep) {
        double score = 0;
        while(rep != null) {
        	for(int i = 0; i < rep.index; i++) {
        		double d2 = table.get(rep.keys[i]);
        		score += d2 * rep.values[i]*rep.factor;
        	}
        	rep = rep.next;
        }
        return score;
    }

    private SparseVector cached;
    
    public void prepareDot(SparseVector y1) {
        //SparseVector.prepareDot(y1);
    	for(SparseVector sv = y1; sv != null; sv = sv.next)
    		sv.prepareDot();
    	cached = y1;
    }

    public double dot(SparseVector y2) {
        //return SparseVector.dot(y2);
        double d = 0;
        for(SparseVector sv1 = cached; sv1 != null; sv1 = sv1.next)   	
            for(SparseVector sv2 = y2; sv2 != null; sv2 = sv2.next) {   	
                double f = sv1.factor * sv2.factor;
                d += f * sv1.dot(sv2);
            }
        return d;
        //return cached.dot(y2);
    }

    public double dot(SparseVector y1, SparseVector y2) {
    	//return SparseVector.dot(y1, y2);
    	for(SparseVector sv1 = y1; sv1 != null; sv1 = sv1.next)
    		sv1.prepareDot();
    	double d = 0;
    	for(SparseVector sv1 = y1; sv1 != null; sv1 = sv1.next)
        	for(SparseVector sv2 = y2; sv2 != null; sv2 = sv2.next) {
                    double f = sv1.factor * sv2.factor;
                    d += f * sv1.dot(sv2);
                }
    	return d;
    }

    public double sqrDiff(SparseVector y1, SparseVector y2) {
    	//return SparseVector.sqrDiff(y1, y2);
    	for(SparseVector sv1 = y1; sv1 != null; sv1 = sv1.next)
    		sv1.prepareDot();
    	for(SparseVector sv2 = y2; sv2 != null; sv2 = sv2.next)
    		sv2.prepareDot();
    	double d11 = 0;
    	for(SparseVector sv1 = y1; sv1 != null; sv1 = sv1.next)
        	for(SparseVector sv2 = y1; sv2 != null; sv2 = sv2.next) {
                    double f = sv1.factor * sv2.factor;
                    d11 += f*sv1.dot(sv2);
                }
    	double d12 = 0;
    	for(SparseVector sv1 = y1; sv1 != null; sv1 = sv1.next)
        	for(SparseVector sv2 = y2; sv2 != null; sv2 = sv2.next) {
                    double f = sv1.factor * sv2.factor;
                    d12 += f*sv1.dot(sv2);
                }
    	double d22 = 0;
    	for(SparseVector sv1 = y2; sv1 != null; sv1 = sv1.next)
        	for(SparseVector sv2 = y2; sv2 != null; sv2 = sv2.next) {
                    double f = sv1.factor * sv2.factor;
                    d22 += f*sv1.dot(sv2);
                }
    	return d11 + d22 - 2*d12;
    }

    public double sqrLength(SparseVector y) {
    	//return SparseVector.sqrLength(y);
    	for(SparseVector sv1 = y; sv1 != null; sv1 = sv1.next)
    		sv1.prepareDot();
    	double d11 = 0;
    	for(SparseVector sv1 = y; sv1 != null; sv1 = sv1.next)
        	for(SparseVector sv2 = y; sv2 != null; sv2 = sv2.next) {
                    double f = sv1.factor * sv2.factor;
                    d11 += f*sv1.dot(sv2);
                }
    	return d11;
    }

//    public double sum() {
//        int[] ks = table.keys();
//        double d = 0;
//        for(int i = 0; i < ks.length; i++)
//            d += (ks[i] % 10) * table.get(ks[i]);
//        return d;
//    }
    
}
