/*
   
 */
package se.lth.cs.nlp.nlputils.ml_long;

import se.lth.cs.nlp.nlputils.core.*;
import java.io.Serializable;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class LinearModelCorrupted extends Model implements Serializable {

    private static final long serialVersionUID = 0L;
    
    private ClosedLongDoubleMap table;
    private int capacity;

    private double wsum, wsqsum;
	private FeatureFilter hidden = null;

	public double meanWeight() {
		int n = table.size();
		if(n < 1)
			return 0;
		return wsum / n;
	}

	public double stddevWeight() {
		// return 1.0/(len(l)-1)*(sum([x*x for x in l]) - 1.0/(len(l))*sum(l)*sum(l))
		int n = table.size();
		if(n < 2)
			return 0;
		return Math.sqrt((wsqsum - wsum/n) / (n-1));
	}

	public void hide(FeatureFilter hidden) {
		this.hidden = hidden;
	}
    
    public LinearModelCorrupted(int size) {
        this.capacity = size;
        table = new ClosedLongDoubleMap(size);
    }
    
    public void update(double d, SparseVector rep) {
        while(rep != null) {
            for(int i = 0; i < rep.index; i++) {

		    //double d2 = d*rep.values[i]*rep.factor;
		    //double oldval = table.add(rep.keys[i], d2);
		    //wsum += d2;
		    //wsqsum -= oldval*oldval;
		    //wsqsum += d2*d2;

		    double d2 = d*rep.values[i]*rep.factor;
		    double oldval = table.add(rep.keys[i], d2);

		    wsum -= Math.abs(oldval);
		    wsum += Math.abs(oldval + d2);
		    wsqsum -= oldval*oldval;
		    wsqsum += (oldval+d2)*(oldval+d2);		    
            }
            rep = rep.next;
        }
    }

	public void update(long index, double delta) {
		//table.add(index, delta);
		if(hidden == null || !hidden.hides(index)) {
			double oldval = table.add(index, delta);

			wsum -= Math.abs(oldval);
			wsum += Math.abs(oldval + delta);
			wsqsum -= oldval*oldval;
			wsqsum += (oldval+delta)*(oldval+delta);
		}
	}
    
    public void set(long index, double value) {
	    double oldval = table.put(index, value);
	    wsum -= Math.abs(oldval);
	    wsum += Math.abs(value);
	    wsqsum -= oldval*oldval;
	    wsqsum += value*value;
    }
    
    private static final double TOLERANCE = 1e-15;
    
    public void updateDiff(double d, SparseVector r1, SparseVector r2, boolean sorted) {
        if(Math.abs(d) < TOLERANCE)
            return;
        if(!sorted) {
            r1.sortIndices();
            r2.sortIndices();
        }
        
        if(r1.next != null || r2.next != null)
            throw new RuntimeException("Stacked vectors unimplemented...");
        int ix1 = 0, ix2 = 0;
        while(ix1 < r1.index || ix2 < r2.index) {
            long k1 = ix1 == r1.index? 0: r1.keys[ix1];
            long k2 = ix2 == r2.index? 0: r2.keys[ix2];
            
            if(k2 == 0 || (k1 != 0 && k1 < k2)) {
                double dd = d*r1.values[ix1]*r1.factor;
                if(dd > TOLERANCE || dd < -TOLERANCE)
			//table.add(k1, dd);
			update(k1, dd);
                ix1++;
            } else if(k1 == 0 || (k2 != 0 && k2 < k1)) {
                double dd = -d*r2.values[ix2]*r2.factor;
                if(dd > TOLERANCE || dd < -TOLERANCE)
			//table.add(k2, dd);
			update(k2, dd);
                ix2++;                
            } else if(k1 == k2) {
                double dd = d*(r1.values[ix1]*r1.factor - r2.values[ix2]*r2.factor);
                if(dd > TOLERANCE || dd < -TOLERANCE)
			//table.add(k1, dd);
			update(k1, dd);
                ix1++;
                ix2++;
            } else {
                throw new RuntimeException("bug, shouldn't be here...");
            }            
        }
        
    }
    
    public void clear() {
        table = new ClosedLongDoubleMap(capacity);
	wsum = 0;
	wsqsum = 0;
    }
        
    public int size() {
        return table.size();
    }
    
    public double score(SparseVector rep) {
        double score = 0;
        while(rep != null) {
        	for(int i = 0; i < rep.index; i++) {
			if(hidden == null || !hidden.hides(rep.keys[i])) {
				double d2 = table.get(rep.keys[i]);
				score += d2 * rep.values[i]*rep.factor;
			}
        	}
        	rep = rep.next;
        }
        return score;
    }

    public double getWeight(long l) {
	    return table.get(l);
    }

    private SparseVector cached;
    
    public void prepareDot(SparseVector y1) {
        //SparseVector.prepareDot(y1);
    	for(SparseVector sv = y1; sv != null; sv = sv.next)
    		sv.prepareDot();
    	cached = y1;
    }

    public double dot(SparseVector y2) {
        //return SparseVector.dot(y2);
        double d = 0;
        for(SparseVector sv1 = cached; sv1 != null; sv1 = sv1.next)   	
            for(SparseVector sv2 = y2; sv2 != null; sv2 = sv2.next) {   	
                double f = sv1.factor * sv2.factor;
                d += f * sv1.dot(sv2);
            }
        return d;
        //return cached.dot(y2);
    }

    public double dot(SparseVector y1, SparseVector y2) {
    	//return SparseVector.dot(y1, y2);
    	for(SparseVector sv1 = y1; sv1 != null; sv1 = sv1.next)
    		sv1.prepareDot();
    	double d = 0;
    	for(SparseVector sv1 = y1; sv1 != null; sv1 = sv1.next)
        	for(SparseVector sv2 = y2; sv2 != null; sv2 = sv2.next) {
                    double f = sv1.factor * sv2.factor;
                    d += f * sv1.dot(sv2);
                }
    	return d;
    }

    public double sqrDiff(SparseVector y1, SparseVector y2) {
    	//return SparseVector.sqrDiff(y1, y2);
    	for(SparseVector sv1 = y1; sv1 != null; sv1 = sv1.next)
    		sv1.prepareDot();
    	for(SparseVector sv2 = y2; sv2 != null; sv2 = sv2.next)
    		sv2.prepareDot();
    	double d11 = 0;
    	for(SparseVector sv1 = y1; sv1 != null; sv1 = sv1.next)
        	for(SparseVector sv2 = y1; sv2 != null; sv2 = sv2.next) {
                    double f = sv1.factor * sv2.factor;
                    d11 += f*sv1.dot(sv2);
                }
    	double d12 = 0;
    	for(SparseVector sv1 = y1; sv1 != null; sv1 = sv1.next)
        	for(SparseVector sv2 = y2; sv2 != null; sv2 = sv2.next) {
                    double f = sv1.factor * sv2.factor;
                    d12 += f*sv1.dot(sv2);
                }
    	double d22 = 0;
    	for(SparseVector sv1 = y2; sv1 != null; sv1 = sv1.next)
        	for(SparseVector sv2 = y2; sv2 != null; sv2 = sv2.next) {
                    double f = sv1.factor * sv2.factor;
                    d22 += f*sv1.dot(sv2);
                }
    	return d11 + d22 - 2*d12;
    }

    public double sqrLength(SparseVector y) {
    	//return SparseVector.sqrLength(y);
    	for(SparseVector sv1 = y; sv1 != null; sv1 = sv1.next)
    		sv1.prepareDot();
    	double d11 = 0;
    	for(SparseVector sv1 = y; sv1 != null; sv1 = sv1.next)
        	for(SparseVector sv2 = y; sv2 != null; sv2 = sv2.next) {
                    double f = sv1.factor * sv2.factor;
                    d11 += f*sv1.dot(sv2);
                }
    	return d11;
    }

//    public double sum() {
//        int[] ks = table.keys();
//        double d = 0;
//        for(int i = 0; i < ks.length; i++)
//            d += (ks[i] % 10) * table.get(ks[i]);
//        return d;
//    }
    
}
