/*
 LinearVectorModel.java
  
 Created on Sep 26, 2006 by Richard Johansson (richard@cs.lth.se).

 $Log: LinearVectorModel.java,v $
 Revision 1.5  2010-07-05 14:04:22  johansson
 Added set.

 Revision 1.4  2010-06-07 11:29:59  johansson
 Added.

 Revision 1.3  2009/04/16 11:59:36  johansson
 Added update on index.

 Revision 1.2  2009/03/06 13:22:50  johansson
 Added modelSqrNorm and getMostProminentFeature.

 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:04  richard
 Branched.

 Revision 1.5  2008/04/10 08:27:54  richard
 Added getDensity.

 Revision 1.4  2008/04/08 11:45:24  richard
 Uses BigDoubleArray.

 Revision 1.3  2007/02/26 08:05:47  richard
 Added factor.

 Revision 1.2  2006/10/25 07:07:39  richard
 SparseVector as a sequence.

 Revision 1.1  2006/09/26 15:07:16  richard
 Added the file.

   
 */
package se.lth.cs.nlp.nlputils.ml_long;

import java.io.Serializable;

import se.lth.cs.nlp.nlputils.core.BigDoubleArray;

import se.lth.cs.nlp.nlputils.core.DoubleIntPair;
import java.util.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class LinearVectorModel extends Model implements Serializable {

    private static final long serialVersionUID = 1L;

    private BigDoubleArray values;

    public LinearVectorModel(int size) {
        values = new BigDoubleArray(size);
    }
    
    public void clear() {
        values.fill(0.0);
    }
    
    public double getDensity() {
        int l = values.length();
        int max = 0;
        int count = 0;
        for(int i = 0; i < l; i++)
            if(values.get(i) != 0) {
                max = i;
                count++;
            }
        if(max == 0)
            return 0;
        else
            return (double) count / max;
    }
    
    public void update(double d, SparseVector rep) {
        int len = values.length();
        while(rep != null) {
            for(int i = 0; i < rep.index; i++)
                if(rep.keys[i] >= len)
                    throw new RuntimeException("Feature index too large");
                else
                    values.modify((int) rep.keys[i], d*rep.values[i]);
                    //values[rep.keys[i]] += d*rep.values[i];
            rep = rep.next;
        }
    }

    public void set(long index, double value) {
        int len = values.length();
        if(index > len)
            throw new RuntimeException("Feature index too large");
        values.set((int) index, value);
    }
    
    public void update(long index, double delta) {
        int len = values.length();
        if(index > len)
            throw new RuntimeException("Feature index too large");
        values.modify((int) index, delta);
    }
    
    public void updateDiff(double d, SparseVector r1, SparseVector r2, boolean sorted) {
        if(!sorted) {
            r1.sortIndices();
            r2.sortIndices();
        }
        if(r1.next != null || r2.next != null)
            throw new RuntimeException("Stacked vectors unimplemented...");
        int ix1 = 0, ix2 = 0;
        while(ix1 < r1.index || ix2 < r2.index) {
            long k1 = ix1 >= r1.index? Long.MAX_VALUE: r1.keys[ix1];
            long k2 = ix2 >= r2.index? Long.MAX_VALUE: r2.keys[ix2]; // hmmm

            if(k1 == 0)
                throw new RuntimeException("k1 = 0");
            if(k2 == 0)
                throw new RuntimeException("k2 = 0");
            
            if(k1 == Long.MAX_VALUE && k2 == Long.MAX_VALUE)
                throw new RuntimeException("???");
            
            if(k2 == Long.MAX_VALUE || k1 < k2) {
                double dd = d*r1.values[ix1]*r1.factor;
                values.modify((int) k1, dd);
                ix1++;
            } else if(k1 == Long.MAX_VALUE || k2 < k1) {
                double dd = -d*r2.values[ix2]*r2.factor;
                values.modify((int) k2, dd);
                ix2++;
            } else if(k1 == k2) {
                double dd = d*(r1.values[ix1]*r1.factor - r2.values[ix2]*r2.factor);
                values.modify((int) k1, dd);
                ix1++;
                ix2++;
            } else {
                throw new RuntimeException("bug, shouldn't be here...");
            }            
        }
    }
    
    public double score(SparseVector rep) {
        double score = 0;
        int len = values.length();
        while(rep != null) {
        	for(int i = 0; i < rep.index; i++)
        		if(rep.keys[i] < len)
        		    score += values.get((int) rep.keys[i]) * rep.values[i];
        	rep = rep.next;
        }
        return score;
    }

    private SparseVector cached;
    
    public void prepareDot(SparseVector y1) {
        //SparseVector.prepareDot(y1);
        for(SparseVector sv = y1; sv != null; sv = sv.next)
                sv.prepareDot();
        cached = y1;
    }

    public double dot(SparseVector y2) {
        //return SparseVector.dot(y2);
        double d = 0;
        for(SparseVector sv1 = cached; sv1 != null; sv1 = sv1.next)     
            for(SparseVector sv2 = y2; sv2 != null; sv2 = sv2.next) {           
                double f = sv1.factor * sv2.factor;
                d += f * sv1.dot(sv2);
            }
        return d;
        //return cached.dot(y2);
    }

    public double dot(SparseVector y1, SparseVector y2) {
        //return SparseVector.dot(y1, y2);
        for(SparseVector sv1 = y1; sv1 != null; sv1 = sv1.next)
                sv1.prepareDot();
        double d = 0;
        for(SparseVector sv1 = y1; sv1 != null; sv1 = sv1.next)
                for(SparseVector sv2 = y2; sv2 != null; sv2 = sv2.next) {
                    double f = sv1.factor * sv2.factor;
                    d += f * sv1.dot(sv2);
                }
        return d;
    }

    public double sqrDiff(SparseVector y1, SparseVector y2) {
        //return SparseVector.sqrDiff(y1, y2);
        for(SparseVector sv1 = y1; sv1 != null; sv1 = sv1.next)
                sv1.prepareDot();
        for(SparseVector sv2 = y2; sv2 != null; sv2 = sv2.next)
                sv2.prepareDot();
        double d11 = 0;
        for(SparseVector sv1 = y1; sv1 != null; sv1 = sv1.next)
                for(SparseVector sv2 = y1; sv2 != null; sv2 = sv2.next) {
                    double f = sv1.factor * sv2.factor;
                    d11 += f*sv1.dot(sv2);
                }
        double d12 = 0;
        for(SparseVector sv1 = y1; sv1 != null; sv1 = sv1.next)
                for(SparseVector sv2 = y2; sv2 != null; sv2 = sv2.next) {
                    double f = sv1.factor * sv2.factor;
                    d12 += f*sv1.dot(sv2);
                }
        double d22 = 0;
        for(SparseVector sv1 = y2; sv1 != null; sv1 = sv1.next)
                for(SparseVector sv2 = y2; sv2 != null; sv2 = sv2.next) {
                    double f = sv1.factor * sv2.factor;
                    d22 += f*sv1.dot(sv2);
                }
        return d11 + d22 - 2*d12;
    }

    public double sqrLength(SparseVector y) {
        //return SparseVector.sqrLength(y);
        for(SparseVector sv1 = y; sv1 != null; sv1 = sv1.next)
                sv1.prepareDot();
        double d11 = 0;
        for(SparseVector sv1 = y; sv1 != null; sv1 = sv1.next)
                for(SparseVector sv2 = y; sv2 != null; sv2 = sv2.next) {
                    double f = sv1.factor * sv2.factor;
                    d11 += f*sv1.dot(sv2);
                }
        return d11;
    }

    public double modelSqrNorm() {
        double out = 0;
        for(int i = values.length() - 1; i >= 0; i--) {
            double d = values.get(i);
            out += d*d;
        }
        return out;
    }
    
    public int getMostProminentFeature() {
        int maxIndex = -1;
        double max = Double.NEGATIVE_INFINITY;
        for(int i = values.length() - 1; i >= 0; i--) {
            double d = Math.abs(values.get(i));
            if(d > max) { 
                max = d;
                maxIndex = i;
            }
        }
        return maxIndex;
    }
    
    public int size() {
    	return values.length();
    }
    
    public LinkedList<DoubleIntPair> getMostProminentFeatures(int N) {
    	LinkedList<DoubleIntPair> l = new LinkedList();
    	for(int i = values.length() - 1; i >= 0; i--) {
            double d = values.get(i);
            double ad = Math.abs(d);

            boolean found = false;
            for(ListIterator<DoubleIntPair> it = l.listIterator(l.size()); it.hasPrevious(); ) {
            	DoubleIntPair prev = it.previous();
            	if(ad < Math.abs(prev.left)) {
            		it.next();
            		it.add(new DoubleIntPair(d, i));
            		found = true;
            		break;
            	}
            }
            if(!found) {
            	l.add(0, new DoubleIntPair(d, i));
            }
            
            if(l.size() > N)
            	l.removeLast();
            
        }    	
    	return l;
    }
    
    
}
