/*
 PairingFeatureListEncoder.java
  
 Created on Apr 1, 2008 by Richard Johansson (richard@cs.lth.se).

 $Log: PairingFeatureListEncoder.java,v $
 Revision 1.3  2010-06-07 11:29:59  johansson
 Added.

 Revision 1.2  2009/03/06 13:24:11  johansson
 Added decode and parse.

 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:04  richard
 Branched.

 Revision 1.5  2008/04/23 09:04:34  richard
 Removed printout.

 Revision 1.4  2008/04/10 07:08:20  richard
 Serializable.

 Revision 1.3  2008/04/03 08:37:52  richard
 Checks.

 Revision 1.2  2008/04/02 13:27:37  richard
 Bug in encoding pairs.

 Revision 1.1  2008/04/02 11:26:04  richard
 Added.


 */
package se.lth.cs.nlp.nlputils.ml_long;

import gnu.trove.TIntIntHashMap;

import java.util.*;
import java.io.Serializable;
import se.lth.cs.nlp.nlputils.core.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class PairingFeatureListEncoder extends FeatureListEncoder 
    implements Serializable {

    private static final long serialVersionUID = 1L;
    
    private SymbolEncoder symbolEncoder = new SymbolEncoder();
    private IntPairEncoder pairEncoder = new IntPairEncoder(1);

    private int[][] fpairs;    
    private int nbits;
    private int maxFeatureValue;

    public void remap(TIntIntHashMap m) {
    	pairEncoder.remap(m);
    }
    
    public void freeze() {
        symbolEncoder.freeze();
        pairEncoder.freeze();
    }
    
    private PairingFeatureListEncoder() {    	
    }
    
    public static HashSet<Pair<String, String>> parsePairString(String arg) {

        arg = arg.trim();

        if(!arg.startsWith("{") || !arg.endsWith("}"))
            throw new IllegalArgumentException("Illegal set format");

        arg = arg.substring(1, arg.length() - 1).trim();

        String[] ss = arg.split(",\\s*");

        HashSet<Pair<String, String>> ps = new HashSet();

        for(int i = 0; i < ss.length; i++) {        
            String s = ss[i];
            if(s.startsWith("(")) {
                if(i == ss.length - 1)
                    throw new IllegalArgumentException("Illegal pair format " + s);
                String s2 = ss[++i];
                if(!s2.endsWith(")"))
                    throw new IllegalArgumentException("Illegal pair format " + s2);
                s = s.substring(1);
                s2 = s2.substring(0, s2.length() - 1).trim();
                if(s.equals("") || s2.equals(""))
                    throw new IllegalArgumentException("Illegal pair format (empty)");
                ps.add(new Pair(s, s2));
            } else
                throw new IllegalArgumentException("Must consist of pairs");
        }

        return ps;
    }
    
    public PairingFeatureListEncoder(FeatureDeclaration decl,
            Collection<Pair<String, String>> pairings) {
        nbits = 32 - Util.msb(decl.size());
        maxFeatureValue = (1 << nbits) - 1; 
        ArrayList<IntPair> ps = new ArrayList();
        for(Pair<String, String> p: pairings) {
            int f1 = decl.getFeatureIndex(p.left);
            if(f1 == MLConstants.EXCLUDED_MAGIC_NUMBER)
                throw new IllegalArgumentException("Illegal feature " + p.left);
            int f2 = decl.getFeatureIndex(p.right);
            if(f2 == MLConstants.EXCLUDED_MAGIC_NUMBER)
                throw new IllegalArgumentException("Illegal feature " + p.right);
            IntPair ip = new IntPair(f1, f2);
            ip.sort();
            ps.add(ip);
        }
        Collections.sort(ps, IntPair.BY_LEFT);
        fpairs = new int[ps.size()][];
        int ix = 0;
        for(IntPair ip: ps)
            fpairs[ix++] = ip.toArray();
        //System.out.println(Arrays.deepToString(fpairs));
    }

    public void encode(FeatureList list, SparseVector sv) {
        sv.clear();
        for(int i = 0; i < list.values.length; i++) {
            Object v = list.values[i];
            if(v instanceof String)
                sv.put(encodeSimple(i, list.values[i]), 1.0);
            else if(v instanceof Double)
                sv.put(encodeSimple(i), (Double) v);
            else if(v instanceof double[]) {
                double[] ds = (double[]) v;
                for(int j = 0; j < ds.length; j++)
                    sv.put(encodeSimple(i, j+1), ds[j]);
            } else 
                for(String s: (HashSet<String>) list.values[i])
                    sv.put(encodeSimple(i, s), 1.0);
        }

        //System.out.println(sv);
        /*for(int i = 0; i < fpairs.length; i++) {
            int fi = fpairs[i][0];
            int fj = fpairs[i][1];
            if(list.values[fi] instanceof String)
                if(list.values[fj] instanceof String)
                    sv.put(encodePair(fi, list.values[fi], fj, list.values[fj]), 1.0);
                else 
                    for(String sj: (HashSet<String>) list.values[fj])
                        sv.put(encodePair(fi, list.values[fi], fj, sj), 1.0);
            else
                if(list.values[fj] instanceof String)
                    for(String si: (HashSet<String>) list.values[fi])
                        sv.put(encodePair(fi, si, fj, list.values[fj]), 1.0);
                else
                    for(String si: (HashSet<String>) list.values[fi])
                        for(String sj: (HashSet<String>) list.values[fj])
                            sv.put(encodePair(fi, si, fj, sj), 1.0);
        }*/
        
        for(int i = 0; i < fpairs.length; i++) {
            int fi = fpairs[i][0];
            int fj = fpairs[i][1];

            // TODO double value, double[] value not implemented
            
            if(list.values[fi] instanceof String)
                if(list.values[fj] instanceof String)
                    sv.put(encodePair(fi, list.values[fi], fj, list.values[fj]), 1.0);
                else
                    for(String sj: (HashSet<String>) list.values[fj])
                        sv.put(encodePair(fi, list.values[fi], fj, sj), 1.0);
            else
                if(list.values[fj] instanceof String)
                    for(String si: (HashSet<String>) list.values[fi])
                        sv.put(encodePair(fi, si, fj, list.values[fj]), 1.0);
                else
                    for(String si: (HashSet<String>) list.values[fi])
                        for(String sj: (HashSet<String>) list.values[fj])
                            sv.put(encodePair(fi, si, fj, sj), 1.0);
        }
        
        
        
        //System.out.println(sv);
        //System.exit(0);
    }

    private int encodeSimple(int i, Object s) {
        int se = symbolEncoder.encode((String) s);
        if(se > maxFeatureValue)
            throw new RuntimeException("Feature encoded value too large!");
        int featureValue = (i << nbits) | se;
        return pairEncoder.encode(0, featureValue);
    }
    
    private int encodeSimple(int i, int f) {
        if(f > maxFeatureValue)
            throw new RuntimeException("Feature encoded value too large!");
        int featureValue = (i << nbits) | f;
        return pairEncoder.encode(0, featureValue);
    }
    
    private int encodeSimple(int i) {
        int featureValue = (i << nbits) | 1;
        return pairEncoder.encode(0, featureValue);
    }
    
    private int encodePair(int i1, Object s1, int i2, Object s2) {
        int se1 = symbolEncoder.encode((String) s1);
        if(se1 > maxFeatureValue)
            throw new RuntimeException("Feature encoded value too large!");
        int se2 = symbolEncoder.encode((String) s2);
        if(se2 > maxFeatureValue)
            throw new RuntimeException("Feature encoded value too large!");
        int featureValue1 = (i1 << nbits) | se1;
        int featureValue2 = (i2 << nbits) | se2;
        return pairEncoder.encode(featureValue1, featureValue2);
    }

    
    
    public Object decode(int encoded) {
        IntPair p = pairEncoder.decode(encoded);
        if(p.left == 0) {
            int feature = p.right >>> nbits;
            int encodedValue = p.right & ((1 << nbits) - 1);
            String value = symbolEncoder.decode(encodedValue);
            return new IntObjPair(feature, value);
        } else {
            int feature1 = p.left >>> nbits;
            int encodedValue1 = p.left & ((1 << nbits) - 1);
            String value1 = symbolEncoder.decode(encodedValue1);            
            int feature2 = p.right >>> nbits;
            int encodedValue2 = p.right & ((1 << nbits) - 1);
            String value2 = symbolEncoder.decode(encodedValue2);
            return new Pair(new IntObjPair(feature1, value1), new IntObjPair(feature2, value2));
        }
    }
    
    public PairingFeatureListEncoder clone() {
    	PairingFeatureListEncoder out = new PairingFeatureListEncoder();
    	out.fpairs = fpairs.clone();
    	out.nbits = nbits;
    	out.maxFeatureValue = maxFeatureValue;
    	out.symbolEncoder = symbolEncoder.clone();
    	out.pairEncoder = pairEncoder.clone();
    	return out;
    }
    
    public static void main(String[] argv) {
        FeatureDeclaration decl = new FeatureDeclaration(new String[] {"F0", "F1"});
        FeatureList fl = new FeatureList(decl);
        SparseVector sv = new SparseVector();

        HashSet<Pair<String, String>> ps = new HashSet();
        ps.add(new Pair("F0", "F1"));
        
        FeatureListEncoder enc = new PairingFeatureListEncoder(decl, ps);
        
        fl.setFeature("F0", "off");
        fl.setFeature("F1", "off");
        enc.encode(fl, sv);
        System.out.println(sv);
        
        fl.setFeature("F0", "on");
        fl.setFeature("F1", "off");        
        enc.encode(fl, sv);
        System.out.println(sv);
        
        fl.setFeature("F0", "off");
        fl.setFeature("F1", "on");
        enc.encode(fl, sv);
        System.out.println(sv);
        
        fl.setFeature("F0", "on");
        fl.setFeature("F1", "on");
        enc.encode(fl, sv);
        System.out.println(sv);

        
    }

}
