/*
 RestrictedEncodedMulticlassDefinition.java

 $Log: RestrictedEncodedMulticlassDefinition.java,v $
 Revision 1.2  2011-02-10 11:50:38  johansson
 Updated.

 Revision 1.1  2010-06-07 11:29:59  johansson
 Added.


 */
package se.lth.cs.nlp.nlputils.ml_long;

import gnu.trove.TIntArrayList;

import java.io.FileOutputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import se.lth.cs.nlp.nlputils.core.*;

import java.util.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class RestrictedEncodedMulticlassDefinition extends MulticlassDefinition<RestrictedInstance<SparseVector, Integer>, Integer> implements Serializable {    

    private static final long serialVersionUID = 1L;

    public RestrictedEncodedMulticlassDefinition() {
        this(true);
    }
    
    public RestrictedEncodedMulticlassDefinition(boolean compress) {
        super();
        enc = new SparseVectorClassEncoder(compress);
    }

    private SparseVectorClassEncoder enc;
    
    public double cost(RestrictedInstance<SparseVector, Integer> x, Integer y, Integer ybar) {
        return y.equals(ybar)? 0: 1;
    }

    public int encodeClass(Integer y) {
        return y;
    }

    public Integer decodeClass(int cl) {
        return cl;
    }

    public void encodeInstance(RestrictedInstance<SparseVector, Integer> x, SparseVector sv) {
    	SparseVector xsv = x.instance;
    	if(sv != xsv) {
            if(xsv.keys.length <= sv.keys.length) {
                sv.index = xsv.index;
                System.arraycopy(xsv.keys, 0, sv.keys, 0, sv.index);
                System.arraycopy(xsv.values, 0, sv.values, 0, sv.index);
            } else {
                sv.clear();
                sv.putAll(xsv);
            }
        }
    }

    public void encode(RestrictedInstance<SparseVector, Integer> x, Integer y, SparseVector sv) {
        if(enc == null)
            throw new RuntimeException("Must specify classes!");
        sv.clear();
        enc.encode(x.instance, y, sv);
    }

    public Integer argmax(RestrictedInstance<SparseVector, Integer> x, Model model, SparseVector sv, double[] result) {
    	double max = Double.NEGATIVE_INFINITY;
    	int maxc = Integer.MIN_VALUE;
    	for(int i = 0; i < x.range.length; i++) {
    		sv.clear();
    		enc.encode(x.instance, x.range[i], sv);
    		double score = model.score(sv);
    		if(score > max) {
    			max = score;
    			maxc = x.range[i];
    		}
    	}
    	if(maxc == Integer.MIN_VALUE) {
    		result[0] = Double.NEGATIVE_INFINITY;
    		return x.range[0];
    	}
        result[0] = max;
        return maxc;
    }

    private static final class Item implements Comparable<Item> {
    	int id;
    	double score;
    	SparseVector rep;
    	public int compareTo(Item i) {
    		return Double.compare(i.score, score);
    	}
    	Item(int id, double score, SparseVector rep) {
    		this.id = id;
    		this.score = score;
    		this.rep = rep;
    	}    	
    }
    
    public int findNBest(RestrictedInstance<SparseVector, Integer> x, Model model, int n, Integer[] ybars, SparseVector[] svs, double[] scores) {
    	ArrayList<Item> items = new ArrayList();
    	for(int i = 0; i < x.range.length; i++) {
    		SparseVector sv = new SparseVector();
            enc.encode(x.instance, x.range[i], sv);
            double score = model.score(sv);
            items.add(new Item(x.range[i], score, sv));
    	}
    	PriorityQueue<Item> q = new PriorityQueue(items);
    	int nout = Math.min(n, q.size());
    	for(int i = 0; i < nout; i++) {
    		Item item = q.poll();
    		ybars[i] = item.id;
    		svs[i] = item.rep;
    		scores[i] = item.score;
    	}
    	return nout;
    }    
    
    public Integer maxLoss(RestrictedInstance<SparseVector, Integer> x, Integer y, Model model, SparseVector sv, double[] result) {
    	double max = Double.NEGATIVE_INFINITY;
        int maxc;        
        if(x.range.length > 1) {
        	maxc = Integer.MIN_VALUE;
        	for(int i = 0; i < x.range.length; i++) {
        		sv.clear();
        		enc.encode(x.instance, x.range[i], sv);
        		double score = model.score(sv) + Math.sqrt(cost(x, y, x.range[i]));
        		if(score > max) {
        			max = score;
        			maxc = x.range[i];
        		}
        	}
        } else
        	maxc = x.range[0];
        sv.clear();
        enc.encode(x.instance, maxc, sv);        
        result[0] = model.score(sv);
        result[1] = cost(x, y, maxc);
        return maxc;
    }

    public Decomposition decompose(SparseVector x, Integer y) {
        throw new RuntimeException("Unimplemented");
    }
    
    
    
    public static void main(String[] argv) {
        try {

            ArrayList<Pair<FeatureList, String>> l = MulticlassDefinition.readDataSet(argv[0], null);
            FeatureDeclaration decl = l.get(0).left.decl;

            String restrictionFeature = argv[2];
            
            ArrayList<Pair<RestrictedInstance<FeatureList, String>, String>> lR 
            = RestrictedMulticlassDefinition.restrictSet(l, restrictionFeature);
            l = null;
            
            ArrayList<Pair<String, String>> pairings = new ArrayList();

            /*
            pairings.add(new Pair("N_PATH", "N_HEAD_POS"));
            pairings.add(new Pair("P_TARGET_LEMMA", "N_HEAD"));
            pairings.add(new Pair("N_PATH", "P_HAS_SUBJ_IN_VC"));
            pairings.add(new Pair("N_PATH", "N_POSITION"));            
            pairings.add(new Pair("N_PATH", "P_OPRD_PARENT_HAS_OBJ"));
            pairings.add(new Pair("N_PATH", "P_TARGET_LEMMA"));
            pairings.add(new Pair("N_RIGHT_POS", "N_HEAD"));
            pairings.add(new Pair("N_PATH", "P_PARENT_WORD"));
            pairings.add(new Pair("N_PATH", "N_HEAD"));
            pairings.add(new Pair("N_PATH", "N_RIGHT_POS"));
            pairings.add(new Pair("N_PATH", "P_PRED_POS"));
            pairings.add(new Pair("P_REL_TO_PARENT", "P_VOICE"));
			*/

            //FeatureListEncoder lenc = new SimpleFeatureListEncoder();
            FeatureListEncoder lenc = new PairingFeatureListEncoder(decl, pairings);
            SymbolEncoder clEnc = new SymbolEncoder();
                        
            ArrayList<Pair<RestrictedInstance<SparseVector, Integer>, Integer>> tsEnc 
            = encodeSet(lR, lenc, clEnc);
            lR = null;
            clEnc.freeze();
            lenc.freeze();
            
            String methodString = argv[3];
            String argString = argv[4];            
            
            LearningAlgorithm method = new AlgorithmFactory().create(methodString, argString);
            RestrictedEncodedMulticlassDefinition def = new RestrictedEncodedMulticlassDefinition();

            long t0 = System.currentTimeMillis();
            Classifier<RestrictedInstance<SparseVector, Integer>, Integer> cl = method.train(def, tsEnc);
            System.out.println("Training time: " + (System.currentTimeMillis() - t0));

            tsEnc = null;

            //sc = new Scanner(new FileInputStream(argv[1]));
            int n = 0, ncorrect = 0;            
            int nAmb = 0, ncorrectAmb = 0;

            long cltime = 0;

            HashMap<Object, int[]> perClassResults = new HashMap();

            ArrayList<Pair<FeatureList, String>> l2 = argv[1].equals("NULL")? null: readDataSet(argv[1], decl);

            if(l2 != null) {
            	System.out.println("Loaded test set.");

                ArrayList<Pair<RestrictedInstance<FeatureList, String>, String>> l2R 
                = RestrictedMulticlassDefinition.restrictSet(l2, restrictionFeature);
                l2 = null;
                
            	ArrayList<String> clDec = clEnc.inverse();            	
            	
            	// TODO något i range kan finnas enbart i testset
            	
            	for(Pair<RestrictedInstance<FeatureList, String>, String> p: l2R) {
            		n++;
            		if(n % 1000 == 0) {
            			System.out.print(".");
            			System.out.flush();
            		}

            		t0 = System.currentTimeMillis();

            		RestrictedInstance<SparseVector, Integer> ri = 
            			encodeRestrictedInstance(p.left, lenc, clEnc);

            		Integer aenc = cl.classify(ri);
            		cltime += System.currentTimeMillis() - t0;

            		String answer = clDec.get(aenc);            		
            		if(answer == null) {
            			int nZeros = CollectionUtils.count(ri.range, 0);
            			if(nZeros > 1)
            				System.out.println("\nWarning: multiple zeros");
            			answer = p.left.range[CollectionUtils.indexOf(ri.range, 0)];
            		}
            		
            		if(answer.equals(p.right))
            			ncorrect++;                

            		if(ri.range.length > 1) {
            			nAmb++;
                		if(answer.equals(p.right))
                			ncorrectAmb++;
            		}
            		
            		int[] c1 = null;
            		if(!answer.equals(MulticlassDefinition.NULL)) {
            			c1 = perClassResults.get(answer);
            			if(c1 == null) {
            				c1 = new int[3];
            				perClassResults.put(answer, c1);
            			}
            			c1[0]++;
            			if(answer.equals(p.right))
            				c1[2]++;
            		}
            		if(!p.right.equals(NULL)) {
            			int[] c2 = perClassResults.get(p.right);
            			if(c2 == null) {
            				c2 = new int[3];
            				perClassResults.put(p.right, c2);
            			}
            			c2[1]++;
            		}

            	}

            	double acc = (double) ncorrect / n;
            	double accAmb = (double) ncorrectAmb / nAmb;

            	System.out.println();           
            	System.out.println("Accuracy (all): " + ncorrect + "/" + n + " = " + acc);
            	System.out.println("Accuracy (amb): " + ncorrectAmb + "/" + nAmb + " = " + accAmb);
            	
            	System.out.println("Classification time: " + cltime);

            	if(false)
            		for(Object c: perClassResults.keySet()) {
            			int[] res = perClassResults.get(c);
            			double pc = (double) res[2] / res[0];
            			double rc = (double) res[2] / res[1];
            			double fc = 2*pc*rc/(pc + rc);
            			pc = Math.round(10000.0*pc)/10000.0;
            			rc = Math.round(10000.0*rc)/10000.0;
            			fc = Math.round(10000.0*fc)/10000.0;
            			System.out.println(c + " (" + res[1] + "): p = " + pc + ", r = " + rc + ", f1 = " + fc);
            		}         
            }
            
            if(argv.length > 5) {
            	System.out.println("Writing model to " + argv[5]);
            	ObjectOutputStream oos = new ObjectOutputStream(new FileOutputStream(argv[5]));
            	oos.writeObject(cl);
            	oos.writeObject(clEnc);
            	oos.writeObject(lenc);
            	oos.close();
            }
            
        } catch(Exception e) {
            e.printStackTrace();
        }
    }

	public static ArrayList<Pair<RestrictedInstance<SparseVector, Integer>, Integer>> encodeSet(
			ArrayList<Pair<RestrictedInstance<FeatureList, String>, String>> lR,
			FeatureListEncoder lenc, SymbolEncoder clEnc) {
		ArrayList<Pair<RestrictedInstance<SparseVector, Integer>, Integer>> tsEnc 
		= new ArrayList();		
        for(Pair<RestrictedInstance<FeatureList, String>, String> p: lR) {
        	RestrictedInstance<SparseVector, Integer> ri 
        	= encodeRestrictedInstance(p.left, lenc, clEnc);
        	int ce = clEnc.encode(p.right);        		
        	tsEnc.add(new Pair(ri, ce));        		
        }
        return tsEnc;
	}
    
	public static RestrictedInstance<SparseVector, Integer> encodeRestrictedInstance(RestrictedInstance<FeatureList, String> inst,
			FeatureListEncoder lenc, SymbolEncoder clEnc) {
		SparseVector sv = new SparseVector(); 	
    	lenc.encode(inst.instance, sv);
    	sv.sortIndices();
    	sv.trim();
    	RestrictedInstance<SparseVector, Integer> ri = new RestrictedInstance();
    	ri.instance = sv;    	
    	ri.range = new Integer[inst.range.length];
    	for(int i = 0; i < ri.range.length; i++)
    		ri.range[i] = clEnc.encode(inst.range[i]);
    	Arrays.sort(ri.range);
		return ri;
	}
	
}