/*
 RestrictedMulticlassDefinition.java

 $Log: RestrictedMulticlassDefinition.java,v $
 Revision 1.1  2010-06-07 11:29:59  johansson
 Added.


 */
package se.lth.cs.nlp.nlputils.ml_long;

import se.lth.cs.nlp.nlputils.core.*;

import java.util.*;
import java.io.*;

import gnu.trove.*;

/**
 * Cost-insensitive multiclass classification with a finite (practical) number
 * of labels.
 * 
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class RestrictedMulticlassDefinition<I, V> extends ProblemDefinition<I, V> implements Serializable {    

    private static final long serialVersionUID = 3L;

    protected FeatureExtractor<I> extractor;
    protected FeatureList list;
    protected FeatureDeclaration decl;

    protected TObjectIntHashMap<V> classIDs = new TObjectIntHashMap();
    protected V[] range;
    
    protected NBestItem<V>[] nbestItems;

    protected FeatureListEncoder listEncoder;
    protected SparseVectorClassEncoder svEncoder = new SparseVectorClassEncoder();
    protected SparseVector sv1 = new SparseVector();
    
    public FeatureExtractor getFeatureExtractor() {
        return extractor;
    }

    public FeatureDeclaration getFeatureDecl() {
        return decl;
    }

    protected RestrictedMulticlassDefinition() {
    }

    public RestrictedMulticlassDefinition(Set<V> range, FeatureDeclaration decl, FeatureExtractor<I> extractor) {
        init(range, decl, extractor, new SimpleFeatureListEncoder());
    }
    
    public RestrictedMulticlassDefinition(List<Pair<I, V>> trainingSet, FeatureDeclaration decl, FeatureExtractor<I> extractor) {
        HashSet<V> range = new HashSet();
        for(Pair<?, V> ex: trainingSet)
            range.add(ex.right);
        init(range, decl, extractor, new SimpleFeatureListEncoder());
    }
    
    public RestrictedMulticlassDefinition(List<Pair<I, V>> trainingSet, 
            FeatureDeclaration decl, 
            FeatureExtractor<I> extractor,
            FeatureListEncoder listEncoder) {
        HashSet<V> range = new HashSet();
        for(Pair<?, V> ex: trainingSet)
            range.add(ex.right);
        init(range, decl, extractor, listEncoder);
    }

    /*

    protected FeatureExtractor<I> extractor;
    protected FeatureList list;
    protected FeatureDeclaration decl;

    protected TObjectIntHashMap<V> classIDs = new TObjectIntHashMap();
    protected V[] range;
    
    protected NBestItem<V>[] nbestItems;

    protected FeatureListEncoder listEncoder;
    protected SparseVectorClassEncoder svEncoder = new SparseVectorClassEncoder();
    protected SparseVector sv1 = new SparseVector();

     */
    
    public static <I, V> RestrictedMulticlassDefinition<I, V> remap(RestrictedMulticlassDefinition<I, V> s,
    		TIntIntHashMap m) {
    	if(s.getClass() != RestrictedMulticlassDefinition.class)
    		throw new IllegalArgumentException("can't handle subclasses!");
    	
    	RestrictedMulticlassDefinition<I, V> out = new RestrictedMulticlassDefinition();
    	out.extractor = s.extractor;
    	out.list = s.list.clone();
    	out.decl = s.decl;
    	out.classIDs = s.classIDs.clone();
    	out.range = s.range.clone();
    	out.nbestItems = s.nbestItems.clone();
    	    	
    	out.listEncoder = s.listEncoder.clone();
    	out.listEncoder.remap(m); // TODO MÅSTE MODIFIERA I STÄLLET FÖR ATT KOPIERA
    	
    	return out;
    }
    
    public int getNClasses() {
        return range.length;
    }
    
    public Set<V> range() {
        HashSet<V> out = new HashSet();
        for(V v: range)
            out.add(v);
        return out;
    }
    
    protected void init(Set<V> range, FeatureDeclaration decl, 
            FeatureExtractor<I> extractor, FeatureListEncoder listEncoder) {
        this.decl = decl;
        this.extractor = extractor;
        this.list = new FeatureList(decl);
        this.listEncoder = listEncoder;
        
        this.range = (V[]) new Object[range.size()];
        this.nbestItems = new NBestItem[range.size()];
        for(int i = 0; i < range.size(); i++)
            this.nbestItems[i] = new NBestItem();
        
        int i = 0;        
        for(V cl: range) {
            classIDs.put(cl, i + 1);
            this.range[i] = cl;
            i++;
        }        
    }
    
    public double cost(I x, V y, V ybar) {
        return y.equals(ybar)? 0: 1;
    }

    private boolean frozen = false;
    
    public void freeze() {
        sorted.clear();
        
        // 100512
        if(frozen)
        	System.err.println("Warning: MulticlassDefinition alreadyFrozen");
        else {
        	if(listEncoder != null)
        		listEncoder.freeze();
        	if(svEncoder != null)
        		svEncoder.freeze();
        }
        // end 100512
    }

    public int encodeClass(V y) {
        int yid = classIDs.get(y);
        return yid == 0? -1: yid;
    }

    public V decodeClass(int cl) {
        return range[cl-1];
    }

    public void encodeInstance(I x, SparseVector sv) {
        extractor.extractFeatures(x, list);
        listEncoder.encode(list, sv);
        //System.out.println("in encodeInstance: x = " + x);
        //sv.sortIndices();
        //System.out.println("in encodeInstance: sv = " + sv);
    }

    public void encodeFeatures(FeatureList features, SparseVector sv) {
        listEncoder.encode(features, sv);
        //System.out.println("in encodeFeatures: fs = " + features);
        //sv.sortIndices();
        //System.out.println("in encodeFeatures: sv = " + sv);
    }
    
    public void encode(I x, V y, SparseVector sv) {
        extractor.extractFeatures(x, list);
        listEncoder.encode(list, sv);
        int cl = encodeClass(y);
        if(cl == -1)
            throw new IllegalArgumentException("The class " + y + " is unknown.");
        svEncoder.encode(sv, cl);
    }

    /* TODO should also return score. */
    public V argmax(I x, Model model, SparseVector sv, double[] result) {
        extractor.extractFeatures(x, list);
        listEncoder.encode(list, sv1);
        int bestyid = -1;
        double bestScore = Double.NEGATIVE_INFINITY;
        for(int yid = 0; yid < range.length; yid++) {
            svEncoder.encode(sv1, yid + 1, sv);
            double score = model.score(sv);
            if(score > bestScore) {
                bestyid = yid;
                bestScore = score;
            }
        }
        sv.clear();
        if(bestyid == -1)
            return null;
        svEncoder.encode(sv1, bestyid + 1, sv);
        result[0] = bestScore;
        return range[bestyid];
    }

    protected static class NBestItem<V> implements Serializable {
        private static final long serialVersionUID = 0L;
        double score;
        int cl;
    }
    
    protected static class NBestComp implements Comparator<NBestItem>, Serializable {
        private static final long serialVersionUID = 0L;
        public int compare(NBestItem t1, NBestItem t2) {
            int c1 = -Double.compare(t1.score, t2.score);
            if(c1 != 0)
                return c1;
            return t1.cl - t2.cl;
        }
    }

    protected static Comparator<NBestItem> ordering = new NBestComp();

    protected TreeSet<NBestItem<V>> sorted = new TreeSet<NBestItem<V>>(ordering);

    /**
     * Returns up to n top-scoring outputs, not necessarily in order.
     * 
     */
    public int findNBest(I x, Model model, int n, V[] ybars, SparseVector[] svs, double[] scores) {
        extractor.extractFeatures(x, list);

        listEncoder.encode(list, sv1);
        
        if(n >= range.length) {
            for(int i = 0; i < range.length; i++) {
                svEncoder.encode(sv1, i + 1, svs[i]);
                scores[i] = model.score(svs[i]);
                ybars[i] = range[i];
            }
            return range.length;
        } else {

            sorted.clear();
            for(int i = 0; i < range.length; i++) {
                svEncoder.encode(sv1, i + 1, svs[0]);
                double score = model.score(svs[0]);
                nbestItems[i].score = score;
                nbestItems[i].cl = i;
                sorted.add(nbestItems[i]);
            }

            int i = 0;
            for(NBestItem<V> p: sorted) {
                scores[i] = p.score;
                ybars[i] = range[p.cl];
                svs[i].clear();
                svEncoder.encode(sv1, p.cl + 1, svs[i]);
                i++;
                if(i == n)
                    return n;
            }
        }
        return 0; // Won't end up here.
    }    
    
    public V maxLoss(I x, V y, Model model, SparseVector sv, double[] result) {
        extractor.extractFeatures(x, list);
        listEncoder.encode(list, sv1);

        int bestyid = -1;
        double maxLoss = Double.NEGATIVE_INFINITY;
        double maxLossScore = 0;
        double maxLossCost = 0;
        for(int yid = 0; yid < range.length; yid++) {
            svEncoder.encode(sv1, yid + 1, sv);
            double score = model.score(sv);
            double cost = cost(x, y, range[yid]);
            double loss = score + cost;
            if(loss > maxLoss) {
                bestyid = yid;
                maxLoss = loss;
                maxLossScore = score;
                maxLossCost = cost;
            }
        }
        sv.clear();
        if(bestyid == -1)
            return null;
        svEncoder.encode(sv1, bestyid + 1, sv);
        result[0] = maxLossScore;
        result[1] = maxLossCost;
        return range[bestyid];
    }

    private HashMap<I, Decomposition> cachedDecompositions = new HashMap();
    
    private class MulticlassDecomposition implements Decomposition {

        private double cachedLogPartition = Double.NaN;
        private DecompositionPart[] parts;
        private double[] thetas;
        
        MulticlassDecomposition(I x, V y) {
            thetas = new double[range.length];
            parts = new DecompositionPart[range.length];
            for (int i = 0; i < range.length; i++) {
                SparseVector sv = new SparseVector();
                encode(x, range[i], sv);
                DecompositionPart p = new MulticlassDecompositionPart(this,
                        i, range[i].equals(y), sv);
                parts[i] = p;
            }            
        }
        
        public Iterator<DecompositionPart> iterator() {
            return CollectionUtils.iterator(parts);
        }
        
        public int size() {
            return parts.length;
        }
        
        public double getLogPartition() {
            if (Double.isNaN(cachedLogPartition)) {
                double maxTheta = Double.NEGATIVE_INFINITY;
                for (int i = 0; i < thetas.length; i++)
                    if (thetas[i] > maxTheta)
                        maxTheta = thetas[i];
                double sumExpTheta = 0;
                for (int i = 0; i < thetas.length; i++)
                    sumExpTheta += Math.exp(thetas[i] - maxTheta);
                cachedLogPartition = maxTheta + Math.log(sumExpTheta);
                if(Double.isInfinite(cachedLogPartition) || Double.isNaN(cachedLogPartition)) {
                    throw new RuntimeException("cachedLogPartition = " + cachedLogPartition);
                }
            }
            return cachedLogPartition;
        }
    }
    
    private class MulticlassDecompositionPart implements DecompositionPart {

        private MulticlassDecomposition decomp;
        
        private double[] thetas;
        private int index;
        private boolean isCorrect;
        private SparseVector sv;
        
        MulticlassDecompositionPart(MulticlassDecomposition decomp, int index,
                boolean isCorrect, SparseVector sv) {
            this.decomp = decomp;
            this.thetas = decomp.thetas;
            this.index = index;
            this.isCorrect = isCorrect;
            this.sv = sv;
        }
        
        public double getTheta() {
            return thetas[index];
        }
        
        public void setTheta(double value) {
            thetas[index] = value;
            decomp.cachedLogPartition = Double.NaN;
        }
        
        public double getCost() {
            return isCorrect? 0.0: 1.0;
        }
        
        public SparseVector encode() {
            return sv;
        }
        
        public double getMarginalProbability() {
            return Math.exp(thetas[index] - decomp.getLogPartition());
        }

    };

    public Decomposition decompose(I x, V y) {
        Decomposition decomp = cachedDecompositions.get(x);
        if (decomp == null) {
            decomp = new MulticlassDecomposition(x, y);
            cachedDecompositions.put(x, decomp);
        }
        return decomp;        
    }
    
    
    public static final FeatureExtractor<FeatureList> ID = new FeatureExtractor<FeatureList>() {
        private static final long serialVersionUID = 1L;

        public void extractFeatures(FeatureList l, FeatureList l2) {
            l2.values = l.values;
        }
    };

    public static final FeatureExtractor<Object[]> COPY = new FeatureExtractor<Object[]>() {        
        private static final long serialVersionUID = 1L;
        public void extractFeatures(Object[] arr, FeatureList l2) {
            System.arraycopy(arr, 0, l2.values, 0, arr.length);
        }
    };
    
    public static ArrayList<Pair<RestrictedInstance<FeatureList, String>, String>> restrictSet(ArrayList<Pair<FeatureList, String>> ts,
    		String restrictionFeature) {
    	FeatureDeclaration decl = ts.get(0).left.decl;
    	int restrIndex = decl.getFeatureIndex(restrictionFeature);
    	ArrayList<Pair<RestrictedInstance<FeatureList, String>, String>> out = new ArrayList();

    	for(Pair<FeatureList, String> p: ts) {
    		FeatureList fl = p.left;
    		String value = p.right;
    		Object f = fl.getFeature(restrIndex);
    		RestrictedInstance<FeatureList, String> ri = new RestrictedInstance();
    		ri.instance = fl;
    		if(f instanceof String) {
    			if(!f.equals(value)) {
    				//throw new IllegalArgumentException("Illegal restriction feature: " + value + " / " + f);
    				System.err.println("Illegal restriction feature (1): " + value + " / " + f);
    				ri.range = new String[2];
    				ri.range[0] = value;
    				ri.range[1] = (String) f;
        			Arrays.sort(ri.range);
    			} else {
    				ri.range = new String[1];
    				ri.range[0] = value;
    			}
    		} else {
    			HashSet<String> s = (HashSet) f;
    			if(!s.contains(value)) {
    				System.err.println("Illegal restriction feature (2): " + value + " / " + s);
    				//throw new IllegalArgumentException("Illegal restriction feature: " + value + " / " + s);
    				HashSet<String> tmp = new HashSet(s);
    				tmp.add(value);
        			ri.range = tmp.toArray(new String[0]);
    			} else
        			ri.range = s.toArray(new String[0]);

    			Arrays.sort(ri.range);
    		}
    		out.add(new Pair(ri, value));
    	}
    	return out;
    }
    
}