/*
 AdditiveUC.java

 Created on Sep 11, 2006 by Richard Johansson (richard@cs.lth.se).

 $Log: AUCUpdate.java,v $
 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:05  richard
 Branched.

 Revision 1.1  2007/09/28 14:42:28  richard
 Refactored to atom-based.

 Revision 1.7  2007/04/27 10:48:42  richard
 Constructor.

 Revision 1.6  2006/09/22 08:07:53  richard
 Renamed specification to definition.

 Revision 1.5  2006/09/21 10:55:26  richard
 Divides the update factor.

 Revision 1.4  2006/09/18 08:06:10  richard
 Renamed encoding methods.

 Revision 1.3  2006/09/17 14:01:16  richard
 Uses createClassifier.

 Revision 1.2  2006/09/13 08:16:02  richard
 All three variants.

 Revision 1.1  2006/09/11 08:25:13  richard
 Added the file.


 */
package se.lth.cs.nlp.nlputils.ml_long.algorithms;

import se.lth.cs.nlp.nlputils.ml_long.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class AUCUpdate<I, V> extends OnlineUpdater<I, V> {

    public static final int UNIFORM_UPDATE = 0;
    public static final int PROPORTIONAL_UPDATE = 1;
    public static final int MAXSCORE_UPDATE = 2;
    public static final int MAXLOSS_UPDATE = 3;

    private static double TOLERANCE = 1e-6;

    private int updateType = UNIFORM_UPDATE;

    private int N = 1;

    private SparseVector yrep = new SparseVector();
    private SparseVector yhatrep = new SparseVector();
    private SparseVector[] reps;

    private V[] ybars;
    private double[] scores;
    private double[] results = new double[2];
    
    double tau;
    double yscore;
    int n;
    V y;
    boolean allZero = false;
    
    public AUCUpdate(String[] argv, int position) {
        for(int i = position; i < argv.length; i++) {
            if(argv[i].equals("N"))
                N = Integer.parseInt(argv[++i]);
            else if(argv[i].equals("updateType"))
                updateType = Integer.parseInt(argv[++i]);
            else
                throw new IllegalArgumentException("Unknown parameter: " + argv[i]);
        }
        if(updateType == UNIFORM_UPDATE || updateType == PROPORTIONAL_UPDATE) {
            this.reps = new SparseVector[N];
            for(int i = 0; i < N; i++)
                reps[i] = new SparseVector();
            this.scores = new double[N];
            this.ybars = (V[]) new Object[N];
        }
    }

    public void calculateUpdate(ProblemDefinition<I, V> spec, Model model, 
            I x, V y) {

        this.y = y;
        spec.encode(x, y, yrep);
        yscore = model.score(yrep);

        switch(updateType) {
        case UNIFORM_UPDATE: {
            n = spec.findNBest(x, model, N, ybars, reps, scores);                	
            int ne = 0;
            for(int j = 0; j < n; j++)
                if(ybars[j] != y && scores[j] >= yscore && !ybars[j].equals(y))
                    ne++;
            if(ne == 0)
                tau = 0;
            else
                tau = -1.0 / ne;
        }
        break;
        case MAXSCORE_UPDATE: {
            V ybar = spec.argmax(x, model, yhatrep, results);
            if(y == ybar || y.equals(ybar))
                tau = 0;
            else
                tau = 1;
        }
        break;
        case MAXLOSS_UPDATE: {
            V ybar = spec.maxLoss(x, y, model, yhatrep, results);
            if(y == ybar || y.equals(ybar))
                tau = 0;
            else
                tau = 1;
        }
        break;
        case PROPORTIONAL_UPDATE: {
            n = spec.findNBest(x, model, N, ybars, reps, scores);                	
            double sum = 0;
            int ne = 0;
            allZero = true;
            for(int j = 0; j < n; j++) {
                if(Math.abs(scores[j]) > TOLERANCE) 
                    allZero = false;
                if(ybars[j] != y && scores[j] >= yscore && !ybars[j].equals(y)) {
                    sum += scores[j] - yscore;
                    ne++;
                }
            }
            if(allZero && ne > 0) {
                tau = 1 / ne;
            } else if(sum < TOLERANCE)
                tau = 0;
            else
                tau = -1.0 / sum;
        }
        break;
        default:
            throw new RuntimeException("Illegal update type!");
        }
    }

    public void updateModel(double d, Model updatedModel) {
        if(tau == 0)
            return;
        updatedModel.update(d*tau, yrep); 
        switch(updateType) {
        case UNIFORM_UPDATE: {
            for(int j = 0; j < n; j++)
                if(ybars[j] != y && scores[j] >= yscore && !ybars[j].equals(y))
                    updatedModel.update(d*tau, reps[j]);
            break;
        }
        case MAXSCORE_UPDATE: 
        case MAXLOSS_UPDATE: {
            updatedModel.update(-d*tau, yhatrep);
            break;
        }
        case PROPORTIONAL_UPDATE: {
            if(!allZero) { 
                for(int j = 0; j < n; j++)
                    if(ybars[j] != y && scores[j] >= yscore && !ybars[j].equals(y)) {
                        double u2 = tau*(scores[j] - yscore);
                        if(u2 > -TOLERANCE)
                            continue;
                        updatedModel.update(d*u2, reps[j]);
                    }
            }
            else {
                for(int j = 0; j < n; j++)
                    if(ybars[j] != y && scores[j] >= yscore && !ybars[j].equals(y))
                        updatedModel.update(d*tau, reps[j]);
            }
            break;
        }
        }
    }
    
    public void roundStatistics() {
    }
    
}
