/*
 CWUpdate.java
  
 Created on Apr 2, 2009 by Richard Johansson (johansson@disi.unitn.it).

 $Log: CWUpdate.java,v $
 Revision 1.3  2010-06-07 11:30:23  johansson
 Major update.

 Revision 1.2  2009/04/16 11:59:09  johansson
 Averaging works.

 Revision 1.1  2009/04/16 11:42:52  johansson
 Added.

   
 */
package se.lth.cs.nlp.nlputils.ml_long.algorithms;

import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.math.CubicInterpolation;
import se.lth.cs.nlp.nlputils.ml_long.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class CWUpdate<I, V> extends OnlineUpdater<I, V> {

    /* Learning parameters. */
    private boolean maximizeLoss = false;

    private double eta = 0.75;

    private double phi, psi, xi, phi2, phi2xi, phi4b4;
    private int size = 10000;
    
    private boolean sorted = false;

    private ClosedLongDoubleMap sdiag;
    
    /* Private data to cache update. */
    private SparseVector yrep, yhatrep, diff;
    private double[] result = new double[2];

    private double alpha, ui_sqrt;
    
    /* Statistics. */
    private double sumCost;
    private int count;
    
    private static final boolean DEBUG = false;
    
    private static final double TOLERANCE = 1e-15;
    
    private static final double A = 1;
    
    public CWUpdate(String[] argv, int offset) {
        for(int i = offset; i < argv.length; i++) {
            if(argv[i].equals("maximizeLoss"))
                maximizeLoss = Boolean.parseBoolean(argv[++i]);
            else if(argv[i].equals("sorted"))
                sorted = Boolean.parseBoolean(argv[++i]);
            //else if(argv[i].equals("sqrt"))
            //    sqrt = Boolean.parseBoolean(argv[++i]);
            else if(argv[i].equals("eta"))
                eta = Double.parseDouble(argv[++i]);
            else if(argv[i].equals("size"))
                size = Integer.parseInt(argv[++i]);
            else
                throw new IllegalArgumentException("Unknown parameter: " + argv[i]);
        }        
        if(eta < 0.5 || eta >= 1)
        	throw new IllegalArgumentException("eta = " + eta);
        
        yrep = new SparseVector();
        yhatrep = new SparseVector();
        diff = new SparseVector();
        
        phi = PhiInv(eta);
        psi = 1.0 + phi*phi / 2;
        xi = 1.0 + phi*phi;
        phi2 = phi*phi;
        phi2xi = phi2*xi;
        phi4b4 = phi2*phi2 / 4;
                
        sdiag = new ClosedLongDoubleMap(size);        
    }

    private static double norm1(SparseVector sv) {
        double sum = 0;
        sv.sortIndices();
        for(int i = 0; i < sv.index; i++)
            sum += Math.abs(sv.values[i]);
        return sum;
    }
    
    public void calculateUpdate(ProblemDefinition<I, V> spec, Model model, 
            I x, V y) {

    	updateSigma();
    	
        if(DEBUG) {
            System.out.println("CalculateUpdate:");
            System.out.println("x = " + x);
            System.out.println("y = " + y);
        }
        
        spec.encode(x, y, yrep);

        if(DEBUG) {
            System.out.println("yrep = " + yrep);
            System.out.println("yrep.1n = " + norm1(yrep));
        }
        
        V yhat;
        if(maximizeLoss)
            yhat = spec.maxLoss(x, y, model, yhatrep, result);
        else
            yhat = spec.argmax(x, model, yhatrep, result);

        if(!sorted) {
        	yrep.sortIndices();
        	yhatrep.sortIndices();        	
        }
        diff.clear();
        SparseVector.sortedSubtract(diff, yrep, yhatrep);
        
        if(DEBUG) {
            System.out.println("yhat = " + yhat);
            System.out.println("yhatrep = " + yhatrep);
            System.out.println("yhatrep.1n = " + norm1(yhatrep));
        }
        
        double yhatscore = result[0];
        double cost      = maximizeLoss? result[1]: spec.cost(x, y, yhat);

        if(DEBUG) {
            System.out.println("yscore = " + model.score(yrep));

            System.out.println("yhatscore = " + yhatscore);
            System.out.println("yhatscore2 = " + model.score(yhatrep));
        }
        
        double vi = 0;
        for(int i = 0; i < diff.index; i++) {
        	double v = diff.values[i];
        	double sd = sdiag.getOrSet(diff.keys[i], A);
        	vi += sd*v*v;
        }

        double mi = model.score(diff);
        
        alpha = Math.sqrt(mi*mi*phi4b4 + vi*phi2xi) - mi*psi;
        if(alpha > TOLERANCE)
        	alpha /= vi*xi;
        else
        	alpha = 0;
        
        ui_sqrt = 0.5 * (Math.sqrt(alpha*alpha*vi*vi*phi2 + 4*vi) - alpha*vi*phi);        

        // not needed
        //beta = alpha*phi / (ui_sqrt + alpha*vi*phi);
        
        if(DEBUG) {
            tmpy = yrep;
            tmpyhat = yhatrep;
            //System.out.println("tau = " + tau);
        }
        
        sumCost += cost;
        count++;
    }

    private SparseVector tmpy, tmpyhat;
    
    long t0;

    private void updateSigma() {
        if(alpha > TOLERANCE) {
        	double f = alpha*phi/ui_sqrt;
        	for(int i = 0; i < diff.index; i++) {
        		long k = diff.keys[i];
        		double v = diff.values[i];        		
        		double sig = sdiag.get(k);
        		sdiag.put(k, 1.0 / (1.0 / sig + f*v*v));        		
        	}
        }
    }
    
    public void updateModel(double d, Model updatedModel) {
        if(alpha > TOLERANCE) {        
        	for(int i = 0; i < diff.index; i++) {
        		long k = diff.keys[i];
        		double v = diff.values[i];        		
        		double sig = sdiag.get(k);        		
        		updatedModel.update(k, d*alpha*sig*v);
        	}
        }
        if(DEBUG) {
            System.out.println("yscore_new = " + updatedModel.score(tmpy));
            System.out.println("yhatscore_new = " + updatedModel.score(tmpyhat));
        }
    }
    
    public void roundStatistics() {
        if(count > 0) {
            long t1 = System.currentTimeMillis();
            int format = Util.HOURS | Util.MINUTES | Util.SECONDS;
            System.out.println("Average cost: " + sumCost / count);           
            System.out.println("Round time: " + Util.toTimeString(t1 - t0, format));
            t0 = t1;
        } else {
            t0 = System.currentTimeMillis();
        }
        sumCost = 0;
        count = 0;
    }

    private static double[][] PHI_INV = {
            { 0.5, 0 },
            { 0.6, 0.253347103135800 },
            { 0.7, 0.524400512708041 },
            { 0.75, 0.674489750196082 },
            { 0.8, 0.841621233572914 },
            { 0.85, 1.03643338949379 },
            { 0.9, 1.28155156554460 },
            { 0.95, 1.64485362695147 },
            { 0.975, 1.95996398454005 },
            { 0.99, 2.32634787404084 },
            { 0.995, 2.57582930354890 },
            { 0.999, 3.09023230616781 },
            { 0.9995, 3.29052673149194 },
            { 0.9999, 3.71901648545567 },
            { 0.99995, 3.89059188641317 },
            { 0.99999, 4.26489079392334 },
            { 0.999995, 4.41717341346829 },
            { 0.999999, 4.75342430881759 },
    };
    
    private static double PhiInv(double p) {
        if(p < 0.5 || p >= 1.0)
            throw new IllegalArgumentException("Illegal p value: " + p);
        return CubicInterpolation.compute(PHI_INV, p);
    }
    
}
