package se.lth.cs.nlp.nlputils.ml_long.algorithms;

import java.util.*;

import se.lth.cs.nlp.nlputils.core.CollectionUtils;
import se.lth.cs.nlp.nlputils.core.Pair;
import se.lth.cs.nlp.nlputils.ml_long.*;

public class KMeans {

    private int k;
    private int seed;
    private int niter;
    
    public KMeans(int k, int niter) {
        this.k = k;
        this.seed = 0;
        this.niter = niter;
    }
    
    public Pair<Model[], int[]> run(ArrayList<SparseVector> trainingSet) {
        Random rand = new Random(seed);
        Model[] cs = new Model[k];
            
        int n = trainingSet.size();
        
        int[] classifications = new int[n];
        int[] hist = new int[k];

        for(int i = 0; i < n; i++) {
            int cl = rand.nextInt(k);
            hist[cl]++;
            classifications[i] = cl;
        }

        int maxIndex = -1;
        for(SparseVector sv: trainingSet) {
            sv.sortIndices();
            if(sv.index == 0)
                throw new IllegalArgumentException("Empty vector");
            long ix = sv.keys[sv.index - 1];
            if(ix > Integer.MAX_VALUE)
                throw new IllegalArgumentException("Feature value too large");
            if(ix > maxIndex)
                maxIndex = (int) ix;
        }
        for(int i = 0; i < k; i++)
            cs[i] = new LinearVectorModel(maxIndex + 1);
        
        double[] normalizers = new double[k];
        double[] cSqrNorms = new double[k];
        
        System.out.println(Arrays.toString(hist));
        
        double prev = Double.NaN;
        for(int i = 0; i < niter; i++) {
            
            /* Compute new centroids. */
            for(int j = 0; j < k; j++) {
              //if(hist[j] == 0)
                //    throw new RuntimeException("zero in hist");
                if(hist[j] == 0)
                    normalizers[j] = -1;
                else
                    normalizers[j] = 1.0 / hist[j];
            }

            // TODO shouldn't we reset hist?????
            Arrays.fill(hist, 0);
            
            for(int j = 0; j < k; j++)
                cs[j].clear();
            for(int j = 0; j < n; j++) {
                int cl = classifications[j];
                cs[cl].update(normalizers[cl], trainingSet.get(j));
            }

            /* Cache centroid norms. */
            for(int j = 0; j < k; j++)
                cSqrNorms[j] = cs[j].modelSqrNorm();

            /* Compute new classifications. */
            Arrays.fill(hist, 0);
            double totalDist = 0;
            for(int j = 0; j < n; j++) {               
                SparseVector sv = trainingSet.get(j);
                int minCl = -1;
                double minDist = Double.POSITIVE_INFINITY;
                for(int l = 0; l < k; l++) {
                    if(normalizers[l] < 0)
                        continue;
                    
                    //double dist = cSqrNorms[l] + SparseVector.sortedSqrLength(sv)
                    //    - 2*cs[l].score(sv);

                    double dist = cs[l].score(sv) / Math.sqrt(cSqrNorms[l] * SparseVector.sortedSqrLength(sv));
                    dist = Math.abs(dist);
                    if(dist > 1)
                        throw new RuntimeException("dist = " + dist);
                    
                    //System.out.println("dist = " + dist);
                    
                    if(dist < minDist) {
                        minDist = dist;
                        minCl = l;
                    }
                }
                classifications[j] = minCl;
                hist[minCl]++;
                totalDist += minDist;
            }

            System.out.println("totalDist = " + totalDist);
            if(totalDist == prev)
                break;
            prev = totalDist;
        }
        return new Pair(cs, classifications);
    }
 
    public static void main(String[] argv) {
        int k = Integer.parseInt(argv[0]);
        int niter = Integer.parseInt(argv[1]);
        String file = argv[2];
        String usedFeatures = argv.length > 3? argv[3].trim(): "";
        String usedPairs = argv.length > 4? argv[4].trim(): "";

        try {
            ArrayList<Pair<FeatureList, String>> goldStandard = MulticlassDefinition.readDataSet(file, null);
            FeatureDeclaration decl = goldStandard.get(0).left.decl;
            FeatureDeclaration subsetDecl;
            
            if(usedFeatures.isEmpty() || usedFeatures.equals("all"))
                subsetDecl = decl;
            else {
                Pair<HashSet<String>, HashSet<Pair<String, String>>> 
                p1 = MulticlassDefinition.parseSubsetAndPairs(usedFeatures);
                if(p1.right.size() > 0)
                    throw new RuntimeException("no pairs here please");
                subsetDecl = decl.subset(p1.left);
                MulticlassDefinition.useSubset(decl, subsetDecl, goldStandard);
            }

            HashSet<Pair<String, String>> ps;            
            if(!usedPairs.isEmpty())
                ps = PairingFeatureListEncoder.parsePairString(usedPairs);
            else
                ps = new HashSet();
            
            FeatureListEncoder fle = new PairingFeatureListEncoder(subsetDecl, ps);
            SymbolEncoder clEnc = new SymbolEncoder();
            ArrayList<SparseVector> svs = new ArrayList();

            int maxCl = -1;
            for(Pair<FeatureList, String> example: goldStandard) {
                SparseVector sv = new SparseVector();
                fle.encode(example.left, sv);
                svs.add(sv);
                int cl = clEnc.encode(example.right);
                if(cl > maxCl)
                    maxCl = cl;
            }
            
            int[] classifications;
            if(false) {
                KMeans kmeans = new KMeans(k, niter);
                Pair<Model[], int[]> p = kmeans.run(svs);
                classifications = p.right;
            } else {
                IterSVMClusterer cls = new IterSVMClusterer(k, niter, 100);
                classifications = cls.run(svs);
            }
            
            long[][] cmatrix = new long[maxCl + 1][k];
            for(int i = 0; i < goldStandard.size(); i++) {
                int cl = clEnc.encode(goldStandard.get(i).right);
                cmatrix[cl][classifications[i]]++;
            }

            for(int i = 1; i < cmatrix.length; i++) {
                System.out.print(clEnc.decode(i));
                for(int j = 0; j < cmatrix[i].length; j++) {
                    System.out.print("\t" + cmatrix[i][j]);
                }                    
                System.out.println();
            }

            double[] classEntropies = new double[maxCl + 1];
            double[] clusterEntropies = new double[k];

            for(int i = 1; i < cmatrix.length; i++) {
                double entropy = 0;                
                long sum = 0;
                for(int j = 0; j < cmatrix[i].length; j++)
                    sum += cmatrix[i][j];
                for(int j = 0; j < cmatrix[i].length; j++) {
                    double pr = (double) cmatrix[i][j] / sum;
                    if(pr > 0)
                        entropy -= pr * Math.log(pr);
                }
                classEntropies[i] = entropy;
            }

            for(int i = 0; i < k; i++) {
                double entropy = 0;
                long sum = 0;
                for(int j = 1; j < cmatrix.length; j++)
                    sum += cmatrix[j][i];
                for(int j = 1; j < cmatrix.length; j++) {
                    double pr = (double) cmatrix[j][i] / sum;                    
                    if(pr > 0)
                        entropy -= pr * Math.log(pr);
                }
                clusterEntropies[i] = entropy;
            }

            System.out.print("class entropies: " + Arrays.toString(classEntropies));
            System.out.println(" (" + CollectionUtils.sum(classEntropies) + ")");
            System.out.print("cluster entropies: " + Arrays.toString(clusterEntropies));
            System.out.println(" (" + CollectionUtils.sum(clusterEntropies) + ")");

            {
                /* Compute Rand index. */

                long a = 0;
                for(int i = 1; i < cmatrix.length; i++)
                    for(int j = 0; j < cmatrix[i].length; j++)
                        a += cmatrix[i][j] * (cmatrix[i][j] - 1);

                long b = 0;
                for(int i = 1; i < cmatrix.length; i++)
                    for(int j = 0; j < cmatrix[i].length; j++)
                        for(int i1 = 1; i1 < cmatrix.length; i1++) {
                            if(i1 == i)
                                continue;
                            for(int j1 = 0; j1 < cmatrix[i1].length; j1++) {
                                if(j1 == j)
                                    continue;                                
                                b += cmatrix[i][j] * cmatrix[i1][j1];
                            }
                        }

                long c = 0;
                for(int i = 1; i < cmatrix.length; i++)
                    for(int j = 0; j < cmatrix[i].length; j++)
                        for(int j1 = 0; j1 < cmatrix[i].length; j1++) {
                            if(j1 == j)
                                continue;                                
                            c += cmatrix[i][j] * cmatrix[i][j1];
                        }
                
                long d = 0;
                for(int i = 1; i < cmatrix.length; i++)
                    for(int j = 0; j < cmatrix[i].length; j++)
                        for(int i1 = 1; i1 < cmatrix.length; i1++) {                        
                            if(i1 == i)
                                continue;
                            d += cmatrix[i][j] * cmatrix[i1][j];
                        }

                double ri = (double) (a + b) / (a + b + c + d);

                System.out.println("Rand index: " + ri);
            }
            
            /*
            for(int i = 0; i < 1000; i++) {
                Pair pa = goldStandard.get(i); 
                System.out.println(pa.right + "\t" + classifications[i] + "\t" + pa.left);
            }*/
            
        } catch(Exception e) {
            e.printStackTrace();
        }
    }
    
}
