/*
 MEC.java
  
 Created on Feb 16, 2007 by Richard Johansson (richard@cs.lth.se).

 $Log: MECUpdate.java,v $
 Revision 1.2  2010-06-07 11:30:23  johansson
 Major update.

 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.2  2008/06/02 12:25:25  richard
 Array bug.

 Revision 1.1  2008/05/16 12:16:05  richard
 Branched.

 Revision 1.1  2007/09/28 14:42:28  richard
 Refactored to atom-based.

 Revision 1.3  2007/05/10 12:58:03  richard
 Moved out optimization.

 Revision 1.2  2007/02/16 15:16:02  richard
 Round statistics.

 Revision 1.1  2007/02/16 12:56:34  richard
 Added the file.

   
 */
package se.lth.cs.nlp.nlputils.ml_long.algorithms;

import java.util.Arrays;

import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.ml_long.*;
import se.lth.cs.nlp.nlputils.optimization.*;
import java.lang.reflect.Array;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class MECUpdate<I, V> extends OnlineUpdater<I, V> {

    private static final double TOLERANCE = 1e-6;
    
    private static final double MAX_EXPSCORE = 1e10;
    private static final double MIN_EXPSCORE = 1e-10;
    
    private int N = 4;
    private boolean verbose = false;
    private double gamma = 1;
    
    private boolean useLaplace = false;
    
    private V[] ybars;
    private double[] scores;
    private SparseVector[] reps;
    private SparseVector yrep = new SparseVector();

    private double[] s;
    private double[] expScores;

    private double[] alpha;
    private int m;
    private double[][] KC;    
    
    private double[] costs;
    
    /* Statistics. */
    private double sumTopCost;
    private double sumExpectedCost;
    private double sumBestCost;
    private int count;

    private UnconstrainedOptimizer optimizer = new GradientDescent();
    private ExpectedCost expectedCostFunction = new ExpectedCost();

    private double expSum;
    private double expSumC;
    
    public MECUpdate(String[] argv, int offset) {
        for(int i = offset; i < argv.length; i++) {
            if(argv[i].equals("N"))
                N = Integer.parseInt(argv[++i]);
            else if(argv[i].equals("gamma"))
                gamma = Double.parseDouble(argv[++i]);
            else if(argv[i].equals("useLaplace"))
                useLaplace = Boolean.parseBoolean(argv[++i]);
            else if(argv[i].equals("verbose"))
                verbose = Boolean.parseBoolean(argv[++i]);
            else
                throw new IllegalArgumentException("Unknown parameter: " + argv[i]);
        }

        alpha = new double[N];        
        costs = new double[N+1];
        s = new double[N];
        
        scores = new double[N + 1];
        expScores = new double[N + 1];
        reps = new SparseVector[N + 1];
        for(int i = 0; i < N + 1; i++)
            reps[i] = new SparseVector();

        KC = new double[N + 1][N + 1];
        //ybars = (V[]) new Object[N + 1];
    }

    public void calculateUpdate(ProblemDefinition<I, V> spec, Model model, 
            I x, V y) {

        if(ybars == null) {
            ybars = (V[]) Array.newInstance(y.getClass(), N + 1);
        }
        
        spec.encode(x, y, yrep);
        double yscore = model.score(yrep);

        m = spec.findNBest(x, model, N, ybars, reps, scores);

        if(verbose) {
            System.out.println("m = " + m);
            System.out.println("scores = " + Arrays.toString(scores));
        }

        double bestCost = Double.MAX_VALUE;
        double bestScore = Double.MIN_VALUE;
        double bestScoreCost = 0;

        int correctIndex = -1;
        for(int j = 0; j < m; j++) {
            double cost = spec.cost(x, y, ybars[j]);
            if(Math.abs(cost) < TOLERANCE) {
                if(correctIndex == -1)
                    correctIndex = j;
                else {
                    //scores[correctIndex] = 0;
                    scores[correctIndex] = yscore;
                    reps[correctIndex] = yrep.copy();
                    ybars[correctIndex] = y;

                    System.out.println("\nWarning! Multiple zero-cost predictions.");

                    for(int k = j+1; k < m; k++) {
                        reps[k-1] = reps[k];
                        scores[k-1] = scores[k];
                        ybars[k-1] = ybars[k];
                    }
                    m--;
                    j--;
                    continue;
                }
            }
            costs[j] = cost;
            
            if(scores[j] > bestScore) {
                bestScore = scores[j];
                bestScoreCost = cost;
            }
            if(cost < bestCost)
                bestCost = cost;
            
        }
        if(correctIndex == -1) {
            //scores[m] = 0;
            scores[m] = yscore;

            SparseVector swap = reps[m];
            reps[m] = yrep;
            yrep = swap;
            
            correctIndex = m;
            costs[m] = 0;
            m++;
        }

        CollectionUtils.swap(scores, 0, correctIndex);
        CollectionUtils.swap(reps, 0, correctIndex);
        CollectionUtils.swap(costs, 0, correctIndex);
        CollectionUtils.swap(ybars, 0, correctIndex);
        
        map(model, correctIndex);

        sumBestCost += bestCost;
        sumTopCost += bestScoreCost;

        count++;
    }

    public void updateModel(double d, Model updatedModel) {
        double sumAlpha = CollectionUtils.sum(alpha, m-1);
        if(verbose) {
            System.out.println("sumAlpha = " + sumAlpha);
            System.out.println("alpha = " + Arrays.toString(alpha));
            System.exit(0);
        }
        updatedModel.update(d*sumAlpha, reps[0]);
        for(int j = 0; j < m-1; j++)
            updatedModel.update(-d*alpha[j], reps[j+1]);
    }
    
    public void roundStatistics() {
        System.out.println("Average predicted cost: " + sumTopCost / count);
        System.out.println("Average expected cost: " + sumExpectedCost / count);
        sumTopCost = 0;
        sumExpectedCost = 0;
        sumBestCost = 0;
        count = 0;
    }

    private void map(Model model, int correctIndex) {

        if(verbose)
            System.out.println("Entering map.");

        for(int i = 0; i < m-1; i++) {
            s[i] = gamma / m / m;
            alpha[i] = 0;
        }
            
        for(int i = 0; i < m; i++) {
            model.prepareDot(reps[i]);
            KC[i][i] = model.sqrLength(reps[i]);
            for(int j = i + 1; j < m; j++) {
                KC[j][i] = KC[i][j] = model.dot(reps[j]);
            }
        }

        expectedCostFunction.compute(alpha);
        sumExpectedCost += expSumC/expSum;
        optimizer.minimize(expectedCostFunction, alpha, m-1);
    }

    private class ExpectedCost extends DiffFunction {

        private void computeScores(double[] x) {
            expSum = 0;
            expSumC = 0;
            for(int k = 0; k < m; k++) {
                double sum = scores[k];
                for(int j = 0; j < m-1; j++)
                    sum += x[j] * (KC[0][k] - KC[j+1][k]);
                expScores[k] = Math.exp(sum);

                if(expScores[k] > MAX_EXPSCORE)
                    expScores[k] = MAX_EXPSCORE;

                if(expScores[k] < MIN_EXPSCORE)
                    expScores[k] = MIN_EXPSCORE;

                expSum += expScores[k];
                if(k > 0)
                    expSumC += expScores[k] * costs[k];
            }

            if(verbose) {
                System.out.println("expSum = " + expSum);
                System.out.println("expSumC = " + expSumC);
            }

        }

        public double compute(double[] x) {

            computeScores(x);
            
            double part1 = -Math.log(expSum + expSumC);
            double part2 = Math.log(expSum);

            double part3 = 0;

            for(int i = 0; i < m-1; i++)
                part3 -= useLaplace? s[i]*Math.abs(x[i]): s[i]*x[i]*x[i];

            if(verbose) {
                System.out.println("computeLogPosterior:");
                System.out.println("part1 = " + part1);   
                System.out.println("part2 = " + part2);   
                System.out.println("part3 = " + part3);   
                double out = Double.isInfinite(part2)? -part2: -part1 - part2 - part3;
                System.out.println("objective is " + out);
            }
            
            return Double.isInfinite(part2)? -part2: -part1 - part2 - part3;
        }

        public void computeGradient(double[] x, double[] g) {

            if(verbose) {
                System.out.println("In computeDerivatives.");
                System.out.println("expSum = " + expSum);
                System.out.println("expScores = " + Arrays.toString(expScores));
            }

            /* First derivatives: */
            for(int i = 0; i < m-1; i++) {
                double diffSum = 0;
                double diffSumC = 0;
                for(int k = 0; k < m; k++) {
                    diffSum += (KC[0][k] - KC[i+1][k]) * expScores[k];
                    diffSumC += costs[k] * (KC[0][k] - KC[i+1][k]) * expScores[k];
                }
                //dalpha[i] = (diffSum- diffSumC)/(expSum - expSumC) - diffSum/expSum
                //    - (useLaplace? s[i]*Math.signum(alpha[i]): 2*alpha[i]*s[i]);
                
                g[i] = (diffSum + diffSumC)/(expSum + expSumC) - diffSum/expSum
                + (useLaplace? s[i]*Math.signum(alpha[i]): 2*alpha[i]*s[i]);

                if(verbose) {
                    System.out.println("diffSum = " + diffSum);
                    System.out.println("diffSumC = " + diffSumC);
                    System.out.println("diffSumC/expSumC = " + diffSumC/expSumC);
                    System.out.println("diffSum/expSum = " + diffSum/expSum);
                }
            }

            for(int i = 0; i < m-1; i++) {
                if(Double.isInfinite(g[i]))
                    throw new RuntimeException("Numerical problems: dalpha is infinite");
                if(Double.isNaN(g[i])) {

                    double diffSumC = 0, diffSum = 0;
                    for(int k = 0; k < m; k++)
                        diffSum += (KC[0][k] - KC[i+1][k]) * expScores[k];
                    for(int k = 0; k < m; k++)
                        diffSumC += costs[k] * (KC[0][k] - KC[i+1][k]) * expScores[k];

                    System.out.println();
                    System.out.println("In computeDerivatives.");
                    System.out.println("m = " + m);
                    System.out.println("i = " + i);
                    System.out.println("KC = " + Arrays.deepToString(KC));             
                    System.out.println("costs = " + Arrays.toString(costs));             
                    System.out.println("s = " + Arrays.toString(s));
                    System.out.println("alpha (prev) = " + Arrays.toString(alpha));
                    System.out.println("expSum = " + expSum);
                    System.out.println("expSumC = " + expSumC);        
                    System.out.println("scores = " + Arrays.toString(scores));
                    System.out.println("expScores = " + Arrays.toString(expScores));
                    System.out.println("dalpha = " + Arrays.toString(g));
                    System.out.println("diffSum[" + i + "]= " + diffSum);
                    System.out.println("diffSumC[" + i + "]= " + diffSumC);

                    throw new RuntimeException("Numerical problems: dalpha is NaN");
                }
            }

            if(verbose) {
                System.out.println("gradient = " + Arrays.toString(g));
            }
            
        }
    }
}
