/*
 MIRA.java
  
 Created on Feb 15, 2007 by Richard Johansson (richard@cs.lth.se).

 $Log: MIRAUpdate.java,v $
 Revision 1.2  2010-06-07 11:30:23  johansson
 Major update.

 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:05  richard
 Branched.

 Revision 1.1  2007/09/28 14:42:29  richard
 Refactored to atom-based.

 Revision 1.18  2007/02/16 15:16:02  richard
 Round statistics.

 Revision 1.17  2007/02/16 10:44:37  richard
 Changed to updater.

   
 */
package se.lth.cs.nlp.nlputils.ml_long.algorithms;

import java.lang.reflect.Array;
import java.util.*;

import se.lth.cs.nlp.nlputils.ml_long.*;


/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class MIRAUpdate<I, V> extends OnlineUpdater<I, V> {

    //private static final double TOLERANCE = 1e-6;
    //private static final double eps = 0.001;
    //private static final double KKT_tolerance = 0.001;
    
    private static final double TOLERANCE = 1e-9;
    private static final double eps = 1e-6;
    private static final double KKT_tolerance = 1e-6;

    /* Learning parameters. */
    private double C = Double.MAX_VALUE;
    private int N = 4;

    /* Private data for interacting with problem definition. */
    private V[] ybars;
    private double[] scores;
    private SparseVector[] reps;
    private SparseVector yrep = new SparseVector();
    
    /* Private data used in SMO. */
    private int m;
    private double[] alpha;
    private double[] errorCache;
    private double[][] KC;
    private double[] Cs;
    private double[] ys;
    private Random random;
    
    /* Statistics. */
    private double sumTopCost;
    private double sumBestCost;
    private int count;    
    
    public MIRAUpdate(String[] argv, int offset) {

        for(int i = offset; i < argv.length; i++) {
            if(argv[i].equals("C"))
                C = Double.parseDouble(argv[++i]);
            else if(argv[i].equals("N"))
                N = Integer.parseInt(argv[++i]);
            else
                throw new IllegalArgumentException("Unknown parameter: " + argv[i]);
        }

        alpha = new double[N + 1];
        errorCache = new double[N + 1];
        ys = new double[N + 1];
        Cs = new double[N + 1];
        scores = new double[N + 1];
        reps = new SparseVector[N + 1];
        for(int i = 0; i < N + 1; i++)
            reps[i] = new SparseVector();
        KC = new double[N + 1][N + 1];
        random = new Random(0);
        
        //ybars = (V[]) new Object[N + 1]; Funkar inte

        
    
    }
    
    public void calculateUpdate(ProblemDefinition<I, V> spec, Model model, 
            I x, V y) {

    	if(ybars == null) {
    		ybars = (V[]) Array.newInstance(y.getClass(), N + 1);
    	}
    	
    	spec.encode(x, y, yrep);
        double yscore = model.score(yrep);

        m = spec.findNBest(x, model, N, ybars, reps, scores);
        int correctIndex = -1;
        double bestCost = Double.MAX_VALUE;
        double bestScore = Double.MIN_VALUE;
        double bestScoreCost = 0;

        for(int j = 0; j < m; j++) {
            double cost = spec.cost(x, y, ybars[j]);
            scores[j] += cost - yscore;

            if(Math.abs(cost) < TOLERANCE) {
                correctIndex = j;
                errorCache[j] = 0;
                Cs[j] = (m-1) * C;
                ys[j] = 1;
            } else {
                errorCache[j] = scores[j];
                Cs[j] = C;
                ys[j] = -1;
            }

            if(scores[j] > bestScore) {
                bestScore = scores[j];
                bestScoreCost = cost;
            }
            if(cost < bestCost)
                bestCost = cost;            
        }
        sumBestCost += bestCost;
        sumTopCost += bestScoreCost;
        if(correctIndex == -1) {
            scores[m] = 0;

            SparseVector swap = reps[m];
            reps[m] = yrep;
            yrep = swap;            
            
            Cs[m] = m * C;
            ys[m] = 1.0;
            errorCache[m] = 0;
            m++;
        }
        smo(model);
        count++;
    }

    public void updateModel(double d, Model updatedModel) {
        for(int j = 0; j < m; j++)
            updatedModel.update(d * ys[j] * alpha[j], reps[j]);
    }

    public void roundStatistics() {
        System.out.println("Average predicted cost: " + sumTopCost / count);
        sumTopCost = 0;
        sumBestCost = 0;
        count = 0;
    }
    
    private void smo(Model model) {

        boolean examineAll = true;
        int numChanged = 0;

        for(int i = 0; i < m; i++) 
            alpha[i] = 0;
        
        /* This is the bottleneck. */
        for(int i = 0; i < m; i++) {
            model.prepareDot(reps[i]);
            KC[i][i] = model.dot(reps[i]); //model.sqrLength(reps[i]); //model.dot(reps[i]);//SparseVector.dot(reps[ii], reps[ii]);
            for(int j = i + 1; j < m; j++) {
                KC[j][i] = KC[i][j] = model.dot(reps[j]);//SparseVector.dot(reps[ii], reps[j]);
            }
        }

//        System.out.println("ybars = " + Arrays.asList(ybars));
//        System.out.println("Cs = " + Arrays.toString(Cs));
//        System.out.println("ys = " + Arrays.toString(ys));
//        System.out.println("scores = " + Arrays.toString(scores));
//        System.out.println("KC = " + Arrays.deepToString(KC));
        
        for(int i = 0; i < m; i++)
            for(int j = i + 1; j < m; j++)
                if(Arrays.equals(KC[i], KC[j])) {
                    System.out.println("Rank deficient!");
                    return;
                }
        
        while(examineAll || numChanged > 0) {
            numChanged = 0;
            if(examineAll) {
                for(int i = 0; i < m; i++) {
                    if(examineExample(i))
                        numChanged++;
                }
            } else {
                for(int i = 0; i < m; i++) {
                    if(alpha[i] > TOLERANCE && alpha[i] < C - TOLERANCE
                            && examineExample(i))
                        numChanged++;
                }
            }
            if(examineAll)
                examineAll = false;
            else if(numChanged == 0)
                examineAll = true;

        }
//        System.out.println("alpha = " + Arrays.toString(alpha));
    }

    private boolean examineExample(int i) {
        double KKT_i = 0;
        
        if(alpha[i] > TOLERANCE)
            KKT_i = Math.max(ys[i]*errorCache[i], 0);
        if(alpha[i] < Cs[i] - TOLERANCE)
            KKT_i += Math.max(-ys[i]*errorCache[i], 0);

        if(KKT_i < KKT_tolerance)
            return false;

        double err_i = errorCache[i];

        int best_j = -1;
        double largest_errdiff = -1;

        /* Heuristic 1. */
        for(int j = 0; j < m; j++) {
            if(j != i && alpha[j] > TOLERANCE && alpha[j] < Cs[j] - TOLERANCE) {
                double errdiff = Math.abs(err_i - errorCache[j]);
                if(errdiff > largest_errdiff) {
                    largest_errdiff = errdiff;
                    best_j = j;
                }
            }
        }
        if(best_j != -1 && takeStep(i, best_j)) {
            return true;
        }

        /* Heuristic 2. */
        int j_start = random.nextInt(m);
        for(int j = j_start; j < m; j++)
            if(j != i && alpha[j] > TOLERANCE && alpha[j] < Cs[j] - TOLERANCE && takeStep(i, j)) {
                return true;
            }
        for(int j = 0; j < j_start; j++)
            if(j != i && alpha[j] > TOLERANCE && alpha[j] < Cs[j] - TOLERANCE && takeStep(i, j)) {
                return true;
            }

        /* Heuristic 3. */
        j_start = random.nextInt(m);
        for(int j = j_start; j < m; j++)
            if(j != i && (alpha[j] < TOLERANCE || alpha[j] > Cs[j] - TOLERANCE) && takeStep(i, j)) {
                return true;
            }
        for(int j = 0; j < j_start; j++)
            if(j != i && (alpha[j] < TOLERANCE || alpha[j] > Cs[j] - TOLERANCE) && takeStep(i, j)) {
                return true;
            }

        return false;
    }

    private boolean takeStep(int i, int j) {
        if(i == j)
            throw new RuntimeException("Shouldn't have i == j!");

        double s = ys[i]*ys[j];
        double L, H;

        if(s == 1) {
            L = Math.max(0, alpha[i] + alpha[j] - Cs[j]);
            H = Math.min(Cs[i], alpha[i] + alpha[j]);
        } else {
            L = Math.max(0, alpha[i] - alpha[j]);
            H = Math.min(Cs[i], Cs[j] + alpha[i] - alpha[j]);
        }

        if(Math.abs(L - H) < TOLERANCE)
            return false;
        
        double khi = KC[i][i] + KC[j][j] - 2*KC[i][j];
        double errdiff = ys[i]*(errorCache[j] - errorCache[i]);

        double alphabar;

        if(khi > TOLERANCE) {
            alphabar = alpha[i] + errdiff / khi;
            alphabar = Math.min(Math.max(alphabar, L), H);
        } else if(errdiff < 0)
            alphabar = H;
        else
            alphabar = L;

        if(Math.abs(alpha[i] - alphabar) < eps*(eps + alphabar + alpha[i]))
            return false;

        double alpha_j_old = alpha[j];
        alpha[j] += s*(alpha[i] - alphabar);
        double alpha_i_old = alpha[i];
        alpha[i] = alphabar;

        for(int k = 0; k < m; k++)
            errorCache[k] += KC[i][k] * ys[i] * (alpha[i] - alpha_i_old)
            + KC[j][k] * ys[j] * (alpha[j] - alpha_j_old);

        return true;
    }
}
