/*
 OnlineAlgorithm.java
  
 Created on Jan 26, 2008 by Richard Johansson (richard@cs.lth.se).

 $Log: OnlineAlgorithm.java,v $
 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:05  richard
 Branched.

 Revision 1.10  2007/10/19 13:30:12  richard
 Options for vector model.

 Revision 1.9  2007/10/08 12:31:34  richard
 Refactored again to remove atoms.

 Revision 1.8.2.1  2007/10/08 12:19:36  richard
 Slowdown.

 Revision 1.8  2007/09/28 14:42:28  richard
 Refactored to atom-based.

 Revision 1.7  2007/04/05 13:33:21  richard
 Offset. printDots.

 Revision 1.6  2007/03/12 12:25:36  richard
 More sensible default.

 Revision 1.5  2007/03/09 11:25:18  richard
 Configurable threshold.

 Revision 1.4  2007/02/20 13:43:13  richard
 Changed limits.

 Revision 1.3  2007/02/20 10:30:11  richard
 Added prune.

 Revision 1.2  2007/02/16 15:16:02  richard
 Round statistics.

 Revision 1.1  2007/02/15 16:13:02  richard
 Added the file.

   
 */
package se.lth.cs.nlp.nlputils.ml_long.algorithms;

import java.lang.reflect.*;
import java.util.*;

import se.lth.cs.nlp.nlputils.ml_long.*;
import se.lth.cs.nlp.nlputils.core.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class OnlineAlgorithm<I, V> extends LearningAlgorithm<ProblemDefinition <I, V>, I, V> {

    private static final String DEFAULT_PATH = "se.lth.cs.nlp.nlputils.ml_long.algorithms.";
    
    private int nRounds = 1;
    private boolean average = true;
    private boolean slowdown = false;
    
    private boolean shuffle = true;
    
    private boolean useQuadratic = false;
    private boolean useNewLinear = false;
    private int modelSize = -1;
    
    private double pruneThreshold = 1.0;
    
    private int suffixRounds = -1;

    private int printDots = 1;
    
    private OnlineUpdater<I, V> updater;
    
    private boolean quiet = false;
    
    private double corruptionRate = -1.0;
    private int featureCorruptionLogProb = -1;
    private boolean antagonistic = false;

    public OnlineAlgorithm(String[] argv, int offset) {
        for(int i = offset; i < argv.length; i++) {
            if(argv[i].equals("nRounds"))
                nRounds = Integer.parseInt(argv[++i]);
            else if(argv[i].equals("average")) {
                average = Boolean.parseBoolean(argv[++i]);
                if(average)
                    slowdown = false;
            } else if(argv[i].equals("slowdown")) {
                slowdown = Boolean.parseBoolean(argv[++i]);
                if(slowdown)
                    average = false;
            } else if(argv[i].equals("suffixRounds"))
                suffixRounds = Integer.parseInt(argv[++i]);
            else if(argv[i].equals("printDots"))
                printDots = Integer.parseInt(argv[++i]);
            else if(argv[i].equals("modelSize")) {
                modelSize = Integer.parseInt(argv[++i]);
                useQuadratic = false;
            } else if(argv[i].equals("useQuadratic"))
                useQuadratic = Boolean.parseBoolean(argv[++i]);
            else if(argv[i].equals("shuffle"))
                shuffle = Boolean.parseBoolean(argv[++i]);
            else if(argv[i].equals("useNewLinear"))
                useNewLinear = Boolean.parseBoolean(argv[++i]);
            else if(argv[i].equals("quiet"))
                quiet = Boolean.parseBoolean(argv[++i]);
            else if(argv[i].equals("pruneThreshold"))
                pruneThreshold = Double.parseDouble(argv[++i]);
            else if(argv[i].equals("corruptionRate"))
		corruptionRate = Double.parseDouble(argv[++i]);
            else if(argv[i].equals("featureCorruptionLogProb"))
		featureCorruptionLogProb = Integer.parseInt(argv[++i]);
            else if(argv[i].equals("antagonistic"))
                antagonistic = Boolean.parseBoolean(argv[++i]);
	    else {
                String name = argv[i];
                try {
                    Class cl;
                    try {
                        cl = Class.forName(name);
                    } catch(ClassNotFoundException e) {
                        cl = Class.forName(DEFAULT_PATH + name);
                    }
                    if(!quiet)
                    	System.out.println("Loaded updater class (" + cl.getName() + ").");
                    Constructor con = cl.getConstructor(new Class[] { String[].class, int.class });
                    updater = (OnlineUpdater<I, V>) con.newInstance(new Object[] {argv, i + 1});
                }
                catch(ClassNotFoundException e) {
                    throw new RuntimeException("Couldn't load class " + name);
                }
                catch(InvocationTargetException e) {
                    throw new RuntimeException("Illegal arguments to " + name + ": " + e.getTargetException().getMessage());
                }
                catch(NoSuchMethodException e) {
                    throw new RuntimeException("The class " + name + " does not have a string array constructor.");
                }
                catch(Exception e) {
                    e.printStackTrace();
                    throw new RuntimeException("Couldn't create!"); 
                }
                break;
            }
        }
        if(suffixRounds > nRounds || suffixRounds < -1)
            throw new IllegalArgumentException("Invalid suffixRounds.");
        if(suffixRounds == -1)
            suffixRounds = nRounds;
    }

    public Classifier<I, V> train(ProblemDefinition<I, V> spec, List<Pair<I, V>> trainingSet) {            

        /* TODO these models should go out */
        Model model, avg;
    	if(useQuadratic) {
            model = null; //new QuadraticModel();
            avg = null; //new QuadraticModel();
            throw new RuntimeException("Quadratic model unimplemented!");
    	} else if(featureCorruptionLogProb > 0) {
		// NB: only avg and stddev!
		model = new LinearModelCorrupted(modelSize);
		avg = new LinearModelCorrupted(modelSize);
	} else if(useNewLinear) { 
            model = new LinearModel2(modelSize);
            avg = new LinearModel2(modelSize);
        } else if(modelSize > 0) {
            model = new LinearVectorModel(modelSize);
            avg = new LinearVectorModel(modelSize);
        } else {
            model = new LinearModel();
            avg = new LinearModel();
    	}
        
        int len = trainingSet.size();

        int nSteps = nRounds * len;
        int suffixStart = len * (nRounds - suffixRounds) + 1;
        int suffixLength = suffixRounds * len;

        Runtime runtime = Runtime.getRuntime();
        
        Random random = new Random(0);
        
        updater.roundStatistics(); // to initialize statistics
        
        for(int i = 1; i <= nRounds; i++) {
        	if(!quiet)
        		System.out.println("Iteration " + i + ".");

            int count = 0;
            for(Pair<I, V> ex: trainingSet) {
                count++;
                if(!quiet && count % printDots == 0) {
                    System.out.print(".");
                    System.out.flush();
                }

		if(featureCorruptionLogProb > 0) {
			FeatureFilter hidden = null;
			if(antagonistic)
				hidden = new AntagonisticFeatureFilter(random,
								       featureCorruptionLogProb,
								       (LinearModelCorrupted) model);
			else
				hidden = new RandomFeatureFilter(random, featureCorruptionLogProb);
			((LinearModelCorrupted) model).hide(hidden);
			if(average)
				((LinearModelCorrupted) avg).hide(hidden);
		}

		Object backup = null;
		if(corruptionRate > 0)
			backup = spec.corruptInputUniform(ex.left,
							  random,
							  corruptionRate);

                updater.calculateUpdate(spec, model, ex.left, ex.right);

                if(!slowdown)
                    updater.updateModel(1.0, model);
                else {
                    int updateNbr = len*(i-1)+(count);
                    double updateWeight = 1.0 / (1.0 + (double) updateNbr / len);
                    updater.updateModel(updateWeight, model);
                }

                if(average) {
                    int updateNbr = len*(i-1)+(count);
                    double updateWeight = updateNbr <= suffixStart? 
                            1.0:
                            ((double) (nSteps - updateNbr + 1) / suffixLength);
                    updater.updateModel(updateWeight, avg);
                }

		if(corruptionRate > 0)
			spec.decorruptInput(ex.left, backup);

                long fm = runtime.freeMemory();
                double megs = (double) fm / (1 << 20);
                
                if(megs < pruneThreshold) {
                    runtime.gc();
                    
                    fm = runtime.freeMemory();
                    megs = (double) fm / (1 << 20);
                    if(megs < 2.0*pruneThreshold) {
                        model.prune();
                        if(average)
                            avg.prune();
                        runtime.gc();
                    }
                }
                
            }
            if(!quiet)
            	System.out.println();
            if(!quiet)
            	updater.roundStatistics();

            if(shuffle && i < nRounds)
                Collections.shuffle(trainingSet, random);
        }

        if(average)
            model = avg;

	if(featureCorruptionLogProb > 0)
		((LinearModelCorrupted) model).hide(null);

        return spec.createClassifier(model);
    }
    
}
