/*
 OnlineEG.java
  
 Created on Oct 8, 2007 by Richard Johansson (richard@cs.lth.se).

 $Log: OnlineEG.java,v $
 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:05  richard
 Branched.

 Revision 1.3  2007/10/26 13:26:46  richard
 Renamed.

 Revision 1.1  2007/10/23 12:53:34  richard
 Added.

   
 */
package se.lth.cs.nlp.nlputils.ml_long.algorithms;

import java.util.*;

import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.ml_long.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class OnlineEG<I, V> extends LearningAlgorithm<ProblemDefinition <I, V>, I, V> {

    private int nRounds = 1;
    private double C = 1;
    private double eta = 1;
    private boolean maxMargin = false;
    
    private int linearVectorSize = -1;
    
    public OnlineEG(String[] argv, int offset) {
        for(int i = offset; i < argv.length; i++) {
            if(argv[i].equals("nRounds"))
                nRounds = Integer.parseInt(argv[++i]);
            else if(argv[i].equals("C"))
                C = Double.parseDouble(argv[++i]);
            else if(argv[i].equals("eta"))
                eta = Double.parseDouble(argv[++i]);
            else if(argv[i].equals("maxMargin"))
                maxMargin = Boolean.parseBoolean(argv[++i]);
            else if(argv[i].equals("linearVectorSize"))
                linearVectorSize = Integer.parseInt(argv[++i]);
            else
                throw new RuntimeException("Unknown argument: " + argv[i]);
        }
    }

    static final double MIN_IMPROVEMENT = 1e-20;
    static final int MAX_NBACKOFFS = 5;
    
    static final double ACCEL = 1.05;
    static final double BRAKE = 0.7;
    
    public Classifier<I, V> train(ProblemDefinition<I, V> spec, List<Pair<I, V>> trainingSet) {            

        Model model;
        if(linearVectorSize > 0) {
            model = new LinearVectorModel(linearVectorSize);
        } else {
            model = new LinearModel();
        }

        int N = trainingSet.size();
        double Cinv = 1.0 / C;
        
        double[][] cachedProbs = new double[N][];
        double[] dObjs = new double[N];
        double[] etas = new double[N];

        int i = 0;
        for(Pair<I, V> ex: trainingSet) {
            SparseVector sv = new SparseVector();
            spec.encode(ex.left, ex.right, sv);
            model.update(1.0, sv);
            
            Decomposition decomp = spec.decompose(ex.left, ex.right);
            double[] probs = new double[decomp.size()];

            double dObj = maxMargin? 0.0: -decomp.getLogPartition();
            
            int j = 0;
            for(DecompositionPart p: decomp) {
                double marginal = p.getMarginalProbability();
                model.update(-marginal, p.encode());
                probs[j] = marginal;
                if(maxMargin)
                    dObj -= marginal * p.getCost();
                else
                    dObj += marginal * p.getTheta();
                j++;
            }
            cachedProbs[i] = probs;
            etas[i] = eta;
            dObjs[i] = dObj;
            i++;
        }

        Random rand = new Random(0);

        for(int round = 0; round < nRounds; round++) {

            double sumEta = 0.0;
            int countEta = 0;
            double roundImprovement = 0.0;
            
            i = 0;
            for(Pair<I, V> ex: trainingSet) {

                Decomposition parts = spec.decompose(ex.left, ex.right);

                double[] oldProbs = cachedProbs[i];
                double[] scores = new double[parts.size()];
                double[] oldThetas = new double[parts.size()];
                double eta_i = etas[i];
                
                int j = 0;
                for(DecompositionPart p: parts) {
                    scores[j] = model.score(p.encode());
                    oldThetas[j] = p.getTheta();
                    j++;
                }

                int countBackoffs = 0;

                while(true) { /* Until there is an improvement in the dual objective. */
                    j = 0;
                    for (DecompositionPart p : parts) {
                        double d = scores[j];
                        double th = oldThetas[j];
                        if (maxMargin)
                            th -= eta_i * (-p.getCost() - d * Cinv);
                        else
                            th -= eta_i * (th - d * Cinv);
                        p.setTheta(th);
                        j++;
                    }

                    SparseVector dw = new SparseVector();

                    double newDObj = maxMargin? 0.0: -parts.getLogPartition();                    
                    j = 0;
                    for (DecompositionPart p : parts) {
                        double newProb = p.getMarginalProbability();
                        dw.putAll(p.encode(), oldProbs[j] - newProb);
                        if(maxMargin)
                            newDObj -= newProb * p.getCost();
                        else
                            newDObj += newProb * p.getTheta();
                        j++;
                    }
                    
                    double diff = SparseVector.sqrLength(dw) + model.score(dw);
                    double improvement = (dObjs[i] - newDObj) - 0.5*diff*Cinv;

                    if(improvement < MIN_IMPROVEMENT) {
                        countBackoffs++;                        
                        if(countBackoffs > MAX_NBACKOFFS) {
                            j = 0;
                            for(DecompositionPart p: parts)
                                p.setTheta(oldThetas[j++]);
                            etas[i] *= BRAKE;
                            break;                            
                        }
                        eta_i *= 0.5;
                    } else {
                        model.update(1.0, dw);                        
                        etas[i] = eta_i*ACCEL;
                        if(!maxMargin && etas[i] > 1.0)
                            etas[i] = 1.0;                        
                        dObjs[i] = newDObj;
                        j = 0;
                        for(DecompositionPart p: parts)
                            oldProbs[j++] = p.getMarginalProbability();
                        sumEta += eta_i;
                        countEta++;
                        roundImprovement += improvement;

                        break;
                    }
                }
                i++;
            }
            
            for(i = 0; i < trainingSet.size(); i++) {
                int ix = rand.nextInt(trainingSet.size());
                CollectionUtils.swap(etas, i, ix);
                CollectionUtils.swap(dObjs, i, ix);
                CollectionUtils.swap(cachedProbs, i, ix);
                Collections.swap(trainingSet, i, ix);
            }

            System.out.println(round + ". roundImprovement = " + roundImprovement + ", avgEta = " + sumEta / N + ", countEta = " + countEta);

            if(countEta == 0)
                break;
            
        }
        System.out.println();
        
        return spec.createClassifier(model);
    }
    
}
