/*
 OnlinePA.java
  
 Created on Feb 16, 2007 by Richard Johansson (richard@cs.lth.se).

 $Log: PAUpdate.java,v $
 Revision 1.3  2010-06-08 09:50:55  johansson
 More debugging.

 Revision 1.2  2010-06-07 11:30:23  johansson
 Major update.

 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.2  2008/06/02 07:11:01  richard
 Added sqrt option.

 Revision 1.1  2008/05/16 12:16:05  richard
 Branched.

 Revision 1.5  2008/02/21 13:24:15  richard
 Added debug flag.

 Revision 1.4  2007/12/10 10:06:26  richard
 Debug.

 Revision 1.3  2007/11/06 09:57:57  richard
 Tolerance.

 Revision 1.2  2007/10/19 13:20:17  richard
 Bug when not using maxLoss.

 Revision 1.1  2007/09/28 14:42:28  richard
 Refactored to atom-based.

 Revision 1.12  2007/02/16 15:15:50  richard
 Round statistics.

 Revision 1.11  2007/02/16 10:43:51  richard
 Changed to updater.

   
 */
package se.lth.cs.nlp.nlputils.ml_long.algorithms;

import se.lth.cs.nlp.nlputils.core.Util;
import se.lth.cs.nlp.nlputils.ml_long.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class PAUpdate<I, V> extends OnlineUpdater<I, V> {

    /* Learning parameters. */
    private boolean maximizeLoss = false;
    private boolean usePA2;
    private double C = Double.MAX_VALUE;

    private boolean sorted = false;
    private boolean sqrt = false;
    
    /* Private data to cache update. */
    private SparseVector yrep, yhatrep;
    private double[] result = new double[2];
    private double tau;
    
    /* Statistics. */
    private double sumCost;
    private int count;
    
    private static final boolean DEBUG = false;
    
    private static final double TOLERANCE = 1e-15;
    
    public PAUpdate(String[] argv, int offset) {
        for(int i = offset; i < argv.length; i++) {
            if(argv[i].equals("maximizeLoss"))
                maximizeLoss = Boolean.parseBoolean(argv[++i]);
            else if(argv[i].equals("usePA2"))
                usePA2 = Boolean.parseBoolean(argv[++i]);
            else if(argv[i].equals("sorted"))
                sorted = Boolean.parseBoolean(argv[++i]);
            else if(argv[i].equals("sqrt"))
                sqrt = Boolean.parseBoolean(argv[++i]);
            else if(argv[i].equals("C"))
                C = Double.parseDouble(argv[++i]);
            else
                throw new IllegalArgumentException("Unknown parameter: " + argv[i]);
        }        
        yrep = new SparseVector();
        yhatrep = new SparseVector();
    }

    private static double norm1(SparseVector sv) {
        double sum = 0;
        sv.sortIndices();
        for(int i = 0; i < sv.index; i++)
            sum += Math.abs(sv.values[i]);
        return sum;
    }
    
    public void calculateUpdate(ProblemDefinition<I, V> spec, Model model, 
            I x, V y) {

        if(DEBUG) {
            System.out.println("CalculateUpdate:");
            System.out.println("x = " + x);
            System.out.println("y = " + y);
        }
        
        spec.encode(x, y, yrep);

        if(DEBUG) {
            System.out.println("yrep = " + yrep);
            System.out.println("yrep.1n = " + norm1(yrep));
        }
        
        V yhat;
        if(maximizeLoss)
            yhat = spec.maxLoss(x, y, model, yhatrep, result);
        else
            yhat = spec.argmax(x, model, yhatrep, result);

        if(DEBUG) {
            System.out.println("yhat = " + yhat);
            System.out.println("yhatrep = " + yhatrep);
            System.out.println("yhatrep.1n = " + norm1(yhatrep));
        }
        
        double yhatscore = result[0];
        double cost      = maximizeLoss? result[1]: spec.cost(x, y, yhat);

        if(DEBUG) {
            System.out.println("yscore = " + model.score(yrep));

            double yhatscore2 = model.score(yhatrep);
            
            if(Math.abs(yhatscore - yhatscore2) > TOLERANCE) {
            	System.out.println("*** WARNING: yhatscore != yhatscore2");
            }
            
            System.out.println("yhatscore = " + yhatscore);
            System.out.println("yhatscore2 = " + yhatscore2);
        }
        
        //System.exit(0);
        
        double loss = yhatscore - model.score(yrep) + (sqrt? Math.sqrt(cost): cost);
        
        if(DEBUG) {
            System.out.println("loss = " + loss);            	
        }
        
        //System.out.println("loss = " + loss);
        if(loss > TOLERANCE) {
            double diff = model.sqrDiff(yhatrep, yrep);

            if(DEBUG) {
                System.out.println("sqrdiff = " + diff);            	
            }
            
            if(usePA2)
                tau = loss / (diff + 0.5 / C);
            else {
                tau = loss / diff;
                if(tau > C)
                    tau = C;
            }
        } else
            tau = 0;

        if(DEBUG) {
            tmpy = yrep;
            tmpyhat = yhatrep;
            System.out.println("tau = " + tau);
        }
        
        sumCost += cost;
        count++;
    }

    private SparseVector tmpy, tmpyhat;
    
    long t0;
    
    public void updateModel(double d, Model updatedModel) {
        if(tau > 0) {
            //updatedModel.update(d*tau, yrep);
            //updatedModel.update(-d*tau, yhatrep);
            updatedModel.updateDiff(d*tau, yrep, yhatrep, sorted);
        }
        if(DEBUG) {
            System.out.println("yscore_new = " + updatedModel.score(tmpy));
            System.out.println("yhatscore_new = " + updatedModel.score(tmpyhat));
        }
    }
    
    public void roundStatistics() {
        if(count > 0) {
            long t1 = System.currentTimeMillis();
            int format = Util.HOURS | Util.MINUTES | Util.SECONDS;
            System.out.println("Average cost: " + sumCost / count);           
            System.out.println("Round time: " + Util.toTimeString(t1 - t0, format));
            t0 = t1;
        } else {
            t0 = System.currentTimeMillis();
        }
        sumCost = 0;
        count = 0;
    }
    
}
