/*
 ClasswiseBinarizer.java
  
 Created on Mar 19, 2008 by Richard Johansson (richard@cs.lth.se).

 $Log: PairwiseBinarizer.java,v $
 Revision 1.1  2010-06-07 11:30:23  johansson
 Major update.

 Revision 1.2  2009/03/06 13:28:31  johansson
 Quiet.

 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:05  richard
 Branched.

 Revision 1.5  2008/04/02 13:33:43  richard
 Quiet.

 Revision 1.4  2008/04/02 11:35:53  richard
 Back-ported to 1.5.

 Revision 1.3  2008/04/02 11:20:34  richard
 New argument handling.

 Revision 1.2  2008/03/26 08:40:54  richard
 Special case for 2 classes.

 Revision 1.1  2008/03/19 15:51:08  richard
 Added.

   
 */
package se.lth.cs.nlp.nlputils.ml_long.algorithms;

import java.util.*;
import java.util.concurrent.*;

import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.ml_long.*;
import se.lth.cs.nlp.nlputils.ml_long.classifiers.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class PairwiseBinarizer<I, V> extends LearningAlgorithm<MulticlassDefinition<I, V>, I, V> {

    private AlgorithmFactory af = new AlgorithmFactory();
    private String[] argv;
    private int offset;
    
    private int nthreads = 1;
    private boolean quiet = false;
    
    public PairwiseBinarizer(String[] argv, int offset) {
        for(int i = offset; i < argv.length; i++) {
            if(argv[i].equals("nThreads"))
                nthreads = Integer.parseInt(argv[++i]);
            else if(argv[i].equals("quiet"))
                quiet = Boolean.parseBoolean(argv[++i]);
            else{
                this.argv = argv;        
                this.offset = i;
                break;
            }
        }
        if(quiet)
            AlgorithmFactory.setVerbosity(0);
    }

    public Classifier<I, V> train(MulticlassDefinition<I, V> spec, List<Pair<I, V>> trainingSet) { 

    	HashSet<V> classes = new HashSet();
        ArrayList<Pair<SparseVector, Integer>> encoded = new ArrayList();
        for(Pair<I, V> p: trainingSet) {
            SparseVector sv = new SparseVector();
            spec.encodeInstance(p.left, sv);
            encoded.add(new Pair(sv, spec.encodeClass(p.right)));
            classes.add(p.right);
        }
                
        V[] clArray = (V[]) classes.toArray();
 
        int m = clArray.length;
        int n = m * (m - 1) / 2;        
        BinaryNumericClassifier[] classifiers = new BinaryNumericClassifier[n];
 
        if(!quiet)
        	System.out.println("ClasswiseBinarizer: number of classes is " + classes.size() + ": " + classes);
        
        if(clArray.length <= 2) {
            encoded = null;
            LearningAlgorithm alg = af.create(argv[offset], argv, offset + 1);
            return alg.train(spec, trainingSet);
        } else {
            ExecutorService threadPool = Executors.newFixedThreadPool(nthreads);

            Future<Classifier>[] results = new Future[n];
            int k = 0;
            for(int i = 0; i < m; i++) {
            	for(int j = i + 1; j < m; j++) {
            		int encCl1 = spec.encodeClass(clArray[i]);
            		int encCl2 = spec.encodeClass(clArray[j]);
            		results[k] = threadPool.submit(new TrainSubclassifierJob(encoded, clArray[i], clArray[j], 
            				encCl1, encCl2));
            		k++;
            	}
            }

            k = 0;
            for(int i = 0; i < m; i++) {
                for(int j = i + 1; j < m; j++) {
                	try {
                		Classifier cl = results[k].get();
                		if(!(cl instanceof BinaryNumericClassifier))
                			throw new RuntimeException("Illegal classifier type: " + cl.getClass().getName());
                		classifiers[k] = (BinaryNumericClassifier) cl;
                	} catch(Exception e) {
                		threadPool.shutdownNow();
                		throw new RuntimeException(e);
                	}
                	k++;
                }
            }
            threadPool.shutdown();
            try {
                threadPool.awaitTermination(10, TimeUnit.SECONDS);
            } catch(Exception e) {
                e.printStackTrace();
            }
            return new PairwiseClassifier(spec, classifiers, clArray);
        }
    }
    
    private class TrainSubclassifierJob implements Callable<Classifier> {
        private ArrayList<Pair<SparseVector, Integer>> ts;
        private V label1;
        private V label2;
        private int encodedLabel1, encodedLabel2;
        TrainSubclassifierJob(ArrayList<Pair<SparseVector, Integer>> ts,
                V label1, V label2, int encodedLabel1, int encodedLabel2) {
            this.ts = ts;
            this.label1 = label1;
            this.label2 = label2;
            this.encodedLabel1 = encodedLabel1;
            this.encodedLabel2 = encodedLabel2;
        }
        public Classifier call() {
            ArrayList<Pair<SparseVector, Integer>> enc2 = new ArrayList();
            int countPositive = 0;
            int countNegative = 0;
            for(Pair<SparseVector, Integer> p: ts)
                if(p.right.equals(encodedLabel1)) {
                    enc2.add(new Pair(p.left, 1));
                    countPositive++;
                } else if(p.right.equals(encodedLabel2)) {
                    enc2.add(new Pair(p.left, 2));
                    countNegative++;
                }

            if(!quiet)
                System.out.println("Training " + label1 + " vs " + label2 + " (" 
                        + countPositive + "/" + countNegative + ")");
            Collections.sort(enc2, Pair.BY_RIGHT);
            EncodedMulticlassDefinition subSpec = new EncodedMulticlassDefinition();
            LearningAlgorithm alg = af.create(argv[offset], argv, offset + 1);
            return alg.train(subSpec, enc2);
        }
    }
    
}
