/*
 PrefBinarizer.java
  
 Created on May 13, 2010 by Richard Johansson (johansson@disi.unitn.it).

 $Log: PrefBinarizer.java,v $
 Revision 1.1  2010-06-07 11:30:23  johansson
 Major update.


 */
package se.lth.cs.nlp.nlputils.ml_long.algorithms;

import java.util.*;

import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.ml_long.*;
import se.lth.cs.nlp.nlputils.ml_long.classifiers.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class PrefBinarizer extends LearningAlgorithm {

    private AlgorithmFactory af = new AlgorithmFactory();
    private String[] argv;
    private int offset;
    
    public PrefBinarizer(String[] argv, int offset) {
    	this.argv = argv;        
    	this.offset = 0;
    }

    private static int LEFT = 1;
    private static int RIGHT = 2;
    
    public Classifier train(ProblemDefinition spec, List trainingSet) { 
    	
    	if(!(spec instanceof MulticlassDefinition))
    		throw new RuntimeException("unsupported problem type");
    	
    	boolean isRestricted;
    	
    	{
    		Pair p = (Pair) trainingSet.get(0);
    		isRestricted = p.left instanceof RestrictedInstance;    	
    	}

    	Object[] fullRange;
    	{
    		HashSet rs = new HashSet();
    		for(Object o: trainingSet) {
    			Pair p = (Pair) o;
    			rs.add(p.right);
    		}
    		fullRange = rs.toArray(new Object[0]);
    	}
        
        if(fullRange.length <= 2) {
            LearningAlgorithm alg = af.create(argv[offset], argv, offset + 1);
            return alg.train(spec, trainingSet);
        }
    	
    	SparseVectorClassEncoder enc = new SparseVectorClassEncoder();
    	IntPairEncoder penc = new IntPairEncoder();
    	
    	ArrayList<Pair<SparseVector, Integer>> encoded = new ArrayList();

    	System.out.println("Binarizing training set, size = " + trainingSet.size());
    	
    	int count = 0;
    	
        for(Object o: trainingSet) {
        	Pair p = (Pair) o;        	
        	
        	int corrEnc = spec.encodeClass(p.right);
        	int corrLeftEnc = penc.encode(LEFT, corrEnc);
        	int corrRightEnc = penc.encode(RIGHT, corrEnc);
        	
            SparseVector svInstance = new SparseVector();
            spec.encodeInstance(p.left, svInstance);
            
            Object[] range;
            if(isRestricted)
            	range = ((RestrictedInstance) p.left).range;
            else
            	range = fullRange;

            for(Object inc: range) {
            	if(inc.equals(p.right))
            		continue;
            	int incEnc = spec.encodeClass(inc);
            	int incLeftEnc = penc.encode(LEFT, incEnc);
            	int incRightEnc = penc.encode(RIGHT, incEnc); 

            	SparseVector svCorrLeft = new SparseVector();
            	enc.encode(svInstance, corrLeftEnc, svCorrLeft);
            	enc.encode(svInstance, incRightEnc, svCorrLeft);
            	svCorrLeft.sortIndices();
            	svCorrLeft.trim();            	
                encoded.add(new Pair(svCorrLeft, 1));
                
            	SparseVector svCorrRight = new SparseVector();
            	enc.encode(svInstance, incLeftEnc, svCorrRight);
            	enc.encode(svInstance, corrRightEnc, svCorrRight);
            	svCorrRight.sortIndices();
            	svCorrRight.trim();            	
                encoded.add(new Pair(svCorrRight, -1));                
            }

            count++;
            if(count % 1000 == 0) {
            	System.out.println("count = " + count + ", enc ts size = "+ encoded.size() 
            			+ " sv enc size = " + enc.size());
            }
        }
 
        penc.freeze();
        enc.freeze();
        
        EncodedMulticlassDefinition subSpec = new EncodedMulticlassDefinition();        
        System.out.println("Training preference-binarized...");       
        LearningAlgorithm alg = af.create(argv[offset], argv, offset + 1);        
        BinaryNumericClassifier cl = (BinaryNumericClassifier) alg.train(subSpec, encoded);
        
        return new PreferenceClassifier(cl, fullRange, isRestricted, enc, penc, 
        		(MulticlassDefinition) spec);
    }
    
    static class PreferenceClassifier<I,V> extends Classifier<I,V> {
        private static final long serialVersionUID = 0L;

        private BinaryNumericClassifier cl;
        private V[] fullRange;
        private boolean isRestricted;
        
        private SparseVectorClassEncoder enc;
        private IntPairEncoder penc;
    	private MulticlassDefinition def;
    	
    	private SparseVector rep = new SparseVector();
    	
    	PreferenceClassifier(BinaryNumericClassifier cl, V[] fullRange,
    			boolean isRestricted, SparseVectorClassEncoder enc,
    			IntPairEncoder penc, MulticlassDefinition def) {
    		this.cl = cl;
    		this.fullRange = fullRange;
    		this.isRestricted = isRestricted;
    		this.enc = enc;
    		this.penc = penc;
    		this.def = def;    		
    	}
    	
        private V apply(SparseVector svInstance, V[] range) {
        	V bestValue = null;
        	double maxScore = Double.NEGATIVE_INFINITY;
        	
        	for(V value: range) {
        		double score = 0;
        		
        		int valEnc = def.encodeClass(value);
        		int valLeftEnc = penc.encode(LEFT, valEnc);
            	int valRightEnc = penc.encode(RIGHT, valEnc);

            	SparseVector sv = new SparseVector();
            	enc.encode(svInstance, valLeftEnc, sv);
            	sv.sortIndices();
            	score += cl.apply(sv);

            	sv.clear();
            	enc.encode(svInstance, valRightEnc, sv);
            	sv.sortIndices();
            	score -= cl.apply(sv);

            	if(score > maxScore) {
            		bestValue = value;
            		maxScore = score;
            	}
        	}
        	return bestValue;
        }
        
        public V classify(I instance) {
        	V[] range;
        	if(isRestricted)
        		range = (V[]) ((RestrictedInstance) instance).range;
        	else
        		range = fullRange;
        		
        	def.encodeInstance(instance, rep);
            return apply(rep, range);
        }
        
        public V classifyByFeatures(FeatureList features) {
            throw new UnsupportedOperationException("Unimplemented");
        }
        
        public V classifyRestricted(I instance, Collection<V> classes) {
            throw new UnsupportedOperationException("Unimplemented");
        }
        
        public V classifyByFeaturesRestricted(FeatureList features, Collection<V> classes) {
            throw new UnsupportedOperationException("Unimplemented");
        }
        
        public void computeScores(I instance, double[] scores) {
            throw new UnsupportedOperationException("Unrestricted scoring not supported.");
        }
        
        public void computeScoresRestricted(I instance, Collection<V> classes, double[] scores) {
            throw new UnsupportedOperationException("Unimplemented");
        }
        
        public void computeScoresByFeatures(FeatureList features, double[] scores) {
            throw new UnsupportedOperationException("Unrestricted scoring not supported.");        
        }

        public void computeScoresByFeaturesRestricted(FeatureList features, Collection<V> classes, double[] scores) {
            throw new UnsupportedOperationException("Unimplemented");
        }
        
    }
    
    
}
