/*
 SPV.java
  
 Created on Feb 16, 2007 by Richard Johansson (richard@cs.lth.se).

 $Log: SPVUpdate.java,v $
 Revision 1.3  2010-06-08 10:01:29  johansson
 Array bug.

 Revision 1.2  2010-06-07 11:30:23  johansson
 Major update.

 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:05  richard
 Branched.

 Revision 1.1  2007/09/28 14:42:28  richard
 Refactored to atom-based.

 Revision 1.2  2007/02/16 15:16:02  richard
 Round statistics.

 Revision 1.1  2007/02/16 13:02:58  richard
 Added the file.

   
 */
package se.lth.cs.nlp.nlputils.ml_long.algorithms;

import java.lang.reflect.Array;
import java.util.Arrays;
import se.lth.cs.nlp.nlputils.ml_long.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class SPVUpdate<I, V> extends OnlineUpdater<I, V> {

    private static final double TOLERANCE = 1e-6;
    
    private static final double MAX_DALPHA_INFNORM = 1e5;
    private static final double MAX_ALPHA_INFNORM = 1e9;
    private static final double MAX_EXPSCORE = 1e10;
    private static final double MAX_STEP = 1e2;

    private static final double MIN_STEPLENGTH = 1e-15;
    private static final double MIN_IMPROVEMENT = 1e-15;
    
    /* Learning parameters. */
    private int N = 4;
    private double gamma = 1;
    private boolean misclassCost = false;    
    private boolean sqrt = false;
    private boolean useLaplace = false;

    private boolean verbose = false;
    
    private V[] ybars;
    private double[] scores;
    private SparseVector[] reps;
    private SparseVector yrep = new SparseVector();
    
    /* Statistics. */
    private double sumTopCost;
    private double sumBestCost;
    private int count;
    
    public SPVUpdate(String[] argv, int offset) {
        for(int i = offset; i < argv.length; i++) {
            if(argv[i].equals("N"))
                N = Integer.parseInt(argv[++i]);
            else if(argv[i].equals("gamma"))
                gamma = Double.parseDouble(argv[++i]);
            else if(argv[i].equals("useLaplace"))
                useLaplace = Boolean.parseBoolean(argv[++i]);
            else if(argv[i].equals("verbose"))
                verbose = Boolean.parseBoolean(argv[++i]);
            else if(argv[i].equals("sqrt"))
                sqrt = Boolean.parseBoolean(argv[++i]);
            else if(argv[i].equals("misclassCost"))
                misclassCost = Boolean.parseBoolean(argv[++i]);
            else
                throw new IllegalArgumentException("Unknown parameter: " + argv[i]);
        }
        init();
    }

    private void init() {
        alpha = new double[N + 1];
        scores = new double[N + 1];
        reps = new SparseVector[N + 1];
        KC = new double[N + 1][N + 1];

        expScores = new double[N + 1];        
        dalpha = new double[N + 1];
        s = new double[N + 1];
        costs = new double[N + 1];
        for(int i = 0; i < N + 1; i++)
            reps[i] = new SparseVector();
    }

    public void calculateUpdate(ProblemDefinition<I, V> spec, Model model, 
            I x, V y) {

        spec.encode(x, y, yrep);
        double yscore = model.score(yrep);

        if(ybars == null) {
    		ybars = (V[]) Array.newInstance(y.getClass(), N + 1);
    	}
        
        m = spec.findNBest(x, model, N, ybars, reps, scores);

        if(verbose) {
            System.out.println("scores = " + Arrays.toString(scores));
        }

        double bestCost = Double.MAX_VALUE;
        double bestScore = Double.MIN_VALUE;
        double bestScoreCost = 0;
        
        int correctIndex = -1;
        for(int j = 0; j < m; j++) {
            double cost = spec.cost(x, y, ybars[j]);
            if(Math.abs(cost) < TOLERANCE) {
                if(correctIndex == -1)
                    correctIndex = j;
                else {
                    //scores[correctIndex] = 0;
                    scores[correctIndex] = yscore;
                    reps[correctIndex] = yrep.copy();
                    ybars[correctIndex] = y;

                    System.out.println("\nWarning! Multiple zero-cost predictions.");

                    for(int k = j+1; k < m; k++) {
                        reps[k-1] = reps[k];
                        scores[k-1] = scores[k];
                        ybars[k-1] = ybars[k];
                    }
                    m--;
                    j--;
                    continue;
                }
            }

            costs[j] = cost;
            
            if(scores[j] > bestScore) {
                bestScore = scores[j];
                bestScoreCost = cost;
            }
            if(cost < bestCost)
                bestCost = cost;

        }
        if(correctIndex == -1) {
            //scores[m] = 0;
            scores[m] = yscore;
            SparseVector swap = reps[m];
            reps[m] = yrep;
            yrep = swap;
            correctIndex = m;
            costs[m] = 0;
            m++;
        }

        map(model, correctIndex);
        
        sumBestCost += bestCost;
        sumTopCost += bestScoreCost;

        count++;
    }

    public void updateModel(double d, Model updatedModel) {
        for(int j = 0; j < m; j++)
            updatedModel.update(d * alpha[j], reps[j]);
    }

    public void roundStatistics() {
        System.out.println("Average predicted cost: " + sumTopCost / count);
        sumTopCost = 0;
        sumBestCost = 0;
        count = 0;
    }

    
    private double startStep = -0.0001;
    
    private double[] alpha;
    private int m;
    private double[][] KC;
    private double[] costs;
    
    private void map(Model model, int correctIndex) {

        if(verbose)
            System.out.println("Entering map.");
        
        for(int i = 0; i < m; i++) {
            double loss = misclassCost? Math.max(0.0, scores[i] - scores[correctIndex]): 0;
            loss += sqrt? Math.sqrt(costs[i]): costs[i];
            s[i] = gamma / m / m / loss;
        }
        for(int i = 0; i < m; i++) {
            alpha[i] = 0;
            dalpha[i] = 0;
        }

        if(verbose) {
            System.out.println("m = " + m);
            System.out.println("scores = " + Arrays.toString(scores));
            System.out.println("costs = " + Arrays.toString(costs));
            System.out.println("s = " + Arrays.toString(s));

        }
        
        for(int i = 0; i < m; i++) {
            //reps[i].prepareDot();
            model.prepareDot(reps[i]);
            KC[i][i] = model.sqrLength(reps[i]); //model.dot(reps[i]);//SparseVector.dot(reps[ii], reps[ii]);
            for(int j = i + 1; j < m; j++) {
                KC[j][i] = KC[i][j] = model.dot(reps[j]);//SparseVector.dot(reps[ii], reps[j]);
            }
        }

        /* Compute intermediate scores. */
        computeScores(correctIndex);
        /* Compute the log posterior. */
        double old_lp = computeLogPosterior(correctIndex);

        if(verbose) {
            System.out.println("KC = " + Arrays.deepToString(KC));
            System.out.println("original lp = " + old_lp);
        }
            
        while(true) { /* Until the log posterior has been maximized. */
            if(verbose) {
                System.out.println("Loop:");
                System.out.println("old_lp = " + old_lp);
                System.out.println("alpha = " + Arrays.toString(alpha));
                double sss = 0;
                for(int jjj = 0; jjj < m; jjj++)
                     sss += alpha[jjj];
                System.out.println("sum(alpha) = " + sss);
            }
            
            /* Compute partial derivatives of the log posterior. */
            computeDerivatives(correctIndex);

            for(int i = 0; i < m; i++) {
                if(Double.isInfinite(dalpha[i]))
                    throw new RuntimeException("Numerical problems: dalpha is infinite");
                if(Double.isNaN(dalpha[i]))
                    throw new RuntimeException("Numerical problems: dalpha is NaN");
            }
            
            if(verbose)
                System.out.println("dalpha = " + Arrays.toString(dalpha));
            
            normalizeDerivatives();

            if(verbose)
                System.out.println("dalpha = " + Arrays.toString(dalpha));

            if(verbose)
                System.out.println("Before: alpha = " + Arrays.toString(alpha));
            
            /* Uppdate alpha. */
            
            double steplength = startStep;
            add_alpha(-startStep);

            if(verbose)
                System.out.println("After: alpha = " + Arrays.toString(alpha));
            
            /* Compute a new log posterior. */
            computeScores(correctIndex);
            double lp = computeLogPosterior(correctIndex);
            double improvement = lp - old_lp;

            if(verbose) {
                System.out.println("1. lp = " + lp);
                System.out.println("1. improvement = " + improvement);
            }
            
            while((Double.isInfinite(improvement) || improvement < MIN_IMPROVEMENT) && steplength < -MIN_STEPLENGTH) {
                steplength *= 0.5;

                /* Compute a new alpha. */                    
                add_alpha(steplength);

                /* Once again compute the log posterior. */
                computeScores(correctIndex);
                lp = computeLogPosterior(correctIndex);
                improvement = lp - old_lp;

                if(verbose) {
                    System.out.println("2. lp = " + lp);
                    System.out.println("2. improvement = " + improvement);
                }
            }

            old_lp = lp;

            double maxalpha = 0;
            for(int j = 0; j < m; j++) {
                double a = Math.abs(alpha[j]);
                if(a > maxalpha)
                    maxalpha = a;
            }

            if(maxalpha > MAX_ALPHA_INFNORM) {
                break;
            }
            
            //if(steplength >= -MIN_STEPLENGTH) /* Log posterior maximized. */
            //  break;
            if(improvement < MIN_IMPROVEMENT)
                break;
            else {
                startStep = 4*steplength < -MAX_STEP ? -MAX_STEP: 4*steplength;
            }
        }

        if(verbose) {
            System.out.println("(correctIndex = " + correctIndex + ")");
            System.out.println("converged, alpha = " + Arrays.toString(alpha));

            double an = 0;
            for(int i = 0; i < m; i++)
                an += dalpha[i]*dalpha[i];
            System.out.println("dalpha = " + Arrays.toString(dalpha));
            System.out.println("norm(dalpha) = " + an);
            
        }
        
        double sum = 0;
        for(int j = 0; j < m; j++) {
             sum += alpha[j];
             alpha[j] = -alpha[j];
        }
        alpha[correctIndex] = sum;
        
        if(verbose) {
            System.out.println("converged, alpha = " + Arrays.toString(alpha));
            System.out.println("sum(alpha) = " + sum);

            System.exit(0);
        }

    }

    private double[] expScores;
    private double[] dalpha;
    private double[] s;
    
    private double expSum;

    private void computeScores(int correct) {
        expSum = 0;
        for(int k = 0; k < m; k++) {
            double sum = scores[k];
            for(int j = 0; j < m; j++)
                if(j != correct)
                    sum += alpha[j] * (KC[correct][k] - KC[j][k]);
            expScores[k] = Math.exp(sum);

            if(expScores[k] > MAX_EXPSCORE)
                expScores[k] = MAX_EXPSCORE;
            
            expSum += expScores[k];
        }
    }

    private double computeLogPosterior(int correct) {
        /* We ignore the constants. */

        double part1 = Math.log(expScores[correct]);
        double part2 = -Math.log(expSum);
        
        double part3 = 0;

        for(int i = 0; i < m; i++) {
            if(i != correct)
                part3 -= useLaplace? s[i]*Math.abs(alpha[i]): s[i]*alpha[i]*alpha[i];
        }
        
        if(verbose) {
            System.out.println("computeLogPosterior:");
            System.out.println("part1 = " + part1);   
            System.out.println("part2 = " + part2);   
            System.out.println("part3 = " + part3);   
        }
        
        return Double.isInfinite(part2)? part2: part1 + part2 + part3;
    }
    
    private void computeDerivatives(int correct) {
        if(verbose) {
            System.out.println("In computeDerivatives.");
            System.out.println("expSum = " + expSum);
            System.out.println("expScores = " + Arrays.toString(expScores));
        }
        
        /* First derivatives: */
        for(int i = 0; i < m; i++) {
            if(i == correct)
                continue;
            double diffSum = 0;
            for(int k = 0; k < m; k++)
                diffSum += (KC[correct][k] - KC[i][k]) * expScores[k];
            dalpha[i] = KC[correct][correct] - KC[correct][i]
                      - diffSum/expSum - (useLaplace? s[i]*Math.signum(alpha[i]): 2*alpha[i]*s[i]);

            if(verbose) {
                System.out.println("delta = " + (KC[correct][correct] - KC[correct][i]));
                System.out.println("diffSum/expSum = " + diffSum/expSum);
            }
        }

        for(int i = 0; i < m; i++) {
            if(Double.isInfinite(dalpha[i]))
                throw new RuntimeException("Numerical problems: dalpha is infinite");
            if(Double.isNaN(dalpha[i])) {

                double diffSum = 0;
                for(int k = 0; k < m; k++)
                    diffSum += (KC[correct][k] - KC[i][k]) * expScores[k];
                
                System.out.println();
                System.out.println("In computeDerivatives.");
                System.out.println("m = " + m);
                System.out.println("KC = " + Arrays.deepToString(KC));             
                System.out.println("costs = " + Arrays.toString(costs));             
                System.out.println("s = " + Arrays.toString(s));
                System.out.println("alpha (prev) = " + Arrays.toString(alpha));
                System.out.println("expSum = " + expSum);
                System.out.println("expScores = " + Arrays.toString(expScores));
                System.out.println("dalpha = " + Arrays.toString(dalpha));
                System.out.println("diffSum[" + i + "]= " + diffSum);
                
                throw new RuntimeException("Numerical problems: dalpha is NaN");
            }
        }

    }

    private void normalizeDerivatives() {
        double max = 0;
        for(int i = 0; i < m; i++) {
            double d = Math.abs(dalpha[i]);
            if(d > max)
                max = d;
        }
        if(max > MAX_DALPHA_INFNORM) {
            max /= MAX_DALPHA_INFNORM;
            for(int i = 0; i < m; i++)
                dalpha[i] /= max;
        }
    }
    
    private void add_alpha(double step) {
        for(int i = 0; i < m; i++)
            alpha[i] += step * dalpha[i];
    }

    
}
