/*
 SplitByFeature.java
  
 Created on Aug 14, 2007 by Richard Johansson (richard@cs.lth.se).

 $Log: SplitByFeature.java,v $
 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:05  richard
 Branched.

 Revision 1.8  2008/04/21 08:41:17  richard
 Added computeScoresRestricted.

 Revision 1.7  2008/04/10 07:03:58  richard
 Added classifyRestricted.

 Revision 1.6  2008/04/08 11:45:58  richard
 Doesn't reuse algorithm instance.

 Revision 1.5  2008/04/07 13:46:42  richard
 Uses unary classifiers.

 Revision 1.4  2007/11/01 10:07:40  richard
 Does not create the training sets simultaneously.

 Revision 1.3  2007/10/08 12:31:34  richard
 Refactored again to remove atoms.

 Revision 1.2.2.1  2007/10/08 12:19:52  richard
 Refactored again to remove atoms.

 Revision 1.2  2007/09/28 14:42:28  richard
 Refactored to atom-based.

 Revision 1.1  2007/08/17 10:25:03  richard
 Added.

   
 */
package se.lth.cs.nlp.nlputils.ml_long.algorithms;

import se.lth.cs.nlp.nlputils.ml_long.*;
import se.lth.cs.nlp.nlputils.ml_long.classifiers.*;
import se.lth.cs.nlp.nlputils.core.*;

import java.util.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class SplitByFeature<I, V> extends LearningAlgorithm<MulticlassDefinition <I, V>, I, V> {

    private AlgorithmFactory af;
    private int offset;
    private String[] argv;
    private String fname;
    
    public SplitByFeature(String[] argv, int offset) {
        fname = argv[offset];
        af = new AlgorithmFactory();
        this.argv = argv;
        this.offset = offset;
    }
    
    public Classifier<I, V> train(MulticlassDefinition<I, V> spec, List<Pair<I, V>> trainingSet) { 
        if(spec.getClass() != MulticlassDefinition.class)
            throw new IllegalArgumentException("Can't use subclass of MulticlassDefinition...");
        System.out.println("Entering SplitByFeature.train");
        FeatureExtractor fe = spec.getFeatureExtractor();
        FeatureDeclaration decl = spec.getFeatureDecl();
        FeatureList fl = new FeatureList(decl);
        int findex = decl.getFeatureIndex(fname);        
        
        HashSet<String> values = new HashSet();       
        Histogram<V> hist = new Histogram<V>();
        
        for(Pair<I, V> ex: trainingSet) {
            fe.extractFeatures(ex.left, fl);
            Object k = fl.getFeature(findex);
            if(k instanceof HashSet)
                throw new IllegalArgumentException("SplitByFeature cannot split by set-valued features!");
            hist.add(ex.right);
            values.add((String) k);
        }

        System.out.println("SplitByFeature: Training subclassifiers.");
        
        HashMap<Object, Classifier<I, V>> cls = new HashMap(values.size());
        for(String k: values) {
            ArrayList<Pair<I, V>> subexs = new ArrayList();
            System.out.print("Training classifier for " + fname + " = " + k);
            HashSet<V> range = new HashSet();
            for(Pair<I, V> ex: trainingSet) {
                fe.extractFeatures(ex.left, fl);                
                Object k2 = fl.getFeature(findex);
                if(k.equals(k2)) {
                    subexs.add(ex);
                    range.add(ex.right);
                }
            }
            System.out.println(", range.size = " + range.size());
            if(range.size() == 1)
                cls.put(k, new UnaryClassifier(range.iterator().next()));
            else {
                LearningAlgorithm alg = af.create(argv[offset+1], argv, offset+2);
                cls.put(k, alg.train(spec, subexs));
            }
        }

        V defv = hist.getMostFrequent();
        
        System.out.println("SplitByFeature: Done, creating classifier.");
        
        return new SplittingClassifier(findex, cls, fe, fl, defv);
    }

    private static class SplittingClassifier<I, V> extends Classifier<I, V> {
        private static final long serialVersionUID = 1L;
        private int findex = -1;
        private HashMap<String, Classifier<I, V>> cls;
        private V defaultOutput;
        private FeatureExtractor<I> fe;
        private FeatureList list;
        SplittingClassifier(int findex, HashMap<String, Classifier<I, V>> cls, 
                            FeatureExtractor fe, FeatureList fl, V defaultOutput) {
            this.findex = findex;
            this.cls = cls;
            this.fe = fe;
            this.list = fl;
            this.defaultOutput = defaultOutput;
        }
        
        public V classifyByFeatures(FeatureList fl) {            
            Object k = fl.getFeature(findex);
            if(k instanceof HashSet)
                throw new IllegalArgumentException("SplitByFeature cannot split by set-valued features!");
            Classifier<I, V> cl = cls.get(k);
            if(cl == null) {
                //System.err.println("default for key " + k);
                return defaultOutput;
            } else {
                //System.out.println("Has classifier for key " + k);
                //System.out.println("its type is " + cl.getClass().getName());
                return cl.classifyByFeatures(fl);
            }
        }
        
        public V classifyByFeaturesRestricted(FeatureList fl, Collection<V> classes) {
            Object k = fl.getFeature(findex);
            if(k instanceof HashSet)
                throw new IllegalArgumentException("SplitByFeature cannot split by set-valued features!");
            Classifier<I, V> cl = cls.get(k);
            if(cl == null) {
                if(classes.contains(defaultOutput))
                    return defaultOutput;
                else
                    return classes.iterator().next();
            } else 
                return cl.classifyByFeaturesRestricted(fl, classes);
        }
        public V classify(I instance) {
            fe.extractFeatures(instance, list);
            Object k = list.getFeature(findex);
            if(k instanceof HashSet)
                throw new IllegalArgumentException("SplitByFeature cannot split by set-valued features!");
            Classifier<I, V> cl = cls.get(k);
            if(cl == null)
                return defaultOutput;
            else 
                return cl.classify(instance);
        }
        
        public V classifyRestricted(I instance, Collection<V> classes) {
            fe.extractFeatures(instance, list);
            Object k = list.getFeature(findex);
            if(k instanceof HashSet)
                throw new IllegalArgumentException("SplitByFeature cannot split by set-valued features!");
            Classifier<I, V> cl = cls.get(k);

            if(cl == null)
                if(classes.contains(defaultOutput))
                    return defaultOutput;
                else
                    return classes.iterator().next();
            else
                return cl.classifyRestricted(instance, classes);
            
        }
        
        public void computeScoresRestricted(I instance, Collection<V> classes, double[] scores) {
            if(classes.size() == 0)
                return;
            fe.extractFeatures(instance, list);
            Object k = list.getFeature(findex);
            if(k instanceof HashSet)
                throw new IllegalArgumentException("SplitByFeature cannot split by set-valued features!");
            Classifier<I, V> cl = cls.get(k);
            if(cl == null) {
                if(classes.contains(defaultOutput)) {
                    int index = 0;
                    for(V v: classes)
                        if(v.equals(defaultOutput))
                            scores[index++] = 0.0;
                        else
                            scores[index++] = Double.NEGATIVE_INFINITY;                    
                } else {
                    double value = Math.log(1.0 / classes.size());
                    for(int i = classes.size()-1; i >= 0; i--)
                        scores[i] = value;
                }
            } else
                cl.computeScoresRestricted(instance, classes, scores);
        }
        
        public void setFeatureExtractor(FeatureExtractor<I> extractor,
                FeatureList list) {
            this.fe = extractor;
            this.list = list;
            for(Classifier cl: cls.values())
                cl.setFeatureExtractor(extractor);
        }
    }
    
}