/*
 BinaryNumericClassifier.java
 
 Created on Mar 27, 2006 by Richard Johansson (richard@cs.lth.se).
 
 $Log: BinaryNumericClassifier.java,v $
 Revision 1.2  2010-06-07 11:30:38  johansson
 Major update.

 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:04  richard
 Branched.

 Revision 1.5  2008/04/21 08:42:31  richard
 Added computeScoresRestricted.

 Revision 1.4  2008/04/10 07:04:41  richard
 Added classifyRestricted.

 Revision 1.3  2008/03/19 15:47:35  richard
 Added setSparseVector.

 Revision 1.2  2007/10/08 12:31:34  richard
 Refactored again to remove atoms.

 Revision 1.1.2.1  2007/10/08 12:20:14  richard
 Refactored again to remove atoms.

 Revision 1.1  2007/09/28 14:42:11  richard
 Moved to new package.

 Revision 1.6  2006/09/22 08:06:39  richard
 Renamed specification to definition.

 Revision 1.5  2006/09/18 09:08:36  richard
 Uses MulticlassDefinition instead of NumericEncoding.

 Revision 1.4  2006/09/07 17:51:19  richard
 New revision.

 Revision 1.3  2006/04/19 09:30:15  richard
 Type parameter.
 
 Revision 1.2  2006/04/10 09:05:56  richard
 Added documentation.
 
 Revision 1.1  2006/04/05 11:38:35  richard
 Added the file.
 
 
 */

package se.lth.cs.nlp.nlputils.ml_long.classifiers;
import se.lth.cs.nlp.nlputils.ml_long.*;
import java.util.Collection;

/**
 * Abstract class for binary numeric classifiers, such as classical
 * linear or quadratic models.
 * 
 * @author Richard Johansson (richard@cs.lth.se)
 */
public abstract class BinaryNumericClassifier<I,V> extends Classifier<I,V> {

    private static final long serialVersionUID = 0L;
    
    /**
     * The range of output values. 
     * <code>classes[0]</code> is returned when <code>apply</code> returns
     * a positive value. Should have a length of 2.
     */
    protected V[] classes;
    protected SparseVector sv = new SparseVector();
    
    protected MulticlassDefinition<I, V> spec;
    
    /**
     * Applies the discriminator function.
     * 
     * @param rep the sparse vector representation of the instance.
     * @return the value of the discriminator function.
     */
    protected abstract double apply();
    
    /**
     * Auxiliary method that encodes the features and calls <code>apply</code>.
     *
     * @param features the feature list to classify.
     * @return the value of the discriminator function.
     */
    public double applyFeatures(FeatureList features) {
        spec.encodeFeatures(features, sv);
        return apply();
    }
    
    public double apply(I instance) {
    	spec.encodeInstance(instance, sv);
        return apply();
    }

    public V classify(I instance) {
        return apply(instance) > 0? classes[0]: classes[1];
    }
    
    public V classifyByFeatures(FeatureList features) {
        return applyFeatures(features) > 0? classes[0]: classes[1];
    }    
    
    public V classifyRestricted(I instance, Collection<V> range) {
        if(apply(instance) > 0) {
            if(range.contains(classes[0]))
                return classes[0];
            if(range.contains(classes[1]))
                return classes[1];
            return range.iterator().next();
        } else {
            if(range.contains(classes[1]))
                return classes[1];
            if(range.contains(classes[0]))
                return classes[0];
            return range.iterator().next();            
        }
    }
    
    public void computeScoresRestricted(I instance, Collection<V> range, double[] scores) {
        double d = apply(instance);
        int index = 0;
        for(V v: range) {
            if(v.equals(classes[0]))
                scores[index] = d;
            else if(v.equals(classes[1]))
                scores[index] = -d;
            else
                scores[index] = Double.NEGATIVE_INFINITY;
            index++;
        }
    }
    
    public void setSparseVector(SparseVector sv) {
        this.sv = sv;
    }
    
}