/*
 ClasswiseClassifier.java
 
 Created on Apr 19, 2006 by Richard Johansson (richard@cs.lth.se).
 
 $Log: ClasswiseClassifier.java,v $
 Revision 1.2  2010-06-07 11:30:38  johansson
 Major update.

 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:04  richard
 Branched.

 Revision 1.4  2008/04/10 07:05:17  richard
 Changes in classifyRestricted.

 Revision 1.3  2008/03/19 15:48:05  richard
 Implemented.

 Revision 1.2  2007/10/08 12:31:34  richard
 Refactored again to remove atoms.

 Revision 1.1.2.1  2007/10/08 12:20:14  richard
 Refactored again to remove atoms.

 Revision 1.1  2007/09/28 14:42:11  richard
 Moved to new package.

 Revision 1.5  2006/09/22 08:06:58  richard
 Renamed specification to definition.

 Revision 1.4  2006/09/18 09:08:36  richard
 Uses MulticlassDefinition instead of NumericEncoding.

 Revision 1.3  2006/09/07 17:51:19  richard
 New revision.

 Revision 1.2  2006/04/19 09:29:53  richard
 Type parameter. Added documentation.
 
 Revision 1.1  2006/04/19 09:21:38  richard
 Added the file.
 
 
 */
package se.lth.cs.nlp.nlputils.ml_long.classifiers;

import se.lth.cs.nlp.nlputils.ml_long.*;
import se.lth.cs.nlp.nlputils.core.*;
import gnu.trove.*;

import java.io.Serializable;
import java.util.*;

/**
 * Multiclass classifier by applying a binary classifier for each class.
 * 
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class ClasswiseClassifier<I,V> extends Classifier<I,V> {
    
    private static final long serialVersionUID = 0L;
    
    private BinaryNumericClassifier[] subClassifiers;
    private V[] outValues;
    private TObjectIntHashMap<V> valueMap;
    SparseVector rep = new SparseVector();
    private MulticlassDefinition<I, V> def;

    public ClasswiseClassifier(MulticlassDefinition<I, V> def, BinaryNumericClassifier[] subClassifiers, V[] outValues) {
        this.subClassifiers = subClassifiers;
        this.outValues = outValues;
        this.def = def;
        for(BinaryNumericClassifier cl: subClassifiers)
            cl.setSparseVector(rep);
        valueMap = CollectionUtils.inverseMap(outValues);
    }    
    
    private V apply() {
        double bestValue = Double.NEGATIVE_INFINITY;
        int bestIndex = -1;
        for(int i = 0; i < subClassifiers.length; i++) {
            double value = subClassifiers[i].apply(rep);
            if(value > bestValue) {
                bestValue = value;
                bestIndex = i;
            }
        }
        return outValues[bestIndex];        
    }

    private V applyRestricted(Collection<V> range) {
        double bestValue = Double.NEGATIVE_INFINITY;
        V best = null;
        for(V v: range) {
            int indexPlus1 = valueMap.get(v);
            if(indexPlus1 > 0) {
                double value = subClassifiers[indexPlus1-1].apply();
                if(value > bestValue) {
                    bestValue = value;
                    best = v;
                }
            }
        }
        if(best == null)
        	return range.iterator().next();
        return best;
    }
    
    private void getScoresRestricted(Collection<V> range, double[] scores) {
        int scIndex = 0;
        for(V v: range) {
            int indexPlus1 = valueMap.get(v);
            if(indexPlus1 > 0)
                scores[scIndex++] = subClassifiers[indexPlus1-1].apply();
            else
                scores[scIndex++] = Double.NEGATIVE_INFINITY;
        }
    }
    
    public V classify(I instance) {    	
    	def.encodeInstance(instance, rep);
    	if(instance instanceof RestrictedInstance) {
    		RestrictedInstance ri = (RestrictedInstance) instance;
    		List<V> range = (List<V>) Arrays.asList(ri.range);
    		return applyRestricted(range);
    	} else
    		return apply();
    }
    
    public V classifyByFeatures(FeatureList features) {
        def.encodeFeatures(features, rep);
        return apply();
    }
    
    public V classifyRestricted(I instance, Collection<V> classes) {
        def.encodeInstance(instance, rep);
        V out = applyRestricted(classes);
        if(out == null)
            return classes.iterator().next();
        else
            return out;
    }
    
    public V classifyByFeaturesRestricted(FeatureList features, Collection<V> classes) {
        def.encodeFeatures(features, rep);
        V out = applyRestricted(classes);
        if(out == null)
            return classes.iterator().next();
        else
            return out;
    }
    
    public void computeScores(I instance, double[] scores) {
        throw new UnsupportedOperationException("Unrestricted scoring not supported.");
    }
    
    public void computeScoresRestricted(I instance, Collection<V> classes, double[] scores) {
        def.encodeInstance(instance, rep);
        getScoresRestricted(classes, scores);
    }
    
    public void computeScoresByFeatures(FeatureList features, double[] scores) {
        throw new UnsupportedOperationException("Unrestricted scoring not supported.");        
    }

    public void computeScoresByFeaturesRestricted(FeatureList features, Collection<V> classes, double[] scores) {
        def.encodeFeatures(features, rep);
        getScoresRestricted(classes, scores);
    }
     
    protected static class NBestItem<V> implements Serializable {
        private static final long serialVersionUID = 0L;
        double score;
        int cl;
    }
    
    protected static class NBestComp implements Comparator<NBestItem>, Serializable {
        private static final long serialVersionUID = 0L;
        public int compare(NBestItem t1, NBestItem t2) {
            int c1 = -Double.compare(t1.score, t2.score);
            if(c1 != 0)
                return c1;
            return t1.cl - t2.cl;
        }
    }

    protected static Comparator<NBestItem> ordering = new NBestComp();
    
    public int nbest(int n, I instance, V[] values, double[] scores) {

    	def.encodeInstance(instance, rep);
    	
    	TreeSet<NBestItem> ts = new TreeSet(ordering);
    	
    	if(instance instanceof RestrictedInstance) {
    		RestrictedInstance ri = (RestrictedInstance) instance;
    		for(Object v: ri.range) {
                int indexPlus1 = valueMap.get((V) v);
                if(indexPlus1 > 0) {
                    NBestItem item = new NBestItem();
                    item.score = subClassifiers[indexPlus1-1].apply();
                    item.cl = indexPlus1-1;
                    ts.add(item);
                }
            }
    	} else {
    		for(int i = 0; i < subClassifiers.length; i++) {
    			NBestItem item = new NBestItem();    			
    			item.score = subClassifiers[i].apply(rep);
    			item.cl = i;
    			ts.add(item);
            }	
    	}
    	
    	//System.out.println(ts.size());
    	
    	int index = 0;
    	for(NBestItem item: ts) {
    		values[index] = outValues[item.cl];
    		scores[index] = item.score;
    		index++;
    		if(index == n || index == values.length)
    			break;
    	}
    	return index;
    }
    
    public int nbest(int n, I instance, V[] values, double[] scores, SparseVector[] reps) {
        throw new UnsupportedOperationException("nbest with reps not supported.");
    }
    
    
}