/*
 GreedyForwardFeatureSearch.java

 Created on Apr 2, 2008 by Richard Johansson (richard@cs.lth.se).

 $Log: GreedyForwardFeatureSearch.java,v $
 Revision 1.1  2010-06-07 11:28:52  johansson
 Added.

 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:04  richard
 Branched.

 Revision 1.6  2008/04/10 07:00:07  richard
 Cross-validates.

 Revision 1.5  2008/04/07 12:11:42  richard
 Update.

 Revision 1.4  2008/04/03 08:38:31  richard
 Specify a start set.

 Revision 1.3  2008/04/03 07:48:18  richard
 Bug in pair set.

 Revision 1.2  2008/04/02 13:26:25  richard
 Skips inverse pairs.

 Revision 1.1  2008/04/02 11:23:52  richard
 Added.


 */
package se.lth.cs.nlp.nlputils.ml_long.featuresearch;

import java.io.IOException;
import java.util.*;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;

import se.lth.cs.nlp.nlputils.core.*;
import it.unitn.disi.nlputils.switchboard.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class GreedyForwardFeatureSearch<I, V> extends FeatureSearch<I, V> {

	private static final double MIN_IMPROVEMENT = 1e-6;

	private FeatureSet startSet;

	private int switchboardPort;

	public GreedyForwardFeatureSearch() {
		switchboardPort = Integer.MIN_VALUE;
	}

	public GreedyForwardFeatureSearch(int switchboardPort) {
		this.switchboardPort = switchboardPort;
	}

	public void setStartSet(FeatureSet startSet) {
		this.startSet = startSet;
	}

	private static class Increment {
		double score;
		double increment;
		Object added;
		FeatureSet newSetXX;

		public String toString() {
			return "(" + added + ", " + increment + ")";    		
		}

		Increment() {    		
		}

		Increment(Object o, double inc) {
			this.increment = inc;
			this.added = o;
		}

	}

	private static abstract class Evaluator {
		abstract double singleEval(FeatureSet fs);
		abstract Increment unigramMax(ArrayList<String> allFeatures,
				double baselineResult,
				FeatureSet baselineFS);
		abstract Increment bigramMax(ArrayList<String> allFeatures,
				double baselineResult,
				FeatureSet baselineFS);
		abstract void stop();
	}

	private static class SimpleEvaluator extends Evaluator {
		Function<FeatureSet, Double> eval;

		SimpleEvaluator(Function<FeatureSet, Double> eval) {
			this.eval = eval;
		}

		double singleEval(FeatureSet fs) {
			return eval.apply(fs);
		}

		Increment unigramMax(ArrayList<String> allFeatures,
				double baselineResult, FeatureSet currentFS) {

			Increment out = new Increment();
			double bestResult = baselineResult;

			for(String feature: allFeatures) {
				if(currentFS.fs.contains(feature))
					continue;
				System.out.println("* Trying " + feature);

				FeatureSet fs = new FeatureSet();
				fs.ps = currentFS.ps;
				fs.fs = new HashSet(currentFS.fs);
				fs.fs.add(feature);

				double result = eval.apply(fs);

				System.out.println("* Result: " + result);
				if(result >= bestResult + MIN_IMPROVEMENT) {
					bestResult = result;

					out.added = feature;
					out.increment = result - baselineResult; 
					out.score = result;

					out.newSetXX = fs;

					System.out.println("* Improvement: " + out);
					/*try {
	                    System.in.read();
	                } catch(Exception e) {}*/
				}
			}

			if(out.added == null)
				return null;
			else
				return out;
		}		

		Increment bigramMax(ArrayList<String> allFeatures, double baselineResult,
				FeatureSet currentFS) {
			Increment out = new Increment();
			double bestResult = baselineResult;

			HashSet<Pair<String, String>> done = new HashSet();

			for(String f1: currentFS.fs) {
				for(String f2: allFeatures) {
					if(f2.equals(f1))
						continue;
					Pair<String, String> p = new Pair(f1, f2);
					Pair<String, String> pinv = new Pair(f2, f1);
					if(currentFS.ps.contains(p) || currentFS.ps.contains(pinv))
						continue;
					if(done.contains(p) || done.contains(pinv))
						continue;
					done.add(p);
					done.add(pinv);

					System.out.println("* Trying " + p);

					FeatureSet newFS = new FeatureSet();
					newFS.fs = new HashSet(currentFS.fs);
					newFS.ps = new HashSet(currentFS.ps);

					newFS.fs.add(f2);
					newFS.ps.add(p);

					double result = eval.apply(newFS);

					System.out.println("* Result: " + result);
					if(result >= bestResult + MIN_IMPROVEMENT) {
						out.newSetXX = newFS;
						out.added = p;
						out.increment = result - baselineResult; 
						out.score = result;
						bestResult = result;
						System.out.println("* Improvement: " + out);
						/*try {
	                        System.in.read();
	                    } catch(Exception e) {}*/
					}

				}
			}

			if(out.added == null)
				return null;
			else
				return out;
		}

		void stop() {			
		}

	}

	private static class SwitchboardEvaluator extends Evaluator {
		Switchboard<FeatureSet, Double> sb;

		SwitchboardEvaluator(int sbPort, Function<FeatureSet, Double> eval) {
			sb = new Switchboard(sbPort, eval.getClass());
			try {
				sb.start();
			} catch(Exception e) {
				throw new RuntimeException(e);				
			}
		}

		double singleEval(FeatureSet fs) {
			try {
				return sb.submit(fs).get();
			} catch(ExecutionException e) {
				throw new RuntimeException(e);
			} catch(InterruptedException e) {
				throw new RuntimeException(e);
			}
		}

		Increment unigramMax(ArrayList<String> allFeatures,
				double baselineResult, FeatureSet currentFS) {

			Increment out = new Increment();
			double bestResult = baselineResult;

			ArrayList<Triple<String, Future<Double>, FeatureSet>> jobs = new ArrayList();

			for(String feature: allFeatures) {
				if(currentFS.fs.contains(feature))
					continue;
				FeatureSet fs = new FeatureSet();
				fs.ps = currentFS.ps;
				fs.fs = new HashSet(currentFS.fs);
				fs.fs.add(feature);
				jobs.add(new Triple(feature, sb.submit(fs), fs));
			}

			for(Triple<String, Future<Double>, FeatureSet> t: jobs) {
				double result = 0;
				try {
					result = t.second.get();
				} catch(Exception e) {
					throw new RuntimeException(e);
				}
				if(result >= bestResult + MIN_IMPROVEMENT) {
					bestResult = result;
					out.added = t.first;
					out.increment = result - baselineResult; 
					out.score = result;
					out.newSetXX = t.third;;
					System.out.println("* Improvement: " + out);
				}
			}

			if(out.added == null)
				return null;
			else
				return out;
		}		

		Increment bigramMax(ArrayList<String> allFeatures, double baselineResult,
				FeatureSet currentFS) {

			Increment out = new Increment();
			double bestResult = baselineResult;

			ArrayList<Triple<Pair<String, String>, Future<Double>, FeatureSet>> jobs = new ArrayList();
			HashSet<Pair<String, String>> done = new HashSet();

			for(String f1: currentFS.fs) {
				for(String f2: allFeatures) {
					if(f2.equals(f1))
						continue;
					Pair<String, String> p = new Pair(f1, f2);
					Pair<String, String> pinv = new Pair(f2, f1);
					if(currentFS.ps.contains(p) || currentFS.ps.contains(pinv))
						continue;
					if(done.contains(p) || done.contains(pinv))
						continue;
					done.add(p);
					done.add(pinv);

					FeatureSet newFS = new FeatureSet();
					newFS.fs = new HashSet(currentFS.fs);
					newFS.ps = new HashSet(currentFS.ps);
					newFS.fs.add(f2);
					newFS.ps.add(p);
					jobs.add(new Triple(p, sb.submit(newFS), newFS));
				}
			}

			for(Triple<Pair<String, String>, Future<Double>, FeatureSet> t: jobs) {
				double result = 0;
				try {
					result = t.second.get();
				} catch(Exception e) {
					throw new RuntimeException(e);
				}
				if(result >= bestResult + MIN_IMPROVEMENT) {
					bestResult = result;
					out.added = t.first;
					out.increment = result - baselineResult; 
					out.score = result;
					out.newSetXX = t.third;;
					System.out.println("* Improvement: " + out);
				}
			}

			if(out.added == null)
				return null;
			else
				return out;
		}

		void stop() {			
			try {
				sb.stop();
			} catch(IOException e) {
				throw new RuntimeException(e);			
			}
		}

	}

	private FeatureSet search(ArrayList<String> allFeatures, 
			Evaluator evaluator) {

		List<Increment> history = new ArrayList();

		FeatureSet currentFS;
		if(startSet == null)
			currentFS = new FeatureSet();
		else {
			currentFS = startSet;
			for(String s: currentFS.fs)
				history.add(new Increment(s, Double.NaN));
			for(Pair p: currentFS.ps)
				history.add(new Increment(p, Double.NaN));
		}

		double bestResult = evaluator.singleEval(currentFS);

		System.out.println("* Baseline value: " + bestResult);

		/* 1. Add single features. */
		while(true) {
			double previousResult = bestResult;

			Increment increment = evaluator.unigramMax(allFeatures,
					previousResult,
					currentFS);

			if(increment == null)
				break;
			else {
				history.add(increment);
				currentFS = increment.newSetXX;
				System.out.println("*** Best result: " + increment);
				System.out.println("*** Current set: " + history);
				bestResult = increment.score;
			}
		}

		/* 2. Add pairs. */
		while(true) {
			double previousResult = bestResult;

			Increment increment = evaluator.bigramMax(allFeatures,
					previousResult,
					currentFS);

			if(increment == null)
				break;
			else {
				history.add(increment);
				currentFS = increment.newSetXX;
				System.out.println("*** Best result: " + increment);
				System.out.println("*** Current set: " + history);
				bestResult = increment.score;
			}

		}

		System.out.println("**** Best result: " + bestResult);
		System.out.println("**** Final set: " + history);

		return currentFS;
	}

	// varför har jag både localSearch och search?????
	
//	public FeatureSet localSearch(ArrayList<String> allFeatures,
//			Function<FeatureSet, Double> eval) {
//		Evaluator evaluator;
//		if(switchboardPort > 0) {
//			evaluator = new SwitchboardEvaluator(switchboardPort, eval);
//		} else {
//			evaluator = new SimpleEvaluator(eval);
//		}
//		FeatureSet out = search(allFeatures, evaluator);
//		evaluator.stop();		
//		return out;
//	}

	public FeatureSet search(ArrayList<String> allFeatures,
			Function<FeatureSet, Double> eval) {
		Evaluator evaluator;
		if(switchboardPort > 0) {
			evaluator = new SwitchboardEvaluator(switchboardPort, eval);
		} else {
			evaluator = new SimpleEvaluator(eval);
		}
		FeatureSet out = search(allFeatures, evaluator);
		evaluator.stop();		
		return out;
	}
	
}
