/*
 LibLinearInterface.java
  
 Created on Apr 1, 2008 by Richard Johansson (richard@cs.lth.se).

 $Log: LibLinearInterface.java,v $
 Revision 1.3  2011-05-26 09:25:38  johansson
 Added support for multiclass SVMs.

 Revision 1.2  2010-06-07 11:27:49  johansson
 Major update.

 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:05  richard
 Branched.

 Revision 1.4  2008/04/21 08:45:28  richard
 Added sparse option.

 Revision 1.3  2008/04/10 07:06:45  richard
 Suppresses messages.

 Revision 1.2  2008/04/08 11:46:51  richard
 Null check.

 Revision 1.1  2008/04/02 11:19:33  richard
 Added.

   
 */
package se.lth.cs.nlp.nlputils.ml_long.glue;

import gnu.trove.TIntIntHashMap;

import java.io.*;
import java.util.*;

import se.lth.cs.nlp.nlputils.core.*;
import se.lth.cs.nlp.nlputils.ml_long.classifiers.*;
import se.lth.cs.nlp.nlputils.ml_long.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
public class LibLinearInterface<I, V> extends LearningAlgorithm<MulticlassDefinition<I, V>, I, V> {
    
    private static final String TRAINING_COMMAND = "ll-train";
    
    private String argString = "";
    private boolean quiet = false;
    private boolean dump = false;
    private boolean sparse = false;
    private boolean compress = false;
    private int sparseThreshold = -1;

    private boolean multiclass = false;
    
    public LibLinearInterface(String[] argv, int offset) {
        for(int i = offset; i < argv.length; i++) {
        	if(argv[i].equals("-s"))
        		if(i < argv.length - 1 && argv[i+1].equals("4"))
        			multiclass = true;
        	if(argv[i].equals("-quiet"))
                quiet = true;
            else if(argv[i].equals("-dump"))
                dump = true;
            else if(argv[i].equals("-sparse"))
                sparse = true;
            else if(argv[i].equals("-compress"))
                compress = true;
            else if(argv[i].equals("-sparseThreshold"))
            	sparseThreshold = Integer.parseInt(argv[++i]);
            else
                argString += " " + argv[i];
        }
    }
    
    private Process process;
    
    public Classifier<I, V> train(MulticlassDefinition<I, V> spec, List<Pair<I, V>> trainingSet) {
        try {
            Histogram<V> hist = new Histogram();
            for(Pair<I, V> ex: trainingSet)
                hist.add(ex.right);
            if(hist.size() > 2 && !multiclass)
                throw new IllegalArgumentException("Only binary classifiers supported if not in multiclass mode");
            if(hist.size() < 2)
                return new UnaryClassifier<I, V>(hist.getMostFrequent());
            

            
            if(multiclass && !quiet) {
            	for(IntObjPair<V> p: hist.asSortedList()) {
            		System.out.println(p.right + "->" + spec.encodeClass(p.right) + ", " + p.left);
            	}
            }
            
            String prefix = System.getProperty("java.io.tmpdir");
            if(!prefix.endsWith(File.separator))
            	prefix = prefix + File.separator;
            String fileName = prefix + "LibLinearInt" + Util.uniqueSymbol();

            PrintWriter pw = new PrintWriter(new FileWriter(fileName));

            SparseVector rep = new SparseVector();
            
            int count = 0;
            for(Pair<I, V> ex: trainingSet) {
                count++;
                
                spec.encodeInstance(ex.left, rep);
                rep.sortIndices(); // TODO give error instead of modifying?
                
                int cl = spec.encodeClass(ex.right);

                //System.out.println("rep = " + rep + ", cl = " + cl);                
                
                pw.print(cl);

                for(int i = 0; i < rep.index; i++)
                    pw.print(" " + rep.keys[i] + ":" + rep.values[i]);

                pw.println();
            }

            pw.close();
            
            if(dump)
                return null;

            // 100512 spec.freeze();

            final String modelFile = fileName + ".model";
            
            String cmdString = TRAINING_COMMAND + argString + " " + fileName + " " + modelFile;

            if(!quiet) {
                System.out.println("LibLinearInterface: training, " + count + " instances.");
                //System.out.println(cmdString);
            }

            process = Runtime.getRuntime().exec(cmdString);

            Thread t1 = new Thread() {
                public void run() {
                    try {
                        BufferedReader in = new BufferedReader(new InputStreamReader(process.getInputStream()));
                        int c;
                        while(true) {
                            c = in.read();
                            if(c == -1)
                                break;
                            if(!quiet) {
                                System.out.print((char) c);
                                System.out.flush();
                            }
                        }
                    } catch(IOException e) {
                        e.printStackTrace();
                    }
                }
            };
            t1.start();
            
            Thread t2 = new Thread() {
                public void run() {
                    Scanner sc = new Scanner(process.getErrorStream());
                    while(sc.hasNextLine())
                        System.out.println(sc.nextLine());
                }
            };
            t2.start();

            int status = process.waitFor();

            if(status != 0)
                throw new IOException("LibLinear was unsuccessful.");

            t1.join();
            t2.join();
            
            Classifier<I, V> cl;
            if(multiclass)
                cl = readMulticlassModel(spec, modelFile, sparse, compress, sparseThreshold);
            else
            	cl = readModel(spec, modelFile, sparse, compress, sparseThreshold);
            
            if(true) {
                if(!new File(fileName).delete())  
                    System.err.println("Could not delete temporary data file.");

                if(!new File(modelFile).delete())
                    System.err.println("Could not delete temporary model file.");
            }
 
            if(!quiet)
                System.out.println("LibLinearInterface: training finished.");
            
            return cl;

        } catch(Exception e) {
            e.printStackTrace();
            throw new RuntimeException(e);
        }
    }
    
    private static <I, V> BinaryNumericClassifier<I, V> readModel(
            MulticlassDefinition<I, V> spec,
            String file,
            boolean sparse,
            boolean compress,
            int sparseThreshold) throws IOException {
        Scanner sc = new Scanner(new File(file));
        
        double bias = 0;
        int nr_feature = -1;

        V[] labels = null;
        
        /* Read preamble. */
        while(sc.hasNextLine()) {
            String line = sc.nextLine().trim();
            if(line.equals("w"))
                break;
            else if(line.startsWith("bias"))
                bias = Double.parseDouble(line.substring("bias ".length()));
            else if(line.startsWith("nr_feature"))
                nr_feature = Integer.parseInt(line.substring("nr_feature ".length()));
            else if(line.startsWith("nr_class")) {
               int nr_class = Integer.parseInt(line.substring("nr_class ".length()));
               if(nr_class > 2)
                   throw new RuntimeException("Only binary classification supported!");
            } else if(line.startsWith("label")) {
                if(spec != null) {
                    String[] tokens = line.split(" ");
                    V l1 = spec.decodeClass(Integer.parseInt(tokens[1]));
                    if(l1 == null) {
                        throw new RuntimeException("null value in definition");
                    }
                    labels = (V[]) java.lang.reflect.Array.newInstance(spec.decodeClass(Integer.parseInt(tokens[1])).getClass(), tokens.length - 1);
                    for(int i = 0; i < labels.length; i++)
                        labels[i] = spec.decodeClass(Integer.parseInt(tokens[i+1]));
                }
            }
        }

        /* Read weight vector. */
        int index = 0;
        double[] w = new double[nr_feature + 1];
        double bias_weight = 0;
        
        int nNonZero = 0;
        
        while(sc.hasNextLine()) {
            double d = Double.parseDouble(sc.nextLine());
            if(index == nr_feature && bias >= 0.0)
                bias_weight = d;
            else if(index >= nr_feature)
                throw new RuntimeException("index >= nr_feature");
            else {
                w[index + 1] = d;
                if(d > TOLERANCE || d < -TOLERANCE)
                	nNonZero++;
            }
            index++;
        }
        if(bias < 0 && index < nr_feature || bias >= 0 && index < nr_feature + 1)
            throw new RuntimeException("did not fill weight vector!");

        sc.close();
        sc = null;
        
        if(sparse || nNonZero < sparseThreshold)
            return new SparseBinaryLinearClassifier(spec, labels, bias*bias_weight, w);
        else {
        	BinaryLinearClassifier out = new BinaryLinearClassifier(spec, labels, bias*bias_weight, w);
        	if(compress) {
        		System.out.println("Compressing binary classifier.");
        		Pair<BinaryLinearClassifier<I, V>, TIntIntHashMap> p
        		= out.compress(COMPRESS_TOLERANCE);
        		out = p.left;
        		p = null;
        		System.out.println("Done.");
        	}
        	return out;
        }
    }

    private static <I, V> ClasswiseClassifier<I, V> readMulticlassModel(
            MulticlassDefinition<I, V> spec,
            String file,
            boolean sparse,
            boolean compress,
            int sparseThreshold) throws IOException {
        Scanner sc = new Scanner(new File(file));
        
        double bias = 0;
        int nr_feature = -1;

        V[] labels = null;

        int nr_class = -1;
        
        /* Read preamble. */
        while(sc.hasNextLine()) {
            String line = sc.nextLine().trim();
            if(line.equals("w"))
                break;
            else if(line.startsWith("bias"))
                bias = Double.parseDouble(line.substring("bias ".length()));
            else if(line.startsWith("nr_feature"))
                nr_feature = Integer.parseInt(line.substring("nr_feature ".length()));
            else if(line.startsWith("nr_class")) {
            	nr_class = Integer.parseInt(line.substring("nr_class ".length()));
            } else if(line.startsWith("label")) {
                if(spec != null) {
                    String[] tokens = line.split(" ");
                    V l1 = spec.decodeClass(Integer.parseInt(tokens[1]));
                    if(l1 == null) {
                        throw new RuntimeException("null value in definition");
                    }
                    labels = (V[]) java.lang.reflect.Array.newInstance(spec.decodeClass(Integer.parseInt(tokens[1])).getClass(), tokens.length - 1);
                    for(int i = 0; i < labels.length; i++)
                        labels[i] = spec.decodeClass(Integer.parseInt(tokens[i+1]));
                }
            }
        }

        /* Read weight vector. */
        int index = 0;
        double[][] w = new double[nr_class][nr_feature + 1];
        double[] bias_weight = new double[nr_class];
        
        int[] nNonZero = new int[nr_class];
        
        while(sc.hasNextLine()) {
        	double[] ds = CollectionUtils.toDoubleArray(sc.nextLine().split("\\s+"));
            if(index == nr_feature && bias >= 0.0)
                bias_weight = ds;
            else if(index >= nr_feature)
                throw new RuntimeException("index >= nr_feature");
            else {
            	for(int i = 0; i < nr_class; i++) {
            		w[i][index+1] = ds[i];
            		if(ds[i] > TOLERANCE || ds[i] < -TOLERANCE)
            			nNonZero[i]++;
            	}
            }
            index++;
        }
        if(bias < 0 && index < nr_feature || bias >= 0 && index < nr_feature + 1)
            throw new RuntimeException("did not fill weight vector!");

        sc.close();
        sc = null;
        
        BinaryNumericClassifier[] classifiers = new BinaryNumericClassifier[labels.length];
        
        for(int i = 0; i < classifiers.length; i++) {
        	EncodedMulticlassDefinition subspec = new EncodedMulticlassDefinition();
            if(sparse || nNonZero[i] < sparseThreshold)
            	classifiers[i] = new SparseBinaryLinearClassifier(subspec, labels, bias*bias_weight[i], w[i]);
            else {
            	BinaryLinearClassifier cl = new BinaryLinearClassifier(subspec, labels, bias*bias_weight[i], w[i]);
            	if(compress) {
            		System.out.println("Compressing binary classifier.");
            		Pair<BinaryLinearClassifier<I, V>, TIntIntHashMap> p
            		= cl.compress(COMPRESS_TOLERANCE);
            		classifiers[i] = p.left;
            		p = null;
            		System.out.println("Done.");
            	} else
            		classifiers[i] = cl;
            }        	
        }
        
        return new ClasswiseClassifier(spec, classifiers, labels);        
    }
    
    private static final double COMPRESS_TOLERANCE = 1e-4;
    private static final double TOLERANCE = 1e-9;
    
    public static void main(String[] argv) {
        String model = argv[0];
        String testfile = argv[1];

        try {        
            BinaryNumericClassifier cl = readModel(null, model, false, false, -1);
            Scanner sc = new Scanner(new File(testfile));
            while(sc.hasNextLine()) {
                String line = sc.nextLine().trim();
                String[] ss = line.split("\\s+");
                SparseVector sv = new SparseVector();
                for(int i = 1; i < ss.length; i++) {
                    int index = ss[i].indexOf(':');
                    int ix = Integer.parseInt(ss[i].substring(0, index));
                    double val = Double.parseDouble(ss[i].substring(index + 1));
                    sv.put(ix, val);
                }
                cl.setSparseVector(sv);
                //System.out.println(cl.apply());
                System.exit(0);
            }
        } catch(Exception e) {
            e.printStackTrace();
        }
    
    }
    
}
