/*
 LibSVMClassifier.java
 
 Created on Mar 27, 2006 by Richard Johansson (richard@cs.lth.se).
 
 $Log: LibSVMClassifier.java,v $
 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:05  richard
 Branched.

 Revision 1.18  2007/10/08 12:31:34  richard
 Refactored again to remove atoms.

 Revision 1.17.2.1  2007/10/08 12:20:40  richard
 Refactored again to remove atoms.

 Revision 1.17  2007/09/28 14:53:31  richard
 Refactored to atom-based.

 Revision 1.16  2007/08/22 10:58:13  richard
 Checks for empty range.

 Revision 1.15  2007/08/22 10:57:28  richard
 Default.

 Revision 1.14  2007/08/20 12:06:23  richard
 Checks for null.

 Revision 1.13  2007/05/02 06:58:13  richard
 Moved libsvm to external.

 Revision 1.12  2007/02/13 16:09:35  richard
 computeProbabilities

 Revision 1.11  2006/11/20 12:43:29  richard
 Renamed to classifyByFeaturesRestricted.

 Revision 1.10  2006/09/22 08:10:01  richard
 Renamed specification to definition.

 Revision 1.9  2006/09/19 13:34:41  richard
 Added canPredictProbs.

 Revision 1.8  2006/09/19 08:28:53  richard
 Probabilities.

 Revision 1.7  2006/09/18 08:11:54  richard
 Implemented classifyRestricted.

 Revision 1.6  2006/09/18 08:06:49  richard
 Adapted to new framework.

 Revision 1.5  2006/09/07 17:51:19  richard
 New revision.

 Revision 1.4  2006/05/30 14:30:18  richard
 Handles the case when the classifier knows nothing about the allowed outputs.
 
 Revision 1.3  2006/04/19 09:30:25  richard
 Type parameter.
 
 Revision 1.2  2006/04/07 14:00:53  richard
 Added restricted classification.
 
 Revision 1.1  2006/04/05 11:45:40  richard
 Added the file.
 
 
 */
package se.lth.cs.nlp.nlputils.ml_long.glue;

import se.lth.cs.nlp.nlputils.external.adapted_libsvm.*;
import se.lth.cs.nlp.nlputils.ml_long.*;

import java.io.IOException;
import java.util.*;

import gnu.trove.TIntIntHashMap;


/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
class LibSVMClassifier<I,V> extends Classifier<I,V> {
    
    private static final long serialVersionUID = 1L;		
    
    private svm_model model;
    private MulticlassDefinition<I, V> spec;
    private double[] probs;
    private TIntIntHashMap labelToIndex;
    private boolean canPredictProbs;
    private SparseVector sv = new SparseVector();
    
    LibSVMClassifier(String modelName, MulticlassDefinition<I, V> spec) throws IOException {
        model = svm.svm_load_model(modelName);
        probs = new double[svm.svm_get_nr_class(model)];
        labelToIndex = new TIntIntHashMap();
        int[] labels = new int[svm.svm_get_nr_class(model)];
        svm.svm_get_labels(model, labels);
        for(int i = 0; i < labels.length; i++)
            labelToIndex.put(labels[i], i + 1);
        this.spec = spec;

        canPredictProbs = svm.svm_check_probability_model(model) == 1;
    }
    
    private svm_node[] encodeSparseVector(SparseVector rep) {
        svm_node[] x = new svm_node[rep.index];
        for(int i = 0; i < rep.index; i++) {
            x[i] = new svm_node();
            x[i].index = (int) rep.keys[i];
            x[i].value = rep.values[i];
        }
        return x;
    }

    public V classify(I x) {
        spec.encodeInstance(x, sv);
        sv.sortIndices();
        svm_node[] xx = encodeSparseVector(sv);
        int answer = (int) svm.svm_predict(model, xx);
        return spec.decodeClass(answer);
    }
    
    public V classifyByFeatures(FeatureList features) {
        spec.encodeFeatures(features, sv);
        sv.sortIndices();
        svm_node[] x = encodeSparseVector(sv);
        int answer = (int) svm.svm_predict(model, x);
        return spec.decodeClass(answer);
    }
    
    public void computeScoresByFeaturesRestricted(FeatureList features, Collection<V> classes, double[] probs) {
        if(!canPredictProbs)
            throw new RuntimeException("This SVM classifier was not trained to predict probabilities.");

        spec.encodeFeatures(features, sv);
        sv.sortIndices();
        
        svm_node[] x = encodeSparseVector(sv);

        svm.svm_predict_probability(model, x, this.probs);

        int index = 0;
        double sum = 0;
        for(V ybar: classes) {
            int cl = spec.encodeClass(ybar);
            int clIndex = labelToIndex.get(cl);
            if(clIndex != 0) {
                probs[index] = this.probs[clIndex - 1];
                sum += probs[index];
            }
            else
                probs[index] = 0;
            index++;
        }

        if(sum > 0)
            for(int i = 0; i < probs.length; i++)
                probs[i] /= sum;
    }

    public void computeScoresRestricted(I instance, Collection<V> classes, double[] probs) {
        if(!canPredictProbs)
            throw new RuntimeException("This SVM classifier was not trained to predict probabilities.");
        spec.encodeInstance(instance, sv);
        sv.sortIndices();
        
        svm_node[] x = encodeSparseVector(sv);

        svm.svm_predict_probability(model, x, this.probs);

        int index = 0;
        double sum = 0;
        for(V ybar: classes) {
            int cl = spec.encodeClass(ybar);
            int clIndex = labelToIndex.get(cl);
            if(clIndex != 0) {
                probs[index] = this.probs[clIndex - 1];
                sum += probs[index];
            }
            else
                probs[index] = 0;
            index++;
        }

        if(sum > 0)
            for(int i = 0; i < probs.length; i++)
                probs[i] /= sum;
    }
    
    private int[] encodeTargets(Collection<V> classes) {
        HashSet<Integer> out = new HashSet<Integer>();
        for(V cl: classes) {
            int o = spec.encodeClass(cl);
            if(o != -1)
                out.add(o);
        }
        int[] outArr = new int[out.size()];
        int n = 0;
        for(Integer i: out)
            outArr[n++] = i;
        return outArr;
    }
    
    public V classifyByFeaturesRestricted(FeatureList features, Collection<V> classes) {
        if(classes == null)
            throw new RuntimeException("classes == null");
        if(classes.size() == 0)
            throw new RuntimeException("no classes");
        if(classes.size() == 1)
            return classes.iterator().next();
        int[] clEnc = encodeTargets(classes);
        if(clEnc.length == 0)
            return classes.iterator().next();
        spec.encodeFeatures(features, sv);
        sv.sortIndices();
        svm_node[] x = encodeSparseVector(sv);
        int answer = (int) svm.svm_predict_restricted(model, x, clEnc);
        return spec.decodeClass(answer);
    }
    
}
