/*
 MulticlassQuadratic.java

 Created on Mar 21, 2006 by Richard Johansson (richard@cs.lth.se).

 $Log: MulticlassQuadratic.java,v $
 Revision 1.1  2009/01/16 10:06:48  johansson
 Added to brenta repository.

 Revision 1.1  2008/05/16 12:16:05  richard
 Branched.

 Revision 1.7  2007/10/08 12:31:34  richard
 Refactored again to remove atoms.

 Revision 1.6.2.1  2007/10/08 12:20:40  richard
 Refactored again to remove atoms.

 Revision 1.6  2007/09/28 14:53:31  richard
 Refactored to atom-based.

 Revision 1.5  2007/08/17 10:26:15  richard
 Adapted to the new ML framework.

 Revision 1.4  2007/05/02 06:58:13  richard
 Moved libsvm to external.

 Revision 1.3  2007/04/02 07:01:23  richard
 Corrected classifyByFeatures.

 Revision 1.2  2007/04/01 13:06:42  richard
 Debug code.

 Revision 1.1  2007/04/01 13:01:34  richard
 Moved to nlputils.


 */
package se.lth.cs.nlp.nlputils.ml_long.glue;

import java.io.*;
import java.util.*;

import se.lth.cs.nlp.nlputils.external.adapted_libsvm.*;
import se.lth.cs.nlp.nlputils.ml_long.*;

/**
 * @author Richard Johansson (richard@cs.lth.se)
 */
class MulticlassQuadratic<I, V> extends Classifier<I, V> implements Serializable {

    private static final long serialVersionUID = 1L;

    private int nClasses;

    private double[][] lC;
    private int[][][] qInd;
    private double[][][] qC;

    private double[] rhos;

    private int[] maxIndices;

    //private NumericEncoding<V> enc;
    private MulticlassDefinition<I, V> def;
    private SparseVector sv = new SparseVector();
    
    private double[] values;

    private int[] labels;
    
    public V classifyByFeatures(FeatureList fl) {        
        //System.out.println("in MulticlassQuadratic.classify");
        def.encodeFeatures(fl, sv);
        sv.sortIndices();
        applySV(sv, values);
        int cl = labels[getClassification(values)];
        return def.decodeClass(cl);
    }

    public V classify(I instance) {
        def.encodeInstance(instance, sv);
        sv.sortIndices();
        applySV(sv, values);
        int cl = labels[getClassification(values)];
        return def.decodeClass(cl);
    }
    
    static <I, V> MulticlassQuadratic<I, V> loadFromLibSVM(String filename, 
            MulticlassDefinition<I, V> def) throws IOException {
        MulticlassQuadratic<I, V> out = loadFromLibSVM(filename);
        out.def = def;
        return out;
    }
    
    static <I, V> MulticlassQuadratic<I, V> loadFromLibSVM(String filename) throws IOException {
        return loadFromLibSVM(filename, false);
    }

    static <I, V> MulticlassQuadratic<I, V> loadFromLibSVM(String filename, boolean verbose) throws IOException {
        FileInputStream fis = new FileInputStream(filename);
        return loadFromLibSVM(fis, verbose);
    }    

    static <I2, V2> MulticlassQuadratic<I2, V2> loadFromLibSVM(InputStream is, boolean verbose) throws IOException {
        MulticlassQuadratic<I2, V2> out = new MulticlassQuadratic<I2, V2>();

        BufferedReader r = new BufferedReader(new InputStreamReader(is));

        double gamma = 1;
        double coef0 = 0;

        int[] nSVs = null;

        String line = r.readLine();
        while(!line.equals("SV")) {

            if(line.startsWith("svm_type"))
                check(line.equals("svm_type c_svc"), "Illegal SVM type.");
            else if(line.startsWith("kernel_type"))
                check(line.equals("kernel_type polynomial"), "Illegal kernel type.");
            else if(line.startsWith("degree"))
                check(line.equals("degree 2"), "Illegal degree.");
            else if(line.startsWith("nr_class")) {
                //check(line.equals("nr_class 2"), "Illegal number of classes.");
                out.nClasses = Integer.parseInt(line.substring(9));
            } else if(line.startsWith("coef0"))
                //check(line.equals("coef0 0"), "Illegal kernel coefficient.");
                coef0 = Double.parseDouble(line.substring(5));
            else if(line.startsWith("label")) {
                String[] tokens = line.split(" ");
                out.labels = new int[tokens.length - 1];
                for(int i = 1; i < tokens.length; i++)
                    out.labels[i-1] = Integer.parseInt(tokens[i]);
                
                //out.label1 = Integer.parseInt(tokens[1]);
                //out.label2 = Integer.parseInt(tokens[2]);
            } else if(line.startsWith("gamma")) {
                gamma = Double.parseDouble(line.substring(5));
            } else if(line.startsWith("rho")) {
                String[] tokens = line.split("\\s+");
                out.rhos = new double[tokens.length - 1];
                for(int i = 0; i < out.rhos.length; i++)
                    out.rhos[i] = Double.parseDouble(tokens[i+1]);
            } else if(line.startsWith("probA")) {
                //out.probA = Double.parseDouble(line.substring(5));
            } else if(line.startsWith("probB")) {
                //out.probB = Double.parseDouble(line.substring(5));
            } else if(line.startsWith("nr_sv")) {
                String[] tokens = line.split("\\s+");
                nSVs = new int[tokens.length - 1];
                for(int i = 0; i < nSVs.length; i++)
                    nSVs[i] = Integer.parseInt(tokens[i+1]);
            }

            line = r.readLine();
        }

        try {
            /* TODO only boolean features? */
            
            int nCl = out.nClasses * (out.nClasses - 1) / 2;

            HashMap<Integer, Double>[] lTables = new HashMap[nCl];
            HashMap<Integer, TreeMap<Integer, Double>>[] qTables = new HashMap[nCl];

            for(int i = 0; i < nCl; i++) {
                lTables[i] = new HashMap<Integer, Double>();
                qTables[i] = new HashMap<Integer, TreeMap<Integer, Double>>();
            }

            out.maxIndices = new int[nCl];

            double[] alphas = new double[out.nClasses - 1];

            int countLines = 0;

            for(int i = 0; i < out.nClasses; i++) {

                for(int j = 0; j < nSVs[i]; j++) {

                    line = r.readLine();
                    int ix = line.indexOf('#');
                    if(ix != -1)
                        line = line.substring(0, ix);
                    line = line.trim();

                    countLines++;
                    if(verbose && countLines % 10 == 0) {
                        System.err.println("countLines = " + countLines);
                    }

                    String[] tokens = line.split("\\s+");

                    //double alpha = Double.parseDouble(tokens[0]);
                    for(int k = 0; k < out.nClasses - 1; k++) {
                        alphas[k] = Double.parseDouble(tokens[k]);

                        int clIndex;
                        if(k < i)
                            clIndex = getClIndex(out.nClasses, k, i);
                        else
                            clIndex = getClIndex(out.nClasses, i, k+1);

                        out.rhos[clIndex] -= alphas[k] * coef0 * coef0;
                    }


                    ArrayList<Integer> indices = new ArrayList<Integer>();

                    for(int k = out.nClasses - 1; k < tokens.length; k++) {
                        String s1 = tokens[k].replaceFirst(":(.*)", "");
                        String s2 = tokens[k].replaceFirst("^(.*):", "");
                        double val = Double.parseDouble(s2);
                        int v2;
                        if(Math.abs(val) < 1e-6)
                            v2 = 0;
                        else if(Math.abs(val - 1) < 1e-6)
                            v2 = 1;
                        else
                            throw new RuntimeException("Illegal feature value");
                        if(v2 == 1)
                            indices.add(Integer.parseInt(s1));
                    }

                    for(int k = 0; k < indices.size(); k++) {
                        int index = indices.get(k);

                        for(int m = 0; m < i; m++) {
                            int clIndex = getClIndex(out.nClasses, m, i);
                            if(index > out.maxIndices[clIndex])
                                out.maxIndices[clIndex] = index;
                        }
                        for(int m = i+1; m < out.nClasses; m++) {
                            int clIndex = getClIndex(out.nClasses, i, m);
                            if(index > out.maxIndices[clIndex])
                                out.maxIndices[clIndex] = index;
                        }

                        double l = 2 * gamma * coef0 + gamma*gamma; // antar boolean
                        double q = 2 * gamma*gamma;

                        for(int m = 0; m < out.nClasses - 1; m++) {

                            int clIndex;
                            if(m < i)
                                clIndex = getClIndex(out.nClasses, m, i);
                            else
                                clIndex = getClIndex(out.nClasses, i, m+1);

                            addToMap(lTables[clIndex], index, alphas[m] * l);

                            TreeMap<Integer, Double> qT = qTables[clIndex].get(index);
                            if(qT == null) {
                                qT = new TreeMap<Integer, Double>();
                                qTables[clIndex].put(index, qT);
                            }

                            for(int j2 = k+1; j2 < indices.size(); j2++)
                                addToMap(qT, indices.get(j2), alphas[m] * q);
                        }
                    }
                }				
            }

            out.lC = new double[nCl][];
            out.qC = new double[nCl][][];
            out.qInd = new int[nCl][][];

            for(int i = 0; i < nCl; i++) {
                out.lC[i] = new double[out.maxIndices[i]+1];
                out.qC[i] = new double[out.maxIndices[i]+1][];
                out.qInd[i] = new int[out.maxIndices[i]+1][];

                for(Map.Entry<Integer, Double> e: lTables[i].entrySet())
                    out.lC[i][e.getKey()] = e.getValue();

                for(Map.Entry<Integer, TreeMap<Integer, Double>> e: qTables[i].entrySet()){
                    int index = e.getKey();
                    TreeMap<Integer, Double> qT = e.getValue();
                    out.qC[i][index] = new double[qT.size()];
                    out.qInd[i][index] = new int[qT.size()];
                    int n = 0;
                    for(Map.Entry<Integer, Double> e2: qT.entrySet()) {
                        out.qInd[i][index][n] = e2.getKey();
                        out.qC[i][index][n] = e2.getValue();
                        n++;
                    }
                }
            }

        } catch(NumberFormatException e) {
            e.printStackTrace();
            throw new RuntimeException("Got exception.");
        }

        out.values = new double[out.nClasses*(out.nClasses-1)/2];

        return out;
    }

    private static void check(boolean b, String message) {
        if(!b)
            throw new RuntimeException("Check failed: " + message);
    }


    private void apply(int[] indices, double[] dfs) {
        int nCl = nClasses * (nClasses - 1) / 2;
        for(int j = 0; j < nCl; j++) {
            double out = -rhos[j];
            for(int i = 0; i < indices.length; i++) {
                int index1 = indices[i];				
                if(index1 > maxIndices[j])
                    break;
                out += lC[j][index1];
                if(qInd[j][index1] == null)
                    continue;
                int[] qInd_i = qInd[j][index1];
                double[] qC_i = qC[j][index1];
                int i1 = i+1, i2 = 0;
                while(i1 < indices.length && i2 < qInd_i.length) {
                    int ind1 = indices[i1];
                    int ind2 = qInd_i[i2];
                    if(ind1 < ind2)
                        i1++;
                    else if(ind1 > ind2)
                        i2++;
                    else {
                        out += qC_i[i2];
                        i1++;
                        i2++;
                    }
                }
            }
            dfs[j] = out;
        }
    }

    private void applySV(SparseVector sv, double[] dfs) {
        /* TODO denna r inte validerad. */
        int nCl = nClasses * (nClasses - 1) / 2;
        for(int j = 0; j < nCl; j++) {
            double out = -rhos[j];
            for(int i = 0; i < sv.index; i++) {
                int index1 = (int) sv.keys[i];
                if(index1 > maxIndices[j])
                    break;
                out += lC[j][index1] * sv.values[i] * sv.values[i];
                if(qInd[j][index1] == null)
                    continue;
                int[] qInd_i = qInd[j][index1];
                double[] qC_i = qC[j][index1];
                int i1 = i+1, i2 = 0;
                while(i1 < sv.index && i2 < qInd_i.length) {
                    int ind1 = (int) sv.keys[i1];
                    int ind2 = qInd_i[i2];
                    if(ind1 < ind2)
                        i1++;
                    else if(ind1 > ind2)
                        i2++;
                    else {
                        out += qC_i[i2] * sv.values[i] * sv.values[i1];
                        i1++;
                        i2++;
                    }
                }
            }
            dfs[j] = out;
        }
    }
    
    private static void addToMap(Map map, Object key, double value) {
        Double dval = (Double) map.get(key);

        //boolean f1 = map == fmap1;
        //System.out.println("Readding: " + f1);

        double d = dval == null? value: dval.doubleValue() + value;
        map.put(key, new Double(d));
    }


    private static int getClIndex(int n, int i1, int i2) {
        return i2 - i1*(i1 - 2*n + 3)/2 - 1;
    }

    private static int getClassification(double[] values) {
        int len = (int) Math.round(0.5 + Math.sqrt(0.25*(1+8.0*values.length)));
        int[] cs = new int[len];
        int ix = 0;
        for(int i = 0; i < len-1; i++)
            for(int j = i+1; j < len; j++)
                if(values[ix++] > 0)
                    cs[i]++;
                else
                    cs[j]++;
        int max = -1;
        int maxindex = -1;
        for(int i = 0; i < len; i++)
            if(cs[i] > max) {
                max = cs[i];
                maxindex = i;
            }        
        return maxindex;
    }

    public static void main(String[] argv) {
        try {

            MulticlassQuadratic ex = loadFromLibSVM(argv[0], false);

            System.out.println("Loaded model 1.");

            int n = ex.nClasses * (ex.nClasses - 1) / 2;

            svm_model model = svm.svm_load_model(argv[0]);

            System.out.println("Loaded model 2.");

            double[] values1 = new double[n], values2 = new double[n];

            int correct1 = 0, correct2 = 0, ntotal = 0;

            long time1 = 0, time2 = 0;

            int[] classes = new int[svm.svm_get_nr_class(model)];
            svm.svm_get_labels(model, classes);              
            System.out.println(Arrays.toString(classes));

            BufferedReader r = new BufferedReader(new FileReader(argv[1]));
            String line = r.readLine();
            while(line != null) {

                String[] tokens = line.split("\\s+");

                int value = Integer.parseInt(tokens[0]);

                ArrayList<Integer> indices = new ArrayList<Integer>();              
                for(int i = 1; i < tokens.length; i++) {
                    String s1 = tokens[i].replaceFirst(":(.*)", "");
                    String s2 = tokens[i].replaceFirst("^(.*):", "");
                    //check(s2.matches("0*[01](\\.0*)?"), "Illegal feature value!");
                    double val = Double.parseDouble(s2);
                    int v2;
                    if(Math.abs(val) < 1e-6)
                        v2 = 0;
                    else if(Math.abs(val - 1) < 1e-6)
                        v2 = 1;
                    else
                        throw new RuntimeException("Illegal feature value");
                    if(v2 == 1)
                        indices.add(Integer.parseInt(s1));
                }

                int[] is = new int[indices.size()];
                for(int i = 0; i < indices.size(); i++)
                    is[i] = indices.get(i);

                long l0 = System.currentTimeMillis();

                ex.apply(is, values1); 
                int p1 = values1[0] < 0? 1: 0;

                long l1 = System.currentTimeMillis();

                svm_node[] nodes = new svm_node[indices.size()];
                for(int i = 0; i < nodes.length; i++) {
                    nodes[i] = new svm_node();
                    nodes[i].index = indices.get(i);
                    nodes[i].value = 1;
                }

                long l2 = System.currentTimeMillis();

                svm.svm_predict_values(model, nodes, values2);
                int p2 = values2[0] < 0? 1: 0;

                long l3 = System.currentTimeMillis();

                /*
                if(Math.abs(values1[4] - values2[4]) > 0.01) {

                    System.out.println(values1[4]);
                    System.out.println(values2[4]);

                    System.out.println(line);
                    System.exit(0);
                }*/

                int c1 = classes[getClassification(values1)];
                int c2 = classes[getClassification(values2)];
                int c3 = (int) svm.svm_predict(model, nodes);

                if(c1 != c2 || c1 != c3) {
                    System.out.println(Arrays.toString(values1));
                    System.out.println(Arrays.toString(values2));

                    System.out.println(c1);
                    System.out.println(c2);

                    throw new RuntimeException("!!!");
                }

                if(c1 == value)                   
                    correct1++;
                if(c2 == value)
                    correct2++;

                ntotal++;

                time1 += l1 - l0;
                time2 += l3 - l2;

                line = r.readLine();
            }

            System.out.println("Correct 1: " + correct1);
            System.out.println("Correct 2: " + correct2);
            System.out.println("Total: " + ntotal);

            System.out.println();
            System.out.println("Time in apply (expanded): " + time1);
            System.out.println("Time in apply (libsvm):   " + time2);

        } catch(Exception e) {
            e.printStackTrace();
        }
    }


}
